with Iirs; use Iirs;

package Sem is
   -- Check the semantic of a design unit.
   -- Also add a few node or change some nodes, when for exemple an
   -- identifier is changed into an access to the type.
   procedure Semantic (Design_Unit: Iir_Design_Unit);

   -- Get the current design unit, ie, the parameter of the procedure semantic.
   function Get_Current_Design_Unit return Iir_Design_Unit;

   -- Get the current subprogram or process.
   function Get_Current_Subprogram return Iir;

   -- Declare an implicit signal.
   procedure Add_Declaration_For_Implicit_Signal (Sig : Iir);

   -- Semantize subtype indication DEF.
   -- If INCOMPLETE is TRUE, then DEF may designate an incomplete type
   -- definition.
   -- This is used by sem_expr for qualified expression and allocators.
   function Sem_Subtype_Indication (Def: Iir; Incomplete : Boolean := False)
     return Iir;

   -- Return FALSE if A_TYPE is an unconstrained array type or subtype.
   function Sem_Is_Constrained (A_Type: Iir) return Boolean;

   --  Add a driver for SIG.
   --  STMT is used in case of error (it is the statement that creates the
   --   driver).
   --  Do nothing if:
   --    The current statement list does not belong to a process,
   --    SIG is a formal signal interface.
   procedure Sem_Add_Driver (Sig : Iir; Stmt : Iir);

   --  Create a default generic or port map aspect that associates all elements
   --  of ENTITY (if any) to elements of COMP with the same name or to
   --  an open association.
   --  If KIND is GENERIC_MAP, apply this on generics, if KIND is PORT_MAP,
   --  apply this on ports.
   --  PARENT is used to report errors.
   type Map_Kind_Type is (Map_Generic, Map_Port);
   function Create_Default_Map_Aspect
     (Comp : Iir; Entity : Iir; Kind : Map_Kind_Type; Parent : Iir)
     return Iir_Association_List;
end Sem;
