with System;
with System.Address_To_Access_Conversions;
with Ada.Unchecked_Deallocation;
with Interfaces; use Interfaces;
with Ortho_Nodes; use Ortho_Nodes;
with Ortho_Ident; use Ortho_Ident;
with Evaluation; use Evaluation;
with Flags;
with Ada.Text_IO;
with Types; use Types;
with Errorout; use Errorout;
with Name_Table; -- use Name_Table;
with Iirs_Utils; use Iirs_Utils;
with Std_Package; use Std_Package;
with Libraries;
with Files_Map;
with Std_Names;
with Configuration;
with Interfaces.C_Streams;

package body Translation is

   --  Ortho type node for STD.BOOLEAN.
   Std_Boolean_Type_Node : O_Tnode;
   Std_Boolean_True_Node : O_Enode;
   Std_Boolean_False_Node : O_Enode;
   --  Ortho type node for string template pointer.
   Std_String_Ptr_Node : O_Tnode;

   --  Ortho type node for std.time.
   Std_Time_Type : O_Tnode;

   --  Ortho type for std.file_open_status.
   Std_File_Open_Status_Type : O_Tnode;

   --  Node for the variable containing the current filename.
   Current_Filename_Node : O_Lnode;
   Current_Filename_Set : Boolean := False;

   --  Global declarations.
   Ghdl_Ptr_Type : O_Tnode;
   Const_Ptr_Type_Node : O_Tnode;
   Sizetype : O_Tnode;
   Ghdl_Integer_Type : O_Tnode;
   Ghdl_Real_Type : O_Tnode;
   Ghdl_Phys_Type : O_Tnode;
   --  Constant character.
   Char_Type_Node : O_Tnode;
   --  Array of char.
   Chararray_Type_Node : O_Tnode;
   --  Pointer to array of char.
   Char_Ptr_Type_Node : O_Tnode;
   --  Constant 0 of type const_char_ptr.
   Null_Const_Char_Ptr_Node : O_Enode;

   Ghdl_Index_Type : O_Tnode;

   --  Type for a file (this is in fact a index in a private table).
   File_Index_Type : O_Tnode;
   File_Index_Ptr_Type : O_Tnode;

   --  Record containing a len and string fields.
   Ghdl_Str_Len_Type_Node : O_Tnode;
   Ghdl_Str_Len_Type_Len_Field : O_Fnode;
   Ghdl_Str_Len_Type_Str_Field : O_Fnode;
   Ghdl_Str_Len_Ptr_Node : O_Tnode;
   Ghdl_Str_Len_Array_Type_Node : O_Tnode;

   --  Location.
   Ghdl_Location_Type_Node : O_Tnode;
   Ghdl_Location_Filename_Node : O_Fnode;
   Ghdl_Location_Line_Node : O_Fnode;
   Ghdl_Location_Col_Node : O_Fnode;
   Ghdl_Location_Ptr_Node : O_Tnode;

   --  Allocate memory for a block.
   Ghdl_Alloc_Ptr : O_Dnode;

   --  Register a process.
   Ghdl_Process_Register : O_Dnode;
   Ghdl_Sensitized_Process_Register : O_Dnode;

   --  Register a sensitivity for a process.
   Ghdl_Process_Add_Sensitivity : O_Dnode;

   --  Register a driver for a process.
   Ghdl_Process_Add_Driver : O_Dnode;

   --  Procedure called in case of check failed.
   Ghdl_Program_Error : O_Dnode;
   Ghdl_Bound_Check_Failed_L0 : O_Dnode;
   Ghdl_Bound_Check_Failed_L1 : O_Dnode;

   --  Procedure called in case of assert failed.
   Ghdl_Assert_Failed : O_Dnode;
   --  Ortho node for default report message.
   Std_Assert_Default_Report_Node : O_Lnode;

   --  NOW variables.
   Ghdl_Now : O_Lnode;

   --  Wait subprograms.
   --  Short forms.
   Ghdl_Process_Wait_Timeout : O_Dnode;
   Ghdl_Process_Wait_Exit : O_Dnode;
   --  Complete form:
   Ghdl_Process_Wait_Set_Timeout : O_Dnode;
   Ghdl_Process_Wait_Add_Sensitivity : O_Dnode;
   Ghdl_Process_Wait_Suspend : O_Dnode;
   Ghdl_Process_Wait_Close : O_Dnode;

   --  bool type.
   Ghdl_Bool_Type_Node : O_Tnode;
   type Enode_Boolean_Array is array (Boolean) of O_Enode;
   Ghdl_Bool_Nodes : Enode_Boolean_Array;
   Ghdl_Bool_False_Node : O_Enode renames Ghdl_Bool_Nodes (False);
   Ghdl_Bool_True_Node : O_Enode renames Ghdl_Bool_Nodes (True);

   --  Comparaison type.
   Ghdl_Compare_Type : O_Tnode;
   Ghdl_Compare_Lt : O_Enode;
   Ghdl_Compare_Eq : O_Enode;
   Ghdl_Compare_Gt : O_Enode;

   --  process_desc type
--    Ghdl_Process_Desc_Type_Node : O_Tnode;
--    Ghdl_Process_Desc_Name_Node : O_Fnode;
--    Ghdl_Process_Desc_Sensitized_Node : O_Fnode;
--    Ghdl_Process_Desc_Passive_Node : O_Fnode;
--    Ghdl_Process_Desc_Proc_Node : O_Fnode;
--    Ghdl_Process_Desc_Signals_Node : O_Fnode;

--    Ghdl_Process_Desc_Ptr_Node : O_Tnode;

   --  Dir type.
   Ghdl_Dir_Type_Node : O_Tnode;
   Ghdl_Dir_To_Node : O_Enode;
   Ghdl_Dir_Downto_Node : O_Enode;

   --  type_kind type.
   Ghdl_Type_Kind_Type_Node : O_Tnode;
   Ghdl_Kind_Bool_Node : O_Enode;
   Ghdl_Kind_E8_Node : O_Enode;
   Ghdl_Kind_E32_Node : O_Enode;
   Ghdl_Kind_I32_Node : O_Enode;
   Ghdl_Kind_I64_Node : O_Enode;
   Ghdl_Kind_F64_Node : O_Enode;
   Ghdl_Kind_P64_Node : O_Enode;
   Ghdl_Kind_Access_Node : O_Enode;
   Ghdl_Kind_Array_Node : O_Enode;
   Ghdl_Kind_Record_Node : O_Enode;
   Ghdl_Kind_Subarray_Node : O_Enode;
   Ghdl_Kind_Se8_Node : O_Enode;
   Ghdl_Kind_Si32_Node : O_Enode;
   Ghdl_Kind_Sf64_Node : O_Enode;
   Ghdl_Kind_Sp64_Node : O_Enode;

   --  Access to any ghdl_desc.
   Ghdl_Desc_Ptr_Type : O_Tnode;
   Ghdl_Desc_Ptr_Array_Type : O_Tnode;
   Ghdl_Desc_Ptr_Array_Ptr_Type : O_Tnode;

   --  ghdl_desc_bool type.
   Ghdl_Desc_Bool_Type : O_Tnode;
   Ghdl_Desc_Bool_Kind_Node : O_Fnode;
   Ghdl_Desc_Bool_False_Node : O_Fnode;
   Ghdl_Desc_Bool_True_Node : O_Fnode;

   --  ghdl_desc_e8 type.
   Ghdl_Desc_E8_Type_Node : O_Tnode;
   Ghdl_Desc_E8_Kind_Node : O_Fnode;
   Ghdl_Desc_E8_Erange_Node : O_Fnode;
   Ghdl_Desc_E8_Values_Node : O_Fnode;

   --  ghdl_desc_i32 type
   Ghdl_Desc_I32_Type_Node : O_Tnode;
   Ghdl_Desc_I32_Kind_Node : O_Fnode;

   --  ghdl_desc_f64_type
   Ghdl_Desc_F64_Type_Node : O_Tnode;
   Ghdl_Desc_F64_Kind_Node : O_Fnode;

   Ghdl_P64_Type : O_Tnode;
   --  ghdl_unit_desc
   Ghdl_Unit_Desc_Type : O_Tnode;
   Ghdl_Unit_Desc_Name_Node : O_Fnode;
   Ghdl_Unit_Desc_Val_Node : O_Fnode;

   Ghdl_Unit_Desc_Ptr : O_Tnode;
   Ghdl_Unit_Desc_Array : O_Tnode;

   --  ghdl_desc_p64_type
   Ghdl_Desc_P64_Type_Node : O_Tnode;
   Ghdl_Desc_P64_Kind_Node : O_Fnode;
   Ghdl_Desc_P64_Base_Name_Node : O_Fnode;
   Ghdl_Desc_P64_Units_Node : O_Fnode;

   --  ghdl_desc_access_type
   Ghdl_Desc_Access_Type_Node : O_Tnode;
   Ghdl_Desc_Access_Kind_Node : O_Fnode;
   Ghdl_Desc_Access_Designated_Node : O_Fnode;

   --  ghdl_desc_array type
   Ghdl_Desc_Array_Type_Node : O_Tnode;
   Ghdl_Desc_Array_Kind_Node : O_Fnode;
   Ghdl_Desc_Array_El_Node : O_Fnode;
   Ghdl_Desc_Array_Dim_Node : O_Fnode;
   Ghdl_Desc_Array_Elsize_Node : O_Fnode;
   Ghdl_Desc_Array_Indexes_Node : O_Fnode;

   --  ghdl_desc_subarray type
   Ghdl_Desc_Subarray_Type_Node : O_Tnode;
   Ghdl_Desc_Subarray_Kind_Node : O_Fnode;
   Ghdl_Desc_Subarray_Basetype_Node : O_Fnode;
   Ghdl_Desc_Subarray_Range_Node : O_Fnode;

   --  ghdl_desc_scalar_subtype
   Ghdl_Desc_Scalar_Subtype_Node : O_Tnode;
   Ghdl_Desc_Scalar_Kind_Node : O_Fnode;
   Ghdl_Desc_Scalar_Basetype_Node : O_Fnode;
   Ghdl_Desc_Scalar_Range_Node : O_Fnode;

   --  ghdl_desc_record_type
   Ghdl_Desc_Record_Type_Node : O_Tnode;
   Ghdl_Desc_Record_Kind_Node : O_Fnode;
   Ghdl_Desc_Record_Nbr_El_Node : O_Fnode;
   Ghdl_Desc_Record_Els_Node : O_Fnode;

   --  Signals.
   Ghdl_Scalar_Bytes : O_Tnode;
   Ghdl_Signal_Type : O_Tnode;
   Ghdl_Signal_Value_Node : O_Fnode;
   Ghdl_Signal_Driving_Value_Node : O_Fnode;
   Ghdl_Signal_Last_Value_Node : O_Fnode;
   Ghdl_Signal_Last_Event_Node : O_Fnode;
   Ghdl_Signal_Last_Active_Node : O_Fnode;
   Ghdl_Signal_Event_Node : O_Fnode;
   Ghdl_Signal_Active_Node : O_Fnode;
   Ghdl_Signal_Ptr : O_Tnode;

   Ghdl_Signal_Simple_Assign_Error : O_Dnode;
   Ghdl_Signal_Start_Assign_Error : O_Dnode;
   Ghdl_Signal_Next_Assign_Error : O_Dnode;

   Ghdl_Signal_Start_Assign_Null : O_Dnode;
   Ghdl_Signal_Next_Assign_Null : O_Dnode;

   Ghdl_Create_Signal_E8 : O_Dnode;
   Ghdl_Signal_Simple_Assign_E8 : O_Dnode;
   Ghdl_Signal_Start_Assign_E8 : O_Dnode;
   Ghdl_Signal_Next_Assign_E8 : O_Dnode;

   Ghdl_Create_Signal_B2 : O_Dnode;
   Ghdl_Signal_Simple_Assign_B2 : O_Dnode;
   Ghdl_Signal_Start_Assign_B2 : O_Dnode;
   Ghdl_Signal_Next_Assign_B2 : O_Dnode;

   Ghdl_Create_Signal_I32 : O_Dnode;
   Ghdl_Signal_Simple_Assign_I32 : O_Dnode;
   Ghdl_Signal_Start_Assign_I32 : O_Dnode;
   Ghdl_Signal_Next_Assign_I32 : O_Dnode;

   Ghdl_Create_Signal_F64 : O_Dnode;
   Ghdl_Signal_Simple_Assign_F64 : O_Dnode;
   Ghdl_Signal_Start_Assign_F64 : O_Dnode;
   Ghdl_Signal_Next_Assign_F64 : O_Dnode;

   Ghdl_Create_Signal_P64 : O_Dnode;
   Ghdl_Signal_Simple_Assign_P64 : O_Dnode;
   Ghdl_Signal_Start_Assign_P64 : O_Dnode;
   Ghdl_Signal_Next_Assign_P64 : O_Dnode;

   Ghdl_Signal_Add_Source : O_Dnode;
   Ghdl_Signal_Effective_Value : O_Dnode;

   Ghdl_Signal_Create_Resolution : O_Dnode;

   Ghdl_Signal_Name : O_Dnode;

   --  Stable signal.
   Ghdl_Create_Stable_Signal : O_Dnode;
   Ghdl_Signal_Register_Stable_Prefix : O_Dnode;

   --  Guard signal.
   Ghdl_Signal_Create_Guard : O_Dnode;
   Ghdl_Signal_Guard_Dependence : O_Dnode;

   --  Stack 2.
   Ghdl_Stack2_Allocate : O_Dnode;
   Ghdl_Stack2_Mark : O_Dnode;
   Ghdl_Stack2_Release : O_Dnode;

   --  Predefined subprograms.
   Ghdl_Memcpy : O_Dnode;
   Ghdl_Deallocate : O_Dnode;
   Ghdl_Malloc : O_Dnode;
   Ghdl_Real_Exp : O_Dnode;
   Ghdl_Integer_Exp : O_Dnode;

   --  Predefined file subprograms.
   Ghdl_Text_File_Elaborate : O_Dnode;
   Ghdl_File_Elaborate : O_Dnode;

   Ghdl_Text_File_Open : O_Dnode;
   Ghdl_File_Open : O_Dnode;

   Ghdl_Text_File_Open_Status : O_Dnode;
   Ghdl_File_Open_Status : O_Dnode;

   Ghdl_Text_Write : O_Dnode;
   Ghdl_Write_Scalar : O_Dnode;

   Ghdl_Read_Scalar : O_Dnode;

   Ghdl_Text_Read_Length : O_Dnode;

   Ghdl_Text_File_Close : O_Dnode;
   Ghdl_File_Close : O_Dnode;

   Ghdl_File_Endfile : O_Dnode;

   --  Hierarchy declaration.
   Ghdl_Name_Kind_Type : O_Tnode;
   Ghdl_Name_Library : O_Enode;
   Ghdl_Name_Entity : O_Enode;
   Ghdl_Name_Architecture : O_Enode;
   Ghdl_Name_Block : O_Enode;
   Ghdl_Name_Generate_Iterative : O_Enode;
   Ghdl_Name_Generate_Conditional : O_Enode;
   Ghdl_Name_Component : O_Enode;
   Ghdl_Name_Package : O_Enode;
   Ghdl_Name_Process : O_Enode;
   Ghdl_Name_Configuration : O_Enode;
   Ghdl_Name_Signal : O_Enode;
   Ghdl_Name_Top : O_Enode;

   Ghdl_Instance_Name_Acc : O_Tnode;
   Ghdl_Instance_Name_Type : O_Tnode;
   Ghdl_Instance_Name_Kind : O_Fnode;
   Ghdl_Instance_Name_Name : O_Fnode;
   Ghdl_Instance_Name_Parent : O_Fnode;
   Ghdl_Instance_Name_Brother : O_Fnode;
   Ghdl_Instance_Name_Child : O_Fnode;

   Ghdl_Hierarchy_Top : O_Lnode;

   type Object_Kind_Type is (Mode_Value, Mode_Signal);

   --  Well known identifiers.
   type Wk_Ident_Type is
     (
      Wkie_This, Wkie_Size, Wkie_Res, Wkie_E, Wkie_Dir_To, Wkie_Dir_Downto,
      Wkie_Left, Wkie_Right, Wkie_Dir, Wkie_Length, Wkie_Kind, Wkie_Dim,
      Wkie_I, Wkie_Instance, Wkie_Arch_Instance, Wkie_Name
      );
   type Wk_Ident_Tree_Array is array (Wk_Ident_Type) of O_Ident;
   Wk_Idents : Wk_Ident_Tree_Array;
   Wki_This : O_Ident renames Wk_Idents (Wkie_This);
   Wki_Size : O_Ident renames Wk_Idents (Wkie_Size);
   Wki_Res : O_Ident renames Wk_Idents (Wkie_Res);
   Wki_E : O_Ident renames Wk_Idents (Wkie_E);
   Wki_Dir_To : O_Ident renames Wk_Idents (Wkie_Dir_To);
   Wki_Dir_Downto : O_Ident renames Wk_Idents (Wkie_Dir_Downto);
   Wki_Left : O_Ident renames Wk_Idents (Wkie_Left);
   Wki_Right : O_Ident renames Wk_Idents (Wkie_Right);
   Wki_Dir : O_Ident renames Wk_Idents (Wkie_Dir);
   Wki_Length : O_Ident renames Wk_Idents (Wkie_Length);
   Wki_Kind : O_Ident renames Wk_Idents (Wkie_Kind);
   Wki_Dim : O_Ident renames Wk_Idents (Wkie_Dim);
   Wki_I : O_Ident renames Wk_Idents (Wkie_I);
   Wki_Instance : O_Ident renames Wk_Idents (Wkie_Instance);
   Wki_Arch_Instance : O_Ident renames Wk_Idents (Wkie_Arch_Instance);
   Wki_Name : O_Ident renames Wk_Idents (Wkie_Name);

   --  ALLOCATION_KIND defines the type of memory storage.
   --  ALLOC_STACK means the object is allocated on the local stack and
   --    deallocated at the end of the function.
   --  ALLOC_MEMORY for object created during design elaboration and whose
   --    life is infinite.
   --  ALLOC_RETURN for unconstrained object returns by function.
   --  ALLOC_HEAP for object created by new.
   --  ALLOC_SYSTEM for system objects.
   type Allocation_Kind is
     (Alloc_Stack, Alloc_Memory, Alloc_Return, Alloc_Heap, Alloc_System);

   package Chap1 is
      --  Declare types for block BLK
      procedure Start_Block_Decl (Blk : Iir);

      procedure Translate_Entity_Declaration (Entity : Iir);

      --  Generate code to initialize generics of instance INSTANCE of ENTITY
      --  using the default values.
      --  This is used when ENTITY is at the top of a design hierarchy.
      procedure Translate_Entity_Init (Entity : Iir; Instance : O_Lnode);

      procedure Translate_Architecture_Declaration (Arch : Iir);
      procedure Translate_Configuration_Declaration (Config : Iir);
   end Chap1;

   package Chap2 is
      --  Subprogram specification being currently translated.  This is used
      --  for the return statement.
      Current_Subprogram : Iir := Null_Iir;

      procedure Translate_Subprogram_Specification (Spec : Iir; Block : Iir);
      procedure Translate_Subprogram_Body (Subprg : Iir);

      procedure Translate_Package_Declaration (Decl : Iir_Package_Declaration);
      procedure Translate_Package_Body (Decl : Iir_Package_Body);

      procedure Elab_Package_Body (Spec : Iir_Package_Declaration; Bod : Iir);

      procedure Elab_Dependence (Design_Unit: Iir_Design_Unit);

      --  Information for instance of a subprogram.
      type Instance_Data is limited private;

      --  Reset instance data DATA.
      procedure Clear_Instance_Data (Data : out Instance_Data);

      --  Add an instance parameter to subprogram SUBPRG for BLOCK.
      procedure Create_Subprg_Instance (Interfaces : in out O_Inter_List;
                                        Subprg : Iir;
                                        Block : Iir);
      procedure Create_Subprg_Instance (Interfaces : in out O_Inter_List;
                                        Data : out Instance_Data;
                                        Block : Iir);

      --  Add an association parameter for the instance of SUBPRG.
      procedure Add_Subprg_Instance_Assoc
        (Assoc : in out O_Assoc_List; Subprg : Iir);
      procedure Add_Subprg_Instance_Assoc
        (Assoc : in out O_Assoc_List; Data : Instance_Data);

      --  Start and finish the use of SUBPRG instance.
      --  If SUBPRG has no instance, then this is a no-op, otherwise scope
      --  is added and removed.
      procedure Start_Subprg_Instance_Use (Subprg : Iir);
      procedure Finish_Subprg_Instance_Use (Subprg : Iir);
      procedure Start_Subprg_Instance_Use (Data : Instance_Data);
      procedure Finish_Subprg_Instance_Use (Data : Instance_Data);

   private
      type Instance_Data is record
         --  Instance parameter.
         Instance_Param : O_Lnode;
         --  Block pointed by the instance parameter.
         Instance_Block : Iir;
      end record;
   end Chap2;

   package Chap5 is
      --  Attribute specification.
      procedure Translate_Attribute_Specification
        (Spec : Iir_Attribute_Specification);
      procedure Elab_Attribute_Specification
        (Spec : Iir_Attribute_Specification);

      --  Disconnection specification.
      procedure Elab_Disconnection_Specification
        (Spec : Iir_Disconnection_Specification);

      --  Elaborate a genrics/ports map aspect list.
      procedure Elab_Unconstrained_Port (Port : Iir; Actual : Iir);
      procedure Elab_Unconstrained_Ports_Of_Port_List (List : Iir_List);
      procedure Elab_Generic_Map_Aspect_List (List : Iir_List);
      procedure Elab_Port_Map_Aspect_List (List : Iir_List);
   end Chap5;


   package Chap8 is
      procedure Translate_Statements_List (List : Iir_List);

      --  Create a case branch for CHOICE.
      --  Used by case statement and aggregates.
      procedure Translate_Case_Choice
        (Choice : Iir; Choice_Type : Iir; Blk : in out O_Case_Block);

      --  Inc or dec by VAL ITERATOR according to DIR.
      --  Used for loop statements.
      procedure Gen_Update_Iterator (Iterator : O_Lnode;
                                     Dir : Iir_Direction;
                                     Val : Unsigned_64;
                                     Itype : Iir);
   end Chap8;

   package Chap9 is
      procedure Translate_Block_Declarations (Block : Iir; Origin : Iir);
      procedure Translate_Block_Subprograms (Block : Iir; Base_Block : Iir);
      procedure Elab_Block_Declarations (Block : Iir; Base_Block : Iir);

      --  Generate code to instantiate an entity.
      --  ASPECT must be an entity_aspect.
      --  MAPPING must be a node with get_port/generic_map_aspect_list.
      --  PARENT is the block in which the instantiation is done.
      --  CONFIG_OVERRIDE, if set, is the configuration to use; if not set, the
      --    configuration to use is determined from ASPECT.
      procedure Translate_Entity_Instantiation
        (Aspect : Iir; Mapping : Iir; Parent : Iir; Config_Override : Iir);

   end Chap9;

   package Chap10 is
      --  There are three data storage kind: global, local or instance.
      --  For example, a constant can have:
      --  * a global storage when declared inside a package.  This storage
      --    can be accessed from any point.
      --  * a local storage when declared in a subprogram.  This storage
      --    can be accessed from the subprogram, is created when the subprogram
      --    is called and destroy when the subprogram exit.
      --  * an instance storage when declared inside a process.  This storage
      --    can be accessed from the process via an instance pointer, is
      --    created during elaboration.
      --procedure Push_Global_Factory (Storage : O_Storage);
      --procedure Pop_Global_Factory;
      procedure Set_Global_Storage (Storage : O_Storage);

      --  Set the global scope handling.
      Global_Storage : O_Storage;

      --  Start to build an instance.
      --  If INSTANCE_TYPE is not O_TNODE_NULL, it must be an uncompleted
      --  record type, that will be completed.
      procedure Push_Instance_Factory (Instance_Type : O_Tnode);
      --  Manually add a field to the current instance being built.
      function Add_Instance_Factory_Field (Name : O_Ident; Ftype : O_Tnode)
        return O_Fnode;
      --  Finish the building of the current instance and return the type
      --  built.
      procedure Pop_Instance_Factory (Instance_Type : out O_Tnode);

      --  Create a new scope, in which variable are created locally
      --  (ie, on the stack).  Always created unlocked.
      procedure Push_Local_Factory;

      --  Destroy a local scope.
      procedure Pop_Local_Factory;

      --  Push_scope defines how to access to a variable stored in an instance.
      --  Variables defined in SCOPE_TYPE can be accessed via field SCOPE_FIELD
      --  in scope SCOPE_PARENT.
      procedure Push_Scope (Scope_Type : O_Tnode;
                            Scope_Field : O_Fnode; Scope_Parent : O_Tnode);
      --  Variables defined in SCOPE_TYPE can be accessed by dereferencing
      --  fiel SCOPE_FIELD defined in SCOPE_PARENT.
      procedure Push_Scope_Via_Field_Ptr
        (Scope_Type : O_Tnode;
         Scope_Field : O_Fnode; Scope_Parent : O_Tnode);
      --  Variables/scopes defined in SCOPE_TYPE can be accessed via
      --  dereference of parameter SCOPE_PARAM.
      procedure Push_Scope (Scope_Type : O_Tnode; Scope_Param : O_Lnode);
      --  No more accesses to SCOPE_TYPE are allowed.
      --  Scopes must be poped in the reverse order they are pushed.
      procedure Pop_Scope (Scope_Type : O_Tnode);

      --  Reset the identifier.
      type Id_Mark_Type is limited private;
      type Local_Identifier_Type is limited private;

      procedure Reset_Identifier_Prefix;
      procedure Push_Identifier_Prefix (Mark : out Id_Mark_Type;
                                        Name : String;
                                        Val : Natural := 0);
      procedure Push_Identifier_Prefix (Mark : out Id_Mark_Type;
                                        Name : Name_Id;
                                        Val : Natural := 0);
      procedure Push_Identifier_Prefix_Uniq (Mark : out Id_Mark_Type);
      procedure Pop_Identifier_Prefix (Mark : in Id_Mark_Type);

      --  Save/restore the local identifier number; this is used by package
      --  body, which has the same prefix as the package declaration, so it
      --  must continue local identifiers numbers.
      --  This is used by subprogram bodies too.
      procedure Save_Local_Identifier (Id : out Local_Identifier_Type);
      procedure Restore_Local_Identifier (Id : Local_Identifier_Type);

      --  Create an identifier from IIR node ID without the prefix.
      function Create_Identifier_Without_Prefix (Id : Iir)
        return O_Ident;
      function Create_Identifier_Without_Prefix (Id : Iir; Str : String)
        return O_Ident;

      --  Create an identifier from the current prefix.
      function Create_Identifier return O_Ident;

      --  Create an identifier from IIR node ID with prefix.
      function Create_Identifier (Id : Iir; Str : String := "")
        return O_Ident;
      function Create_Identifier (Id : Iir; Val : Natural; Str : String := "")
        return O_Ident;
      function Create_Identifier (Id : Name_Id; Str : String := "")
        return O_Ident;
      --  Create a prefixed identifier from a string.
      function Create_Identifier (Str : String) return O_Ident;

      --  Create an identifier for a variable.
      --  IE, if the variable is global, prepend the prefix,
      --   if the variable belong to an instance, no prefix is added.
      type Var_Ident_Type is private;
      function Create_Var_Identifier (Id : Iir; Str : String)
        return Var_Ident_Type;
      function Create_Var_Identifier (Id : Name_Id; Str : String)
        return Var_Ident_Type;
      function Create_Var_Identifier (Id : Iir)
        return Var_Ident_Type;
      function Create_Var_Identifier (Id : String)
        return Var_Ident_Type;
      function Create_Uniq_Identifier return Var_Ident_Type;

      type Var_Type (<>) is limited private;
      type Var_Acc is access Var_Type;

      --  Create a variable in the current scope.
      --  If the current scope is the global scope, then a variable is
      --   created at the top level (using decl_global_storage).
      --  If the current scope is not the global scope, then a field is added
      --   to the current scope.
      function Create_Var
        (Name : Var_Ident_Type;
         Vtype : O_Tnode;
         Storage : O_Storage := Global_Storage)
        return Var_Acc;

      --  Create a global variable.
      function Create_Global_Var
        (Name : O_Ident; Vtype : O_Tnode; Storage : O_Storage)
        return Var_Acc;

      --  Create a global constant and initialize it to INITIAL_VALUE.
      function Create_Global_Const
        (Name : O_Ident;
         Vtype : O_Tnode;
         Storage : O_Storage;
         Initial_Value : O_Enode)
      return Var_Acc;

      --  Return the (real) reference to a variable created by Create_Var.
      function Get_Var (Var : Var_Acc) return O_Lnode;

      --  Return a reference to the instance of type ITYPE.
      function Get_Instance_Ref (Itype : O_Tnode) return O_Lnode;

      --  Return the address of the instance for block BLOCK.
      function Get_Instance_Access (Block : Iir) return O_Enode;

      --  Return the storage for the variable VAR.
      function Get_Alloc_Kind_For_Var (Var : Var_Acc) return Allocation_Kind;

      --  Return TRUE iff VAR is stable, ie get_var (VAR) can be referenced
      --  several times.
      function Is_Var_Stable (Var : Var_Acc) return Boolean;
   private
      type Local_Identifier_Type is new Natural;
      type Id_Mark_Type is record
         Len : Natural;
         Local_Id : Local_Identifier_Type;
      end record;

      type Var_Ident_Type is record
         Id : O_Ident;
      end record;

      --  Kind of variable:
      --  VAR_GLOBAL: the variable is a global variable (static or not).
      --  VAR_LOCAL: the variable is on the stack.
      --  VAR_SCOPE: the variable is in the instance record.
      type Var_Kind is (Var_Global, Var_Scope, Var_Local);

      --  An instance contains all the data (variable, signals, constant...)
      --  which are declared by an entity and an architecture.
      --  (An architecture inherits the data of its entity).
      --
      --  The processes and implicit guard signals of an entity/architecture
      --  are translated into functions.  The first argument of these functions
      --  is a pointer to the instance.

      type Inst_Build_Kind_Type is (Local, Global, Instance);
      type Inst_Build_Type (Kind : Inst_Build_Kind_Type);
      type Inst_Build_Acc is access Inst_Build_Type;
      type Inst_Build_Type (Kind : Inst_Build_Kind_Type) is record
         Prev : Inst_Build_Acc;
         Prev_Id_Start : Natural;
         case Kind is
            when Local =>
               --  Previous global storage.
               Prev_Global_Storage : O_Storage;
            when Global =>
               null;
            when Instance =>
               Elements : O_Element_List;
               Vars : Var_Acc;
         end case;
      end record;

      type Var_Type (Kind : Var_Kind) is record
         case Kind is
            when Var_Global
              | Var_Local =>
               E : O_Lnode;
            when Var_Scope =>
               I_Field : O_Fnode;
               I_Type : O_Tnode;
               I_Link : Var_Acc;
         end case;
      end record;
   end Chap10;
   use Chap10;

   package Chap14 is
      function Translate_Length_Array_Attribute (Expr : Iir; Rtype : Iir)
        return O_Enode;
      function Translate_Low_Array_Attribute (Expr : Iir) return O_Enode;
      function Translate_High_Array_Attribute (Expr : Iir) return O_Enode;
      function Translate_Range_Array_Attribute (Expr : Iir) return O_Lnode;
      function Translate_Right_Array_Attribute (Expr : Iir) return O_Enode;
      function Translate_Left_Array_Attribute (Expr : Iir) return O_Enode;

      --  Return the value of the left bound/right bound/direction of scalar
      --  type ATYPE.
      function Translate_Left_Type_Attribute (Atype : Iir) return O_Enode;
      function Translate_Right_Type_Attribute (Atype : Iir) return O_Enode;
      function Translate_Dir_Type_Attribute (Atype : Iir) return O_Enode;

      function Translate_High_Type_Attribute (Atype : Iir) return O_Enode;
      function Translate_Low_Type_Attribute (Atype : Iir) return O_Enode;

      function Translate_Val_Attribute (Attr : Iir) return O_Enode;
      function Translate_Pos_Attribute (Attr : Iir; Res_Type : Iir)
                                       return O_Enode;

      function Translate_Succ_Pred_Attribute (Attr : Iir) return O_Enode;

      function Translate_Event_Attribute (Attr : Iir) return O_Enode;
      function Translate_Active_Attribute (Attr : Iir) return O_Enode;
      function Translate_Last_Value_Attribute (Attr : Iir) return O_Enode;

      function Translate_Last_Time_Attribute (Prefix : Iir; Field : O_Fnode)
        return O_Enode;
   end Chap14;

   type Ortho_Info_Kind is
     (
      Kind_Type,
      Kind_Expr,
      Kind_Subprg,
      Kind_Object,
      Kind_Alias,
      Kind_Iterator,
      Kind_Interface,
      Kind_Disconnect,
      Kind_Process,
      Kind_Loop,
      Kind_Block,
      Kind_Component,
      Kind_Field,
      Kind_Package,
      Kind_Config
      );

   type O_Fnode_Arr is array (Natural range <>) of O_Fnode;
   type O_Fnode_Arr_Acc is access O_Fnode_Arr;
   type Ortho_Info_Type_Kind is
     (
      Kind_Type_Scalar,
      Kind_Type_Array,
      Kind_Type_Record
      );
   type O_Tnode_Array is array (Object_Kind_Type) of O_Tnode;
   type O_Fnode_Array is array (Object_Kind_Type) of O_Fnode;
   type O_Lnode_Array is array (Object_Kind_Type) of O_Lnode;
   type O_Dnode_Array is array (Object_Kind_Type) of O_Dnode;
   type Var_Acc_Array is array (Object_Kind_Type) of Var_Acc;
   type Instance_Data_Array is array (Object_Kind_Type) of Chap2.Instance_Data;

   type Ortho_Info_Type_Type (Kind : Ortho_Info_Type_Kind := Kind_Type_Scalar)
   is record
      case Kind is
         when Kind_Type_Scalar =>
            --  For scalar types:
            --  Ortho type for the range record type.
            Range_Type : O_Tnode;

            --  Ortho type for an access to the range record type.
            Range_Ptr_Type : O_Tnode;

            --  Tree for the range record declaration.
            Range_Var : Var_Acc;

            --  Fields of TYPE_RANGE_TYPE.
            Range_Left : O_Fnode;
            Range_Right : O_Fnode;
            Range_Dir : O_Fnode;
            Range_Length : O_Fnode;

         when Kind_Type_Array =>
            Base_Type : O_Tnode_Array;
            Base_Ptr_Type : O_Tnode_Array;
            Bounds_Type : O_Tnode;
            Bounds_Ptr_Type : O_Tnode;

            Base_Field : O_Fnode_Array;
            Bounds_Field : O_Fnode_Array;

            --  Field declaration for each dimension (1 based).
            Bounds_Vector : O_Fnode_Arr_Acc;

            --  True if the array bounds are static.
            Static_Bounds : Boolean;

            --  Variable containing the bounds for a constrained array.
            Array_Bounds : Var_Acc;

            --  Variable containing a 1 length bound for unidimensional
            --  unconstrained arrays.
            Array_1bound : Var_Acc;

            --  Variable containing the description for each index.
            Array_Index_Desc : Var_Acc;

         when Kind_Type_Record =>
            --  Variable containing the description for each element.
            Record_El_Desc : Var_Acc;
      end case;
   end record;

   Ortho_Info_Type_Scalar_Init : constant Ortho_Info_Type_Type :=
     (Kind => Kind_Type_Scalar,
      Range_Type => O_Tnode_Null,
      Range_Ptr_Type => O_Tnode_Null,
      Range_Var => null,
      Range_Left => O_Fnode_Null,
      Range_Right => O_Fnode_Null,
      Range_Dir => O_Fnode_Null,
      Range_Length => O_Fnode_Null);

   Ortho_Info_Type_Array_Init : constant Ortho_Info_Type_Type :=
     (Kind => Kind_Type_Array,
      Base_Type => (O_Tnode_Null, O_Tnode_Null),
      Base_Ptr_Type => (O_Tnode_Null, O_Tnode_Null),
      Bounds_Type => O_Tnode_Null,
      Bounds_Ptr_Type => O_Tnode_Null,
      Base_Field => (O_Fnode_Null, O_Fnode_Null),
      Bounds_Field => (O_Fnode_Null, O_Fnode_Null),
      Bounds_Vector => null,
      Static_Bounds => False,
      Array_Bounds => null,
      Array_1bound => null,
      Array_Index_Desc => null);

   Ortho_Info_Type_Record_Init : constant Ortho_Info_Type_Type :=
     (Kind => Kind_Type_Record,
      Record_El_Desc => null);

   --  Mode of the type; roughly speaking, this corresponds to its size
   --  (for scalars) or its layout (for composite types).
   --  Used to select library subprograms for signals.
   type Type_Mode_Type is
     (
      --  Unknown mode.
      Type_Mode_Unknown,
      --  Boolean type, with 2 elements.
      Type_Mode_B2,
      --  Enumeration with at most 256 elements.
      Type_Mode_E8,
      --  Enumeration with more than 256 elements.
      Type_Mode_E32,
      --  Integer type.
      Type_Mode_I32,
      --  Physical type.
      Type_Mode_P64,
      --  Floating point type.
      Type_Mode_F64,
      --  File type.
      Type_Mode_File,
      --  Thin access.
      Type_Mode_Acc,

      --  Fat access.
      Type_Mode_Fat_Acc,

      --  Record.
      Type_Mode_Record,
      --  Constrained array type (length is known at compile-time).
      Type_Mode_Array,
      --  Array pointer type (used for constrained array whose length is
      --  known at run-time).
      Type_Mode_Ptr_Array,
      --  Fat array type (used for unconstrained array).
      Type_Mode_Fat_Array);

   --  Scalar types.
   subtype Type_Mode_Scalar is Type_Mode_Type
     range Type_Mode_B2 .. Type_Mode_F64;

   subtype Type_Mode_Non_Composite is Type_Mode_Type
     range Type_Mode_B2 .. Type_Mode_Fat_Acc;

   --  Composite types, with the vhdl meaning: record and arrays.
   subtype Type_Mode_Composite is Type_Mode_Type
     range Type_Mode_Record .. Type_Mode_Fat_Array;

   --  Array types.
   subtype Type_Mode_Arrays is Type_Mode_Type range
     Type_Mode_Array .. Type_Mode_Fat_Array;

   --  Thin types, ie types whose length is a scalar.
   subtype Type_Mode_Thin is Type_Mode_Type
     range Type_Mode_B2 .. Type_Mode_Acc;

   --  Fat types, ie types whose length is longer than a scalar.
   subtype Type_Mode_Fat is Type_Mode_Type
     range Type_Mode_Fat_Acc .. Type_Mode_Fat_Array;

   --  These parameters are passed by value, ie the argument of the subprogram
   --  is the value of the object.
   subtype Type_Mode_By_Value is Type_Mode_Type
     range Type_Mode_B2 .. Type_Mode_Acc;

   --  These parameters are passed by copy, ie a copy of the object is created
   --  and the reference of the copy is passed.  If the object will not be
   --  modified by the subprogram, the object can be passed by reference.
   subtype Type_Mode_By_Copy is Type_Mode_Type
     range Type_Mode_Fat_Acc .. Type_Mode_Fat_Acc;

   --  The parameters are passed by reference, ie the argument of the
   --  subprogram is an address to the object.
   subtype Type_Mode_By_Ref is Type_Mode_Type
     range Type_Mode_Record .. Type_Mode_Fat_Array;

   --  Additional informations for a resolving function.
   type Subprg_Resolv_Info is record
      Resolv_Func : O_Dnode;
      --  Base block which the function was defined in.
      Resolv_Block : Iir;
      --  Parameter nodes.
      Var_Res : O_Lnode;
      Var_Length : O_Lnode;
      Var_Vals : O_Lnode;
      Var_Instance : O_Lnode;
   end record;
   type Subprg_Resolv_Info_Acc is access Subprg_Resolv_Info;

   --  Additional info for complex types.
   type Complex_Type_Info is record
      --  Variable containing the size of the type.
      --  This is defined only for types whose size is only known at
      --  running time (and not a compile-time).
      Size_Var : Var_Acc_Array;

      Builder_Need_Func : Boolean;

      --  Parameters for type builders.
      --  NOTE: this is only set for types (and *not* for subtypes).
      Builder_Instance : Instance_Data_Array;
      Builder_Base_Param : O_Lnode_Array;
      Builder_Bound_Param : O_Lnode_Array;
      Builder_Func : O_Dnode_Array;
   end record;
   type Complex_Type_Info_Acc is access Complex_Type_Info;

   type Ortho_Info_Type (Kind : Ortho_Info_Kind) is record
      case Kind is
         when Kind_Type =>
            --  Mode of the type.
            Type_Mode : Type_Mode_Type := Type_Mode_Unknown;

            --  Additionnal info for complex types.
            C : Complex_Type_Info_Acc := null;

            --  Ortho node which represents the type.
            Ortho_Type : O_Tnode_Array;
            --  Ortho pointer to the type.
            Ortho_Ptr_Type : O_Tnode_Array;
            --  Type of type desc.
            Type_Desc_Type : O_Tnode;
            --  Tree for the descriptor declaration.
            Type_Desc : Var_Acc;

            --  True if type_desc is static (ie, values are known at compile
            --  time).  This is different from type_staticness, since
            --  array type are never static.
            Type_Desc_Static : Boolean := False;

            --  More info according to the type.
            T : Ortho_Info_Type_Type;
         when Kind_Expr =>
            --  Ortho tree which represents the expression, used for
            --  enumeration literals.
            Expr_Node : O_Enode;
         when Kind_Subprg =>
            --  Subprogram declaration node.
            Ortho_Func : O_Dnode;

            --  True if the function can return a value stored in the secondary
            --  stack.  In this case, the caller must deallocate the area
            --  allocated by the callee when the value was used.
            Use_Stack2 : Boolean := False;

            --  For a function:
            --    If the return value is not composite, then this field
            --      must be O_LNODE_NULL.
            --    If the return value is a composite type, then the caller must
            --    give to the callee an area to put the result.  This area is
            --    given via an (hidden to the user) interface.  Furthermore,
            --    the function is translated into a procedure.
            --  For a procedure:
            --    If there are copy-out interfaces, they are gathered in a
            --    record and a pointer to the record is passed to the
            --    procedure.  RES_INTERFACE is the interface for this pointer.
            Res_Interface : O_Lnode := O_Lnode_Null;

            --  For a procedure with a result interface:
            --    Type definition for the record.
            Res_Record_Type : O_Tnode := O_Tnode_Null;
            --    Type definition for access to the record.
            Res_Record_Ptr : O_Tnode := O_Tnode_Null;

            --  Instance parameter or O_Lnode_null if the subprogram is global.
            Subprg_Instance : Chap2.Instance_Data;

            --  Overload count, to be able to define differents name for
            --  overloaded subprograms.
            --  0 means not overloaded.
            Subprg_Overload : Natural;

            Subprg_Resolv : Subprg_Resolv_Info_Acc := null;

            --  Local identifier number, set by spec, continued by body.
            Subprg_Local_Id : Local_Identifier_Type;
         when Kind_Object =>
            --  The object itself.
            Object_Var : Var_Acc;
            --  Name of the object (used for signals).
            Object_Name : Var_Acc;
            --  Function to compute the value of object (used for implicit
            --   guard signal declaration).
            Object_Function : O_Dnode;
            --  For unconstrained signals: subtype description.
            Object_Type_Desc_Var : Var_Acc;
         when Kind_Alias =>
            Alias_Var : Var_Acc;
            Alias_Kind : Object_Kind_Type;
         when Kind_Iterator =>
            Iterator_Var : Var_Acc;
         when Kind_Interface =>
            --  Ortho node for the interface.
            Interface_Node : O_Lnode;
            --  Field of the result record for copy-out arguments of procedure.
            Interface_Field : O_Fnode;
         when Kind_Disconnect =>
            --  Variable which contains the time_expression of the
            --  disconnection specification
            Disconnect_Var : Var_Acc;
         when Kind_Process =>
            --  Type of process declarations.
            Process_Decls_Type : O_Tnode;

            --  Field in the parent block for the declarations in the process.
            Process_Parent_Field : O_Fnode;

            --  Subprogram for the process.
            Process_Subprg : O_Dnode;

            --  Field for instance name.
            Process_Name : O_Fnode;

         when Kind_Loop =>
            --  Labels for the loop.
            --  Used for exit/next from while-loop, and to exit from for-loop.
            Label_Exit : O_Snode;
            --  Used to next from for-loop, with an exit statment.
            Label_Next : O_Snode;
         when Kind_Block =>
            --  Instance type (ortho record) for declarations contained in the
            --  block/entity/architecture.
            Block_Decls_Type : O_Tnode;
            Block_Decls_Ptr_Type : O_Tnode;

            --  Field in the instance type containing the name of the block.
            Block_Name : O_Fnode;

            --  For an entity: must be o_fnode_null.
            --  For an architecture: the entity field.
            --  For a block, a component or a generate block: field in the
            --    parent instance which contains the declarations for this
            --    block.
            Block_Parent_Field : O_Fnode;

            --  For a generate block: field in the block providing a chain to
            --  the previous block (note: this may not be the parent, but
            --  is a parent).
            Block_Origin_Field : O_Fnode;
            --  For an iterative block: range field of the iterator in the
            --  parent block.
            Block_Range_Field : O_Fnode;
            --  For an iterative block: boolean field set when the block
            --  is configured.  This is used to check if the block was already
            --  configured since index and slice are not compelled to be
            --  locally static.
            Block_Configured_Field : O_Fnode;

            --  Address of the iterator type descriptor.  This field is just
            --  after the name.
            Block_Iterator_TD_Field : O_Fnode;

            --  Address of the iterator.  This field is just after the type
            --  descriptor
            Block_Iterator_Addr_Field : O_Fnode;

            --  For iterative generate block: array of instances.
            Block_Decls_Array_Type : O_Tnode;
            Block_Decls_Array_Ptr_Type : O_Tnode;

            --  Ortho type/field DECL for the declarations instance.
            --  For an entity or an architecture, this is a pointer to the
            --   instance.
            --  For a block, this is a field of the entity/arch instance.
            Block_Decls : O_Enode;

            --  Subprogram which elaborates the block (for entity or arch).
            Block_Elab_Subprg : O_Dnode;
            --  Size of the block instance.
            Block_Instance_Size : O_Lnode;
         when Kind_Component =>
            --  Instance for the component.
            Comp_Type : O_Tnode;
            Comp_Ptr_Type : O_Tnode;
            --  Name field in the instance.
            Comp_Name : O_Fnode;
         when Kind_Config =>
            Config_Subprg : O_Dnode;
         when Kind_Field =>
            --  Node for a record element declaration.
            Field_Node : O_Fnode_Array;
         when Kind_Package =>
            --  Subprogram which elaborate the package spec/body.
            --  External units should call the body elaborator.
            --  The spec elaborator is called only from the body elaborator.
            Package_Elab_Spec_Subprg : O_Dnode;
            Package_Elab_Body_Subprg : O_Dnode;

            --  Variable set to true when the package is elaborated.
            Package_Elab_Var : O_Lnode;

            --  Package name.
            Package_Name : O_Lnode;

            --  Local id, set by package declaration, continued by package
            --  body.
            Package_Local_Id : Local_Identifier_Type;
      end case;
   end record;

   package Addr_To_Info_Cv is new System.Address_To_Access_Conversions
     (Object => Ortho_Info_Type);
   use Addr_To_Info_Cv;
   subtype Ortho_Info_Acc is Addr_To_Info_Cv.Object_Pointer;

   procedure Free is new Ada.Unchecked_Deallocation
     (Name => Ortho_Info_Acc, Object => Ortho_Info_Type);

   subtype Loop_Info_Acc is Ortho_Info_Acc (Kind_Loop);
   subtype Block_Info_Acc is Ortho_Info_Acc (Kind_Block);
   subtype Comp_Info_Acc is Ortho_Info_Acc (Kind_Component);
   subtype Proc_Info_Acc is Ortho_Info_Acc (Kind_Process);
   subtype Type_Info_Acc is Ortho_Info_Acc (Kind_Type);
   subtype Subprg_Info_Acc is Ortho_Info_Acc (Kind_Subprg);
   subtype Object_Info_Acc is Ortho_Info_Acc (Kind_Object);
   subtype Alias_Info_Acc is Ortho_Info_Acc (Kind_Alias);
   subtype Config_Info_Acc is Ortho_Info_Acc (Kind_Config);
   subtype Field_Info_Acc is Ortho_Info_Acc (Kind_Field);

   procedure Set_Info (Target : Iir; Info : Ortho_Info_Acc)
   is
      use System;
   begin
      if Get_Back_End_Info (Target) /= Null_Address then
         raise Internal_Error;
      end if;
      Set_Back_End_Info (Target, Addr_To_Info_Cv.To_Address (Info));
   end Set_Info;

   function Get_Info (Target : Iir) return Ortho_Info_Acc is
   begin
      return Addr_To_Info_Cv.To_Pointer (Get_Back_End_Info (Target));
   end Get_Info;

   --  Create an ortho_info field of kind KIND for iir node TARGET, and
   --  return it.
   function Add_Info (Target : Iir; Kind : Ortho_Info_Kind)
     return Ortho_Info_Acc
   is
      Res : Ortho_Info_Acc;
   begin
      Res := new Ortho_Info_Type (Kind);
      Set_Info (Target, Res);
      return Res;
   end Add_Info;

   procedure Free_Info (Target : Iir)
   is
      Info : Ortho_Info_Acc;
   begin
      Info := Get_Info (Target);
      Free (Info);
      Set_Back_End_Info (Target, System.Null_Address);
   end Free_Info;

   procedure Free_Type_Info (Info : in out Type_Info_Acc) is
   begin
      --  FIXME: free C field.
      Free (Info);
   end Free_Type_Info;

   procedure Set_Ortho_Expr (Target : Iir; Expr : O_Enode)
   is
      use Addr_To_Info_Cv;
      Info : Ortho_Info_Acc;
   begin
      Info := Add_Info (Target, Kind_Expr);
      Info.Expr_Node := Expr;
   end Set_Ortho_Expr;

   function Get_Ortho_Expr (Target : Iir) return O_Enode is
   begin
      return Get_Info (Target).Expr_Node;
   end Get_Ortho_Expr;

   function Get_Ortho_Type (Target : Iir; Is_Sig : Object_Kind_Type)
     return O_Tnode is
   begin
      return Get_Info (Target).Ortho_Type (Is_Sig);
   end Get_Ortho_Type;

   --  Return true is INFO is a type info for a composite type, ie:
   --  * a record
   --  * an array (fat or thin)
   --  * a fat pointer.
   function Is_Composite (Info : Type_Info_Acc) return Boolean;
   pragma Inline (Is_Composite);

   function Is_Composite (Info : Type_Info_Acc) return Boolean is
   begin
      return Info.Type_Mode in Type_Mode_Fat;
   end Is_Composite;

   --  Convert an o_lnode to an o_enode, either by taking value or address.
   function L2e_Node (L : O_Lnode;
                      Type_Info : Type_Info_Acc;
                      Kind : Object_Kind_Type)
     return O_Enode is
   begin
      case Type_Info.Type_Mode is
         when Type_Mode_Unknown =>
            raise Internal_Error;
         when Type_Mode_Scalar
           | Type_Mode_Acc
           | Type_Mode_Ptr_Array
           | Type_Mode_File =>
            return New_Value (L);
         when Type_Mode_Fat_Array
           | Type_Mode_Fat_Acc =>
            return New_Address (L, Type_Info.Ortho_Ptr_Type (Kind));
         when Type_Mode_Record
           | Type_Mode_Array =>
            return New_Address (L, Type_Info.Ortho_Ptr_Type (Kind));
      end case;
   end L2e_Node;

   --  Get Lnode from a variable pointer.
   function Ptr2l_Node (Var_Ptr : O_Lnode; Info : Type_Info_Acc) return O_Lnode
   is
   begin
      case Info.Type_Mode is
         when Type_Mode_Fat_Array
           | Type_Mode_Array
           | Type_Mode_Record
           | Type_Mode_Fat_Acc =>
            return New_Access_Element (New_Value (Var_Ptr));
         when Type_Mode_Ptr_Array =>
            return Var_Ptr;
         when others =>
            raise Internal_Error;
      end case;
   end Ptr2l_Node;

   function Get_Bounds_Ptr (Info : Type_Info_Acc) return O_Enode is
   begin
      case Info.Type_Mode is
         when Type_Mode_Array
           | Type_Mode_Ptr_Array =>
            return New_Address (Get_Var (Info.T.Array_Bounds),
                                Info.T.Bounds_Ptr_Type);
         when others =>
            raise Internal_Error;
      end case;
   end Get_Bounds_Ptr;

   --  In order to simplify the handling of Enode/Lnode, let's introduce
   --  Mnode (yes, another node).
   --  An Mnode is a typed union, containing either an Lnode or a Enode.
   --  See Mstate for a description of the union.
   --  The real data is contained insisde a record, so that the discriminant
   --  can be changed.
   type Mnode;

   --  State of an Mmode.
   type Mstate is
     (
      --  The Mnode contains an Enode, which can be either a value or a
      --  pointer.
      --  This Mnode can be used only once.
      Mstate_E,

      --  The Mnode contains an Lnode representing a value.
      --  This Lnode can be used only once.
      Mstate_Lv,

      --  The Mnode contains an Lnode representing a pointer.
      --  This Lnode can be used only once.
      Mstate_Lp,

      --  The Mnode contains an Lnode for a variable representing a value.
      --  This Lnode may be used several times.
      Mstate_Vv,

      --  The Mnode contains an Lnode for a variable representing a pointer.
      --  This Lnode may be used several times.
      Mstate_Vp,

      --  Null Mnode.
      Mstate_Null,

      --  The Mnode is invalid (such as already used).
      Mstate_Bad);

   type Mnode1 (State : Mstate := Mstate_Bad) is record
      --  True if the object is composite (its value cannot be read directly).
      Comp : Boolean;

      --  Additionnal informations about the objects: kind and type.
      K : Object_Kind_Type;
      T : Type_Info_Acc;

      --  Ortho type of the object.
      Vtype : O_Tnode;

      --  Type for a pointer to the object.
      Ptype : O_Tnode;

      case State is
         when Mstate_E =>
            E : O_Enode;
         when Mstate_Lv
           | Mstate_Vv =>
            Lv : O_Lnode;
         when Mstate_Lp
           | Mstate_Vp =>
            Lp : O_Lnode;
         when Mstate_Bad
           | Mstate_Null =>
            null;
      end case;
   end record;
   --pragma Pack (Mnode1);

   type Mnode is record
      M1 : Mnode1;
   end record;

   --  Null Mnode.
   Mnode_Null : constant Mnode := Mnode'(M1 => (State => Mstate_Null,
                                                Comp => False,
                                                K => Mode_Value,
                                                Ptype => O_Tnode_Null,
                                                Vtype => O_Tnode_Null,
                                                T => null));


   --  Object kind of a Mnode
   function Get_Object_Kind (M : Mnode) return Object_Kind_Type;

   --  Transform VAR to Mnode.
   function Get_Var
     (Var : Var_Acc; Vtype : Type_Info_Acc; Mode : Object_Kind_Type)
     return Mnode;

   --  Return a stabilized node for M.
   --  The former M is not usuable anymore.
   function Stabilize (M : Mnode; Can_Copy : Boolean := False) return Mnode;

   --  Stabilize M.
   procedure Stabilize (M : in out Mnode);

   --  Create a temporary of type INFO and kind KIND.
   function Create_Temp (Info : Type_Info_Acc;
                         Kind : Object_Kind_Type := Mode_Value)
                        return Mnode;

   package Chap3 is
      --  Translate the subtype of an object, since an object can define
      --  a subtype.
      --  This can be done only for a declaration.
      --  DECL must have an identifier and a type.
      procedure Translate_Object_Subtype
        (Decl : Iir; With_Vars : Boolean := True);
      procedure Elab_Object_Subtype (Def : Iir);

      --  Translate the subtype of a literal.
      --  This can be done not at declaration time, ie no variables are created
      --  for this subtype.
      --procedure Translate_Literal_Subtype (Def : Iir);

      --  Translation of a type definition:
      --  1. Create corresponding Ortho type.
      --  2. Create bounds type
      --  3. Create bounds declaration
      --  4. Create bounds constructor
      --  5. Create type descriptor declaration
      --  6. Create type descriptor constructor

      --procedure Translate_Type_Definition
      --  (Def : Iir; With_Vars : Boolean := True);

      procedure Translate_Named_Type_Definition (Def : Iir; Id : Name_Id);
      procedure Translate_Anonymous_Type_Definition (Def : Iir);

      --  Translate subprograms for types.
      procedure Translate_Type_Subprograms
        (Decl : Iir_Type_Declaration; Block : Iir);

      procedure Create_Type_Definition_Type_Range (Def : Iir);

      --  Same as Translate_type_definition only for std.standard.boolean and
      --  std.standard.bit.
      procedure Translate_Bool_Type_Definition (Def : Iir);

      --  Translate_type_definition_Elab do 4 and 6.
      --  It generates code to do type elaboration.
      procedure Elab_Type_Declaration (Decl : Iir_Type_Declaration);
      procedure Elab_Subtype_Declaration (Decl : Iir_Subtype_Declaration);

      --  Return the address of the DEF type desc.
      function Get_Type_Desc_Addr (A_Type : Iir) return O_Enode;

      --  Set an array subtype descriptor VAR with BASETYPE and BOUNDS.
      procedure Set_Array_Subtype_Desc
        (Var : Var_Acc; Basetype : Iir; Bounds : O_Lnode);

      --  Builders.
      --  A complex type is a type whose size is not locally static.
      --
      --  The most simple example is an unidimensionnl array whose range
      --  depends on generics.
      --
      --  We call first order complex type any array whose bounds are not
      --  locally static and whose sub-element size is locally static.
      --
      --  First order complex type objects are represented by a pointer to an
      --  array of sub-element, and the storage area for the array is
      --  allocated at run-time.
      --
      --  Since a sub-element type may be a complex type, a type may be
      --  complex because one of its sub-element type is complex.
      --  EG, a record type whose one element is a complex array.
      --
      --  A type may be complex either because it is a first order complex
      --  type (ie an array whose bounds are not locally static) or because
      --  one of its sub-element type is such a type (this is recursive).
      --
      --  We call second order complex type a complex type that is not of first
      --  order.
      --  We call third order complex type a second order complex type which is
      --  an array whose bounds are not locally static.
      --
      --  In a complex type, sub-element of first order complex type are
      --  represented by a pointer.
      --  Any complex type object (constant, signal, variable, port, generic)
      --  is represented by a pointer.
      --
      --  Creation of a second or third order complex type object consists in
      --  allocating the memory and building the object.
      --  Building a object consists in setting internal pointers.
      --
      --  A complex type has always a non-null INFO.C, and its size is computed
      --  during elaboration.
      --
      --  For a second or third order complex type, INFO.C.BUILDER_NEED_FUNC
      --  is set to TRUE.

      --  Call builder for variable pointed VAR of type VAR_TYPE.
      procedure Gen_Call_Type_Builder (Var : Mnode; Var_Type : Iir);

      --  Build variable given by GET_FIELD_LNODE: ie set internals
      --  fields.
      generic
         with function Get_Field_Lnode return O_Lnode;
      procedure Builder_Update_Field
        (Field_Type : Iir; Mem : O_Lnode; Kind : Object_Kind_Type);

      --  Functions for fat array.
      --  Fat array are array whose size is not known at compilation time.
      --  This corresponds to an unconstrained array or a non locally static
      --  constrained array.
      --  A fat array is a structure containing 2 fields:
      --  * base: a pointer to the data of the array.
      --  * bounds: a pointer to a structure containing as many fields as
      --    number of dimensions; these fields are a structure describing the
      --    range of the dimension.

      --  Index array BASE of type ATYPE with INDEX.
      --  INDEX must be of type ghdl_index_type, thus no bounds checks are
      --  performed.
      function Index_Base (Base : Mnode; Atype : Iir; Index : O_Enode)
        return Mnode;

      --  Get the length of the array.
      function Get_Array_Length (Arr : Mnode; Atype : Iir) return O_Enode;

      function Get_Array_Type_Length (Atype : Iir) return O_Enode;

      --  Get the base of array ARR.
      function Get_Array_Base (Arr : Mnode) return Mnode;

      --  Get the bounds of array ARR.
      function Get_Array_Bounds (Arr : Mnode) return Mnode;

      --  Get length of range R.
      function Range_To_Length (R : Mnode) return Mnode;

      --  Get direction of range R.
      function Range_To_Dir (R : Mnode) return Mnode;

      --  Get left/right bounds for range R.
      function Range_To_Left (R : Mnode) return Mnode;
      function Range_To_Right (R : Mnode) return Mnode;

      --  Get range for dimension DIM (1 based) of array bounds B or type
      --  ATYPE.
      function Bounds_To_Range (B : Mnode; Atype : Iir; Dim : Positive)
        return Mnode;

      --  Get the range of dimension DIM (1 based) of array ARR of type ATYPE.
      function Get_Array_Range (Arr : Mnode; Atype : Iir; Dim : Positive)
        return Mnode;

      --  Get array bounds for type ATYPE.
      function Get_Array_Type_Bounds (Atype : Iir) return Mnode;

      --  PTR must be a variable pointing to a bounds of type ATYPE.
      function Get_Bounds_Ptr_Length (Ptr : O_Lnode; Atype : Iir)
        return O_Enode;

      --  Return the a pointer to the array base from variable PTR
      --  containing a pointer to array.
      function Get_Array_Ptr_Base_Ptr
        (Ptr : O_Lnode; Atype : Iir; Is_Sig : Object_Kind_Type)
        return O_Lnode;

      --  Return pointer to range DIM of array pointed by PTR.
      function Get_Array_Ptr_Range_Ptr (Ptr : O_Lnode;
                                        Array_Type : Iir;
                                        Dim : Natural;
                                        Is_Sig : Object_Kind_Type)
        return O_Enode;


      function Get_Array_Bounds_Ptr
        (Arr : O_Lnode; Arr_Type : Iir; Is_Sig : Object_Kind_Type)
        return O_Enode;

      --  Return the bounds field of a fat array from variable PTR containing a
      --  pointer to a fat array.
      function Get_Array_Ptr_Bounds_Ptr
        (Ptr : O_Lnode; Atype : Iir; Is_Sig : Object_Kind_Type)
        return O_Enode;

      --  Performs deallocation of PARAM (the parameter of a deallocate call).
      procedure Translate_Object_Deallocation (Param : Iir);

      --  Allocate an object of type DTYPE.
      --  For an unconstrained array, BOUNDS is a pointer to the boundaries of
      --  the object, which are copied.
      --  Return a pointer to the object or a pointer to the fat pointer.
      function Translate_Object_Allocation
        (Alloc_Kind : Allocation_Kind;
         Obj_Type : Iir;
         Bounds : O_Enode;
         Kind : Object_Kind_Type)
        return Mnode;

      --  Copy SRC to DEST.
      --  Both have the same type, OTYPE.
      --  Furthermore, arrays are of the same length.
      procedure Translate_Object_Copy
        (Dest : Mnode;
         Src : O_Enode;
         Obj_Type : Iir;
         Kind : Object_Kind_Type);

      --  Get size (in bytes with type ghdl_index_type) of object OBJ.
      --  For an unconstrained array, OBJ must be really an object, otherwise,
      --  it may be a null_mnode, created by T2M.
      function Get_Object_Size (Obj : Mnode; Obj_Type : Iir) return O_Enode;

      --  Allocate the base of a fat array, whose length is determined from
      --  the bounds.
      --  RES_PTR is a pointer to the fat pointer (must be a variable that
      --  can be referenced several times).
      --  ARR_TYPE is the type of the array.
      procedure Allocate_Fat_Array_Base (Alloc_Kind : Allocation_Kind;
                                         Res : Mnode;
                                         Arr_Type : Iir;
                                         Is_Sig : Object_Kind_Type);

      --  Create the bounds for SUB_TYPE.
      --  SUB_TYPE is expected to be a non-static, anonymous array type.
      procedure Create_Array_Subtype (Sub_Type : Iir);

      --  Return TRUE if VALUE is not is the range specified by ATYPE.
      function Not_In_Range (Value : O_Lnode; Atype : Iir) return O_Enode;

      --  Return TRUE if base type of ATYPE is larger than its bounds, ie
      --  if a value of type ATYPE may be out of range.
      function Need_Range_Check (Atype : Iir) return Boolean;

      --  Generate an error if VALUE (computed from EXPR which may be NULL_IIR
      --  if not from a tree) is not in range specified by ATYPE.
      procedure Check_Range (Value : O_Lnode; Expr : Iir; Atype : Iir);

      --  Check bounds length of L match bounds length of R.
      --  If L_TYPE (resp. R_TYPE) is not a thin array, then L_NODE
      --    (resp. R_NODE) are not used (and may be o_lnode_null).
      --  If L_TYPE (resp. T_TYPE) is a fat array, then L_NODE (resp. R_NODE)
      --    must be a variable pointing to the array.
      procedure Check_Array_Match (L_Type : Iir;
                                   L_Node : O_Lnode;
                                   L_Mode : Object_Kind_Type;
                                   R_Type : Iir;
                                   R_Node : O_Lnode;
                                   R_Mode : Object_Kind_Type;
                                   Loc : Iir);

      --  Create a subtype range to be stored into the location pointed by
      --  RANGE_PTR from length LENGTH, which is of type INDEX_TYPE.
      --  This is done according to rules 7.2.4 of LRM93, ie:
      --  direction and left bound of the range is the same of INDEX_TYPE.
      --  LENGTH and RANGE_PTR are variables.
      procedure Create_Range_From_Length
        (Index_Type : Iir; Length : O_Lnode; Range_Ptr : O_Lnode);

   end Chap3;

   package Chap4 is
      --  Translate of a type declaration corresponds to the translation of
      --  its definition.
      procedure Translate_Type_Declaration (Decl : Iir_Type_Declaration);
      procedure Translate_Subtype_Declaration (Decl : Iir_Subtype_Declaration);
      procedure Translate_Bool_Type_Declaration (Decl : Iir_Type_Declaration);

      --  Translate declaration DECL, which must not be a subprogram
      --  specification.
      procedure Translate_Declaration (Decl : Iir);

      --  Translate declarations, except subprograms spec and bodies.
      --  For a package body, FIRST_LIST is the declaration list of the package
      --  declaration.
      procedure Translate_Declaration_List
        (Decl_List : Iir_List; First_List : Iir_List);

      --  Translate subprograms in declaration list DECL_LIST.
      --  For a global subprograms belonging to an instance (ie, subprograms
      --  declared in a block, entity or architecture), BLOCK is the info
      --  for the base block to which the subprograms belong; null if none;
      --  It is used to add an instance parameter.
      procedure Translate_Declaration_List_Subprograms
        (Decl_List : Iir_List; Block : Iir);

      --  Create code to elaborate declarations.
      procedure Elab_Declaration_List (Decl_List : Iir_List; Parent : Iir);

      --  Translate port or generic list.
      procedure Translate_Port_List (List : Iir_List);
      procedure Translate_Generic_List (List : Iir_List);

      --  Create code to elaborate a generic/port list.
      procedure Elab_Generic_List (List : Iir_List);
      procedure Elab_Port_List (List : Iir_List; Parent : Iir);

      --  Generate code to create object OBJ and initialize it with value VAL.
      procedure Elab_Object_Value (Obj : Iir; Value : Iir);

      --  Get the ortho type for an object of type TINFO.
      function Get_Object_Type (Tinfo : Type_Info_Acc; Kind : Object_Kind_Type)
        return O_Tnode;

      --  Get the ortho type for an element of type TINFO.
      function Get_Element_Type (Tinfo : Type_Info_Acc;
                                 Kind : Object_Kind_Type)
        return O_Tnode;

      --  Allocate (and build) a complex object of type OBJ_TYPE.
      --  VAR is the object to be allocated.
      procedure Allocate_Complex_Object (Obj_Type : Iir;
                                         Alloc_Kind : Allocation_Kind;
                                         Var : in out Mnode);

      --function Translate_Interface_Declaration
      --  (Decl : Iir; Subprg : Iir) return Tree;

      --  Create a record that describe thes location of an IIR node and
      --  returns the address of it.
      function Get_Location (N : Iir) return O_Enode;
   end Chap4;

   package Chap6 is
      --  Translate NAME.
      --  RES contains a lnode for the result. This is the object.
      --    RES can be a tree, so it may be referenced only once.
      --  SIG is true if RES is a signal object.
      function Translate_Name (Name : Iir) return Mnode;

      --  Get record element EL of PREFIX.
      function Translate_Selected_Element (Prefix : Mnode;
                                           El : Iir_Element_Declaration)
        return Mnode;

      --  Get direction/length/left bound/right bound of dimension DIM of
      --  array ARR whose type if ARR_TYPE.
      --  For a thin array, ARR is the array;
      --  For a fat array, ARR is the fat array (ie the record with base and
      --    bounds pointer) and not a pointer.
      function Get_Array_Bound_Dir (Arr : O_Lnode;
                                    Arr_Type : Iir;
                                    Dim : Natural;
                                    Is_Sig : Object_Kind_Type)
        return O_Enode;
      function Get_Array_Bound_Length (Arr : O_Lnode;
                                       Arr_Type : Iir;
                                       Dim : Natural;
                                       Is_Sig : Object_Kind_Type)

        return O_Enode;
      function Get_Array_Ptr_Bound_Length (Ptr : O_Lnode;
                                           Arr_Type : Iir;
                                           Dim : Natural;
                                           Is_Sig : Object_Kind_Type)
        return O_Enode;
      function Get_Array_Bound_Left (Arr : O_Lnode;
                                     Arr_Type : Iir;
                                     Dim : Natural;
                                     Is_Sig : Object_Kind_Type)
        return O_Enode;
      function Get_Array_Bound_Right (Arr : O_Lnode;
                                      Arr_Type : Iir;
                                      Dim : Natural;
                                      Is_Sig : Object_Kind_Type)
        return O_Enode;

      --  Extract from fat array FAT_ARRAY the range corresponding to dimension
      --  DIM.
      function Fat_Array_To_Range (Fat_Array : O_Lnode;
                                   Array_Type : Iir;
                                   Dim : Natural;
                                   Is_Sig : Object_Kind_Type)
        return O_Lnode;

      procedure Gen_Bound_Error (Loc : Iir);

      --  Generate code to emit a program error.
      procedure Gen_Program_Error (Loc : Iir);

      --  Generate code to emit a failure if COND is TRUE, indicating an
      --  index violation for dimension DIM of an array.  LOC is usually
      --  the expression which has computed the index and is used only for
      --  its location.
      procedure Check_Bound_Error (Cond : O_Enode; Loc : Iir; Dim : Natural);

      --  Get the offset in the range pointed by RANGE_PTR of INDEX.
      --  This checks INDEX belongs to the range.
      function Translate_Index_To_Offset (Range_Ptr : O_Lnode;
                                          Index : O_Enode;
                                          Index_Type : Iir;
                                          Loc : Iir)
        return O_Enode;

      --  Fill the NAME field of DECL instance.
      procedure Set_Instance_Name (Decl : Iir);
      --  Get address of instance name of BLK (entity/arch/block/comp inst).
      function Get_Instance_Name_Ref (Blk : Iir) return O_Enode;

      --  Fill the name fields of PARENT and CHILD so that CHILD is a child
      --  of PARENT.
      procedure Link_Instance_Name (Parent : Iir; Child : Iir);
   end Chap6;

   package Chap7 is
      --  Generic function to extract a value from a signal.
      generic
         with function Read_Value (Sig : O_Enode; Sig_Type : Iir)
           return O_Enode;
      function Translate_Signal_Value (Sig : O_Enode; Sig_Type : Iir)
        return O_Enode;

      --  Translate expression EXPR into ortho tree.
      function Translate_Expression (Expr : Iir; Rtype : Iir := Null_Iir)
        return O_Enode;

      --  Translate call to function IMP.
      function Translate_Function_Call (Imp : Iir; Assocs : Iir_List)
        return O_Enode;

      --  Translate range and return an lvalue containing the range.
      --  The node returned can be used only one time.
      function Translate_Range (Arange : Iir; Range_Type : Iir)
        return O_Lnode;

      --  Translate range expression EXPR and store the result into the node
      --  pointed by RES_PTR, of type RANGE_TYPE.
      procedure Translate_Range_Ptr
        (Res_Ptr : O_Lnode; Arange : Iir; Range_Type : Iir);
      function Translate_Static_Range (Arange : Iir; Range_Type : Iir)
        return O_Enode;

      --  Same as Translate_Range_Ptr, but for a discrete range (ie: ARANGE
      --  can be a discrete subtype indication).
      procedure Translate_Discrete_Range_Ptr (Res_Ptr : O_Lnode; Arange : Iir);

      --  Translate the static expression EXPR into an ortho expression whose
      --  type must be RTYPE.  Therefore, an implicite conversion might
      --  occurs.
      function Translate_Static_Expression (Expr : Iir; Rtype : Iir)
        return O_Enode;

      --  Convert (if necessary) EXPR of type EXPR_TYPE to type ATYPE.
      function Translate_Implicit_Conv
        (Expr : O_Enode;
         Expr_Type : Iir;
         Atype : Iir;
         Is_Sig : Object_Kind_Type)
        return O_Enode;

      --  Convert range EXPR into ortho tree.
      --  If RANGE_TYPE /= NULL_IIR, convert bounds to RANGE_TYPE.
      --function Translate_Range (Expr : Iir; Range_Type : Iir) return O_Enode;
      function Translate_Static_Range_Left
        (Expr : Iir; Range_Type : Iir := Null_Iir)
        return O_Enode;
      function Translate_Static_Range_Right
        (Expr : Iir; Range_Type : Iir := Null_Iir)
        return O_Enode;
      function Translate_Static_Range_Dir (Expr : Iir) return O_Enode;
      function Translate_Static_Range_Length (Expr : Iir) return O_Enode;

      --  These functions evaluates left bound/right bound/length of the
      --  range expression EXPR.
      function Translate_Range_Expression_Left (Expr : Iir;
                                                Range_Type : Iir := Null_Iir)
        return O_Enode;
      function Translate_Range_Expression_Right (Expr : Iir;
                                                 Range_Type : Iir := Null_Iir)
        return O_Enode;
      function Translate_Range_Expression_Length (Expr : Iir) return O_Enode;

      --  Get the length of any range expression (ie maybe an attribute).
      function Translate_Range_Length (Expr : Iir) return O_Enode;

      --  Assign AGGR to TARGET of type TARGET_TYPE.
      procedure Translate_Aggregate
        (Target : Mnode; Target_Type : Iir; Aggr : Iir);

      --  Translate implicit functions defined by a type.
      procedure Translate_Implicit_Functions_List
        (List : Iir_List; Block : Iir);

      --  Assign EXPR to TARGET.
      --  FIXME: do the checks.
      procedure Translate_Assign
        (Target : Mnode; Expr : Iir; Target_Type : Iir);

      --  Find the declaration of the predefined function IMP in type
      --  definition BASE_TYPE.
      function Find_Predefined_Function
        (Base_Type : Iir; Imp : Iir_Predefined_Functions)
        return Iir;

      function Translate_Lib_Operator (Left, Right : O_Enode; Func : O_Dnode)
        return O_Enode;
   end Chap7;

   package Helpers is
      function Is_Unconstrained_Type (Atype : Iir) return Boolean;

      --  Return the value of field FIELD of lnode L that is contains
      --   a pointer to a record.
      --  This is equivalent to:
      --  new_value (new_selected_element (new_access_element (new_value (l)),
      --                                   field))
      function New_Value_Selected_Acc_Value (L : O_Lnode; Field : O_Fnode)
        return O_Enode;
      function New_Selected_Acc_Value (L : O_Lnode; Field : O_Fnode)
        return O_Lnode;

      --  Equivalent to:
      --  new_value (new_indexed_element (new_access_element (L), I);
      function New_Value_Indexed_Access (L : O_Enode; I : O_Enode)
        return O_Enode;
      function New_Indexed_Access (L : O_Enode; I : O_Enode) return O_Lnode;

      --  Equivalent to new_access_element (new_value (l))
      function New_Acc_Value (L : O_Lnode) return O_Lnode;

      --  Copy a fat pointer.
      --  D and S are stabilized fat pointers.
      procedure Copy_Fat_Pointer (D : Mnode; S: Mnode);

      --  Copy a fat access.
      --  D and S are variable containing address of the fat pointer.
      --  PTR_TYPE is the type of the fat access.
      procedure Copy_Fat_Access (D : O_Lnode; S : O_Lnode; Ptr_Type : Iir);

      --  Generate code to initialize a ghdl_index_type variable V to 0.
      procedure Init_Var (V : O_Lnode);

      --  Generate code to increment a ghdl_index_type variable V.
      procedure Inc_Var (V : O_Lnode);

      --  Generate code to exit from loop LABEL iff COND is true.
      procedure Gen_Exit_When (Label : O_Snode; Cond : O_Enode);

      --  Create a uniq identifier.
      function Create_Uniq_Identifier return O_Ident;

      --  Create a region for temporary variables.
      procedure Open_Temp;
      --  Create a temporary variable.
      function Create_Temp (Atype : O_Tnode) return O_Lnode;
      --  Create a temporary variable of ATYPE and initialize it with VALUE.
      function Create_Temp_Init (Atype : O_Tnode; Value : O_Enode)
        return O_Lnode;
      --  Create a temporary variable of ATYPE and initialize it with the
      --  address of NAME.
      function Create_Temp_Ptr (Atype : O_Tnode; Name : O_Lnode)
        return O_Lnode;
      --  Create a temporary variable for ATYPE and assign it with address
      --  of NAME.
      function Create_Temp_Ptr
        (Atype : Iir; Name : O_Lnode; Is_Sig : Object_Kind_Type)
        return O_Lnode;
      --  Create a mark in the temporary region for the stack2.
      --  FIXME: maybe a flag must be added to CLOSE_TEMP where it is known
      --   stack2 can be released.
      procedure Create_Temp_Stack2_Mark;
      --  Close the temporary region.
      procedure Close_Temp;

      function Create_String_Len (Name : Name_Id; Id : O_Ident) return O_Enode;

      procedure Gen_Memcpy (Dest : O_Enode; Src : O_Enode; Length : O_Enode);

      --  Allocate SIZE bytes aligned on the biggest alignment and return a
      --  pointer of type PTYPE.
      function Gen_Alloc
        (Kind : Allocation_Kind; Size : O_Enode; Ptype : O_Tnode)
        return O_Enode;

      --  Allocate on the heap LENGTH bytes aligned on the biggest alignment,
      --  and returns a pointer of type PTYPE.
      --function Gen_Malloc (Length : O_Enode; Ptype : O_Tnode) return O_Enode;

      --  Handle a composite type TARG/TARG_TYPE and apply DO_NON_COMPOSITE
      --  on each non composite type.
      --  There is a generic parameter DATA which may be updated
      --    before indexing an array by UPDATE_DATA_ARRAY.
      generic
         type Data_Type is private;
         type Composite_Data_Type is private;
         with procedure Do_Non_Composite (Targ : Mnode;
                                          Targ_Type : Iir;
                                          Data : Data_Type);

         --  This function should extract the base of DATA.
         with function Prepare_Data_Array (Data : Data_Type; Targ_Type : Iir)
                                          return Composite_Data_Type;

         --  This function should index DATA.
         with function Update_Data_Array (Data : Composite_Data_Type;
                                          Targ_Type : Iir;
                                          Index : O_Lnode)
                                         return Data_Type;

         --  This function should stabilize DATA.
         with function Prepare_Data_Record (Data : Data_Type; Targ_Type : Iir)
                                           return Composite_Data_Type;

         --  This functon should extract field EL of DATA.
         with function Update_Data_Record (Data : Composite_Data_Type;
                                           Targ_Type : Iir;
                                           El : Iir_Element_Declaration)
           return Data_Type;
      procedure Foreach_Non_Composite (Targ : Mnode;
                                       Targ_Type : Iir;
                                       Data : Data_Type);

      --  Call a procedure (DATA_TYPE) for each signal of TARG.
      procedure Register_Signal
        (Targ : Mnode; Targ_Type : Iir; Proc : O_Dnode);

      --  Call PROC for each scalar signal of list LIST.
      procedure Register_Signal_List (List : Iir_List; Proc : O_Dnode);

      --  Often used subprograms for Foreach_non_composite
      --  when DATA_TYPE is o_enode.
      function Gen_Oenode_Prepare_Data_Composite (Val : O_Enode;
                                                  Targ_Type : Iir)
        return Mnode;
      function Gen_Oenode_Update_Data_Array (Val : Mnode;
                                             Targ_Type : Iir;
                                             Index : O_Lnode)
        return O_Enode;
      function Gen_Oenode_Update_Data_Record
        (Val : Mnode; Targ_Type : Iir; El : Iir_Element_Declaration)
        return O_Enode;

      type Hexstr_Type is array (Integer range 0 .. 15) of Character;
      N2hex : constant Hexstr_Type := "0123456789abcdef";
   private
   end Helpers;
   use Helpers;

   function Get_Type_Info (M : Mnode) return Type_Info_Acc is
   begin
      return M.M1.T;
   end Get_Type_Info;

   function Get_Object_Kind (M : Mnode) return Object_Kind_Type is
   begin
      return M.M1.K;
   end Get_Object_Kind;

   function E2M (E : O_Enode; T : Type_Info_Acc; Kind : Object_Kind_Type)
     return Mnode is
   begin
      return Mnode'(M1 => (State => Mstate_E,
                           Comp => T.Type_Mode in Type_Mode_Fat,
                           K => Kind, T => T, E => E,
                           Vtype => T.Ortho_Type (Kind),
                           Ptype => T.Ortho_Ptr_Type (Kind)));
   end E2M;

   function Lv2M (L : O_Lnode; T : Type_Info_Acc; Kind : Object_Kind_Type)
     return Mnode is
   begin
      return Mnode'(M1 => (State => Mstate_Lv,
                           Comp => T.Type_Mode in Type_Mode_Fat,
                           K => Kind, T => T, Lv => L,
                           Vtype => T.Ortho_Type (Kind),
                           Ptype => T.Ortho_Ptr_Type (Kind)));
   end Lv2M;

   function Lv2M (L : O_Lnode;
                  Comp : Boolean;
                  Vtype : O_Tnode;
                  Ptype : O_Tnode;
                  T : Type_Info_Acc; Kind : Object_Kind_Type)
     return Mnode is
   begin
      return Mnode'(M1 => (State => Mstate_Lv,
                           Comp => Comp,
                           K => Kind, T => T, Lv => L,
                           Vtype => Vtype, Ptype => Ptype));
   end Lv2M;

   function Lp2M (L : O_Lnode; T : Type_Info_Acc; Kind : Object_Kind_Type)
     return Mnode is
   begin
      return Mnode'(M1 => (State => Mstate_Lp,
                           Comp => T.Type_Mode in Type_Mode_Fat,
                           K => Kind, T => T, Lp => L,
                           Vtype => T.Ortho_Type (Kind),
                           Ptype => T.Ortho_Ptr_Type (Kind)));
   end Lp2M;

   function Lp2M (L : O_Lnode;
                  T : Type_Info_Acc;
                  Kind : Object_Kind_Type;
                  Vtype : O_Tnode;
                  Ptype : O_Tnode)
     return Mnode is
   begin
      return Mnode'(M1 => (State => Mstate_Lp,
                           Comp => T.Type_Mode in Type_Mode_Fat,
                           K => Kind, T => T, Lp => L,
                           Vtype => Vtype, Ptype => Ptype));
   end Lp2M;

   function Lv2M (L : O_Lnode;
                  T : Type_Info_Acc;
                  Kind : Object_Kind_Type;
                  Vtype : O_Tnode;
                  Ptype : O_Tnode)
     return Mnode is
   begin
      return Mnode'(M1 => (State => Mstate_Lv,
                           Comp => T.Type_Mode in Type_Mode_Fat,
                           K => Kind, T => T, Lv => L,
                           Vtype => Vtype, Ptype => Ptype));
   end Lv2M;

   function Vv2M (L : O_Lnode;
                  T : Type_Info_Acc;
                  Kind : Object_Kind_Type)
     return Mnode is
   begin
      return Mnode'(M1 => (State => Mstate_Vv,
                           Comp => T.Type_Mode in Type_Mode_Fat,
                           K => Kind, T => T, Lv => L,
                           Vtype => T.Ortho_Type (Kind),
                           Ptype => T.Ortho_Ptr_Type (Kind)));
   end Vv2M;

   function Vv2M (L : O_Lnode;
                  T : Type_Info_Acc;
                  Kind : Object_Kind_Type;
                  Vtype : O_Tnode;
                  Ptype : O_Tnode)
     return Mnode is
   begin
      return Mnode'(M1 => (State => Mstate_Vv,
                           Comp => T.Type_Mode in Type_Mode_Fat,
                           K => Kind, T => T, Lv => L,
                           Vtype => Vtype,
                           Ptype => Ptype));
   end Vv2M;

   function Vp2M (L : O_Lnode;
                  T : Type_Info_Acc;
                  Kind : Object_Kind_Type;
                  Vtype : O_Tnode;
                  Ptype : O_Tnode)
     return Mnode is
   begin
      return Mnode'(M1 => (State => Mstate_Vp,
                           Comp => T.Type_Mode in Type_Mode_Fat,
                           K => Kind, T => T, Lp => L,
                           Vtype => Vtype, Ptype => Ptype));
   end Vp2M;

   function Vp2M (L : O_Lnode;
                  T : Type_Info_Acc;
                  Kind : Object_Kind_Type)
     return Mnode is
   begin
      return Mnode'(M1 => (State => Mstate_Vp,
                           Comp => T.Type_Mode in Type_Mode_Fat,
                           K => Kind, T => T, Lp => L,
                           Vtype => T.Ortho_Type (Kind),
                           Ptype => T.Ortho_Ptr_Type (Kind)));
   end Vp2M;

   function M2Lv (M : Mnode) return O_Lnode is
   begin
      case M.M1.State is
         when Mstate_E =>
            case Get_Type_Info (M).Type_Mode is
               when Type_Mode_Thin =>
                  --  Scalar to var is not possible.
                  --  FIXME: This is not coherent with the fact that this
                  --  conversion is possible when M is stabilized.
                  raise Internal_Error;
               when Type_Mode_Fat =>
                  return New_Access_Element (M.M1.E);
               when Type_Mode_Unknown =>
                  raise Internal_Error;
            end case;
         when Mstate_Lp
           | Mstate_Vp =>
            return New_Acc_Value (M.M1.Lp);
         when Mstate_Lv
           | Mstate_Vv =>
            return M.M1.Lv;
         when Mstate_Null
           | Mstate_Bad =>
            raise Internal_Error;
      end case;
   end M2Lv;

   --  Convert M to an O_LNODE using the type convention.
   function M2L (M : Mnode) return O_Lnode is
   begin
      case M.M1.State is
         when Mstate_E =>
            case Get_Type_Info (M).Type_Mode is
               when Type_Mode_Thin =>
                  --  Scalar to var is not possible.
                  --  FIXME: This is not coherent with the fact that this
                  --  conversion is possible when M is stabilized.
                  raise Internal_Error;
               when Type_Mode_Fat =>
                  return New_Access_Element (M.M1.E);
               when Type_Mode_Unknown =>
                  raise Internal_Error;
            end case;
         when Mstate_Lp
           | Mstate_Vp =>
            case Get_Type_Info (M).Type_Mode is
               when Type_Mode_Thin
                 | Type_Mode_Fat_Acc
                 | Type_Mode_Record
                 | Type_Mode_Array
                 | Type_Mode_Fat_Array =>
                  return New_Acc_Value (M.M1.Lp);
               when Type_Mode_Ptr_Array =>
                  return M.M1.Lp;
               when others =>
                  raise Internal_Error;
            end case;
         when Mstate_Lv
           | Mstate_Vv =>
            case Get_Type_Info (M).Type_Mode is
               when Type_Mode_Thin
                 | Type_Mode_Fat_Acc
                 | Type_Mode_Record
                 | Type_Mode_Array
                 | Type_Mode_Fat_Array =>
                  return M.M1.Lv;
               when Type_Mode_Ptr_Array =>
                  --  FIXME: this is bad, since this creates a new node.
                  return Create_Temp_Init (M.M1.Ptype,
                                           New_Address (M.M1.Lv, M.M1.Ptype));
               when others =>
                  raise Internal_Error;
            end case;
         when Mstate_Null
           | Mstate_Bad =>
            raise Internal_Error;
      end case;
   end M2L;

   function M2Lp (M : Mnode) return O_Lnode is
   begin
      case M.M1.State is
         when Mstate_E =>
            raise Internal_Error;
         when Mstate_Lp
           | Mstate_Vp =>
            return M.M1.Lp;
         when Mstate_Lv
           | Mstate_Vv =>
            if Get_Type_Info (M).Type_Mode in Type_Mode_Fat then
               return Create_Temp_Init (M.M1.Ptype,
                                        New_Address (M.M1.Lv, M.M1.Ptype));
            else
               raise Internal_Error;
            end if;
         when Mstate_Null
           | Mstate_Bad =>
            raise Internal_Error;
      end case;
   end M2Lp;

   function M2Vp (M : Mnode) return O_Lnode is
   begin
      case M.M1.State is
         when Mstate_Vp =>
            return M.M1.Lp;
         when others =>
            raise Internal_Error;
      end case;
   end M2Vp;

   function T2M (Atype : Iir; Kind : Object_Kind_Type) return Mnode
   is
      T : Type_Info_Acc;
   begin
      T := Get_Info (Atype);
      return Mnode'(M1 => (State => Mstate_Null,
                           Comp => T.Type_Mode in Type_Mode_Fat,
                           K => Kind, T => T,
                           Vtype => T.Ortho_Type (Kind),
                           Ptype => T.Ortho_Ptr_Type (Kind)));
   end T2M;

   function Stabilize (M : Mnode; Can_Copy : Boolean := False) return Mnode
   is
      L : O_Lnode;
      K : Object_Kind_Type;
   begin
      K := M.M1.K;
      case M.M1.State is
         when Mstate_E =>
            if M.M1.Comp then
               L := Create_Temp_Init (M.M1.Ptype, M.M1.E);
               return Mnode'(M1 => (State => Mstate_Vp,
                                    Comp => M.M1.Comp,
                                    K => K, T => M.M1.T, Lp => L,
                                    Vtype => M.M1.Vtype, Ptype => M.M1.Ptype));
            else
               L := Create_Temp_Init (M.M1.Vtype, M.M1.E);
               return Mnode'(M1 => (State => Mstate_Vv,
                                    Comp => M.M1.Comp,
                                    K => K, T => M.M1.T, Lv => L,
                                    Vtype => M.M1.Vtype, Ptype => M.M1.Ptype));
            end if;
         when Mstate_Lp =>
            L := Create_Temp_Init (M.M1.Ptype, New_Value (M.M1.Lp));
            return Mnode'(M1 => (State => Mstate_Vp,
                                 Comp => M.M1.Comp, K => K, T => M.M1.T,
                                 Lp => L,
                                 Vtype => M.M1.Vtype,
                                 Ptype => M.M1.Ptype));
         when Mstate_Lv =>
            if M.M1.Ptype = O_Tnode_Null then
               if not Can_Copy then
                  raise Internal_Error;
               end if;
               L := Create_Temp_Init (M.M1.Vtype, New_Value (M.M1.Lv));
               return Mnode'(M1 => (State => Mstate_Vv,
                                    Comp => M.M1.Comp, K => K,
                                    T => M.M1.T, Lv => L,
                                    Vtype => M.M1.Vtype, Ptype => M.M1.Ptype));

            else
               L := Create_Temp_Ptr (M.M1.Ptype, M.M1.Lv);
               return Mnode'(M1 => (State => Mstate_Vp,
                                    Comp => M.M1.Comp,
                                    K => K, T => M.M1.T, Lp => L,
                                    Vtype => M.M1.Vtype, Ptype => M.M1.Ptype));
            end if;
         when Mstate_Vp
           | Mstate_Vv =>
            return M;
         when Mstate_Bad
           | Mstate_Null =>
            raise Internal_Error;
      end case;
   end Stabilize;

   procedure Stabilize (M : in out Mnode) is
   begin
      M := Stabilize (M);
   end Stabilize;

   function M2E (M : Mnode) return O_Enode is
   begin
      case M.M1.State is
         when Mstate_E =>
            return M.M1.E;
         when Mstate_Lp
           | Mstate_Vp =>
            case M.M1.T.Type_Mode is
               when Type_Mode_Unknown =>
                  raise Internal_Error;
               when Type_Mode_Thin =>
                  return New_Value (New_Acc_Value (M.M1.Lp));
               when Type_Mode_Fat =>
                  return New_Value (M.M1.Lp);
            end case;
         when Mstate_Lv
           | Mstate_Vv =>
            case M.M1.T.Type_Mode is
               when Type_Mode_Unknown =>
                  raise Internal_Error;
               when Type_Mode_Thin =>
                  return New_Value (M.M1.Lv);
               when Type_Mode_Fat =>
                  return New_Address (M.M1.Lv, M.M1.Ptype);
            end case;
         when Mstate_Bad
           | Mstate_Null =>
            raise Internal_Error;
      end case;
   end M2E;

   function M2Addr (M : Mnode) return O_Enode is
   begin
      case M.M1.State is
         when Mstate_Lp
           | Mstate_Vp =>
            return New_Value (M.M1.Lp);
         when Mstate_Lv
           | Mstate_Vv =>
            return New_Address (M.M1.Lv, M.M1.Ptype);
         when Mstate_E =>
            if M.M1.Comp then
               return M.M1.E;
            else
               raise Internal_Error;
            end if;
         when Mstate_Bad
           | Mstate_Null =>
            raise Internal_Error;
      end case;
   end M2Addr;

   function Is_Null (M : Mnode) return Boolean is
   begin
      return M.M1.State = Mstate_Null;
   end Is_Null;

   function Is_Stable (M : Mnode) return Boolean is
   begin
      case M.M1.State is
         when Mstate_Vp
           | Mstate_Vv =>
            return True;
         when others =>
            return False;
      end case;
   end Is_Stable;

   function Varv2M
     (Var : Var_Acc; Vtype : Type_Info_Acc; Mode : Object_Kind_Type)
     return Mnode is
   begin
      return Lv2M (Get_Var (Var), Vtype, Mode);
   end Varv2M;

   function Varv2M (Var : Var_Acc;
                    Var_Type : Type_Info_Acc;
                    Mode : Object_Kind_Type;
                    Vtype : O_Tnode;
                    Ptype : O_Tnode)
     return Mnode is
   begin
      return Lv2M (Get_Var (Var), Var_Type, Mode, Vtype, Ptype);
   end Varv2M;

   --  Convert a Lnode for a sub object) to an MNODE.
   function Lo2M (L : O_Lnode; Vtype : Type_Info_Acc; Mode : Object_Kind_Type)
     return Mnode is
   begin
      case Vtype.Type_Mode is
         when Type_Mode_Scalar
           | Type_Mode_Acc
           | Type_Mode_File
           | Type_Mode_Fat_Array
           | Type_Mode_Fat_Acc =>
            return Lv2M (L, Vtype, Mode);
         when Type_Mode_Ptr_Array
           | Type_Mode_Array
           | Type_Mode_Record =>
            if Vtype.C = null then
               return Lv2M (L, Vtype, Mode);
            else
               return Lp2M (L, Vtype, Mode);
            end if;
         when Type_Mode_Unknown =>
            raise Internal_Error;
      end case;
   end Lo2M;

   function Get_Var
     (Var : Var_Acc; Vtype : Type_Info_Acc; Mode : Object_Kind_Type)
     return Mnode
   is
      L : O_Lnode;
      Stable : Boolean;
   begin
      --  FIXME: there may be Vv2M and Vp2M.
      L := Get_Var (Var);
      Stable := Is_Var_Stable (Var);
      case Vtype.Type_Mode is
         when Type_Mode_Scalar
           | Type_Mode_Acc
           | Type_Mode_File
           | Type_Mode_Fat_Array
           | Type_Mode_Fat_Acc =>
            if Stable then
               return Vv2M (L, Vtype, Mode);
            else
               return Lv2M (L, Vtype, Mode);
            end if;
         when Type_Mode_Ptr_Array
           | Type_Mode_Array
           | Type_Mode_Record =>
            if Vtype.C = null then
               if Stable then
                  return Vv2M (L, Vtype, Mode);
               else
                  return Lv2M (L, Vtype, Mode);
               end if;
            else
               if Stable then
                  return Vp2M (L, Vtype, Mode);
               else
                  return Lp2M (L, Vtype, Mode);
               end if;
            end if;
         when Type_Mode_Unknown =>
            raise Internal_Error;
      end case;
   end Get_Var;

   function Create_Temp (Info : Type_Info_Acc;
                         Kind : Object_Kind_Type := Mode_Value)
                        return Mnode is
   begin
      if Info.C /= null and then Info.Type_Mode /= Type_Mode_Fat_Array then
         --  For a complex and constrained object, we just allocate
         --  a pointer to the object.
         return Vp2M (Create_Temp (Info.Ortho_Ptr_Type (Kind)), Info, Kind);
      else
         return Vv2M (Create_Temp (Info.Ortho_Type (Kind)), Info, Kind);
      end if;
   end Create_Temp;

   package body Helpers is
      function Is_Unconstrained_Type (Atype : Iir) return Boolean is
      begin
         return
           Get_Kind (Atype) in Iir_Kinds_Unconstrained_Array_Type_Definition;
      end Is_Unconstrained_Type;

      function New_Value_Selected_Acc_Value (L : O_Lnode; Field : O_Fnode)
        return O_Enode is
      begin
         return New_Value
           (New_Selected_Element (New_Access_Element (New_Value (L)), Field));
      end New_Value_Selected_Acc_Value;

      function New_Selected_Acc_Value (L : O_Lnode; Field : O_Fnode)
        return O_Lnode is
      begin
         return New_Selected_Element
           (New_Access_Element (New_Value (L)), Field);
      end New_Selected_Acc_Value;

      function New_Value_Indexed_Access (L : O_Enode; I : O_Enode)
        return O_Enode is
      begin
         return New_Value (New_Indexed_Element (New_Access_Element (L), I));
      end New_Value_Indexed_Access;

      function New_Indexed_Access (L : O_Enode; I : O_Enode) return O_Lnode is
      begin
         return New_Indexed_Element (New_Access_Element (L), I);
      end New_Indexed_Access;

      function New_Acc_Value (L : O_Lnode) return O_Lnode is
      begin
         return New_Access_Element (New_Value (L));
      end New_Acc_Value;

      procedure Copy_Fat_Pointer (D : Mnode; S: Mnode)
      is
      begin
         New_Assign_Stmt (M2Lp (Chap3.Get_Array_Base (D)),
                          M2Addr (Chap3.Get_Array_Base (S)));
         New_Assign_Stmt (M2Lp (Chap3.Get_Array_Bounds (D)),
                          M2Addr (Chap3.Get_Array_Bounds (S)));
      end Copy_Fat_Pointer;

      procedure Copy_Fat_Pointer
        (D : O_Lnode; S : O_Lnode; Ftype : Iir; Is_Sig : Object_Kind_Type)
      is
         Info : Type_Info_Acc := Get_Info (Ftype);
      begin
         New_Assign_Stmt
           (New_Selected_Acc_Value (D, Info.T.Base_Field (Is_Sig)),
            New_Value_Selected_Acc_Value (S, Info.T.Base_Field (Is_Sig)));
         New_Assign_Stmt
           (New_Selected_Acc_Value (D, Info.T.Bounds_Field (Is_Sig)),
            New_Value_Selected_Acc_Value (S, Info.T.Bounds_Field (Is_Sig)));
      end Copy_Fat_Pointer;

      procedure Copy_Fat_Access (D : O_Lnode; S : O_Lnode; Ptr_Type : Iir)
      is
      begin
         Copy_Fat_Pointer (D, S, Get_Designated_Type (Ptr_Type), Mode_Value);
      end Copy_Fat_Access;

      procedure Inc_Var (V : O_Lnode) is
      begin
         New_Assign_Stmt
           (V, New_Dyadic_Op (ON_Add_Ov,
                              New_Value (V),
                              New_Unsigned_Literal (Ghdl_Index_Type, 1)));
      end Inc_Var;

      procedure Init_Var (V : O_Lnode) is
      begin
         New_Assign_Stmt (V, New_Unsigned_Literal (Ghdl_Index_Type, 0));
      end Init_Var;

      procedure Gen_Exit_When (Label : O_Snode; Cond : O_Enode)
      is
         If_Blk : O_If_Block;
      begin
         Start_If_Stmt (If_Blk, Cond);
         New_Exit_Stmt (Label);
         Finish_If_Stmt (If_Blk);
      end Gen_Exit_When;

      Uniq_Id : Natural := 0;

      function Create_Uniq_Identifier return O_Ident
      is
         Str : String (1 .. 12);
         Val : Natural;
      begin
         Str (1 .. 3) := "_UI";
         Val := Uniq_Id;
         Uniq_Id := Uniq_Id + 1;
         for I in reverse 4 .. 11 loop
            Str (I) := N2hex (Val mod 16);
            Val := Val / 16;
         end loop;
         --Str (12) := Nul;
         return Get_Identifier (Str (1 .. 11));
      end Create_Uniq_Identifier;

      --  Create a temporary variable.
      type Temp_Level_Type;
      type Temp_Level_Acc is access Temp_Level_Type;
      type Temp_Level_Type is record
         Prev : Temp_Level_Acc;
         Level : Natural;
         Id : Natural;
         Emitted : Boolean;
         Stack2_Mark : O_Lnode;
      end record;
      --  Current level.
      Temp_Level : Temp_Level_Acc := null;

      --  List of unused temp_level_type structures.  To be faster, they are
      --  never deallocated.
      Old_Level : Temp_Level_Acc := null;

      procedure Open_Temp is
         L : Temp_Level_Acc;
      begin
         if Old_Level /= null then
            L := Old_Level;
            Old_Level := L.Prev;
         else
            L := new Temp_Level_Type;
         end if;
         L.all := (Prev => Temp_Level,
                   Level => 0,
                   Id => 0,
                   Emitted => False,
                   Stack2_Mark => O_Lnode_Null);
         if Temp_Level /= null then
            L.Level := Temp_Level.Level + 1;
         end if;
         Temp_Level := L;
      end Open_Temp;

      procedure Close_Temp is
         L : Temp_Level_Acc;
         Constr : O_Assoc_List;
      begin
         if Temp_Level = null then
            --  OPEN_TEMP was not called.
            raise Internal_Error;
         end if;
         if Temp_Level.Stack2_Mark /= O_Lnode_Null then
            Start_Association (Constr, Ghdl_Stack2_Release);
            New_Association (Constr, New_Value (Temp_Level.Stack2_Mark));
            New_Procedure_Call (Constr);
         end if;
         if Temp_Level.Emitted then
            Finish_Declare_Stmt;
         end if;
         L := Temp_Level;
         Temp_Level := L.Prev;
         L.Prev := Old_Level;
         Old_Level := L;
      end Close_Temp;

      procedure Create_Temp_Stack2_Mark
      is
         Constr : O_Assoc_List;
      begin
         if Temp_Level.Stack2_Mark /= O_Lnode_Null then
            --  Only the first mark in a region is registred.
            --  The release operation frees the memory allocated after the
            --  first mark.
            return;
         end if;
         Temp_Level.Stack2_Mark := Create_Temp (Ghdl_Ptr_Type);
         Start_Association (Constr, Ghdl_Stack2_Mark);
         New_Assign_Stmt (Temp_Level.Stack2_Mark, New_Function_Call (Constr));
      end Create_Temp_Stack2_Mark;

      function Create_Temp (Atype : O_Tnode) return O_Lnode
      is
         Str : String (1 .. 12);
         Val : Natural;
         Res : O_Lnode;
         P : Natural;
      begin
         if Temp_Level = null then
            --  OPEN_TEMP was never called.
            raise Internal_Error;
            --  This is an hack, just to allow array subtype to array type
            --  conversion.
            --New_Var_Decl
            --  (Res, Create_Uniq_Identifier, O_Storage_Private, Atype);
            --return Res;
         else
            if not Temp_Level.Emitted then
               Temp_Level.Emitted := True;
               Start_Declare_Stmt;
            end if;
         end if;
         Val := Temp_Level.Id;
         Temp_Level.Id := Temp_Level.Id + 1;
         P := Str'Last;
         loop
            Str (P) := Character'Val (Val mod 10 + Character'Pos ('0'));
            Val := Val / 10;
            P := P - 1;
            exit when Val = 0;
         end loop;
         Str (P) := '_';
         P := P - 1;
         Val := Temp_Level.Level;
         loop
            Str (P) := Character'Val (Val mod 10 + Character'Pos ('0'));
            Val := Val / 10;
            P := P - 1;
            exit when Val = 0;
         end loop;
         Str (P) := 'T';
         --Str (12) := Nul;
         New_Var_Decl
           (Res, Get_Identifier (Str (P .. Str'Last)), O_Storage_Local, Atype);
         return Res;
      end Create_Temp;

      function Create_Temp_Init (Atype : O_Tnode; Value : O_Enode)
        return O_Lnode
      is
         Res : O_Lnode;
      begin
         Res := Create_Temp (Atype);
         New_Assign_Stmt (Res, Value);
         return Res;
      end Create_Temp_Init;

      function Create_Temp_Ptr (Atype : O_Tnode; Name : O_Lnode)
        return O_Lnode is
      begin
         return Create_Temp_Init (Atype, New_Address (Name, Atype));
      end Create_Temp_Ptr;

      function Create_Temp_Ptr
        (Atype : Iir; Name : O_Lnode; Is_Sig : Object_Kind_Type)
        return O_Lnode
      is
         Temp_Type : O_Tnode;
      begin
         Temp_Type := Get_Info (Atype).Ortho_Ptr_Type (Is_Sig);
         return Create_Temp_Init (Temp_Type, New_Address (Name, Temp_Type));
      end Create_Temp_Ptr;

      --  Convert NAME into a STRING_CST.
      function Create_String (Name : Name_Id; Id : O_Ident) return O_Enode
      is
         Str : String := Name_Table.Image (Name);
         Atype : O_Tnode;
         Res : O_Enode;
         List : O_Array_Aggr_List;
         Const : O_Lnode;
      begin
         Atype := New_Constrained_Array_Type
           (Chararray_Type_Node,
            New_Unsigned_Literal (Ghdl_Index_Type, Str'Length));
         Start_Array_Aggr (List, Atype);
         for I in Str'Range loop
            New_Array_Aggr_El
              (List,
               New_Unsigned_Literal (Char_Type_Node, Character'Pos (Str (I))));
         end loop;
         Finish_Array_Aggr (List, Res);
         New_Const_Decl (Const, Id, O_Storage_Private, Atype, Res);
         return New_Address (Const, Char_Ptr_Type_Node);
      end Create_String;

      function Create_String_Len (Name : Name_Id; Id : O_Ident)
        return O_Enode
      is
         Str : O_Enode;
         Str_Len : O_Enode;
         List : O_Record_Aggr_List;
         Res : O_Enode;
      begin
         if Name_Table.Is_Character (Name) then
            raise Internal_Error;
         end if;
         Str := Create_String (Name, Id);
         Str_Len := New_Unsigned_Literal
           (Ghdl_Index_Type,
            Unsigned_64 (Name_Table.Get_Name_Length (Name)));
         Start_Record_Aggr (List, Ghdl_Str_Len_Type_Node);
         New_Record_Aggr_El (List, Str_Len);
         New_Record_Aggr_El (List, Str);
         Finish_Record_Aggr (List, Res);
         return Res;
      end Create_String_Len;

      procedure Gen_Memcpy (Dest : O_Enode; Src : O_Enode; Length : O_Enode)
      is
         Constr : O_Assoc_List;
      begin
         Start_Association (Constr, Ghdl_Memcpy);
         New_Association (Constr, New_Convert_Ov (Dest, Ghdl_Ptr_Type));
         New_Association (Constr, New_Convert_Ov (Src, Ghdl_Ptr_Type));
         New_Association (Constr, Length);
         New_Procedure_Call (Constr);
      end Gen_Memcpy;

--       function Gen_Malloc (Length : O_Enode; Ptype : O_Tnode) return O_Enode
--       is
--          Constr : O_Assoc_List;
--       begin
--          Start_Association (Constr, Ghdl_Malloc);
--          New_Association (Constr, Length);
--          return New_Convert_Ov (New_Function_Call (Constr), Ptype);
--       end Gen_Malloc;

      function Gen_Alloc
        (Kind : Allocation_Kind; Size : O_Enode; Ptype : O_Tnode)
        return O_Enode
      is
         Constr : O_Assoc_List;
      begin
         case Kind is
            when Alloc_Memory
              | Alloc_System
              | Alloc_Heap =>
               Start_Association (Constr, Ghdl_Malloc);
               New_Association (Constr, Size);
               return New_Convert_Ov (New_Function_Call (Constr), Ptype);
            when Alloc_Stack =>
               return New_Alloca (Ptype, Size);
            when Alloc_Return =>
               Start_Association (Constr, Ghdl_Stack2_Allocate);
               New_Association (Constr, Size);
               return New_Convert_Ov (New_Function_Call (Constr), Ptype);
         end case;
      end Gen_Alloc;

      procedure Foreach_Non_Composite (Targ : Mnode;
                                       Targ_Type : Iir;
                                       Data : Data_Type)
      is
         Type_Info : Type_Info_Acc;
      begin
         Type_Info := Get_Info (Targ_Type);
         case Type_Info.Type_Mode is
            when Type_Mode_B2
              | Type_Mode_E8
              | Type_Mode_E32
              | Type_Mode_I32
              | Type_Mode_P64
              | Type_Mode_F64 =>
               Do_Non_Composite (Targ, Targ_Type, Data);
            when Type_Mode_Array =>
               declare
                  Var_I : O_Lnode;
                  Var_Array : Mnode;
                  Loop_Blk : O_Loop_Block;
                  Label : O_Snode;
                  If_Blk : O_If_Block;
                  Composite_Data : Composite_Data_Type;
                  Sub_Data : Data_Type;
               begin
                  Open_Temp;
                  Var_Array := Stabilize (Targ);
                  if True then
                     Var_I := Create_Temp (Ghdl_Index_Type);
                  else
                     New_Var_Decl
                       (Var_I, Wki_I, O_Storage_Local, Ghdl_Index_Type);
                  end if;
                  Init_Var (Var_I);
                  Composite_Data := Prepare_Data_Array (Data, Targ_Type);
                  Start_Loop_Stmt (Loop_Blk, Label);
                  Gen_Exit_When
                    (Label,
                     New_Compare_Op (ON_Ge,
                                     New_Value (Var_I),
                                     Chap3.Get_Array_Type_Length (Targ_Type),
                                     Ghdl_Bool_Type_Node));
                  Sub_Data := Update_Data_Array
                    (Composite_Data, Targ_Type, Var_I);
                  Foreach_Non_Composite
                    (Chap3.Index_Base (Var_Array, Targ_Type,
                                       New_Value (Var_I)),
                     Get_Element_Subtype (Targ_Type),
                     Sub_Data);
                  Inc_Var (Var_I);
                  Finish_Loop_Stmt (Loop_Blk);
                  Close_Temp;
               end;
            when Type_Mode_Fat_Array
              | Type_Mode_Ptr_Array =>
               declare
                  Var_Array : Mnode;
                  Var_Base : Mnode;
                  Var_Length : O_Lnode;
                  Var_I : O_Lnode;
                  Loop_Blk : O_Loop_Block;
                  Label : O_Snode;
                  If_Blk : O_If_Block;
                  Sub_Data : Data_Type;
                  Composite_Data : Composite_Data_Type;
               begin
                  Open_Temp;
                  Var_Array := Stabilize (Targ);
                  Var_Length := Create_Temp (Ghdl_Index_Type);
                  Var_Base := Stabilize (Chap3.Get_Array_Base (Var_Array));
                  New_Assign_Stmt
                    (Var_Length,
                     Chap3.Get_Array_Length (Var_Array, Targ_Type));
                  Composite_Data := Prepare_Data_Array (Data, Targ_Type);
                  if True then
                     Var_I := Create_Temp (Ghdl_Index_Type);
                  else
                     New_Var_Decl
                       (Var_I, Wki_I, O_Storage_Local, Ghdl_Index_Type);
                  end if;
                  Init_Var (Var_I);
                  Start_Loop_Stmt (Loop_Blk, Label);
                  Gen_Exit_When
                    (Label, New_Compare_Op (ON_Ge,
                                            New_Value (Var_I),
                                            New_Value (Var_Length),
                                            Ghdl_Bool_Type_Node));
                  Sub_Data := Update_Data_Array
                    (Composite_Data, Targ_Type, Var_I);
                  Foreach_Non_Composite
                    (Chap3.Index_Base (Var_Base, Targ_Type,
                                       New_Value (Var_I)),
                     Get_Element_Subtype (Targ_Type),
                     Sub_Data);
                  Inc_Var (Var_I);
                  Finish_Loop_Stmt (Loop_Blk);
                  Close_Temp;
               end;
            when Type_Mode_Record =>
               declare
                  Var_Record : Mnode;
                  Sub_Data : Data_Type;
                  Composite_Data : Composite_Data_Type;
                  List : Iir_List;
                  El : Iir_Element_Declaration;
               begin
                  Open_Temp;
                  Var_Record := Stabilize (Targ);
                  Composite_Data := Prepare_Data_Record (Data, Targ_Type);
                  List := Get_Element_Declaration_List
                    (Get_Base_Type (Targ_Type));
                  for I in Natural loop
                     El := Get_Nth_Element (List, I);
                     exit when El = Null_Iir;
                     Sub_Data := Update_Data_Record
                       (Composite_Data, Targ_Type, El);
                     Foreach_Non_Composite
                       (Chap6.Translate_Selected_Element (Var_Record, El),
                        Get_Type (El),
                        Sub_Data);
                  end loop;
                  Close_Temp;
               end;
            when others =>
               Error_Kind ("foreach_non_composite/"
                           & Type_Mode_Type'Image (Type_Info.Type_Mode),
                           Targ_Type);
         end case;
      end Foreach_Non_Composite;

      procedure Register_Non_Composite_Signal (Targ : Mnode;
                                               Targ_Type : Iir;
                                               Proc : O_Dnode)
      is
         Constr : O_Assoc_List;
      begin
         Start_Association (Constr, Proc);
         New_Association
           (Constr, New_Convert_Ov (New_Value (M2Lv (Targ)), Ghdl_Signal_Ptr));
         New_Procedure_Call (Constr);
      end Register_Non_Composite_Signal;

      function Register_Update_Data_Array
        (Data : O_Dnode; Targ_Type : Iir; Index : O_Lnode)
        return O_Dnode
      is
      begin
         return Data;
      end Register_Update_Data_Array;

      function Register_Prepare_Data_Composite (Data : O_Dnode;
                                                Targ_Type : Iir)
        return O_Dnode
      is
      begin
         return Data;
      end Register_Prepare_Data_Composite;

      function Register_Update_Data_Record
        (Data : O_Dnode; Targ_Type : Iir; El : Iir_Element_Declaration)
        return O_Dnode
      is
      begin
         return Data;
      end Register_Update_Data_Record;

      procedure Register_Signal_1 is new Foreach_Non_Composite
        (Data_Type => O_Dnode,
         Composite_Data_Type => O_Dnode,
         Do_Non_Composite => Register_Non_Composite_Signal,
         Prepare_Data_Array => Register_Prepare_Data_Composite,
         Update_Data_Array => Register_Update_Data_Array,
         Prepare_Data_Record => Register_Prepare_Data_Composite,
         Update_Data_Record => Register_Update_Data_Record);

      procedure Register_Signal (Targ : Mnode;
                                 Targ_Type : Iir;
                                 Proc : O_Dnode)
        renames Register_Signal_1;

      procedure Register_Signal_List (List : Iir_List; Proc : O_Dnode)
      is
         El : Iir;
         Sig : Mnode;
      begin
         if List = Null_Iir_List then
            return;
         end if;
         for I in Natural loop
            El := Get_Nth_Element (List, I);
            exit when El = Null_Iir;
            Open_Temp;
            Sig := Chap6.Translate_Name (El);
            Register_Signal (Sig, Get_Type (El), Proc);
            Close_Temp;
         end loop;
      end Register_Signal_List;

      function Gen_Oenode_Prepare_Data_Composite (Val : O_Enode;
                                                  Targ_Type : Iir)
        return Mnode
      is
         Res : Mnode;
         Type_Info : Type_Info_Acc;
      begin
         Type_Info := Get_Info (Targ_Type);
         Res := E2M (Val, Type_Info, Mode_Value);
         case Type_Info.Type_Mode is
            when Type_Mode_Array
              | Type_Mode_Ptr_Array
              | Type_Mode_Fat_Array =>
               Res := Chap3.Get_Array_Base (Res);
            when Type_Mode_Record =>
               Res := Stabilize (Res);
            when others =>
               --  Not a composite type!
               raise Internal_Error;
         end case;
         return Res;
      end Gen_Oenode_Prepare_Data_Composite;

      function Gen_Oenode_Update_Data_Array (Val : Mnode;
                                             Targ_Type : Iir;
                                             Index : O_Lnode)
        return O_Enode
      is
      begin
         return M2E (Chap3.Index_Base (Val, Targ_Type, New_Value (Index)));
      end Gen_Oenode_Update_Data_Array;

      function Gen_Oenode_Update_Data_Record
        (Val : Mnode; Targ_Type : Iir; El : Iir_Element_Declaration)
        return O_Enode
      is
      begin
         return M2E (Chap6.Translate_Selected_Element (Val, El));
      end Gen_Oenode_Update_Data_Record;
   end Helpers;

   package body Chap1 is
      procedure Start_Block_Decl (Blk : Iir)
      is
         Info : Block_Info_Acc;
      begin
         Info := Get_Info (Blk);
         New_Uncomplete_Record_Type (Info.Block_Decls_Type);
         New_Type_Decl (Create_Identifier ("INSTTYPE"), Info.Block_Decls_Type);
         Info.Block_Decls_Ptr_Type := New_Access_Type (Info.Block_Decls_Type);
         New_Type_Decl (Create_Identifier ("INSTPTR"),
                        Info.Block_Decls_Ptr_Type);
      end Start_Block_Decl;

      procedure Translate_Entity_Init (Entity : Iir; Instance : O_Lnode)
      is
         Info : Block_Info_Acc;
         List : Iir_List;
         El : Iir;
         El_Type : Iir;
      begin
         Info := Get_Info (Entity);

         Push_Local_Factory;

         List := Get_Generic_List (Entity);
         if List /= Null_Iir_List then
            for I in Natural loop
               El := Get_Nth_Element (List, I);
               exit when El = Null_Iir;
               Chap4.Elab_Object_Value (El, Get_Default_Value (El));
            end loop;
         end if;

         --  FIXME: unconstrained signals with default values.
         List := Get_Port_List (Entity);
         if List /= Null_Iir_List then
            for I in Natural loop
               El := Get_Nth_Element (List, I);
               exit when El = Null_Iir;
               El_Type := Get_Type (El);
               if Get_Kind (El_Type)
                 in Iir_Kinds_Unconstrained_Array_Type_Definition
               then
                  Chap5.Elab_Unconstrained_Port (El, Get_Default_Value (El));
               end if;
            end loop;
         end if;

         Pop_Local_Factory;
      end Translate_Entity_Init;

      procedure Translate_Entity_Declaration (Entity : Iir)
      is
         Info : Block_Info_Acc;
         Interface_List : O_Inter_List;
         Instance : O_Lnode;
      begin
         Info := Add_Info (Entity, Kind_Block);
         Chap1.Start_Block_Decl (Entity);
         Push_Instance_Factory (Info.Block_Decls_Type);

         --  Entity name.
         Info.Block_Name := Add_Instance_Factory_Field
           (Wki_Name, Ghdl_Instance_Name_Type);

         --  generics, ports.
         Chap4.Translate_Generic_List (Get_Generic_List (Entity));
         Chap4.Translate_Port_List (Get_Port_List (Entity));

         Chap9.Translate_Block_Declarations (Entity, Entity);

         Pop_Instance_Factory (Info.Block_Decls_Type);

         --  Entity elaborator.
         Start_Procedure_Decl (Interface_List, Create_Identifier ("ELAB"),
                               Global_Storage);
         New_Interface_Decl (Interface_List, Instance, Wki_Instance,
                             Info.Block_Decls_Ptr_Type);
         Finish_Subprogram_Decl (Interface_List, Info.Block_Elab_Subprg);

         if Global_Storage = O_Storage_External then
            return;
         end if;

         --  Entity process subprograms.
         Chap9.Translate_Block_Subprograms (Entity, Entity);

         --  Elaborator Body.
         Start_Subprogram_Body (Info.Block_Elab_Subprg);
         Push_Local_Factory;
         Push_Scope (Info.Block_Decls_Type, Instance);
         --  Set entity name.
         New_Debug_Line_Stmt (Get_Line_Number (Entity));
         Chap6.Set_Instance_Name (Entity);
         Chap2.Elab_Dependence (Get_Parent (Entity));
         Chap9.Elab_Block_Declarations (Entity, Entity);
         Pop_Scope (Info.Block_Decls_Type);
         Pop_Local_Factory;
         Finish_Subprogram_Body;

         --  Default value if any.
         if False then --Is_Entity_Declaration_Top (Entity) then
            declare
               Init_Subprg : O_Dnode;
            begin
               Start_Procedure_Decl
                 (Interface_List, Create_Identifier ("_INIT"), Global_Storage);
               New_Interface_Decl (Interface_List, Instance, Wki_Instance,
                                   Info.Block_Decls_Ptr_Type);
               Finish_Subprogram_Decl (Interface_List, Init_Subprg);

               Start_Subprogram_Body (Init_Subprg);
               Translate_Entity_Init (Entity, Instance);
               Finish_Subprogram_Body;
            end;
         end if;
      end Translate_Entity_Declaration;

      --  Push scope for architecture ARCH via INSTANCE, and for its
      --  entity via the entity field of the instance.
      procedure Push_Architecture_Scope (Arch : Iir; Instance : O_Lnode)
      is
         Arch_Info : Block_Info_Acc;
         Entity : Iir;
         Entity_Info : Block_Info_Acc;
      begin
         Arch_Info := Get_Info (Arch);
         Entity := Get_Entity (Arch);
         Entity_Info := Get_Info (Entity);

         Push_Scope (Arch_Info.Block_Decls_Type, Instance);
         Push_Scope (Entity_Info.Block_Decls_Type,
                     Arch_Info.Block_Parent_Field, Arch_Info.Block_Decls_Type);
      end Push_Architecture_Scope;

      --  Pop scopes created by Push_Architecture_Scope.
      procedure Pop_Architecture_Scope (Arch : Iir)
      is
         Arch_Info : Block_Info_Acc;
         Entity : Iir;
         Entity_Info : Block_Info_Acc;
      begin
         Arch_Info := Get_Info (Arch);
         Entity := Get_Entity (Arch);
         Entity_Info := Get_Info (Entity);

         Pop_Scope (Entity_Info.Block_Decls_Type);
         Pop_Scope (Arch_Info.Block_Decls_Type);
      end Pop_Architecture_Scope;

      procedure Translate_Architecture_Declaration (Arch : Iir)
      is
         Info : Block_Info_Acc;
         Entity : Iir;
         Entity_Info : Block_Info_Acc;
         Interface_List : O_Inter_List;
         Constr : O_Assoc_List;
         Instance : O_Lnode;
         Size : O_Enode;
         Var_Arch_Instance : O_Lnode;
      begin
         Info := Add_Info (Arch, Kind_Block);
         Start_Block_Decl (Arch);
         Push_Instance_Factory (Info.Block_Decls_Type);

         Entity := Get_Entity (Arch);
         Entity_Info := Get_Info (Entity);
         Info.Block_Parent_Field := Add_Instance_Factory_Field
           (Get_Identifier ("ENTITY"), Entity_Info.Block_Decls_Type);

         --  Architecture name.
         Info.Block_Name := Add_Instance_Factory_Field
           (Wki_Name, Ghdl_Instance_Name_Type);

         Chap9.Translate_Block_Declarations (Arch, Arch);

         Pop_Instance_Factory (Info.Block_Decls_Type);

         --  Declare the constant containing the size of the instance.
         if Global_Storage = O_Storage_External then
            Size := O_Enode_Null;
         else
            Size := New_Sizeof (Info.Block_Decls_Type, Ghdl_Index_Type);
         end if;
         New_Const_Decl
           (Info.Block_Instance_Size, Create_Identifier ("INSTSIZE"),
            Global_Storage, Ghdl_Index_Type, Size);

         --  Elaborator.
         Start_Procedure_Decl
           (Interface_List, Create_Identifier ("ELAB"), Global_Storage);
         New_Interface_Decl
           (Interface_List, Instance, Wki_Instance,
            Entity_Info.Block_Decls_Ptr_Type);
         Finish_Subprogram_Decl (Interface_List, Info.Block_Elab_Subprg);

         if Global_Storage = O_Storage_External then
            return;
         end if;

         --  Create process subprograms.
         Push_Scope (Entity_Info.Block_Decls_Type,
                     Info.Block_Parent_Field, Info.Block_Decls_Type);
         Chap9.Translate_Block_Subprograms (Arch, Arch);
         Pop_Scope (Entity_Info.Block_Decls_Type);

         --  Elaborator body.
         Start_Subprogram_Body (Info.Block_Elab_Subprg);
         Push_Local_Factory;

         ----  Allocate memory for the instance.
         --New_Assign_Stmt
         --  (Instance, Gen_Alloc (Alloc_Memory,
         --                        New_Sizeof (Info.Block_Decls_Type,
         --                                    Ghdl_Index_Type),
         --                        Info.Block_Decls_Ptr_Type));
         New_Var_Decl (Var_Arch_Instance, Wki_Arch_Instance,
                       O_Storage_Local, Info.Block_Decls_Ptr_Type);
         New_Assign_Stmt
           (Var_Arch_Instance,
            New_Convert_Ov (New_Value (Instance), Info.Block_Decls_Ptr_Type));

         --  Call entity elaborator.
         Start_Association (Constr, Entity_Info.Block_Elab_Subprg);
         New_Association (Constr, New_Value (Instance));
         New_Procedure_Call (Constr);

         Push_Architecture_Scope (Arch, Var_Arch_Instance);

         New_Debug_Line_Stmt (Get_Line_Number (Arch));
         Chap6.Set_Instance_Name (Arch);
         Chap2.Elab_Dependence (Get_Parent (Arch));

         Chap9.Elab_Block_Declarations (Arch, Arch);
         --Chap6.Leave_Simple_Name (Ghdl_Leave_Architecture);

         Pop_Architecture_Scope (Arch);

         Pop_Local_Factory;
         Finish_Subprogram_Body;
      end Translate_Architecture_Declaration;

      procedure Translate_Component_Configuration_Decl
        (Cfg : Iir; Arch : Iir_Architecture_Declaration; Num : in out Natural)
      is
         Inter_List : O_Inter_List;
         Comp : Iir_Component_Declaration;
         Comp_Info : Comp_Info_Acc;
         Info : Config_Info_Acc;
         Instance : O_Lnode;
         Mark : Id_Mark_Type;

         Arch_Info : Block_Info_Acc;
         Arch_Param : O_Lnode;

         Block : Iir_Block_Configuration;
         Binding : Iir_Binding_Indication;
         Entity_Aspect : Iir;
         Conf_Override : Iir;
         Conf_Info : Config_Info_Acc;
      begin
         Binding := Get_Binding_Indication (Cfg);
         if Binding = Null_Iir then
            --  This is an unbound component configuration, since this is a
            --  no-op, it is not translated.
            return;
         end if;

         Entity_Aspect := Get_Entity_Aspect (Binding);

         Comp := Get_Component_Name (Cfg);
         Comp_Info := Get_Info (Comp);

         if Get_Kind (Cfg) = Iir_Kind_Component_Configuration then
            Block := Get_Block_Configuration (Cfg);
         else
            Block := Null_Iir;
         end if;

         if Block /= Null_Iir then
            Push_Identifier_Prefix (Mark, Get_Identifier (Comp), Num);
            Num := Num + 1;
            Translate_Configuration_Declaration (Cfg);
            Pop_Identifier_Prefix (Mark);
            Conf_Override := Cfg;
            Conf_Info := Get_Info (Cfg);
            Set_Back_End_Info (Cfg, System.Null_Address);
         else
            Conf_Info := null;
            Conf_Override := Null_Iir;
         end if;
         Info := Add_Info (Cfg, Kind_Config);

         Arch_Info := Get_Info (Arch);

         Start_Procedure_Decl
           (Inter_List, Create_Identifier (Comp, Num), O_Storage_Private);
         New_Interface_Decl (Inter_List, Instance, Wki_Instance,
                             Comp_Info.Comp_Ptr_Type);
         New_Interface_Decl (Inter_List, Arch_Param, Get_Identifier ("ARCH"),
                             Arch_Info.Block_Decls_Ptr_Type);
         Finish_Subprogram_Decl (Inter_List, Info.Config_Subprg);
         Num := Num + 1;

         --  Extract the entity/architecture.

         Start_Subprogram_Body (Info.Config_Subprg);
         Push_Local_Factory;

         Push_Architecture_Scope (Arch, Arch_Param);
         Push_Scope (Comp_Info.Comp_Type, Instance);

         if Conf_Info /= null then
            Set_Back_End_Info (Cfg, System.Null_Address);
            Set_Info (Cfg, Conf_Info);
         end if;
         Chap9.Translate_Entity_Instantiation
           (Entity_Aspect, Binding, Comp, Conf_Override);
         if Conf_Info /= null then
            Set_Back_End_Info (Cfg, System.Null_Address);
            Set_Info (Cfg, Info);
         end if;

         Pop_Scope (Comp_Info.Comp_Type);
         Pop_Architecture_Scope (Arch);
         Pop_Local_Factory;
         Finish_Subprogram_Body;
      end Translate_Component_Configuration_Decl;

      --  Create subprogram specifications for each configuration_specification
      --  in BLOCK_CONFIG and its sub-blocks.
      --  ARCH is the architecture being configured.
      --  NUM is an integer used to generate uniq names.
      procedure Translate_Block_Configuration_Decls
        (Block_Config : Iir_Block_Configuration;
         Arch : Iir_Architecture_Declaration;
         Num : in out Integer)
      is
         El : Iir;
         Mark : Id_Mark_Type;
         List : Iir_List;
         Blk : Iir;
      begin
         List := Get_Configuration_Item_List (Block_Config);
         for I in Natural loop
            El := Get_Nth_Element (List, I);
            exit when El = Null_Iir;
            case Get_Kind (El) is
               when Iir_Kind_Component_Configuration
                 | Iir_Kind_Configuration_Specification =>
                  Translate_Component_Configuration_Decl (El, Arch, Num);
               when Iir_Kind_Block_Configuration =>
                  Blk := Get_Block_From_Block_Specification
                    (Get_Block_Specification (El));
                  Push_Identifier_Prefix (Mark, Get_Identifier (Blk));
                  Translate_Block_Configuration_Decls (El, Arch, Num);
                  Pop_Identifier_Prefix (Mark);
               when others =>
                  Error_Kind ("translate_block_configuration_decls(1)", El);
            end case;
         end loop;
      end Translate_Block_Configuration_Decls;

      procedure Translate_Component_Configuration_Call
        (Cfg : Iir;
         Arch : Iir_Architecture_Declaration;
         Block_Info : Block_Info_Acc)
      is
         Cfg_Info : Config_Info_Acc;
         Arch_Info : Block_Info_Acc;
      begin
         if Get_Binding_Indication (Cfg) = Null_Iir then
            --  Unbound component configuration, nothing to do.
            return;
         end if;

         Cfg_Info := Get_Info (Cfg);
         Arch_Info := Get_Info (Arch);

         --  Call the subprogram for the instantiation list.
         declare
            List : Iir_List;
            El : Iir;
         begin
            List := Get_Instantiation_List (Cfg);
            for I in Natural loop
               El := Get_Nth_Element (List, I);
               exit when El = Null_Iir;
               case Get_Kind (El) is
                  when Iir_Kind_Component_Instantiation_Statement =>
                     declare
                        Assoc : O_Assoc_List;
                        Info : Block_Info_Acc;
                        V : O_Lnode;
                     begin
                        Info := Get_Info (El);
                        Start_Association (Assoc, Cfg_Info.Config_Subprg);
                        V := Get_Instance_Ref (Block_Info.Block_Decls_Type);
                        V := New_Selected_Element (V, Info.Block_Parent_Field);
                        New_Association
                          (Assoc, New_Address (V, Info.Block_Decls_Ptr_Type));
                        V := Get_Instance_Ref (Arch_Info.Block_Decls_Type);
                        New_Association
                          (Assoc,
                           New_Address (V, Arch_Info.Block_Decls_Ptr_Type));
                        New_Procedure_Call (Assoc);
                     end;
                  when others =>
                     Error_Kind ("translate_component_configuration", El);
               end case;
            end loop;
         end;
      end Translate_Component_Configuration_Call;

      procedure Translate_Block_Configuration_Calls
        (Block_Config : Iir_Block_Configuration;
         Arch : Iir_Architecture_Declaration;
         Info : Block_Info_Acc);

      procedure Translate_Generate_Block_Configuration_Calls
        (Block_Config : Iir_Block_Configuration;
         Arch : Iir_Architecture_Declaration;
         Parent_Info : Block_Info_Acc)
      is
         Spec : Iir;
         Block : Iir_Generate_Statement;
         Scheme : Iir;
         Info : Block_Info_Acc;

         --  Generate a call for a iterative generate block whose index is
         --  INDEX.
         --  FAILS is true if it is an error if the block is already
         --  configured.
         procedure Gen_Subblock_Call (Index : O_Enode; Fails : Boolean)
         is
            Var_Inst : O_Lnode;
            If_Blk : O_If_Block;
         begin
            Open_Temp;
            Var_Inst := Create_Temp (Info.Block_Decls_Ptr_Type);
            New_Assign_Stmt
              (Var_Inst,
               New_Address (New_Indexed_Element
                            (New_Acc_Value
                             (New_Selected_Element
                              (Get_Instance_Ref (Parent_Info.Block_Decls_Type),
                               Info.Block_Parent_Field)),
                             Index),
                            Info.Block_Decls_Ptr_Type));
            --  Configure only if not yet configured.
            Start_If_Stmt
              (If_Blk,
               New_Compare_Op (ON_Eq,
                               New_Value_Selected_Acc_Value
                               (Var_Inst, Info.Block_Configured_Field),
                               Ghdl_Bool_False_Node,
                               Ghdl_Bool_Type_Node));
            --  Mark the block as configured.
            New_Assign_Stmt
              (New_Selected_Acc_Value (Var_Inst, Info.Block_Configured_Field),
               Ghdl_Bool_True_Node);
            Push_Scope (Info.Block_Decls_Type, Var_Inst);
            Push_Scope_Via_Field_Ptr
              (Parent_Info.Block_Decls_Type,
               Info.Block_Origin_Field,
               Info.Block_Decls_Type);
            Translate_Block_Configuration_Calls (Block_Config, Arch, Info);
            Pop_Scope (Parent_Info.Block_Decls_Type);
            Pop_Scope (Info.Block_Decls_Type);

            if Fails then
               New_Else_Stmt (If_Blk);
               --  Already configured.
               Chap6.Gen_Program_Error (Block_Config);
            end if;

            Finish_If_Stmt (If_Blk);
            Close_Temp;
         end Gen_Subblock_Call;

         Type_Info : Type_Info_Acc;
         Iter_Type : Iir;
      begin
         Spec := Get_Block_Specification (Block_Config);
         Block := Get_Block_From_Block_Specification (Spec);
         Info := Get_Info (Block);
         Scheme := Get_Generation_Scheme (Block);
         if Get_Kind (Scheme) = Iir_Kind_Iterator_Declaration then
            Iter_Type := Get_Base_Type (Get_Type (Scheme));
            Type_Info := Get_Info (Iter_Type);
            case Get_Kind (Spec) is
               when Iir_Kind_Generate_Statement
                 | Iir_Kind_Selected_Name =>
                  --  Apply for all/remaining blocks.
                  declare
                     Var_I : O_Lnode;
                     Loop_Blk : O_Loop_Block;
                     Label : O_Snode;
                  begin
                     Start_Declare_Stmt;
                     New_Var_Decl (Var_I, Wki_I, O_Storage_Local,
                                   Ghdl_Index_Type);
                     Init_Var (Var_I);
                     Start_Loop_Stmt (Loop_Blk, Label);
                     Gen_Exit_When
                       (Label,
                        New_Compare_Op
                        (ON_Eq,
                         New_Value (Var_I),
                         New_Value (New_Selected_Element
                                    (New_Selected_Element
                                     (Get_Instance_Ref
                                      (Parent_Info.Block_Decls_Type),
                                      Info.Block_Range_Field),
                                     Type_Info.T.Range_Length)),
                         Ghdl_Bool_Type_Node));
                     --  Selected_name is for default configurations, so
                     --  program should not fail if a block is already
                     --  configured but continue silently.
                     Gen_Subblock_Call
                       (New_Value (Var_I),
                        Get_Kind (Spec) /= Iir_Kind_Selected_Name);
                     Inc_Var (Var_I);
                     Finish_Loop_Stmt (Loop_Blk);
                     Finish_Declare_Stmt;
                  end;
               when Iir_Kind_Indexed_Name =>
                  declare
                     Range_Ptr : O_Lnode;
                  begin
                     Open_Temp;
                     Range_Ptr := Create_Temp_Ptr
                       (Type_Info.T.Range_Ptr_Type,
                        New_Selected_Element
                        (Get_Instance_Ref (Parent_Info.Block_Decls_Type),
                         Info.Block_Range_Field));
                     Gen_Subblock_Call
                       (Chap6.Translate_Index_To_Offset
                        (Range_Ptr,
                         Chap7.Translate_Expression
                         (Get_First_Element (Get_Index_List (Spec)),
                          Iter_Type),
                         Iter_Type,
                         Spec),
                        True);
                     Close_Temp;
                  end;
               when Iir_Kind_Slice_Name =>
                  declare
                     Range_Ptr : O_Lnode;
                     Slice : O_Lnode;
                     Slice_Ptr : O_Lnode;
                     Left, Right : O_Lnode;
                     Index : O_Lnode;
                     High : O_Lnode;
                     If_Blk : O_If_Block;
                     Loop_Blk : O_Loop_Block;
                     Label : O_Snode;
                  begin
                     Open_Temp;
                     Range_Ptr := Create_Temp_Ptr
                       (Type_Info.T.Range_Ptr_Type,
                        New_Selected_Element
                        (Get_Instance_Ref (Parent_Info.Block_Decls_Type),
                         Info.Block_Range_Field));
                     Slice := Create_Temp (Type_Info.T.Range_Type);
                     Slice_Ptr := Create_Temp_Ptr
                       (Type_Info.T.Range_Ptr_Type, Slice);
                     Chap7.Translate_Discrete_Range_Ptr
                       (Slice_Ptr, Get_Suffix (Spec));
                     Left := Create_Temp_Init
                       (Ghdl_Index_Type,
                        Chap6.Translate_Index_To_Offset
                        (Range_Ptr,
                         New_Value (New_Selected_Element
                                    (Slice, Type_Info.T.Range_Left)),
                         Iter_Type,
                         Spec));
                     Right := Create_Temp_Init
                       (Ghdl_Index_Type,
                        Chap6.Translate_Index_To_Offset
                        (Range_Ptr,
                         New_Value (New_Selected_Element
                                    (Slice, Type_Info.T.Range_Right)),
                         Iter_Type,
                         Spec));
                     Index := Create_Temp (Ghdl_Index_Type);
                     High := Create_Temp (Ghdl_Index_Type);
                     Start_If_Stmt
                       (If_Blk,
                        New_Compare_Op (ON_Eq,
                                        New_Value_Selected_Acc_Value
                                        (Range_Ptr, Type_Info.T.Range_Dir),
                                        New_Value
                                        (New_Selected_Element
                                         (Slice, Type_Info.T.Range_Dir)),
                                        Ghdl_Bool_Type_Node));
                     --  Same direction, so left to right.
                     New_Assign_Stmt (Index, New_Value (Left));
                     New_Assign_Stmt (High, New_Value (Right));
                     New_Else_Stmt (If_Blk);
                     --  Opposite direction, so right to left.
                     New_Assign_Stmt (Index, New_Value (Right));
                     New_Assign_Stmt (High, New_Value (Left));
                     Finish_If_Stmt (If_Blk);

                     --  Loop.
                     Start_Loop_Stmt (Loop_Blk, Label);
                     Gen_Exit_When
                       (Label, New_Compare_Op (ON_Gt,
                                               New_Value (Index),
                                               New_Value (High),
                                               Ghdl_Bool_Type_Node));
                     Open_Temp;
                     Gen_Subblock_Call (New_Value (Index), True);
                     Close_Temp;
                     Inc_Var (Index);
                     Finish_Loop_Stmt (Loop_Blk);
                     Close_Temp;
                  end;
               when others =>
                  Error_Kind
                    ("translate_generate_block_configuration_calls", Spec);
            end case;
         else
            --  Conditional generate statement.
            declare
               Var : O_Lnode;
               If_Blk : O_If_Block;
            begin
               --  Configure the block only if it was created.
               Open_Temp;
               Var := Create_Temp_Init
                 (Info.Block_Decls_Ptr_Type,
                  New_Value (New_Selected_Element
                             (Get_Instance_Ref (Parent_Info.Block_Decls_Type),
                              Info.Block_Parent_Field)));
               Start_If_Stmt
                 (If_Blk,
                  New_Compare_Op (ON_Neq,
                                  New_Value (Var),
                                  New_Null_Access (Info.Block_Decls_Ptr_Type),
                                  Ghdl_Bool_Type_Node));
               Push_Scope (Info.Block_Decls_Type, Var);
               Push_Scope_Via_Field_Ptr (Parent_Info.Block_Decls_Type,
                                         Info.Block_Origin_Field,
                                         Info.Block_Decls_Type);
               Translate_Block_Configuration_Calls (Block_Config, Arch, Info);
               Pop_Scope (Parent_Info.Block_Decls_Type);
               Pop_Scope (Info.Block_Decls_Type);
               Finish_If_Stmt (If_Blk);
               Close_Temp;
            end;
         end if;
      end Translate_Generate_Block_Configuration_Calls;

      procedure Translate_Block_Configuration_Calls
        (Block_Config : Iir_Block_Configuration;
         Arch : Iir_Architecture_Declaration;
         Info : Block_Info_Acc)
      is
         El : Iir;
         List : Iir_List;
      begin
         List := Get_Configuration_Item_List (Block_Config);

         for I in Natural loop
            El := Get_Nth_Element (List, I);
            exit when El = Null_Iir;
            case Get_Kind (El) is
               when Iir_Kind_Component_Configuration
                 | Iir_Kind_Configuration_Specification =>
                  Translate_Component_Configuration_Call (El, Arch, Info);
               when Iir_Kind_Block_Configuration =>
                  declare
                     Block : Iir;
                     Block_Info : Block_Info_Acc;
                  begin
                     Block := Get_Block_Specification (El);
                     if Get_Kind (Block) = Iir_Kind_Block_Statement then
                        Block_Info := Get_Info (Block);
                        Push_Scope (Block_Info.Block_Decls_Type,
                                    Block_Info.Block_Parent_Field,
                                    Info.Block_Decls_Type);
                        Translate_Block_Configuration_Calls
                          (El, Arch, Block_Info);
                        Pop_Scope (Block_Info.Block_Decls_Type);
                     else
                        Translate_Generate_Block_Configuration_Calls
                          (El, Arch, Info);
                     end if;
                  end;
               when others =>
                  Error_Kind ("translate_block_configuration_calls(2)", El);
            end case;
         end loop;
      end Translate_Block_Configuration_Calls;

      procedure Translate_Configuration_Declaration (Config : Iir)
      is
         Interface_List : O_Inter_List;
         Block_Config : Iir_Block_Configuration;
         Arch : Iir_Architecture_Declaration;
         Arch_Info : Block_Info_Acc;
         Config_Info : Config_Info_Acc;
         Instance : O_Lnode;
         Num : Natural;
      begin
         Config_Info := Add_Info (Config, Kind_Config);
         Block_Config := Get_Block_Configuration (Config);
         Arch := Get_Block_Specification (Block_Config);
         Arch_Info := Get_Info (Arch);

         --  Configurator.
         Start_Procedure_Decl
           (Interface_List, Create_Identifier, Global_Storage);
         New_Interface_Decl (Interface_List, Instance, Wki_Instance,
                             Arch_Info.Block_Decls_Ptr_Type);
         Finish_Subprogram_Decl (Interface_List, Config_Info.Config_Subprg);

         if Global_Storage = O_Storage_External then
            return;
         end if;

         --  Declare subprograms for configuration.
         Num := 0;
         Translate_Block_Configuration_Decls (Block_Config, Arch, Num);

         --  Body.
         Start_Subprogram_Body (Config_Info.Config_Subprg);
         Push_Local_Factory;

         Push_Architecture_Scope (Arch, Instance);

         --  FIXME: translate declarations.

         Translate_Block_Configuration_Calls (Block_Config, Arch, Arch_Info);

         Pop_Architecture_Scope (Arch);
         Pop_Local_Factory;
         Finish_Subprogram_Body;
      end Translate_Configuration_Declaration;
   end Chap1;

   package body Chap2 is
      procedure Elab_Package (Spec : Iir_Package_Declaration);

      type Name_Xlat_Array is array (Name_Id range <>) of Name_Id;
      Operator_Name_Xlat : Name_Xlat_Array (Std_Names.Name_Id_Operators) :=
        (others => Null_Identifier);

      function Do_Operator_Name_Xlat (Id : Name_Id) return Name_Id
      is
         use Std_Names;

         type Name_String_Xlat_Array is array (Name_Id range <>) of
           String (1 .. 4);
         Operator_String_Xlat : constant
           Name_String_Xlat_Array (Name_Id_Operators) :=
             (Name_Op_Equality => "OPEq",
              Name_Op_Inequality => "OPNe",
              Name_Op_Less => "OPLt",
              Name_Op_Less_Equal => "OPLe",
              Name_Op_Greater => "OPGt",
              Name_Op_Greater_Equal => "OPGe",
              Name_Op_Plus => "OPPl",
              Name_Op_Minus => "OPMi",
              Name_Op_Mul => "OPMu",
              Name_Op_Div => "OPDi",
              Name_Op_Exp => "OPEx",
              Name_Op_Concatenation => "OPCc");

         Res : Name_Id;
      begin
         Res := Name_Table.Get_Identifier (Operator_String_Xlat (Id));
         Operator_Name_Xlat (Id) := Res;
         return Res;
      end Do_Operator_Name_Xlat;

      --  Set the identifier prefix with the subprogram identifier and
      --  overload number if any.
      procedure Push_Subprg_Identifier
        (Spec : Iir; Info : Subprg_Info_Acc; Mark : out Id_Mark_Type)
      is
         Id : Name_Id;
      begin
         --  FIXME: name_shift_operators, name_logical_operators,
         --   name_word_operators, name_mod, name_rem
         Id := Get_Identifier (Spec);
         if Id in Std_Names.Name_Id_Operators then
            if Operator_Name_Xlat (Id) /= Null_Identifier then
               Id := Operator_Name_Xlat (Id);
            else
               Id := Do_Operator_Name_Xlat (Id);
            end if;
         end if;

         Push_Identifier_Prefix (Mark, Id, Info.Subprg_Overload);
      end Push_Subprg_Identifier;

      procedure Translate_Subprogram_Specification (Spec : Iir; Block : Iir)
      is
         Interfaces : Iir_Interface_List;
         Inter : Iir;
         Inter_Type : Iir;
         Inter_Kind : Iir_Kind;
         Info : Subprg_Info_Acc;
         Arg_Info : Ortho_Info_Acc;
         Tinfo : Type_Info_Acc;
         Interface_List : O_Inter_List;
         Has_Result_Record : Boolean;
         El_List : O_Element_List;
         Mark : Id_Mark_Type;
         Is_Func : Boolean;
         Rtype : Iir;
         Mode : Object_Kind_Type;
         Interface_Type : O_Tnode;
      begin
         Info := Get_Info (Spec);
         Info.Res_Interface := O_Lnode_Null;
         Interfaces := Get_Interface_Declaration (Spec);
         Is_Func := Get_Kind (Spec) in Iir_Kinds_Function_Specification;

         --  Set the identifier prefix with the subprogram identifier and
         --  overload number if any.
         Push_Subprg_Identifier (Spec, Info, Mark);

         --  Translate interface types.
         for I in Natural loop
            Inter := Get_Nth_Element (Interfaces, I);
            exit when Inter = Null_Iir;
            Inter_Type := Get_Type (Inter);
            Chap3.Translate_Object_Subtype (Inter);
         end loop;

         if Is_Func then
            --  If the result of a function is a composite type for ortho,
            --  the result is allocated by the caller and an access to it is
            --  given to the function.
            Rtype := Get_Return_Type (Spec);
            Info.Use_Stack2 := False;
            Tinfo := Get_Info (Rtype);

            if Is_Composite (Tinfo) then
               Start_Procedure_Decl
                 (Interface_List, Create_Identifier, Global_Storage);
               New_Interface_Decl
                 (Interface_List, Info.Res_Interface,
                  Get_Identifier ("RESULT"),
                  Tinfo.Ortho_Ptr_Type (Mode_Value));
               --  Furthermore, if the result type is unconstrained, the
               --  function will allocate it on a secondary stack.
               if Get_Kind (Rtype)
                 in Iir_Kinds_Unconstrained_Array_Type_Definition
               then
                  Info.Use_Stack2 := True;
               end if;
            else
               Start_Function_Decl
                 (Interface_List, Create_Identifier, Global_Storage,
                  Tinfo.Ortho_Type (Mode_Value));
            end if;
         else
            --  Create info for each interface of the procedure.
            --  For parameters passed via copy and that needs a copy-out,
            --  gather them in a record.  An access to the record is then
            --  passed to the procedure.
            Has_Result_Record := False;
            for I in Natural loop
               Inter := Get_Nth_Element (Interfaces, I);
               exit when Inter = Null_Iir;
               Arg_Info := Add_Info (Inter, Kind_Interface);
               Inter_Type := Get_Type (Inter);
               Inter_Kind := Get_Kind (Inter_Type);
               Tinfo := Get_Info (Inter_Type);
               if Get_Kind (Inter) = Iir_Kind_Variable_Interface_Declaration
                 and then Get_Mode (Inter) in Iir_Out_Mode .. Iir_Inout_Mode
                 and then Tinfo.Type_Mode not in Type_Mode_By_Ref
               then
                  --  This interface is done via the result record.
                  if not Has_Result_Record then
                     --  Create the record.
                     Start_Record_Type (El_List);
                     Has_Result_Record := True;
                  end if;
                  --  Add a field to the record.
                  Tinfo := Get_Info (Inter_Type);
                  New_Record_Field (El_List, Arg_Info.Interface_Field,
                                    Create_Identifier_Without_Prefix (Inter),
                                    Tinfo.Ortho_Type (Mode_Value));
               else
                  Arg_Info.Interface_Field := O_Fnode_Null;
               end if;
            end loop;
            if Has_Result_Record then
               --  Declare the record type and an access to the record.
               Finish_Record_Type (El_List, Info.Res_Record_Type);
               New_Type_Decl (Create_Identifier ("RESTYPE"),
                              Info.Res_Record_Type);
               Info.Res_Record_Ptr := New_Access_Type (Info.Res_Record_Type);
               New_Type_Decl (Create_Identifier ("RESPTR"),
                              Info.Res_Record_Ptr);
            end if;

            Start_Procedure_Decl
              (Interface_List, Create_Identifier, Global_Storage);

            if Has_Result_Record then
               --  Add the record parameter.
               New_Interface_Decl (Interface_List, Info.Res_Interface,
                                   Get_Identifier ("RESULT"),
                                   Info.Res_Record_Ptr);
            end if;
         end if;

         --  Instance parameter if any.
         Chap2.Create_Subprg_Instance (Interface_List, Spec, Block);

         for I in Natural loop
            Inter := Get_Nth_Element (Interfaces, I);
            exit when Inter = null;
            Arg_Info := Get_Info (Inter);
            --  Translate the interface if not in the record.
            if Is_Func or else Arg_Info.Interface_Field = O_Fnode_Null then
               case Get_Kind (Inter) is
                  when Iir_Kind_Constant_Interface_Declaration
                    | Iir_Kind_Variable_Interface_Declaration
                    | Iir_Kind_File_Interface_Declaration =>
                     Mode := Mode_Value;
                  when Iir_Kind_Signal_Interface_Declaration =>
                     Mode := Mode_Signal;
                  when others =>
                     Error_Kind
                       ("translate_subprogram_specification(2)", Inter);
               end case;
               Tinfo := Get_Info (Get_Type (Inter));
               if Is_Func then
                  Arg_Info := Add_Info (Inter, Kind_Interface);
                  Arg_Info.Interface_Field := O_Fnode_Null;
               end if;
               case Tinfo.Type_Mode is
                  when Type_Mode_Unknown =>
                     raise Internal_Error;
                  when Type_Mode_By_Value =>
                     Interface_Type := Tinfo.Ortho_Type (Mode);
                  when Type_Mode_By_Copy
                    | Type_Mode_By_Ref =>
                     Interface_Type := Tinfo.Ortho_Ptr_Type (Mode);
               end case;
               New_Interface_Decl
                 (Interface_List, Arg_Info.Interface_Node,
                  Create_Identifier_Without_Prefix (Inter),
                  Interface_Type);
            else
               --  Parameter is passed by the result record.
               Arg_Info.Interface_Node := Info.Res_Interface;
            end if;
         end loop;
         Finish_Subprogram_Decl (Interface_List, Info.Ortho_Func);

         Save_Local_Identifier (Info.Subprg_Local_Id);
         Pop_Identifier_Prefix (Mark);
      end Translate_Subprogram_Specification;

      procedure Translate_Subprogram_Body (Subprg : Iir)
      is
         Spec : Iir;
         Func_Decl : O_Dnode;
         Info : Ortho_Info_Acc;
         Old_Subprogram : Iir;
         Mark : Id_Mark_Type;
         Decls : Iir_List;
      begin
         Spec := Get_Subprogram_Spec (Subprg);
         Info := Get_Info (Spec);
         Func_Decl := Info.Ortho_Func;

         --  Set the identifier prefix with the subprogram identifier and
         --  overload number if any.
         Push_Subprg_Identifier (Spec, Info, Mark);
         Restore_Local_Identifier (Info.Subprg_Local_Id);

         Start_Subprogram_Body (Func_Decl);

         Start_Subprg_Instance_Use (Spec);

         Push_Local_Factory;

         Decls := Get_Subprogram_Declaration (Subprg);
         Chap4.Translate_Declaration_List (Decls, Null_Iir_List);
         Chap4.Translate_Declaration_List_Subprograms (Decls, Null_Iir);

         Chap4.Elab_Declaration_List (Decls, Subprg);

         Old_Subprogram := Current_Subprogram;
         Current_Subprogram := Spec;
         Chap8.Translate_Statements_List (Get_Subprogram_Statements (Subprg));
         Current_Subprogram := Old_Subprogram;

         Pop_Local_Factory;

         Finish_Subprg_Instance_Use (Spec);

         Finish_Subprogram_Body;

         Pop_Identifier_Prefix (Mark);
      end Translate_Subprogram_Body;

      procedure Translate_Package_Declaration (Decl : Iir_Package_Declaration)
      is
         Decl_List : Iir_List;
         Info : Ortho_Info_Acc;
         I_List : O_Inter_List;
         --Storage : O_Storage;
      begin
         Decl_List := Get_Declaration_List (Decl);
         Chap4.Translate_Declaration_List (Decl_List, Null_Iir_List);
         Chap4.Translate_Declaration_List_Subprograms (Decl_List, Null_Iir);

--          if Chap10.Global_Storage = O_Storage_Public
--            and then not Get_Need_Body (Decl)
--          then
--             Storage := O_Storage_Public;
--          else
--             Storage := O_Storage_External;
--          end if;

         Info := Add_Info (Decl, Kind_Package);

         Start_Procedure_Decl
           (I_List, Create_Identifier ("ELAB_SPEC"), Global_Storage);
         Finish_Subprogram_Decl (I_List, Info.Package_Elab_Spec_Subprg);

         Start_Procedure_Decl
           (I_List, Create_Identifier ("ELAB_BODY"), Global_Storage);
         Finish_Subprogram_Decl (I_List, Info.Package_Elab_Body_Subprg);

         New_Var_Decl (Info.Package_Elab_Var, Create_Identifier ("ELABORATED"),
                       Chap10.Global_Storage, Ghdl_Bool_Type_Node);

         New_Var_Decl (Info.Package_Name, Create_Identifier ("NAME"),
                       Global_Storage, Ghdl_Instance_Name_Type);

         if Global_Storage = O_Storage_Public then
            Elab_Package (Decl);
         end if;
         Save_Local_Identifier (Info.Package_Local_Id);
      end Translate_Package_Declaration;

      procedure Translate_Package_Body (Decl : Iir_Package_Body)
      is
         Decl_List : Iir_List;
         Pkg : Iir_Package_Declaration;
      begin
         Pkg := Get_Package (Decl);
         Restore_Local_Identifier (Get_Info (Pkg).Package_Local_Id);
         Decl_List := Get_Declaration_List (Decl);
         Chap4.Translate_Declaration_List (Decl_List,
                                           Get_Declaration_List (Pkg));
         Chap4.Translate_Declaration_List_Subprograms (Decl_List, Null_Iir);
         Elab_Package_Body (Pkg, Decl);
      end Translate_Package_Body;

      procedure Elab_Package (Spec : Iir_Package_Declaration)
      is
         Info : Ortho_Info_Acc;
         If_Blk : O_If_Block;
      begin
         Info := Get_Info (Spec);
         Start_Subprogram_Body (Info.Package_Elab_Spec_Subprg);
         Push_Local_Factory;

         Elab_Dependence (Get_Parent (Spec));
         Chap4.Elab_Declaration_List (Get_Declaration_List (Spec), Spec);

         Pop_Local_Factory;
         Finish_Subprogram_Body;
      end Elab_Package;

      procedure Elab_Package_Body (Spec : Iir_Package_Declaration; Bod : Iir)
      is
         Info : Ortho_Info_Acc;
         If_Blk : O_If_Block;
         Constr : O_Assoc_List;
      begin
         Info := Get_Info (Spec);
         Start_Subprogram_Body (Info.Package_Elab_Body_Subprg);
         Push_Local_Factory;

         --  If the package was already elaborated, return now,
         --  else mark the package as elaborated.
         Start_If_Stmt (If_Blk, New_Value (Info.Package_Elab_Var));
         New_Return_Stmt;
         New_Else_Stmt (If_Blk);
         New_Assign_Stmt (Info.Package_Elab_Var, Ghdl_Bool_True_Node);
         Finish_If_Stmt (If_Blk);

         Chap6.Set_Instance_Name (Spec);
         Chap6.Link_Instance_Name (Null_Iir, Spec);

         --  Elab Spec.
         Start_Association (Constr, Info.Package_Elab_Spec_Subprg);
         New_Procedure_Call (Constr);

         if Bod /= Null_Iir then
            Elab_Dependence (Get_Parent (Bod));
            Chap4.Elab_Declaration_List (Get_Declaration_List (Bod), Bod);
         end if;

         Pop_Local_Factory;
         Finish_Subprogram_Body;
      end Elab_Package_Body;

      procedure Elab_Dependence (Design_Unit: Iir_Design_Unit)
      is
         Depend_List: Iir_Design_Unit_List;
         Design: Iir;
         Library_Unit: Iir;
         Info : Ortho_Info_Acc;
         If_Blk : O_If_Block;
         Constr : O_Assoc_List;
      begin
         Depend_List := Get_Dependence_List (Design_Unit);

         for I in Natural loop
            Design := Get_Nth_Element (Depend_List, I);
            exit when Design = Null_Iir;
            if Get_Kind (Design) = Iir_Kind_Design_Unit then
               Library_Unit := Get_Library_Unit (Design);
               case Get_Kind (Library_Unit) is
                  when Iir_Kind_Package_Declaration =>
                     if Library_Unit /= Standard_Package then
                        Info := Get_Info (Library_Unit);
                        Start_If_Stmt
                          (If_Blk, New_Monadic_Op
                           (ON_Not, New_Value (Info.Package_Elab_Var)));
                        -- Elaborates only non-elaborated packages.
                        Start_Association (Constr,
                                           Info.Package_Elab_Body_Subprg);
                        New_Procedure_Call (Constr);
                        Finish_If_Stmt (If_Blk);
                     end if;
                  when Iir_Kind_Entity_Declaration =>
                     --  FIXME: architecture already elaborates its entity.
                     null;
                  when Iir_Kind_Configuration_Declaration =>
                     null;
                  when others =>
                     Error_Kind ("elab_dependence", Library_Unit);
               end case;
            end if;
         end loop;
      end Elab_Dependence;

      procedure Clear_Instance_Data (Data : out Instance_Data) is
      begin
         Data.Instance_Block := Null_Iir;
         Data.Instance_Param := O_Lnode_Null;
      end Clear_Instance_Data;

      procedure Create_Subprg_Instance (Interfaces : in out O_Inter_List;
                                        Data : out Instance_Data;
                                        Block : Iir)
      is
         Block_Info : Block_Info_Acc;
      begin
         if Block /= Null_Iir then
            Block_Info := Get_Info (Block);
            Data.Instance_Block := Block;
            New_Interface_Decl (Interfaces, Data.Instance_Param, Wki_Instance,
                                Block_Info.Block_Decls_Ptr_Type);
         else
            Clear_Instance_Data (Data);
         end if;
      end Create_Subprg_Instance;

      procedure Create_Subprg_Instance (Interfaces : in out O_Inter_List;
                                        Subprg : Iir;
                                        Block : Iir)
      is
      begin
         Create_Subprg_Instance
           (Interfaces, Get_Info (Subprg).Subprg_Instance, Block);
      end Create_Subprg_Instance;

      procedure Add_Subprg_Instance_Assoc
        (Assoc : in out O_Assoc_List; Data : Instance_Data)
      is
      begin
         if Data.Instance_Block /= Null_Iir then
            New_Association (Assoc, Get_Instance_Access (Data.Instance_Block));
         end if;
      end Add_Subprg_Instance_Assoc;

      procedure Add_Subprg_Instance_Assoc
        (Assoc : in out O_Assoc_List; Subprg : Iir)
      is
      begin
         Add_Subprg_Instance_Assoc (Assoc, Get_Info (Subprg).Subprg_Instance);
      end Add_Subprg_Instance_Assoc;

      procedure Start_Subprg_Instance_Use (Data : Instance_Data)
      is
      begin
         if Data.Instance_Block /= Null_Iir then
            Push_Scope (Get_Info (Data.Instance_Block).Block_Decls_Type,
                        Data.Instance_Param);
         end if;
      end Start_Subprg_Instance_Use;

      procedure Start_Subprg_Instance_Use (Subprg : Iir)
      is
      begin
         Start_Subprg_Instance_Use (Get_Info (Subprg).Subprg_Instance);
      end Start_Subprg_Instance_Use;

      procedure Finish_Subprg_Instance_Use (Data : Instance_Data)
      is
      begin
         if Data.Instance_Block /= Null_Iir then
            Pop_Scope (Get_Info (Data.Instance_Block).Block_Decls_Type);
         end if;
      end Finish_Subprg_Instance_Use;

      procedure Finish_Subprg_Instance_Use (Subprg : Iir)
      is
      begin
         Finish_Subprg_Instance_Use (Get_Info (Subprg).Subprg_Instance);
      end Finish_Subprg_Instance_Use;
   end Chap2;

   package body Chap3 is
      procedure Translate_Type_Definition
        (Def : Iir; With_Vars : Boolean := True);

      function Create_Static_Type_Definition_Type_Range (Def : Iir)
        return O_Enode;
      procedure Create_Scalar_Type_Range (Def : Iir; Target : O_Lnode);

      type Type_Completion_Kind is
         (Incomplete_Type, Completed_Type, Full_Type);

      --  Finish a type definition: declare the type, define and declare a
      --   pointer to the type.
      procedure Finish_Type_Definition
        (Info : Type_Info_Acc;
         Is_Scalar : Boolean;
         Completion : Type_Completion_Kind := Full_Type)
      is
      begin
         if Completion >= Completed_Type then
            if Info.Type_Mode = Type_Mode_Unknown then
               --  Internal check: type_mode field must be set.
               raise Internal_Error;
            end if;
         end if;

         --  Declare the type.
         if Completion /= Completed_Type then
            New_Type_Decl
              (Create_Identifier, Info.Ortho_Type (Mode_Value));
         end if;
         if Completion = Incomplete_Type then
            return;
         end if;

         --  Create an access to the type and declare it.
         Info.Ortho_Ptr_Type (Mode_Value) :=
           New_Access_Type (Info.Ortho_Type (Mode_Value));
         New_Type_Decl (Create_Identifier ("PTR"),
                        Info.Ortho_Ptr_Type (Mode_Value));

         --  Signal type.
         if Info.Type_Mode in Type_Mode_Scalar then
            Info.Ortho_Type (Mode_Signal) :=
              New_Access_Type (Info.Ortho_Type (Mode_Value));
         end if;
         if Info.Ortho_Type (Mode_Signal) /= O_Tnode_Null then
            New_Type_Decl (Create_Identifier ("SIG"),
                           Info.Ortho_Type (Mode_Signal));
         end if;

         --  Signal pointer type.
         if Info.Type_Mode in Type_Mode_Composite
           and then Info.Ortho_Type (Mode_Signal) /= O_Tnode_Null
         then
            Info.Ortho_Ptr_Type (Mode_Signal) :=
              New_Access_Type (Info.Ortho_Type (Mode_Signal));
            New_Type_Decl (Create_Identifier ("SIGPTR"),
                           Info.Ortho_Ptr_Type (Mode_Signal));
         else
            Info.Ortho_Ptr_Type (Mode_Signal) := O_Tnode_Null;
         end if;
      end Finish_Type_Definition;

      --  Return the address of the DEF type desc.
      function Get_Type_Desc_Addr (A_Type : Iir) return O_Enode is
      begin
         return New_Unchecked_Address (Get_Var (Get_Info (A_Type).Type_Desc),
                                       Ghdl_Desc_Ptr_Type);
      end Get_Type_Desc_Addr;

      function Create_Static_Scalar_Subtype_Desc (Def : Iir; Kind : O_Enode)
        return O_Enode
      is
         Info : Type_Info_Acc;
         List : O_Record_Aggr_List;
         Val : O_Enode;
         Range_O : O_Enode;
      begin
         Info := Get_Info (Def);
         Range_O := New_Unchecked_Address
           (Get_Var (Info.T.Range_Var), Const_Ptr_Type_Node);
         Start_Record_Aggr (List, Ghdl_Desc_Scalar_Subtype_Node);
         New_Record_Aggr_El (List, Kind);
         New_Record_Aggr_El (List, Get_Type_Desc_Addr (Get_Base_Type (Def)));
         New_Record_Aggr_El (List, Range_O);
         Finish_Record_Aggr (List, Val);
         return Val;
      end Create_Static_Scalar_Subtype_Desc;

      procedure Create_Scalar_Subtype_Desc
        (Def : Iir; Name : Name_Id; Kind : O_Enode)
      is
         Info : Type_Info_Acc;
         Var : Var_Acc;
      begin
         Info := Get_Info (Def);
         Var := Info.Type_Desc;
         New_Assign_Stmt
           (New_Selected_Element (Get_Var (Var),
                                  Ghdl_Desc_Scalar_Kind_Node),
            Kind);
         New_Assign_Stmt
           (New_Selected_Element (Get_Var (Var),
                                  Ghdl_Desc_Scalar_Basetype_Node),
            Get_Type_Desc_Addr (Get_Base_Type (Def)));
         New_Assign_Stmt
           (New_Selected_Element (Get_Var (Var),
                                  Ghdl_Desc_Scalar_Range_Node),
            New_Unchecked_Address
            (Get_Var (Info.T.Range_Var), Const_Ptr_Type_Node));
      end Create_Scalar_Subtype_Desc;

      procedure Create_Size_Var (Info : Type_Info_Acc) is
      begin
         Info.C := new Complex_Type_Info;
         Info.C.Size_Var (Mode_Value) := Create_Var
           (Create_Var_Identifier ("SIZE"), Ghdl_Index_Type);
         Info.C.Size_Var (Mode_Signal) := Create_Var
           (Create_Var_Identifier ("SIGSIZE"), Ghdl_Index_Type);
      end Create_Size_Var;

      --  A builder set internal fields of object pointed by BASE_PTR, using
      --  memory from BASE_PTR and returns a pointer to the next memory byte
      --  to be used.
      procedure Create_Builder_Subprogram_Decl (Info : Type_Info_Acc;
                                                Name : Name_Id;
                                                Block : Iir;
                                                Kind : Object_Kind_Type)
      is
         Interface_List : O_Inter_List;
         Ident : O_Ident;
         Ptype : O_Tnode;
      begin
         case Kind is
            when Mode_Value =>
               Ident := Create_Identifier (Name, "_BUILDER");
            when Mode_Signal =>
               Ident := Create_Identifier (Name, "_SIGBUILDER");
         end case;
         Start_Function_Decl
           (Interface_List, Ident, Global_Storage, Char_Ptr_Type_Node);
         Chap2.Create_Subprg_Instance
           (Interface_List, Info.C.Builder_Instance (Kind), Block);
         case Info.Type_Mode is
            when Type_Mode_Fat_Array =>
               Ptype := Info.T.Base_Ptr_Type (Kind);
            when Type_Mode_Record =>
               Ptype := Info.Ortho_Ptr_Type (Kind);
            when others =>
               raise Internal_Error;
         end case;
         New_Interface_Decl
           (Interface_List, Info.C.Builder_Base_Param (Kind),
            Get_Identifier ("base_ptr"), Ptype);
         --  Add parameter for array bounds.
         if Info.Type_Mode in Type_Mode_Arrays then
            New_Interface_Decl
              (Interface_List, Info.C.Builder_Bound_Param (Kind),
               Get_Identifier ("bound"), Info.T.Bounds_Ptr_Type);
         end if;
         Finish_Subprogram_Decl (Interface_List, Info.C.Builder_Func (Kind));
      end Create_Builder_Subprogram_Decl;

      function Gen_Call_Type_Builder (Var_Ptr : O_Lnode;
                                      Var_Type : Iir;
                                      Kind : Object_Kind_Type)
        return O_Enode
      is
         Tinfo : Type_Info_Acc;
         Binfo : Type_Info_Acc;
         Assoc : O_Assoc_List;
      begin
         Tinfo := Get_Info (Var_Type);
         --  Build the field
         Binfo := Get_Info (Get_Base_Type (Var_Type));
         Start_Association (Assoc, Binfo.C.Builder_Func (Kind));
         Chap2.Add_Subprg_Instance_Assoc
           (Assoc, Binfo.C.Builder_Instance (Kind));
         case Tinfo.Type_Mode is
            when Type_Mode_Record
              | Type_Mode_Array =>
               New_Association (Assoc, New_Value (Var_Ptr));
            when Type_Mode_Ptr_Array =>
               --New_Association (Assoc, New_Value (New_Acc_Value (Var_Ptr)));
               New_Association (Assoc, New_Value (Var_Ptr));
            when Type_Mode_Fat_Array =>
               --  Note: a fat array can only be at the top of a complex type;
               --  the bounds must have been set.
               New_Association (Assoc, New_Value_Selected_Acc_Value
                                (Var_Ptr, Tinfo.T.Base_Field (Kind)));
            when others =>
               raise Internal_Error;
         end case;
         case Tinfo.Type_Mode is
            when Type_Mode_Array
              | Type_Mode_Ptr_Array =>
               New_Association
                 (Assoc,
                  Get_Array_Bounds_Ptr (O_Lnode_Null, Var_Type, Kind));
            when Type_Mode_Fat_Array =>
               New_Association
                 (Assoc, Get_Array_Ptr_Bounds_Ptr (Var_Ptr, Var_Type, Kind));
            when Type_Mode_Record =>
               null;
            when others =>
               raise Internal_Error;
         end case;
         return New_Function_Call (Assoc);
      end Gen_Call_Type_Builder;

      procedure Gen_Call_Type_Builder (Var : Mnode; Var_Type : Iir)
      is
         Mem : O_Lnode;
         V : Mnode;
      begin
         Open_Temp;
         V := Stabilize (Var);
         Mem := Create_Temp (Char_Ptr_Type_Node);
         New_Assign_Stmt
           (Mem,
            Gen_Call_Type_Builder (M2Lp (V), Var_Type, Get_Object_Kind (Var)));
         Close_Temp;
      end Gen_Call_Type_Builder;

      procedure Builder_Update_Field
        (Field_Type : Iir; Mem : O_Lnode; Kind : Object_Kind_Type)
      is
         Tinfo : Type_Info_Acc;
         Var_Ptr : O_Lnode;
      begin
         Tinfo := Get_Info (Field_Type);
         if Tinfo.C /= null then
            if Tinfo.C.Builder_Need_Func then
               --  This is a complex type.
               Start_Declare_Stmt;
               New_Var_Decl (Var_Ptr, Get_Identifier ("var_ptr"),
                             O_Storage_Local, Tinfo.Ortho_Ptr_Type (Kind));

               --  Allocate memory.
               --  Set the field with mem.
               --  FIXME: alignment ???
               New_Assign_Stmt
                 (Var_Ptr, New_Convert_Ov (New_Value (Mem),
                                           Tinfo.Ortho_Ptr_Type (Kind)));
               New_Assign_Stmt (Get_Field_Lnode, New_Value (Var_Ptr));

               --  Build second/third-order complex type.
               New_Assign_Stmt
                 (Mem, Gen_Call_Type_Builder (Var_Ptr, Field_Type, Kind));

               Finish_Declare_Stmt;
            else
               --  Allocate memory.
               New_Assign_Stmt (Get_Field_Lnode,
                                New_Convert_Ov (New_Value (Mem),
                                                Tinfo.Ortho_Ptr_Type (Kind)));
               --  Allocate memory for first order complex type.
               New_Assign_Stmt
                 (Mem,
                  New_Address
                  (New_Slice (New_Acc_Value (Mem),
                              Chararray_Type_Node,
                              New_Value
                              (Get_Var (Tinfo.C.Size_Var (Kind)))),
                   Char_Ptr_Type_Node));
            end if;
         end if;
      end Builder_Update_Field;

      ------------------
      --  Enumeration --
      ------------------
      function Create_Enumeration_Literal_Name_Val (El : Iir)
        return O_Enode
      is
         Name : Name_Id;
      begin
         Name := Get_Identifier (El);

         if Name_Table.Is_Character (Name) then
            declare
               Constr : O_Record_Aggr_List;
               Res : O_Enode;
            begin
               Start_Record_Aggr (Constr, Ghdl_Str_Len_Type_Node);
               New_Record_Aggr_El
                 (Constr, New_Unsigned_Literal
                  (Ghdl_Index_Type,
                   Character'Pos (Name_Table.Get_Character (Name))));
               New_Record_Aggr_El (Constr, Null_Const_Char_Ptr_Node);
               Finish_Record_Aggr (Constr, Res);
               return Res;
            end;
         else
            return Create_String_Len
              (Name, Create_Identifier (Name, "_STR"));
         end if;
      end Create_Enumeration_Literal_Name_Val;

      function Create_Enumeration_Val_Array
        (Def : Iir_Enumeration_Type_Definition)
        return O_Lnode
      is
         Array_Type : O_Tnode;
         Val : O_Enode;
         Res : O_Lnode;
         List : O_Array_Aggr_List;
         En_List : Iir;
         El : Iir_Enumeration_Literal;
      begin
         En_List := Get_Enumeration_Literal_List (Def);
         Array_Type := New_Constrained_Array_Type
           (Ghdl_Str_Len_Array_Type_Node,
            New_Unsigned_Literal (Ghdl_Index_Type,
                                  Unsigned_64 (Get_Nbr_Elements (En_List))));
         Start_Array_Aggr (List, Array_Type);
         for I in Natural loop
            El := Get_Nth_Element (En_List, I);
            exit when El = null;
            New_Array_Aggr_El (List, Create_Enumeration_Literal_Name_Val (El));
         end loop;
         Finish_Array_Aggr (List, Val);
         New_Const_Decl (Res, Create_Identifier ("NAMEARRAY"),
                         O_Storage_Private, Array_Type, Val);
         return Res;
      end Create_Enumeration_Val_Array;

      function Create_Static_Enumeration_Type_Desc
        (Def : Iir_Enumeration_Type_Definition)
        return O_Enode
      is
         Info : Ortho_Info_Acc;
         Res : O_Enode;
         List : O_Record_Aggr_List;
      begin
         Info := Get_Info (Def);
         Start_Record_Aggr (List, Ghdl_Desc_E8_Type_Node);
         New_Record_Aggr_El (List, Ghdl_Kind_E8_Node);
         New_Record_Aggr_El
           (List, New_Unchecked_Address (Get_Var (Info.T.Range_Var),
                                         Ghdl_Ptr_Type));
         New_Record_Aggr_El
           (List,
            New_Unchecked_Address (Create_Enumeration_Val_Array (Def),
                                   Ghdl_Str_Len_Ptr_Node));
         Finish_Record_Aggr (List, Res);
         return Res;
      end Create_Static_Enumeration_Type_Desc;

      function Translate_Enumeration_Literal (Lit : Iir_Enumeration_Literal)
        return O_Ident
      is
         El_Str : String (1 .. 4);
         Id : Name_Id;
         N : Integer;
         C : Character;
      begin
         Id := Get_Identifier (Lit);
         if Name_Table.Is_Character (Id) then
            C := Name_Table.Get_Character (Id);
            El_Str (1) := 'C';
            case C is
               when 'A' .. 'Z'
                 | 'a' .. 'z'
                 | '0' .. '9' =>
                  El_Str (2) := '_';
                  El_Str (3) := C;
               when others =>
                  N := Character'Pos (Name_Table.Get_Character (Id));
                  El_Str (2) := N2hex (N / 16);
                  El_Str (3) := N2hex (N mod 16);
            end case;
            return Get_Identifier (El_Str (1 .. 3));
         else
            return Create_Identifier_Without_Prefix (Lit);
         end if;
      end Translate_Enumeration_Literal;

      procedure Translate_Enumeration_Type
        (Def : Iir_Enumeration_Type_Definition)
      is
         El_List : Iir;
         El : Iir_Enumeration_Literal;
         Constr : O_Enum_List;
         Lit_Name : O_Ident;
         Val : O_Enode;
         Info : Type_Info_Acc;
         Nbr : Natural;
         Size : Bitsize_Type;
      begin
         El_List := Get_Enumeration_Literal_List (Def);
         Nbr := Get_Nbr_Elements (El_List);
         if Nbr <= 256 then
            Size := 8;
         else
            Size := 32;
         end if;
         Start_Enum_Type (Constr, Size);
         for I in Natural loop
            El := Get_Nth_Element (El_List, I);
            exit when El = null;

            Lit_Name := Translate_Enumeration_Literal (El);
            New_Enum_Literal (Constr, Lit_Name, Val);
            Set_Ortho_Expr (El, Val);
         end loop;
         Info := Get_Info (Def);
         Finish_Enum_Type (Constr, Info.Ortho_Type (Mode_Value));
         if Nbr <= 256 then
            Info.Type_Mode := Type_Mode_E8;
         else
            Info.Type_Mode := Type_Mode_E32;
         end if;
         Finish_Type_Definition (Info, True);
      end Translate_Enumeration_Type;

      procedure Translate_Bool_Type (Def : Iir_Enumeration_Type_Definition)
      is
         Info : Type_Info_Acc;
         El_List : Iir_List;
         True_Lit, False_Lit : Iir_Enumeration_Literal;
         False_Node, True_Node : O_Enode;
      begin
         Info := Get_Info (Def);
         El_List := Get_Enumeration_Literal_List (Def);
         if Get_Nbr_Elements (El_List) /= 2 then
            raise Internal_Error;
         end if;
         False_Lit := Get_Nth_Element (El_List, 0);
         True_Lit := Get_Nth_Element (El_List, 1);
         New_Boolean_Type
           (Info.Ortho_Type (Mode_Value),
            Translate_Enumeration_Literal (False_Lit), False_Node,
            Translate_Enumeration_Literal (True_Lit), True_Node);
         Info.Type_Mode := Type_Mode_B2;
         Set_Ortho_Expr (False_Lit, False_Node);
         Set_Ortho_Expr (True_Lit, True_Node);
         Finish_Type_Definition (Info, True);
      end Translate_Bool_Type;

      function Create_Bool_Type_Desc (Def : Iir_Enumeration_Type_Definition)
        return O_Enode
      is
         Info : Ortho_Info_Acc;
         Res : O_Enode;
         List : O_Record_Aggr_List;
         El_List : Iir_List;
      begin
         Info := Get_Info (Def);
         El_List := Get_Enumeration_Literal_List (Def);
         if Get_Nbr_Elements (El_List) /= 2 then
            raise Internal_Error;
         end if;
         Start_Record_Aggr (List, Ghdl_Desc_Bool_Type);
         New_Record_Aggr_El (List, Ghdl_Kind_Bool_Node);
         for I in 0 .. 1 loop
            New_Record_Aggr_El (List, Create_Enumeration_Literal_Name_Val
                                (Get_Nth_Element (El_List, I)));
         end loop;
         Finish_Record_Aggr (List, Res);
         return Res;
      end Create_Bool_Type_Desc;

      ---------------
      --  Integer  --
      ---------------
      procedure Translate_Integer_Type
        (Def : Iir_Integer_Type_Definition)
      is
         Info : Type_Info_Acc;
      begin
         --  FIXME: should check precision
         Info := Get_Info (Def);
         Info.Ortho_Type (Mode_Value) := New_Signed_Type (32);
         Info.Type_Mode := Type_Mode_I32;
         Finish_Type_Definition (Info, True);
      end Translate_Integer_Type;

      function Create_Integer_Type_Desc (Def : Iir_Integer_Type_Definition)
        return O_Enode
      is
         List : O_Record_Aggr_List;
         Val : O_Enode;
      begin
         Start_Record_Aggr (List, Ghdl_Desc_I32_Type_Node);
         New_Record_Aggr_El (List, Ghdl_Kind_I32_Node);
         Finish_Record_Aggr (List, Val);
         return Val;
      end Create_Integer_Type_Desc;

      ----------------------
      --  Floating types  --
      ----------------------
      procedure Translate_Floating_Type (Def : Iir_Floating_Type_Definition)
      is
         Info : Type_Info_Acc;
      begin
         --  FIXME: should check precision
         Info := Get_Info (Def);
         Info.Type_Mode := Type_Mode_F64;
         Info.Ortho_Type (Mode_Value) := New_Float_Type;
         Finish_Type_Definition (Info, True);
      end Translate_Floating_Type;

      function Create_Floating_Type_Desc (Def : Iir_Floating_Type_Definition)
        return O_Enode
      is
         List : O_Record_Aggr_List;
         Val : O_Enode;
      begin
         Start_Record_Aggr (List, Ghdl_Desc_F64_Type_Node);
         New_Record_Aggr_El (List, Ghdl_Kind_F64_Node);
         Finish_Record_Aggr (List, Val);
         return Val;
      end Create_Floating_Type_Desc;

      ----------------
      --  Physical  --
      ----------------
      procedure Translate_Physical_Type (Def : Iir_Physical_Type_Definition)
      is
         Info : Type_Info_Acc;
      begin
         Info := Get_Info (Def);
         Info.Ortho_Type (Mode_Value) := New_Signed_Type (64);
         Info.Type_Mode := Type_Mode_P64;
         Finish_Type_Definition (Info, True);
      end Translate_Physical_Type;

      function Create_Physical_Type_Desc (Def : Iir_Physical_Type_Definition)
        return O_Enode
      is
         List : O_Record_Aggr_List;
         Arr_List : O_Array_Aggr_List;
         Val : O_Enode;
         Unit_List : Iir_Unit_List;
         Unit : Iir_Physical_Unit;
         Arr_Type : O_Tnode;
         Units : O_Lnode;
         Primary_Name : Name_Id;
      begin
         --  Create the list of secondary units.
         Unit_List := Get_Unit_List (Def);
         Arr_Type := New_Constrained_Array_Type
           (Ghdl_Unit_Desc_Array,
            New_Unsigned_Literal (Ghdl_Index_Type,
                                  Unsigned_64 (Get_Nbr_Elements (Unit_List))));
         Start_Array_Aggr (Arr_List, Arr_Type);
         for I in Natural loop
            Unit := Get_Nth_Element (Unit_List, I);
            exit when Unit = null;
            Start_Record_Aggr (List, Ghdl_Unit_Desc_Type);
            New_Record_Aggr_El
              (List,
               Create_String_Len (Get_Identifier (Unit),
                                  Create_Identifier (Unit, "_STR")));
            --  FIXME: type conversion may overflow.
            New_Record_Aggr_El
              (List, New_Signed_Literal
               (Ghdl_P64_Type, Integer_64 (Get_Value (Unit))));
            Finish_Record_Aggr (List, Val);
            New_Array_Aggr_El (Arr_List, Val);
         end loop;
         Finish_Array_Aggr (Arr_List, Val);
         New_Const_Decl (Units, Create_Identifier ("_TU"),
                           O_Storage_Private, Arr_Type, Val);
         Start_Record_Aggr (List, Ghdl_Desc_P64_Type_Node);
         New_Record_Aggr_El (List, Ghdl_Kind_P64_Node);
         Primary_Name := Get_Identifier (Get_Primary_Unit (Def));
         New_Record_Aggr_El
           (List, Create_String_Len
            (Primary_Name, Create_Identifier (Primary_Name, "_STR")));
         New_Record_Aggr_El
           (List, New_Unchecked_Address (Units, Ghdl_Unit_Desc_Ptr));
         Finish_Record_Aggr (List, Val);
         return Val;
      end Create_Physical_Type_Desc;

      --------------
      --  Access  --
      --------------
      procedure Translate_Access_Type (Def : Iir_Access_Type_Definition)
      is
         D_Type : Iir;
         D_Info : Type_Info_Acc;
         Dtype : O_Tnode;
         Def_Info : Type_Info_Acc;
      begin
         D_Type := Get_Designated_Type (Def);
         D_Info := Get_Info (D_Type);
         Def_Info := Get_Info (Def);

         if D_Info.Type_Mode = Type_Mode_Fat_Array then
            --  An access type to an unconstrained type definition is a fat
            --  pointer.
            Def_Info.Type_Mode := Type_Mode_Fat_Acc;
            Def_Info.Ortho_Type := D_Info.Ortho_Type;
            Def_Info.T := D_Info.T;
            Def_Info.Ortho_Ptr_Type (Mode_Value) :=
              New_Access_Type (Def_Info.Ortho_Type (Mode_Value));
            New_Type_Decl (Create_Identifier ("PTR"),
                           Def_Info.Ortho_Ptr_Type (Mode_Value));
         else
            --  Otherwise, it is a thin pointer.
            Def_Info.Type_Mode := Type_Mode_Acc;
            if D_Info.Type_Mode in Type_Mode_Arrays then
               --  The designated type cannot be a sub array inside ortho.
               Dtype := D_Info.T.Base_Type (Mode_Value);
            else
               Dtype := D_Info.Ortho_Type (Mode_Value);
            end if;
            Def_Info.Ortho_Type (Mode_Value) := New_Access_Type (Dtype);
            Def_Info.Ortho_Type (Mode_Signal) := O_Tnode_Null;
            Finish_Type_Definition (Def_Info, False);
         end if;
      end Translate_Access_Type;

      function Create_Static_Access_Type_Desc
        (Def : Iir_Access_Type_Definition)
        return O_Enode
      is
         Rec_Constr : O_Record_Aggr_List;
         Res : O_Enode;
      begin
         --  Create the type desc.
         Start_Record_Aggr (Rec_Constr, Ghdl_Desc_Access_Type_Node);
         New_Record_Aggr_El (Rec_Constr, Ghdl_Kind_Access_Node);
         New_Record_Aggr_El
           (Rec_Constr, Get_Type_Desc_Addr (Get_Designated_Type (Def)));
         Finish_Record_Aggr (Rec_Constr, Res);
         return Res;
      end Create_Static_Access_Type_Desc;

      ------------
      --  File  --
      ------------
      procedure Translate_File_Type (Def : Iir_File_Type_Definition)
      is
         Info : Type_Info_Acc;
      begin
         Info := Get_Info (Def);
         Info.Ortho_Type (Mode_Value) := File_Index_Type;
         Info.Ortho_Ptr_Type (Mode_Value) := File_Index_Ptr_Type;
         Info.Type_Mode := Type_Mode_File;
      end Translate_File_Type;

      -------------
      --  Array  --
      -------------
      procedure Translate_Array_Type_Bounds
        (Def : Iir_Array_Type_Definition; Info : Type_Info_Acc)
      is
         Constr : O_Element_List;
         Dim : String (1 .. 8);
         N : Natural;
         P : Natural;
         Index_List : Iir_List;
         Index : Iir;
         Mark : Id_Mark_Type;
      begin
         Start_Record_Type (Constr);
         Index_List := Get_Index_List (Def);
         Info.T.Bounds_Vector :=
           new O_Fnode_Arr (1 .. Get_Nbr_Elements (Index_List));
         for I in Natural loop
            Index := Get_Nth_Element (Index_List, I);
            exit when Index = Null_Iir;
            if Get_Info (Index) = null then
               Push_Identifier_Prefix (Mark, "DIM", I + 1);
               Translate_Type_Definition (Index, True);
               Pop_Identifier_Prefix (Mark);
            end if;
            N := I + 1;
            P := Dim'Last;
            loop
               Dim (P) := Character'Val (Character'Pos ('0') + N mod 10);
               P := P - 1;
               N := N / 10;
               exit when N = 0;
            end loop;
            P := P - 3;
            Dim (P .. P + 3) := "dim_";
            New_Record_Field (Constr, Info.T.Bounds_Vector (I + 1),
                              Get_Identifier (Dim (P .. Dim'Last)),
                              Get_Info (Get_Base_Type (Index)).T.Range_Type);
         end loop;
         Finish_Record_Type (Constr, Info.T.Bounds_Type);
         New_Type_Decl (Create_Identifier ("BOUND"),
                        Info.T.Bounds_Type);
         Info.T.Bounds_Ptr_Type := New_Access_Type (Info.T.Bounds_Type);
         New_Type_Decl (Create_Identifier ("BOUNDP"),
                        Info.T.Bounds_Ptr_Type);
      end Translate_Array_Type_Bounds;

      procedure Translate_Array_Type_Base
        (Def : Iir_Array_Type_Definition; Info : Type_Info_Acc)
      is
         El_Type : Iir;
         El_Tinfo : Type_Info_Acc;
         Id, Idptr : O_Ident;
      begin
         El_Type := Get_Element_Subtype (Def);
         Translate_Type_Definition (El_Type, True);
         El_Tinfo := Get_Info (El_Type);
         for Kind in Object_Kind_Type loop
            case Kind is
               when Mode_Value =>
                  --  For the values.
                  Id := Create_Identifier ("BASE");
                  Idptr := Create_Identifier ("BASEP");
               when Mode_Signal =>
                  --  For the signals
                  Id := Create_Identifier ("SIGBASE");
                  Idptr := Create_Identifier ("SIGBASEP");
            end case;
            Info.T.Base_Type (Kind) :=
              New_Array_Type (Chap4.Get_Element_Type (El_Tinfo, Kind),
                              Ghdl_Index_Type);
            New_Type_Decl (Id, Info.T.Base_Type (Kind));
            Info.T.Base_Ptr_Type (Kind) :=
              New_Access_Type (Info.T.Base_Type (Kind));
            New_Type_Decl (Idptr, Info.T.Base_Ptr_Type (Kind));
         end loop;
      end Translate_Array_Type_Base;

      --  For unidimensional arrays: create a constant bounds whose length
      --  is 1, for concatenation with element.
      procedure Translate_Static_Unidimensional_Array_Length_One
        (Def : Iir_Array_Type_Definition)
      is
         Indexes : Iir_List;
         Index_Type : Iir;
         Index_Base_Type : Iir;
         Constr : O_Record_Aggr_List;
         Constr1 : O_Record_Aggr_List;
         Arr_Info : Type_Info_Acc;
         Tinfo : Type_Info_Acc;
         Irange : Iir;
         Res1 : O_Enode;
         Res : O_Enode;
      begin
         Indexes := Get_Index_List (Def);
         if Get_Nbr_Elements (Indexes) /= 1 then
            return;
         end if;
         Index_Type := Get_First_Element (Indexes);
         Arr_Info := Get_Info (Def);
         if Get_Type_Staticness (Index_Type) = Locally then
            if Global_Storage /= O_Storage_External then
               Index_Base_Type := Get_Base_Type (Index_Type);
               Tinfo := Get_Info (Index_Base_Type);
               Irange := Get_Range_Constraint (Index_Type);
               Start_Record_Aggr (Constr, Arr_Info.T.Bounds_Type);
               Start_Record_Aggr (Constr1, Tinfo.T.Range_Type);
               New_Record_Aggr_El
                 (Constr1,
                  Chap7.Translate_Static_Range_Left (Irange, Index_Base_Type));
               New_Record_Aggr_El
                 (Constr1,
                  Chap7.Translate_Static_Range_Left (Irange, Index_Base_Type));
               New_Record_Aggr_El
                 (Constr1, Chap7.Translate_Static_Range_Dir (Irange));
               New_Record_Aggr_El
                 (Constr1, New_Unsigned_Literal (Ghdl_Index_Type, 1));
               Finish_Record_Aggr (Constr1, Res1);
               New_Record_Aggr_El (Constr, Res1);
               Finish_Record_Aggr (Constr, Res);
            else
               Res := O_Enode_Null;
            end if;
            Arr_Info.T.Array_1bound := Create_Global_Const
              (Create_Identifier ("BR1"),
               Arr_Info.T.Bounds_Type, Global_Storage, Res);
         else
            Arr_Info.T.Array_1bound := Create_Var
              (Create_Var_Identifier ("BR1"),
               Arr_Info.T.Bounds_Type, Global_Storage);
         end if;
      end Translate_Static_Unidimensional_Array_Length_One;

      procedure Translate_Array_Type (Def : Iir_Array_Type_Definition)
      is
         Constr : O_Element_List;
         Info : Type_Info_Acc;
         El_Tinfo : Type_Info_Acc;
      begin
         Info := Get_Info (Def);
         Translate_Array_Type_Base (Def, Info);
         Translate_Array_Type_Bounds (Def, Info);
         for I in Object_Kind_Type loop
            Start_Record_Type (Constr);
            New_Record_Field
              (Constr, Info.T.Base_Field (I), Get_Identifier ("BASE"),
               Info.T.Base_Ptr_Type (I));
            New_Record_Field
              (Constr, Info.T.Bounds_Field (I), Get_Identifier ("BOUNDS"),
               Info.T.Bounds_Ptr_Type);
            Finish_Record_Type (Constr, Info.Ortho_Type (I));
         end loop;
         Info.Type_Mode := Type_Mode_Fat_Array;
         Finish_Type_Definition (Info, False);

         Translate_Static_Unidimensional_Array_Length_One (Def);

         El_Tinfo := Get_Info (Get_Element_Subtype (Def));
         if El_Tinfo.C /= null then
            --  This is a complex type.
            Info.C := new Complex_Type_Info;
            --  No size variable for unconstrained array type.
            Info.C.Size_Var (Mode_Value) := null;
            Info.C.Size_Var (Mode_Signal) := null;
            Info.C.Builder_Need_Func := True;
         end if;
      end Translate_Array_Type;

      function Create_Static_Array_Index_Desc_Array
        (Index_List : Iir_List; Array_Type : O_Tnode)
        return O_Enode
      is
         Constr : O_Array_Aggr_List;
         El : Iir;
         Res : O_Enode;
      begin
         Start_Array_Aggr (Constr, Array_Type);
         for I in Natural loop
            El := Get_Nth_Element (Index_List, I);
            exit when El = Null_Iir;
            New_Array_Aggr_El
              (Constr, Get_Type_Desc_Addr (Get_Base_Type (El)));
         end loop;
         Finish_Array_Aggr (Constr, Res);
         return Res;
      end Create_Static_Array_Index_Desc_Array;

      procedure Create_Array_Index_Desc_Array (Def : Iir)
      is
         El : Iir;
         Index_List : Iir_List;
         Type_Info : Type_Info_Acc;
      begin
         Index_List := Get_Index_List (Def);
         Type_Info := Get_Info (Def);
         for I in Natural loop
            El := Get_Nth_Element (Index_List, I);
            exit when El = Null_Iir;
            New_Assign_Stmt
              (New_Indexed_Element (Get_Var (Type_Info.T.Array_Index_Desc),
                                    New_Unsigned_Literal (Ghdl_Index_Type,
                                                          Unsigned_64 (I))),
               Get_Type_Desc_Addr (Get_Base_Type (El)));
         end loop;
      end Create_Array_Index_Desc_Array;

      function Create_Static_Array_Type_Desc (Def : Iir_Array_Type_Definition)
        return O_Enode
      is
         Constr : O_Record_Aggr_List;
         Val : O_Enode;
         El_Subtype : Iir;
      begin
         Start_Record_Aggr (Constr, Ghdl_Desc_Array_Type_Node);
         New_Record_Aggr_El (Constr, Ghdl_Kind_Array_Node);
         Val := Get_Type_Desc_Addr (Get_Element_Subtype (Def));
         New_Record_Aggr_El (Constr, Val);
         El_Subtype := Get_Element_Subtype (Def);
         if Is_Unconstrained_Type_Definition (El_Subtype) then
            Val := New_Unsigned_Literal (Sizetype, 0);
         else
            --Val := Size_In_Bytes (Get_Ortho_Type (El_Subtype));
            Val := New_Unsigned_Literal (Sizetype, 0);
         end if;
         New_Record_Aggr_El (Constr, Val);
         New_Record_Aggr_El
           (Constr,
            New_Signed_Literal
            (Ghdl_Integer_Type,
             Integer_64 (Get_Nbr_Elements (Get_Index_List (Def)))));
         New_Record_Aggr_El
           (Constr,
            New_Address (Get_Var (Get_Info (Def).T.Array_Index_Desc),
                         Ghdl_Desc_Ptr_Array_Ptr_Type));
         Finish_Record_Aggr (Constr, Val);
         return Val;
      end Create_Static_Array_Type_Desc;

      procedure Create_Array_Type_Desc (Def : Iir_Array_Type_Definition)
      is
         Type_Info : Type_Info_Acc;
      begin
         Type_Info := Get_Info (Def);
         New_Assign_Stmt
           (New_Selected_Element (Get_Var (Type_Info.Type_Desc),
                                  Ghdl_Desc_Array_Kind_Node),
            Ghdl_Kind_Array_Node);
         New_Assign_Stmt
           (New_Selected_Element (Get_Var (Type_Info.Type_Desc),
                                  Ghdl_Desc_Array_El_Node),
            Get_Type_Desc_Addr (Get_Element_Subtype (Def)));
         New_Assign_Stmt
           (New_Selected_Element (Get_Var (Type_Info.Type_Desc),
                                  Ghdl_Desc_Array_Dim_Node),
            New_Signed_Literal
            (Ghdl_Integer_Type,
             Integer_64 (Get_Nbr_Elements (Get_Index_List (Def)))));
         New_Assign_Stmt
           (New_Selected_Element (Get_Var (Type_Info.Type_Desc),
                                  Ghdl_Desc_Array_Elsize_Node),
            New_Unsigned_Literal (Sizetype, 0));
         New_Assign_Stmt
           (New_Selected_Element (Get_Var (Type_Info.Type_Desc),
                                  Ghdl_Desc_Array_Indexes_Node),
            New_Address (Get_Var (Type_Info.T.Array_Index_Desc),
                         Ghdl_Desc_Ptr_Array_Ptr_Type));
      end Create_Array_Type_Desc;

      --  Get the length of DEF, ie the number of elements.
      --  If the length is not statically defined, returns -1.
      function Get_Array_Subtype_Length (Def : Iir_Array_Subtype_Definition)
        return Iir_Int64
      is
         Index_List : Iir_List;
         Index : Iir;
         Len : Iir_Int64;
      begin
         Index_List := Get_Index_List (Def);
         --  Check if the bounds of the array are locally static.
         Len := 1;
         for I in Natural loop
            Index := Get_Nth_Element (Index_List, I);
            exit when Index = Null_Iir;
            if Get_Type_Staticness (Index) /= Locally then
               return -1;
            end if;
            Len := Len * Eval_Discrete_Type_Length (Index);
         end loop;
         return Len;
      end Get_Array_Subtype_Length;

      procedure Translate_Array_Subtype (Def : Iir_Array_Subtype_Definition)
      is
         Info : Type_Info_Acc;
         Binfo : Type_Info_Acc;
         Len : Iir_Int64;
         Ptr : O_Tnode;
         Id : O_Ident;
      begin
         Info := Get_Info (Def);
         Binfo := Get_Info (Get_Base_Type (Def));
         Len := Get_Array_Subtype_Length (Def);
         if Len < 0 then
            --  Length of the array is not known at compile time.
            Info.Type_Mode := Type_Mode_Ptr_Array;
            Info.Ortho_Type := Binfo.T.Base_Ptr_Type;
            Info.Ortho_Ptr_Type := Binfo.T.Base_Ptr_Type;
         else
            --  Length is known.  Create a constrained array.
            Info.Type_Mode := Type_Mode_Array;
            for I in Object_Kind_Type loop
               case I is
                  when Mode_Value =>
                     Id := Create_Identifier;
                  when Mode_Signal =>
                     Id := Create_Identifier ("SIG");
               end case;
               Info.Ortho_Type (I) := New_Constrained_Array_Type
                 (Binfo.T.Base_Type (I),
                  New_Unsigned_Literal (Ghdl_Index_Type, Unsigned_64 (Len)));
               New_Type_Decl (Id, Info.Ortho_Type (I));
               --Ptr := New_Access_Type (Info.Ortho_Type);
               --New_Type_Decl (Create_Identifier (Name, "_ARGT"), Ptr);
               Ptr := Binfo.T.Base_Ptr_Type (I);
               Info.Ortho_Ptr_Type (I) := Ptr;
            end loop;
         end if;

         --  Create a size variable if the length is not known or if
         --  the element size is not known at compile-time.
         if Binfo.C /= null then
            --  The base type is a complex type, so is the type.
            Create_Size_Var (Info);
            Info.C.Builder_Need_Func := True;
         elsif Len < 0 then
            --  This may creates complex types.
            Create_Size_Var (Info);
            Info.C.Builder_Need_Func := False;
         end if;
      end Translate_Array_Subtype;

      function Create_Static_Array_Subtype_Bounds
        (Def : Iir_Array_Subtype_Definition)
        return O_Enode
      is
         Index_List : Iir_List;
         Index : Iir;
         List : O_Record_Aggr_List;
         Res : O_Enode;
         Baseinfo : Type_Info_Acc;
      begin
         Index_List := Get_Index_List (Def);
         Baseinfo := Get_Info (Get_Base_Type (Def));
         Start_Record_Aggr (List, Baseinfo.T.Bounds_Type);
         for I in Natural loop
            Index := Get_Nth_Element (Index_List, I);
            exit when Index = null;
            New_Record_Aggr_El
              (List, Create_Static_Type_Definition_Type_Range (Index));
         end loop;
         Finish_Record_Aggr (List, Res);
         return Res;
      end Create_Static_Array_Subtype_Bounds;

      procedure Create_Array_Subtype_Bounds
        (Def : Iir_Array_Subtype_Definition; Target : O_Lnode)
      is
         Index_List : Iir_List;
         Index : Iir;
         Baseinfo : Type_Info_Acc;
         Targ : Mnode;
      begin
         Baseinfo := Get_Info (Get_Base_Type (Def));
         Targ := Lv2M (Target, True,
                       Baseinfo.T.Bounds_Type,
                       Baseinfo.T.Bounds_Ptr_Type,
                       null, Mode_Value);
         Index_List := Get_Index_List (Def);
         Open_Temp;
         if Get_Nbr_Elements (Index_List) > 1 then
            Targ := Stabilize (Targ);
         end if;
         for I in Natural loop
            Index := Get_Nth_Element (Index_List, I);
            exit when Index = Null_Iir;
            declare
               Index_Type : Iir;
               Index_Info : Type_Info_Acc;
               D : O_Lnode;
            begin
               Index_Type := Get_Base_Type (Index);
               Index_Info := Get_Info (Index_Type);
               Open_Temp;
               D := Create_Temp_Ptr
                 (Index_Info.T.Range_Ptr_Type,
                  New_Selected_Element (M2Lv (Targ),
                                        Baseinfo.T.Bounds_Vector (I + 1)));
               Chap7.Translate_Discrete_Range_Ptr (D, Index);
               Close_Temp;
            end;
         end loop;
         Close_Temp;
      end Create_Array_Subtype_Bounds;

      function Create_Static_Array_Subtype_Desc
        (Def : Iir_Array_Subtype_Definition)
        return O_Enode
      is
         Info : Type_Info_Acc;
         Base_Info : Type_Info_Acc;
         List : O_Record_Aggr_List;
         Val : O_Enode;
         Basetype_O : O_Enode;
         Range_O : O_Enode;
      begin
         Info := Get_Info (Def);
         Base_Info := Get_Info (Get_Base_Type (Def));
         Basetype_O := Get_Type_Desc_Addr (Get_Base_Type (Def));
         Range_O := New_Unchecked_Address
           (Get_Var (Info.T.Array_Bounds), Const_Ptr_Type_Node);
         --Range_O := New_Null_Access (Const_Ptr_Type_Node);
         Start_Record_Aggr (List, Ghdl_Desc_Subarray_Type_Node);
         New_Record_Aggr_El (List, Ghdl_Kind_Subarray_Node);
         New_Record_Aggr_El (List, Basetype_O);
         New_Record_Aggr_El (List, Range_O);
         Finish_Record_Aggr (List, Val);
         return Val;
      end Create_Static_Array_Subtype_Desc;

      procedure Set_Array_Subtype_Desc
        (Var : Var_Acc; Basetype : Iir; Bounds : O_Lnode)
      is
      begin
         New_Assign_Stmt
           (New_Selected_Element (Get_Var (Var),
                                  Ghdl_Desc_Subarray_Range_Node),
            New_Unchecked_Address (Bounds, Const_Ptr_Type_Node));

         New_Assign_Stmt
           (New_Selected_Element (Get_Var (Var),
                                  Ghdl_Desc_Subarray_Kind_Node),
            Ghdl_Kind_Subarray_Node);
         New_Assign_Stmt
           (New_Selected_Element (Get_Var (Var),
                                  Ghdl_Desc_Subarray_Basetype_Node),
            Get_Type_Desc_Addr (Basetype));
      end Set_Array_Subtype_Desc;

      procedure Create_Array_Subtype_Desc
        (Def : Iir_Array_Subtype_Definition; Name : Name_Id)
      is
         Info : Type_Info_Acc;
      begin
         Info := Get_Info (Def);
         Set_Array_Subtype_Desc (Info.Type_Desc,
                                 Get_Base_Type (Def),
                                 Get_Var (Info.T.Array_Bounds));
      end Create_Array_Subtype_Desc;

      --  Get staticness of the array bounds.
      function Get_Array_Bounds_Staticness (Def : Iir) return Iir_Staticness
      is
         List : Iir_List;
         El : Iir;
      begin
         List := Get_Index_List (Def);
         for I in Natural loop
            El := Get_Nth_Element (List, I);
            exit when El = Null_Iir;
            if Get_Type_Staticness (El) /= Locally then
               return Globally;
            end if;
         end loop;
         return Locally;
      end Get_Array_Bounds_Staticness;

      function Get_Array_Type_Desc_Static (Def : Iir) return Boolean is
      begin
         if not Get_Info (Get_Element_Subtype (Def)).Type_Desc_Static then
            return False;
         end if;
         return Get_Array_Bounds_Staticness (Def) = Locally;
      end Get_Array_Type_Desc_Static;

      procedure Create_Array_Type_Index_Desc_Array_Var (Def : Iir)
      is
         Info : Type_Info_Acc;
         Val : O_Enode;
         Index_List : Iir_List;
         Length : O_Enode;
         Array_Type : O_Tnode;
      begin
         Info := Get_Info (Def);
         if Info.T.Array_Index_Desc /= null then
            return;
         end if;

         --  Create array index desc type.
         Index_List := Get_Index_List (Def);
         Length := New_Unsigned_Literal
           (Ghdl_Index_Type, Unsigned_64 (Get_Nbr_Elements (Index_List)));
         Array_Type := New_Constrained_Array_Type
           (Ghdl_Desc_Ptr_Array_Type, Length);
         New_Type_Decl (Create_Identifier ("IDATYPE"), Array_Type);

         if Info.Type_Desc_Static then
            if Global_Storage = O_Storage_External then
               --  Do not create the value of the type desc, since it
               --  is never dereferenced in a static type desc.
               Val := O_Enode_Null;
            else
               Val := Create_Static_Array_Index_Desc_Array
                 (Index_List, Array_Type);
            end if;
            Info.T.Array_Index_Desc := Create_Global_Const
              (Create_Identifier ("IDA"),
               Array_Type, Global_Storage, Val);
         else
            Info.T.Array_Index_Desc := Create_Var
              (Create_Var_Identifier ("IDA"), Array_Type);
         end if;
      end Create_Array_Type_Index_Desc_Array_Var;

      --  Create a variable containing the bounds for array subtype DEF.
      procedure Create_Array_Subtype_Bounds_Var
        (Def : Iir; Elab_Now : Boolean)
      is
         Info : Type_Info_Acc;
         Base_Info : Type_Info_Acc;
         Val : O_Enode;
      begin
         Info := Get_Info (Def);
         if Info.T.Array_Bounds /= null then
            return;
         end if;
         Base_Info := Get_Info (Get_Base_Type (Def));
         case Get_Array_Bounds_Staticness (Def) is
            when None
              | Globally =>
               Info.T.Static_Bounds := False;
               Info.T.Array_Bounds := Create_Var
                 (Create_Var_Identifier ("STB"), Base_Info.T.Bounds_Type);
               if Elab_Now then
                  Create_Array_Subtype_Bounds
                    (Def, Get_Var (Info.T.Array_Bounds));
               end if;
            when Locally =>
               Info.T.Static_Bounds := True;
               if Global_Storage = O_Storage_External then
                  --  Do not create the value of the type desc, since it
                  --  is never dereferenced in a static type desc.
                  Val := O_Enode_Null;
               else
                  Val := Create_Static_Array_Subtype_Bounds (Def);
               end if;
               Info.T.Array_Bounds := Create_Global_Const
                 (Create_Identifier ("STB"),
                  Base_Info.T.Bounds_Type, Global_Storage, Val);

            when Unknown =>
               raise Internal_Error;
         end case;
      end Create_Array_Subtype_Bounds_Var;

      procedure Create_Array_Type_Builder
        (Def : Iir_Array_Type_Definition; Kind : Object_Kind_Type)
      is
         Base : O_Lnode;
         Var_I : O_Lnode;

         function Get_Field_Lnode return O_Lnode is
         begin
            return New_Indexed_Element (New_Acc_Value (Base),
                                        New_Value (Var_I));
         end Get_Field_Lnode;

         procedure Update_Field is new Builder_Update_Field (Get_Field_Lnode);

         Mem : O_Lnode;
         Info : Type_Info_Acc;
         El_Info : Type_Info_Acc;
         If_Blk : O_If_Block;
         Var_Length : O_Lnode;
         Loop_Blk : O_Loop_Block;
         Label : O_Snode;
      begin
         Info := Get_Info (Def);
         Start_Subprogram_Body (Info.C.Builder_Func (Kind));
         Chap2.Start_Subprg_Instance_Use (Info.C.Builder_Instance (Kind));

         --  Aliased
         Base := Info.C.Builder_Base_Param (Kind);

         --  Compute length of the array.
         New_Var_Decl (Var_Length, Wki_Length, O_Storage_Local,
                       Ghdl_Index_Type);
         New_Var_Decl (Mem, Get_Identifier ("mem"), O_Storage_Local,
                       Char_Ptr_Type_Node);
         New_Assign_Stmt
           (Var_Length,
            Chap3.Get_Bounds_Ptr_Length (Info.C.Builder_Bound_Param (Kind),
                                         Def));

         --  Reserve the size of the array vector.
         El_Info := Get_Info (Get_Element_Subtype (Def));
         New_Assign_Stmt
           (Mem,
            New_Address
            (New_Slice
             (New_Access_Element
              (New_Convert_Ov (New_Value (Base),
                               Char_Ptr_Type_Node)),
              Chararray_Type_Node,
              New_Dyadic_Op (ON_Mul_Ov,
                             New_Value (Var_Length),
                             New_Sizeof (El_Info.Ortho_Ptr_Type (Kind),
                                         Ghdl_Index_Type))),
             Char_Ptr_Type_Node));

         --  Set each index of the array.
         New_Var_Decl (Var_I, Wki_I, O_Storage_Local, Ghdl_Index_Type);
         Init_Var (Var_I);
         Start_Loop_Stmt (Loop_Blk, Label);
         Gen_Exit_When (Label,
                        New_Compare_Op (ON_Eq,
                                        New_Value (Var_I),
                                        New_Value (Var_Length),
                                        Ghdl_Bool_Type_Node));
         Update_Field (Get_Element_Subtype (Def), Mem, Kind);
         Inc_Var (Var_I);
         Finish_Loop_Stmt (Loop_Blk);

         New_Return_Stmt (New_Value (Mem));

         Chap2.Finish_Subprg_Instance_Use (Info.C.Builder_Instance (Kind));
         Finish_Subprogram_Body;
      end Create_Array_Type_Builder;

      --------------
      --  record  --
      --------------
      function Type_To_Last_Object_Kind (Def : Iir) return Object_Kind_Type is
      begin
         if Get_Signal_Type_Flag (Def) then
            return Mode_Signal;
         else
            return Mode_Value;
         end if;
      end Type_To_Last_Object_Kind;

      procedure Translate_Record_Type (Def : Iir_Record_Type_Definition;
                                       Completion : Type_Completion_Kind)
      is
         El_List : O_Element_List;
         List : Iir_List;
         El : Iir_Element_Declaration;
         Info : Type_Info_Acc;
         Field_Info : Ortho_Info_Acc;
         El_Type : Iir;
         El_Tinfo : Type_Info_Acc;

         --  True a size variable will be created since the size of the record
         --  is not known at compile-time,
         Need_Size : Boolean;

         Mark : Id_Mark_Type;
      begin
         Info := Get_Info (Def);
         List := Get_Element_Declaration_List (Def);
         Need_Size := False;
         for I in Natural loop
            El := Get_Nth_Element (List, I);
            exit when El = Null_Iir;

            El_Type := Get_Type (El);
            if Get_Info (El_Type) = null then
               Push_Identifier_Prefix (Mark, Get_Identifier (El));
               Translate_Type_Definition (El_Type);
               Pop_Identifier_Prefix (Mark);
            end if;
            if not Need_Size and then Get_Info (El_Type).C /= null then
               Need_Size := True;
            end if;
            Field_Info := Add_Info (El, Kind_Field);
         end loop;

         for Kind in Mode_Value .. Type_To_Last_Object_Kind (Def) loop
            if Completion = Completed_Type then
               Start_Uncomplete_Record_Type (Info.Ortho_Type (Kind), El_List);
            else
               Start_Record_Type (El_List);
            end if;
            for I in Natural loop
               El := Get_Nth_Element (List, I);
               exit when El = Null_Iir;

               Field_Info := Get_Info (El);
               El_Tinfo := Get_Info (Get_Type (El));
               New_Record_Field (El_List, Field_Info.Field_Node (Kind),
                                 Create_Identifier_Without_Prefix (El),
                                 Chap4.Get_Element_Type (El_Tinfo, Kind));
            end loop;
            Finish_Record_Type (El_List, Info.Ortho_Type (Kind));
         end loop;
         if Get_Signal_Type_Flag (Def) = False then
            Info.Ortho_Type (Mode_Signal) := O_Tnode_Null;
         end if;
         Info.Type_Mode := Type_Mode_Record;
         Finish_Type_Definition (Info, False, Completion);

         if Need_Size then
            Create_Size_Var (Info);
            Info.C.Builder_Need_Func := True;
         end if;
      end Translate_Record_Type;

      function Create_Static_Record_El_Desc_Array (El_List : Iir_List;
                                                   Array_Type : O_Tnode)
        return O_Enode
      is
         Arr_Constr : O_Array_Aggr_List;
         Val : O_Enode;
         El : Iir;
      begin
         Start_Array_Aggr (Arr_Constr, Array_Type);
         for I in Natural loop
            El := Get_Nth_Element (El_List, I);
            exit when El = Null_Iir;
            New_Array_Aggr_El (Arr_Constr, Get_Type_Desc_Addr (Get_Type (El)));
         end loop;
         Finish_Array_Aggr (Arr_Constr, Val);
         return Val;
      end Create_Static_Record_El_Desc_Array;

      procedure Create_Record_El_Desc_Array (Def : Iir_Record_Type_Definition)
      is
         El : Iir;
         Index_List : Iir_List;
         Type_Info : Type_Info_Acc;
      begin
         Type_Info := Get_Info (Def);
         Index_List := Get_Element_Declaration_List (Def);
         for I in Natural loop
            El := Get_Nth_Element (Index_List, I);
            exit when El = Null_Iir;
            New_Assign_Stmt
              (New_Indexed_Element (Get_Var (Type_Info.T.Record_El_Desc),
                                    New_Unsigned_Literal (Ghdl_Index_Type,
                                                          Unsigned_64 (I))),
               Get_Type_Desc_Addr (Get_Type (El)));
         end loop;
      end Create_Record_El_Desc_Array;

      function Create_Static_Record_Type_Desc
        (Def : Iir_Record_Type_Definition)
        return O_Enode
      is
         El_List : Iir_List;
         Rec_Constr : O_Record_Aggr_List;
         Res : O_Enode;
      begin
         El_List := Get_Element_Declaration_List (Def);

         --  Create the type desc.
         Start_Record_Aggr (Rec_Constr, Ghdl_Desc_Record_Type_Node);
         New_Record_Aggr_El (Rec_Constr, Ghdl_Kind_Record_Node);
         New_Record_Aggr_El
           (Rec_Constr,
            New_Signed_Literal (Ghdl_Integer_Type,
                                Integer_64 (Get_Nbr_Elements (El_List))));
         New_Record_Aggr_El
           (Rec_Constr, New_Address (Get_Var (Get_Info (Def).T.Record_El_Desc),
                                     Ghdl_Desc_Ptr_Array_Ptr_Type));
         Finish_Record_Aggr (Rec_Constr, Res);
         return Res;
      end Create_Static_Record_Type_Desc;

      procedure Create_Record_Type_Desc (Def : Iir_Record_Type_Definition)
      is
         Type_Info : Type_Info_Acc;
         El_List : Iir_List;
      begin
         Type_Info := Get_Info (Def);
         El_List := Get_Element_Declaration_List (Def);
         New_Assign_Stmt
           (New_Selected_Element (Get_Var (Type_Info.Type_Desc),
                                  Ghdl_Desc_Record_Kind_Node),
            Ghdl_Kind_Record_Node);
         New_Assign_Stmt
           (New_Selected_Element (Get_Var (Type_Info.Type_Desc),
                                  Ghdl_Desc_Record_Nbr_El_Node),
            New_Signed_Literal (Ghdl_Integer_Type,
                                Integer_64 (Get_Nbr_Elements (El_List))));
         New_Assign_Stmt
           (New_Selected_Element (Get_Var (Type_Info.Type_Desc),
                                  Ghdl_Desc_Record_Els_Node),
            New_Address (Get_Var (Type_Info.T.Record_El_Desc),
                         Ghdl_Desc_Ptr_Array_Ptr_Type));
      end Create_Record_Type_Desc;

      procedure Create_Record_Type_El_Desc_Array_Var
        (Def : Iir_Record_Type_Definition)
      is
         Info : Type_Info_Acc;
         Val : O_Enode;
         El_List : Iir_List;
         Length : O_Enode;
         Array_Type : O_Tnode;
      begin
         Info := Get_Info (Def);
         if Info.T.Record_El_Desc /= null then
            return;
         end if;

         --  Create array index desc type.
         El_List := Get_Element_Declaration_List (Def);
         Length := New_Unsigned_Literal
           (Ghdl_Index_Type, Unsigned_64 (Get_Nbr_Elements (El_List)));
         Array_Type := New_Constrained_Array_Type
           (Ghdl_Desc_Ptr_Array_Type, Length);
         New_Type_Decl (Create_Identifier ("ELATYPE"), Array_Type);

         if Info.Type_Desc_Static then
            if Global_Storage = O_Storage_External then
               --  Do not create the value of the type desc, since it
               --  is never dereferenced in a static type desc.
               Val := O_Enode_Null;
            else
               Val := Create_Static_Record_El_Desc_Array
                 (El_List, Array_Type);
            end if;
            Info.T.Record_El_Desc := Create_Global_Const
              (Create_Identifier ("ELA"),
               Array_Type, Global_Storage, Val);
         else
            Info.T.Record_El_Desc := Create_Var
              (Create_Var_Identifier ("ELA"), Array_Type);
         end if;
      end Create_Record_Type_El_Desc_Array_Var;

      function Get_Record_Type_Desc_Static (Def : Iir) return Boolean
      is
         List : Iir_List;
         El : Iir;
      begin
         List := Get_Element_Declaration_List (Def);
         for I in Natural loop
            El := Get_Nth_Element (List, I);
            exit when El = Null_Iir;
            if not Get_Info (Get_Type (El)).Type_Desc_Static then
               return False;
            end if;
         end loop;
         return True;
      end Get_Record_Type_Desc_Static;

      procedure Create_Record_Type_Builder
        (Def : Iir_Record_Type_Definition; Kind : Object_Kind_Type)
      is
         Base : O_Lnode;
         El : Iir_Element_Declaration;

         function Get_Field_Lnode
           return O_Lnode
         is
         begin
            return New_Selected_Element (New_Acc_Value (Base),
                                         Get_Info (El).Field_Node (Kind));
         end Get_Field_Lnode;

         procedure Update_Field is new
           Builder_Update_Field (Get_Field_Lnode);

         Info : Type_Info_Acc;
         El_List : Iir_List;
         Mem : O_Lnode;
         El_Type : Iir;
      begin
         Info := Get_Info (Def);
         El_List := Get_Element_Declaration_List (Def);
         Start_Subprogram_Body (Info.C.Builder_Func (Kind));
         Chap2.Start_Subprg_Instance_Use (Info.C.Builder_Instance (Kind));

         --  Aliases.
         Base := Info.C.Builder_Base_Param (Kind);

         New_Var_Decl (Mem, Get_Identifier ("mem"), O_Storage_Local,
                       Char_Ptr_Type_Node);

         --  Reserve memory for the record, ie:
         --  MEM = BASE + SIZEOF (record).
         New_Assign_Stmt
           (Mem,
            New_Address
            (New_Slice (New_Access_Element
                        (New_Convert_Ov (New_Value (Base),
                                         Char_Ptr_Type_Node)),
                        Chararray_Type_Node,
                        New_Sizeof (Info.Ortho_Type (Kind), Ghdl_Index_Type)),
             Char_Ptr_Type_Node));

         --  Set memory for each complex element.
         for I in Natural loop
            El := Get_Nth_Element (El_List, I);
            exit when El = Null_Iir;
            El_Type := Get_Type (El);
            if Get_Info (El_Type).C /= null then
               --  Complex type.
               Update_Field (El_Type, Mem, Kind);
            end if;
         end loop;
         Chap2.Finish_Subprg_Instance_Use (Info.C.Builder_Instance (Kind));
         New_Return_Stmt (New_Value (Mem));
         Finish_Subprogram_Body;
      end Create_Record_Type_Builder;

      ------------------------
      --  Incomplete types  --
      ------------------------
      procedure Translate_Incomplete_Type (Def : Iir)
      is
         Ftype : Iir;
         Info : Type_Info_Acc;
      begin
         Info := Get_Info (Def);
         Ftype := Get_Type (Get_Type_Declarator (Def));
         case Get_Kind (Ftype) is
            when Iir_Kind_Record_Type_Definition =>
               Info.Type_Mode := Type_Mode_Unknown;
               New_Uncomplete_Record_Type (Info.Ortho_Type (Mode_Value));
            when others =>
               Error_Kind ("translate_incomplete_type", Ftype);
         end case;
         Set_Info (Ftype, Info);
         Finish_Type_Definition (Info, False, Incomplete_Type);
      end Translate_Incomplete_Type;

      ---------------
      --  Scalars  --
      ---------------

      --  Create a type_range structure.
      procedure Create_Scalar_Type_Range (Def : Iir; Target : O_Lnode)
      is
         T_Info : Type_Info_Acc;
         Base_Type : Iir;
         Expr : Iir;
         V : O_Lnode;
      begin
         Base_Type := Get_Base_Type (Def);
         T_Info := Get_Info (Base_Type);
         Expr := Get_Range_Constraint (Def);
         Open_Temp;
         V := Create_Temp_Ptr (T_Info.T.Range_Ptr_Type, Target);
         Chap7.Translate_Range_Ptr (V, Expr, Def);
         Close_Temp;
      end Create_Scalar_Type_Range;

      function Create_Static_Scalar_Type_Range (Def : Iir) return O_Enode is
      begin
         return Chap7.Translate_Static_Range (Get_Range_Constraint (Def),
                                              Get_Base_Type (Def));
      end Create_Static_Scalar_Type_Range;

      procedure Create_Scalar_Type_Range_Type
        (Def : Iir; With_Length : Boolean)
      is
         Constr : O_Element_List;
         Info : Ortho_Info_Acc;
      begin
         Info := Get_Info (Def);
         Start_Record_Type (Constr);
         New_Record_Field
           (Constr, Info.T.Range_Left, Wki_Left,
            Info.Ortho_Type (Mode_Value));
         New_Record_Field
           (Constr, Info.T.Range_Right, Wki_Right,
            Info.Ortho_Type (Mode_Value));
         New_Record_Field
           (Constr, Info.T.Range_Dir, Wki_Dir, Ghdl_Dir_Type_Node);
         if With_Length then
            New_Record_Field
              (Constr, Info.T.Range_Length, Wki_Length, Ghdl_Index_Type);
         else
            Info.T.Range_Length := O_Fnode_Null;
         end if;
         Finish_Record_Type (Constr, Info.T.Range_Type);
         New_Type_Decl (Create_Identifier ("TRT"), Info.T.Range_Type);
         Info.T.Range_Ptr_Type := New_Access_Type (Info.T.Range_Type);
         New_Type_Decl (Create_Identifier ("TRPTR"),
                        Info.T.Range_Ptr_Type);
      end Create_Scalar_Type_Range_Type;

      function Create_Static_Type_Definition_Type_Range (Def : Iir)
        return O_Enode
      is
      begin
         case Get_Kind (Def) is
            when Iir_Kind_Enumeration_Type_Definition
              | Iir_Kind_Enumeration_Subtype_Definition
              | Iir_Kind_Integer_Subtype_Definition
              | Iir_Kind_Floating_Subtype_Definition
              | Iir_Kind_Physical_Subtype_Definition =>
               return Create_Static_Scalar_Type_Range (Def);

            when Iir_Kind_Array_Subtype_Definition =>
               return Create_Static_Array_Subtype_Bounds (Def);

            when Iir_Kind_Array_Type_Definition =>
               return O_Enode_Null;

            when others =>
               Error_Kind ("create_static_type_definition_type_range", Def);
         end case;
      end Create_Static_Type_Definition_Type_Range;

      procedure Create_Type_Definition_Type_Range (Def : Iir)
      is
         Target : O_Lnode;
         Info : Type_Info_Acc;
      begin
         case Get_Kind (Def) is
            when Iir_Kind_Enumeration_Type_Definition
              | Iir_Kind_Enumeration_Subtype_Definition
              | Iir_Kind_Integer_Subtype_Definition
              | Iir_Kind_Floating_Subtype_Definition
              | Iir_Kind_Physical_Subtype_Definition =>
               Target := Get_Var (Get_Info (Def).T.Range_Var);
               Create_Scalar_Type_Range (Def, Target);

            when Iir_Kind_Array_Subtype_Definition =>
               Info := Get_Info (Def);
               if not Info.T.Static_Bounds then
                  Target := Get_Var (Info.T.Array_Bounds);
                  Create_Array_Subtype_Bounds (Def, Target);
               end if;

            when Iir_Kind_Array_Type_Definition =>
               --  FIXME: create unidimensional array bound of length 1
               return;
            when Iir_Kind_Access_Type_Definition
              | Iir_Kind_File_Type_Definition
              | Iir_Kind_Record_Type_Definition =>
               return;

            when others =>
               Error_Kind ("create_type_definition_type_range", Def);
         end case;
      end Create_Type_Definition_Type_Range;

      function Get_Additionnal_Size (Def : Iir; Kind : Object_Kind_Type)
        return O_Enode
      is
         Info : Type_Info_Acc;
      begin
         Info := Get_Info (Def);

         if Info.C = null then
            --  Short-cut.
            return O_Enode_Null;
         else
            return New_Value (Get_Var (Info.C.Size_Var (Kind)));
         end if;
      end Get_Additionnal_Size;

      procedure Create_Type_Definition_Size_Var (Def : Iir)
      is
         Info : Type_Info_Acc;
         Res : O_Enode;
         Add : O_Enode;
      begin
         Info := Get_Info (Def);
         if Info.C = null then
            return;
         end if;
         for Kind in Object_Kind_Type loop
            if Info.C.Size_Var (Kind) /= null then
               Open_Temp;
               case Info.Type_Mode is
                  when Type_Mode_Non_Composite
                    | Type_Mode_Fat_Array
                    | Type_Mode_Unknown =>
                     raise Internal_Error;
                  when Type_Mode_Array =>
                     Res := New_Sizeof (Info.Ortho_Type (Kind),
                                        Ghdl_Index_Type);
                     Add := Get_Additionnal_Size
                       (Get_Element_Subtype (Def), Kind);
                     if Add /= O_Enode_Null then
                        Add := New_Dyadic_Op
                          (ON_Mul_Ov, Get_Array_Type_Length (Def), Add);
                        Res := New_Dyadic_Op (ON_Add_Ov, Res, Add);
                     end if;
                  when Type_Mode_Record =>
                     declare
                        List : Iir_List;
                        El : Iir_Element_Declaration;
                        N_Res : O_Enode;
                     begin
                        List := Get_Element_Declaration_List
                          (Get_Base_Type (Def));
                        Res := New_Sizeof (Info.Ortho_Type (Kind),
                                           Ghdl_Index_Type);
                        for I in Natural loop
                           El := Get_Nth_Element (List, I);
                           exit when El = Null_Iir;
                           N_Res := Get_Additionnal_Size (Get_Type (El), Kind);
                           if N_Res /= O_Enode_Null then
                              Res := New_Dyadic_Op (ON_Add_Ov, Res, N_Res);
                           end if;
                        end loop;
                     end;
                  when Type_Mode_Ptr_Array =>
                     --  If element is a composite type then
                     --    Return length * (sizeof (element)
                     --                     + sizeof (element_ptr))
                     --  else
                     --    Return length * sizeof (element)
                     --  end if
                     declare
                        El_Type : Iir;
                        El_Tinfo : Type_Info_Acc;
                     begin
                        El_Type := Get_Element_Subtype (Def);
                        El_Tinfo := Get_Info (El_Type);
                        Res := Chap3.Get_Object_Size
                          (T2M (El_Type, Kind), El_Type);
                        if El_Tinfo.C /= null then
                           Res := New_Dyadic_Op
                             (ON_Add_Ov,
                              Res,
                              New_Sizeof (El_Tinfo.Ortho_Ptr_Type (Kind),
                                          Ghdl_Index_Type));
                        end if;
                        Res := New_Dyadic_Op
                          (ON_Mul_Ov,
                           Chap3.Get_Array_Type_Length (Def),
                           Res);
                     end;
               end case;
               New_Assign_Stmt (Get_Var (Info.C.Size_Var (Kind)), Res);
               Close_Temp;
            end if;
         end loop;
      end Create_Type_Definition_Size_Var;

      procedure Create_Type_Definition_Type_Desc (Def : Iir; Name : Name_Id) is
      begin
         --  Create type desc.
         case Get_Kind (Def) is
            when Iir_Kind_Enumeration_Type_Definition
              | Iir_Kind_Integer_Type_Definition
              | Iir_Kind_Floating_Type_Definition
              | Iir_Kind_Physical_Type_Definition =>
               Error_Kind ("create_type_definition_type_desc", Def);

            when Iir_Kind_Record_Type_Definition =>
               Create_Record_El_Desc_Array (Def);
               Create_Record_Type_Desc (Def);

            when Iir_Kind_Array_Type_Definition =>
               Create_Array_Index_Desc_Array (Def);
               Create_Array_Type_Desc (Def);

            when Iir_Kind_Enumeration_Subtype_Definition =>
               Create_Scalar_Subtype_Desc (Def, Name, Ghdl_Kind_Se8_Node);

            when Iir_Kind_Integer_Subtype_Definition =>
               Create_Scalar_Subtype_Desc (Def, Name, Ghdl_Kind_Si32_Node);

            when Iir_Kind_Floating_Subtype_Definition =>
               Create_Scalar_Subtype_Desc (Def, Name, Ghdl_Kind_Sf64_Node);

            when Iir_Kind_Physical_Subtype_Definition =>
               Create_Scalar_Subtype_Desc (Def, Name, Ghdl_Kind_Sp64_Node);

            when Iir_Kind_Array_Subtype_Definition =>
               Create_Array_Subtype_Desc (Def, Name);

            when Iir_Kind_Access_Type_Definition
              | Iir_Kind_File_Type_Definition =>
               return;

            when others =>
               Error_Kind ("create_type_definition_type_desc", Def);
         end case;
      end Create_Type_Definition_Type_Desc;

      --  Only when type_desc_static of DEF is true: compute the
      --  type_desc value.
      function Get_Static_Type_Definition_Type_Desc (Def : Iir)
        return O_Enode
      is
      begin
         --  Create type desc.
         case Get_Kind (Def) is
            when Iir_Kind_Enumeration_Type_Definition =>
               return Create_Static_Enumeration_Type_Desc (Def);

            when Iir_Kind_Enumeration_Subtype_Definition =>
               return Create_Static_Scalar_Subtype_Desc
                 (Def, Ghdl_Kind_Se8_Node);

            when Iir_Kind_Integer_Type_Definition =>
               return Create_Integer_Type_Desc (Def);

            when Iir_Kind_Integer_Subtype_Definition =>
               return Create_Static_Scalar_Subtype_Desc
                 (Def, Ghdl_Kind_Si32_Node);

            when Iir_Kind_Floating_Type_Definition =>
               return Create_Floating_Type_Desc (Def);

            when Iir_Kind_Floating_Subtype_Definition =>
               return Create_Static_Scalar_Subtype_Desc
                 (Def, Ghdl_Kind_Sf64_Node);

            when Iir_Kind_Array_Subtype_Definition =>
               return Create_Static_Array_Subtype_Desc (Def);

            when Iir_Kind_Physical_Type_Definition =>
               return Create_Physical_Type_Desc (Def);

            when Iir_Kind_Physical_Subtype_Definition =>
               return Create_Static_Scalar_Subtype_Desc
                 (Def, Ghdl_Kind_Sp64_Node);

            when Iir_Kind_Array_Type_Definition =>
               return Create_Static_Array_Type_Desc (Def);

            when Iir_Kind_Record_Type_Definition =>
               return Create_Static_Record_Type_Desc (Def);

            when Iir_Kind_Access_Type_Definition =>
               return Create_Static_Access_Type_Desc (Def);

            when others =>
               Error_Kind ("get_static_type_definition_type_desc", Def);
         end case;
      end Get_Static_Type_Definition_Type_Desc;

      procedure Create_Type_Range_Var (Def : Iir)
      is
         Info : Type_Info_Acc;
         Base_Info : Type_Info_Acc;
         Val : O_Enode;
         Suffix : String (1 .. 3) := "xTR";
      begin
         Info := Get_Info (Def);
         case Get_Kind (Def) is
            when Iir_Kinds_Subtype_Definition =>
               Suffix (1) := 'S'; -- "STR";
            when Iir_Kind_Enumeration_Type_Definition =>
               Suffix (1) := 'B'; -- "BTR";
            when others =>
               raise Internal_Error;
         end case;
         Base_Info := Get_Info (Get_Base_Type (Def));
         case Get_Type_Staticness (Def) is
            when None
              | Globally =>
               Info.T.Range_Var := Create_Var
                 (Create_Var_Identifier (Suffix), Base_Info.T.Range_Type);
            when Locally =>
               if Global_Storage = O_Storage_External then
                  --  Do not create the value of the type desc, since it
                  --  is never dereferenced in a static type desc.
                  Val := O_Enode_Null;
               else
                  Val := Create_Static_Type_Definition_Type_Range (Def);
               end if;
               Info.T.Range_Var := Create_Global_Const
                 (Create_Identifier (Suffix),
                  Base_Info.T.Range_Type, Global_Storage, Val);
            when Unknown =>
               raise Internal_Error;
         end case;
      end Create_Type_Range_Var;


      --  Call HANDLE_A_SUBTYPE for all type/subtypes declared with DEF
      --  (of course, this is a noop if DEF is not a composite type).
      generic
         with procedure Handle_A_Subtype (Atype : Iir);
      procedure Handle_Anonymous_Subtypes (Def : Iir);

      procedure Handle_Anonymous_Subtypes (Def : Iir) is
      begin
         case Get_Kind (Def) is
            when Iir_Kind_Array_Type_Definition
              | Iir_Kind_Array_Subtype_Definition =>
               declare
                  Asub : Iir;
               begin
                  Asub := Get_Element_Subtype (Def);
                  if Is_Anonymous_Type_Definition (Asub) then
                     Handle_A_Subtype (Asub);
                  end if;
               end;
            when Iir_Kind_Record_Type_Definition =>
               declare
                  List : Iir_List;
                  El : Iir;
                  Asub : Iir;
               begin
                  List := Get_Element_Declaration_List (Def);
                  for I in Natural loop
                     El := Get_Nth_Element (List, I);
                     exit when El = Null_Iir;
                     Asub := Get_Type (El);
                     if Is_Anonymous_Type_Definition (Asub) then
                        Handle_A_Subtype (Asub);
                     end if;
                  end loop;
               end;
            when others =>
               null;
         end case;
      end Handle_Anonymous_Subtypes;

      procedure Create_Type_Desc_Var (Def : Iir);
      procedure Create_Type_Desc_Var_For_A_Subtype (Def : Iir) is
      begin
         Create_Type_Desc_Var (Def);
      end Create_Type_Desc_Var_For_A_Subtype;

      procedure Create_Type_Desc_Var_Depend is new Handle_Anonymous_Subtypes
        (Handle_A_Subtype => Create_Type_Desc_Var_For_A_Subtype);

      procedure Create_Type_Desc_Var (Def : Iir)
      is
         Info : Type_Info_Acc;
         Val : O_Enode;
         Suffix : String (1 .. 3) := "xTD";
      begin
         Info := Get_Info (Def);
         if Info.Type_Desc /= null then
            return;
         end if;

         Create_Type_Desc_Var_Depend (Def);

         if Get_Kind (Def) in Iir_Kinds_Subtype_Definition then
            Suffix (1) := 'S'; -- "STD"
         else
            Suffix (1) := 'B'; -- "BTD"
         end if;

         -- Compute type desc staticness.
         if not Info.Type_Desc_Static then
            Info.Type_Desc := Create_Var (Create_Var_Identifier (Suffix),
                                          Info.Type_Desc_Type);
         else
            if Global_Storage = O_Storage_External then
               --  Do not create the value of the type desc, since it
               --  is never dereferenced in a static type desc.
               Val := O_Enode_Null;
            else
               Val := Get_Static_Type_Definition_Type_Desc (Def);
            end if;
            Info.Type_Desc := Create_Global_Const
              (Create_Identifier (Suffix),
               Info.Type_Desc_Type, Global_Storage, Val);
         end if;
      end Create_Type_Desc_Var;

      procedure Create_Subtype_Info_From_Type (Subtype_Info : Type_Info_Acc;
                                               Base_Info : Type_Info_Acc)
      is
      begin
         Subtype_Info.Ortho_Type := Base_Info.Ortho_Type;
         Subtype_Info.Ortho_Ptr_Type := Base_Info.Ortho_Ptr_Type;
         --Subtype_Info.Sig_Ref_Type := Base_Info.Sig_Ref_Type;
         Subtype_Info.Type_Mode := Base_Info.Type_Mode;
      end Create_Subtype_Info_From_Type;

      --  Note: boolean types are translated by translate_bool_type_definition!
      procedure Translate_Type_Definition
        (Def : Iir; With_Vars : Boolean := True)
      is
         Info, Base_Info : Type_Info_Acc;
         Base_Type : Iir;
         Completion : Type_Completion_Kind;
      begin
         --  If the definition is already translated, return now.
         Info := Get_Info (Def);
         if Info /= null then
            if Info.Type_Mode /= Type_Mode_Unknown then
               return;
            else
               Completion := Completed_Type;
            end if;
         else
            Info := Add_Info (Def, Kind_Type);
            Completion := Full_Type;
         end if;

         Base_Type := Get_Base_Type (Def);
         Base_Info := Get_Info (Base_Type);

         case Get_Kind (Def) is
            when Iir_Kind_Enumeration_Type_Definition =>
               Translate_Enumeration_Type (Def);
               Create_Scalar_Type_Range_Type (Def, True);
               Info.Type_Desc_Type := Ghdl_Desc_E8_Type_Node;
               Info.Type_Desc_Static := True;
               Create_Type_Range_Var (Def);
               Create_Type_Desc_Var (Def);

            when Iir_Kind_Enumeration_Subtype_Definition =>
               Create_Subtype_Info_From_Type (Info, Base_Info);
               Info.T := Base_Info.T;
               Info.Type_Desc_Type := Ghdl_Desc_Scalar_Subtype_Node;
               Info.Type_Desc_Static :=
                 Get_Expr_Staticness (Get_Range_Constraint (Def)) = Locally;
               if With_Vars then
                  Create_Type_Range_Var (Def);
                  Create_Type_Desc_Var (Def);
               end if;

            when Iir_Kind_Integer_Type_Definition =>
               Translate_Integer_Type (Def);
               --Info.Sig_Ref_Type := Ghdl_Scalar_Signal_Index;
               Create_Scalar_Type_Range_Type (Def, True);
               Info.Type_Desc_Type := Ghdl_Desc_I32_Type_Node;
               Info.Type_Desc_Static := True;
               Create_Type_Desc_Var (Def);

            when Iir_Kind_Integer_Subtype_Definition =>
               Translate_Type_Definition (Base_Type);
               Base_Info := Get_Info (Base_Type);
               Create_Subtype_Info_From_Type (Info, Base_Info);
               Info.T := Base_Info.T;
               Info.Type_Desc_Type := Ghdl_Desc_Scalar_Subtype_Node;
               Info.Type_Desc_Static :=
                 Get_Expr_Staticness (Get_Range_Constraint (Def)) = Locally;
               if With_Vars then
                  Create_Type_Range_Var (Def);
                  Create_Type_Desc_Var (Def);
               end if;

            when Iir_Kind_Physical_Type_Definition =>
               Translate_Physical_Type (Def);
               --Info.Sig_Ref_Type := Ghdl_Scalar_Signal_Index;
               Create_Scalar_Type_Range_Type (Def, False);
               Info.Type_Desc_Type := Ghdl_Desc_P64_Type_Node;
               Info.Type_Desc_Static := True;
               Create_Type_Desc_Var (Def);

            when Iir_Kind_Physical_Subtype_Definition =>
               Translate_Type_Definition (Base_Type);
               Base_Info := Get_Info (Base_Type);
               Create_Subtype_Info_From_Type (Info, Base_Info);
               Info.T := Base_Info.T;
               Info.Type_Desc_Type := Ghdl_Desc_Scalar_Subtype_Node;
               Info.Type_Desc_Static :=
                 Get_Expr_Staticness (Get_Range_Constraint (Def)) = Locally;
               if With_Vars then
                  Create_Type_Range_Var (Def);
                  Create_Type_Desc_Var (Def);
               end if;

            when Iir_Kind_Floating_Type_Definition =>
               Translate_Floating_Type (Def);
               --Info.Sig_Ref_Type := Ghdl_Scalar_Signal_Index;
               Info.Type_Desc_Type := Ghdl_Desc_F64_Type_Node;
               Info.Type_Desc_Static := True;
               Create_Scalar_Type_Range_Type (Def, False);
               Create_Type_Desc_Var (Def);

            when Iir_Kind_Floating_Subtype_Definition =>
               Translate_Type_Definition (Base_Type);
               Base_Info := Get_Info (Base_Type);
               Create_Subtype_Info_From_Type (Info, Base_Info);
               Info.T := Base_Info.T;
               Info.Type_Desc_Type := Ghdl_Desc_Scalar_Subtype_Node;
               Info.Type_Desc_Static :=
                 Get_Expr_Staticness (Get_Range_Constraint (Def)) = Locally;
               if With_Vars then
                  Create_Type_Range_Var (Def);
                  Create_Type_Desc_Var (Def);
               end if;

            when Iir_Kind_Array_Type_Definition =>
               declare
                  El_Type : Iir;
                  Mark : Id_Mark_Type;
               begin
                  El_Type := Get_Element_Subtype (Def);
                  if Get_Info (El_Type) = null then
                     Push_Identifier_Prefix (Mark, "ET");
                     Translate_Type_Definition (El_Type);
                     Pop_Identifier_Prefix (Mark);
                  end if;
               end;
               Info.T := Ortho_Info_Type_Array_Init;
               Translate_Array_Type (Def);
            -- Info.Type_Range_Type := Create_Array_Type_Bounds_Type (Def, Id);
               Info.Type_Desc_Type := Ghdl_Desc_Array_Type_Node;
               Info.Type_Desc_Static := Get_Array_Type_Desc_Static (Def);
               Create_Array_Type_Index_Desc_Array_Var (Def);
               Create_Type_Desc_Var (Def);

            when Iir_Kind_Unconstrained_Array_Subtype_Definition =>
               Free_Info (Def);
               Set_Info (Def, Base_Info);

            when Iir_Kind_Array_Subtype_Definition =>
               Base_Info := Get_Info (Base_Type);
               if Base_Info = null then
                  declare
                     Mark : Id_Mark_Type;
                  begin
                     Push_Identifier_Prefix (Mark, "BT");
                     Translate_Type_Definition (Base_Type);
                     Pop_Identifier_Prefix (Mark);
                     Base_Info := Get_Info (Base_Type);
                  end;
               end if;
               Translate_Array_Subtype (Def);
               Info.T := Base_Info.T;
            --   Info.Sig_Ref_Type :=
            --     Translate_Array_Subtype_Sig_Ref (Def, In_Expr);
               --Info.Type_Range_Type := Base_Info.Type_Range_Type;
               Info.Type_Desc_Type := Ghdl_Desc_Subarray_Type_Node;
               Info.Type_Desc_Static := Get_Array_Type_Desc_Static (Def);
               if With_Vars then
                  Create_Array_Subtype_Bounds_Var (Def, False);
                  Create_Type_Desc_Var (Def);
               end if;

            when Iir_Kind_Record_Type_Definition =>
               Translate_Record_Type (Def, Completion);
               Info.Type_Desc_Type := Ghdl_Desc_Record_Type_Node;
               Info.Type_Desc_Static := Get_Record_Type_Desc_Static (Def);
               Info.T := Ortho_Info_Type_Record_Init;
               Create_Record_Type_El_Desc_Array_Var (Def);
               if With_Vars then
                  Create_Type_Desc_Var (Def);
               end if;

            when Iir_Kind_Record_Subtype_Definition =>
               Free_Info (Def);
               Set_Info (Def, Base_Info);

            when Iir_Kind_Access_Type_Definition =>
               Translate_Access_Type (Def);
               Info.Type_Desc_Type := Ghdl_Desc_Access_Type_Node;
               Info.Type_Desc_Static :=
                 Get_Info (Get_Designated_Type (Def)).Type_Desc_Static;
               if With_Vars then
                  Create_Type_Desc_Var (Def);
               end if;

            when Iir_Kind_File_Type_Definition =>
               Translate_File_Type (Def);
               Info.Type_Desc_Type := O_Tnode_Null;

            when Iir_Kind_Incomplete_Type_Definition =>
               Translate_Incomplete_Type (Def);

            when others =>
               Error_Kind ("translate_type_definition", Def);
         end case;
      end Translate_Type_Definition;

      procedure Translate_Bool_Type_Definition (Def : Iir)
      is
         Decl : Iir;
         Id : Name_Id;
         Info, Base_Info : Type_Info_Acc;
         Base_Type : Iir;
         Td_Name : O_Ident;
         Val : O_Enode;
      begin
         --  If the definition is already translated, return now.
         Info := Get_Info (Def);
         if Info /= null then
            raise Internal_Error;
         end if;

         Info := Add_Info (Def, Kind_Type);
         Base_Type := Get_Base_Type (Def);
         Decl := Get_Type_Declarator (Def);

         Id := Get_Identifier (Decl);

         if Get_Kind (Def) /= Iir_Kind_Enumeration_Type_Definition then
            raise Internal_Error;
         end if;
         Translate_Bool_Type (Def);
         Info.Type_Desc_Type := Ghdl_Desc_Bool_Type;
         Info.Type_Desc_Static := True;

         --  Backward annotation.
         Base_Info := Get_Info (Base_Type);

         Td_Name := Create_Identifier ("BTD");
         if Get_Type_Staticness (Def) /= Locally then
            raise Internal_Error;
         end if;
         if Global_Storage = O_Storage_External then
            --  Do not create the value of the type desc, since it
            --  is never dereferenced in a static type desc.
            Val := O_Enode_Null;
         else
            Val := Create_Bool_Type_Desc (Def);
         end if;
         Info.Type_Desc := Create_Global_Const
           (Td_Name, Info.Type_Desc_Type, Global_Storage, Val);

         --  This is usually done in translate_type_definition, but boolean
         --  types are not handled by translate_type_definition.
         Create_Scalar_Type_Range_Type (Def, True);
      end Translate_Bool_Type_Definition;

      procedure Translate_Type_Subprograms
        (Decl : Iir_Type_Declaration; Block : Iir)
      is
         Def : Iir;
         Tinfo : Type_Info_Acc;
         Id : Name_Id;
      begin
         Def := Get_Type (Decl);

         if Get_Kind (Def) in Iir_Kinds_Subtype_Definition then
            --  Also elaborate the base type, iff DEF and its BASE_TYPE have
            --  been declared by the same type declarator.  This avoids several
            --  elaboration of the same type.
            Def := Get_Base_Type (Def);
            if Get_Type_Declarator (Def) /= Decl then
               --  Can this happen ??
               raise Internal_Error;
            end if;
         end if;

         Tinfo := Get_Info (Def);
         if Tinfo.C = null or else Tinfo.C.Builder_Need_Func = False then
            return;
         end if;

         --  Declare subprograms.
         Id := Get_Identifier (Decl);
         Create_Builder_Subprogram_Decl (Tinfo, Id, Block, Mode_Value);
         Create_Builder_Subprogram_Decl (Tinfo, Id, Block, Mode_Signal);

         if Global_Storage = O_Storage_External then
            return;
         end if;

         --  Define subprograms.
         case Get_Kind (Def) is
            when Iir_Kind_Array_Type_Definition =>
               Create_Array_Type_Builder (Def, Mode_Value);
               Create_Array_Type_Builder (Def, Mode_Signal);
            when Iir_Kind_Record_Type_Definition =>
               Create_Record_Type_Builder (Def, Mode_Value);
               Create_Record_Type_Builder (Def, Mode_Signal);
            when others =>
               Error_Kind ("translate_type_subprograms", Def);
         end case;
      end Translate_Type_Subprograms;

      --  Initialize the objects related to a type (type range and type
      --  descriptor).
      procedure Elab_Type_Definition (Def : Iir);
      procedure Elab_Type_Definition_Depend is new Handle_Anonymous_Subtypes
        (Handle_A_Subtype => Elab_Type_Definition);
      procedure Elab_Type_Definition (Def : Iir)
      is
         Info : Type_Info_Acc;
         Id : Name_Id;
         Decl : Iir;
         Base_Type : Iir;
      begin
         Info := Get_Info (Def);
         if Info.Type_Desc_Static then
            --  Preelaborated type.
            return;
         end if;

         Decl := Get_Type_Declarator (Def);

         case Get_Kind (Def) is
            when Iir_Kinds_Subtype_Definition =>
               --  Also elaborate the base type, iff DEF and its BASE_TYPE
               --  have been declared by the same type declarator.  This
               --  avoids several elaboration of the same type.
               Base_Type := Get_Base_Type (Def);
               if Get_Type_Declarator (Base_Type) = Decl then
                  Elab_Type_Definition (Base_Type);
               end if;
            when Iir_Kind_Incomplete_Type_Definition =>
               --  Nothing to do.
               return;
            when others =>
               null;
         end case;

         Elab_Type_Definition_Depend (Def);

         if Decl = Null_Iir then
            Id := Null_Identifier;
         else
            Id := Get_Identifier (Decl);
         end if;

         Create_Type_Definition_Type_Range (Def);
         Create_Type_Definition_Size_Var (Def);
         Create_Type_Definition_Type_Desc (Def, Id);
      end Elab_Type_Definition;

      procedure Translate_Named_Type_Definition (Def : Iir; Id : Name_Id)
      is
         Mark : Id_Mark_Type;
      begin
         Push_Identifier_Prefix (Mark, Id);
         Chap3.Translate_Type_Definition (Def);
         Pop_Identifier_Prefix (Mark);
      end Translate_Named_Type_Definition;

      procedure Translate_Anonymous_Type_Definition (Def : Iir) is
         Mark : Id_Mark_Type;
      begin
         Push_Identifier_Prefix_Uniq (Mark);
         Chap3.Translate_Type_Definition (Def, False);
         Pop_Identifier_Prefix (Mark);
      end Translate_Anonymous_Type_Definition;

      procedure Translate_Object_Subtype (Decl : Iir;
                                          With_Vars : Boolean := True)
      is
         Mark : Id_Mark_Type;
         Mark2 : Id_Mark_Type;
         Def : Iir;
      begin
         Def := Get_Type (Decl);
         if Is_Anonymous_Type_Definition (Def) then
            Push_Identifier_Prefix (Mark, Get_Identifier (Decl));
            Push_Identifier_Prefix (Mark2, "OT");
            Chap3.Translate_Type_Definition (Def, With_Vars);
            Pop_Identifier_Prefix (Mark2);
            Pop_Identifier_Prefix (Mark);
         end if;
      end Translate_Object_Subtype;

      procedure Elab_Object_Subtype (Def : Iir) is
      begin
         if Is_Anonymous_Type_Definition (Def) then
            Elab_Type_Definition (Def);
         end if;
      end Elab_Object_Subtype;

      procedure Elab_Type_Declaration (Decl : Iir_Type_Declaration)
      is
      begin
         Elab_Type_Definition (Get_Type (Decl));
      end Elab_Type_Declaration;

      procedure Elab_Subtype_Declaration (Decl : Iir_Subtype_Declaration)
      is
      begin
         Elab_Type_Definition (Get_Type (Decl));
      end Elab_Subtype_Declaration;

      function Get_Thin_Array_Length (Atype : Iir) return O_Enode
      is
         Index_List : Iir_List;
         Nbr_Dim : Natural;
         Val : Iir_Int64;
         Rng : Iir;
      begin
         Index_List := Get_Index_List (Atype);
         Nbr_Dim := Get_Nbr_Elements (Index_List);
         Val := 1;
         for I in 0 .. Nbr_Dim - 1 loop
            Rng := Get_Range_Constraint
              (Get_Nth_Element (Index_List, I));
            Val := Val * Eval_Discrete_Range_Length (Rng);
         end loop;
         return New_Unsigned_Literal (Ghdl_Index_Type, Unsigned_64 (Val));
      end Get_Thin_Array_Length;

      function Get_Bounds_Ptr_Length (Ptr : O_Lnode; Atype : Iir)
        return O_Enode
      is
         Index_List : Iir_List;
         Index_Type : Iir;
         Nbr_Dim : Natural;
         Dim_Length : O_Enode;
         Res : O_Enode;
         Type_Info : Type_Info_Acc;
         Index_Info : Type_Info_Acc;
      begin
         Index_List := Get_Index_List (Atype);
         Nbr_Dim := Get_Nbr_Elements (Index_List);

         Type_Info := Get_Info (Get_Base_Type (Atype));
         for Dim in 1 .. Nbr_Dim loop
            Index_Type := Get_Nth_Element (Index_List, Dim - 1);
            Index_Info := Get_Info (Get_Base_Type (Index_Type));
            Dim_Length := New_Value
              (New_Selected_Element
               (New_Selected_Element (New_Acc_Value (Ptr),
                                      Type_Info.T.Bounds_Vector (Dim)),
                Index_Info.T.Range_Length));
            if Dim = 1 then
               Res := Dim_Length;
            else
               Res := New_Dyadic_Op (ON_Mul_Ov, Res, Dim_Length);
            end if;
         end loop;
         return Res;
      end Get_Bounds_Ptr_Length;

      function Bounds_To_Range (B : Mnode; Atype : Iir; Dim : Positive)
        return Mnode
      is
         Tinfo : Type_Info_Acc;
         Index_Type : Iir;
         Iinfo : Type_Info_Acc;
      begin
         Tinfo := Get_Type_Info (B);
         Index_Type := Get_Nth_Element
           (Get_Index_List (Get_Base_Type (Atype)), Dim - 1);
         Iinfo := Get_Info (Get_Base_Type (Index_Type));
         return Lv2M (New_Selected_Element (M2Lv (B),
                                            Tinfo.T.Bounds_Vector (Dim)),
                      Iinfo,
                      Get_Object_Kind (B),
                      Iinfo.T.Range_Type,
                      Iinfo.T.Range_Ptr_Type);
      end Bounds_To_Range;

      function Range_To_Length (R : Mnode) return Mnode
      is
         Tinfo : Type_Info_Acc;
      begin
         Tinfo := Get_Type_Info (R);
         return Lv2M (New_Selected_Element (M2Lv (R),
                                            Tinfo.T.Range_Length),
                      Tinfo,
                      Mode_Value);
      end Range_To_Length;

      function Range_To_Dir (R : Mnode) return Mnode
      is
         Tinfo : Type_Info_Acc;
      begin
         Tinfo := Get_Type_Info (R);
         return Lv2M (New_Selected_Element (M2Lv (R),
                                            Tinfo.T.Range_Dir),
                      Tinfo,
                      Mode_Value);
      end Range_To_Dir;

      function Range_To_Left (R : Mnode) return Mnode
      is
         Tinfo : Type_Info_Acc;
      begin
         Tinfo := Get_Type_Info (R);
         return Lv2M (New_Selected_Element (M2Lv (R),
                                            Tinfo.T.Range_Left),
                      Tinfo,
                      Mode_Value);
      end Range_To_Left;

      function Range_To_Right (R : Mnode) return Mnode
      is
         Tinfo : Type_Info_Acc;
      begin
         Tinfo := Get_Type_Info (R);
         return Lv2M (New_Selected_Element (M2Lv (R),
                                            Tinfo.T.Range_Right),
                      Tinfo,
                      Mode_Value);
      end Range_To_Right;

      function Get_Array_Type_Bounds (Info : Type_Info_Acc) return Mnode
      is
      begin
         case Info.Type_Mode is
            when Type_Mode_Fat_Array =>
               raise Internal_Error;
            when Type_Mode_Array
              | Type_Mode_Ptr_Array =>
               return Varv2M (Info.T.Array_Bounds,
                              Info, Mode_Value,
                              Info.T.Bounds_Type,
                              Info.T.Bounds_Ptr_Type);
            when others =>
               raise Internal_Error;
         end case;
      end Get_Array_Type_Bounds;

      function Get_Array_Type_Bounds (Atype : Iir) return Mnode is
      begin
         return Get_Array_Type_Bounds (Get_Info (Atype));
      end Get_Array_Type_Bounds;

      function Get_Array_Bounds (Arr : Mnode) return Mnode
      is
         Info : Type_Info_Acc;
      begin
         Info := Get_Type_Info (Arr);
         case Info.Type_Mode is
            when Type_Mode_Fat_Array
              | Type_Mode_Fat_Acc =>
               declare
                  Kind : Object_Kind_Type;
               begin
                  Kind := Get_Object_Kind (Arr);
                  return Lp2M
                    (New_Selected_Element (M2Lv (Arr),
                                           Info.T.Bounds_Field (Kind)),
                     Info,
                     Mode_Value,
                     Info.T.Bounds_Type,
                     Info.T.Bounds_Ptr_Type);
               end;
            when Type_Mode_Array
              | Type_Mode_Ptr_Array =>
               return Get_Array_Type_Bounds (Info);
            when others =>
               raise Internal_Error;
         end case;
      end Get_Array_Bounds;

      function Get_Array_Range (Arr : Mnode; Atype : Iir; Dim : Positive)
                               return Mnode is
      begin
         return Bounds_To_Range (Get_Array_Bounds (Arr), Atype, Dim);
      end Get_Array_Range;

      function Get_Array_Type_Length (Atype : Iir) return O_Enode
      is
         Index_List : Iir_List;
         Index_Type : Iir;
         Nbr_Dim : Natural;
         Dim_Length : O_Enode;
         Res : O_Enode;
         Type_Info : Type_Info_Acc;
         Binfo : Type_Info_Acc;
         Index_Info : Type_Info_Acc;
         Bounds : Mnode;
      begin
         Index_List := Get_Index_List (Atype);
         Nbr_Dim := Get_Nbr_Elements (Index_List);

         --  Handle thin array case.
         Type_Info := Get_Info (Atype);
         case Type_Info.Type_Mode is
            when Type_Mode_Ptr_Array =>
               Bounds := Get_Array_Type_Bounds (Atype);
               if Nbr_Dim > 1 then
                  Bounds := Stabilize (Bounds);
               end if;
            when Type_Mode_Array =>
               return Get_Thin_Array_Length (Atype);
            when others =>
               raise Internal_Error;
         end case;

         Binfo := Get_Info (Get_Base_Type (Atype));
         for Dim in 1 .. Nbr_Dim loop
            Index_Type := Get_Nth_Element (Index_List, Dim - 1);
            Index_Info := Get_Info (Get_Base_Type (Index_Type));
            Dim_Length :=
              M2E (Range_To_Length (Bounds_To_Range (Bounds, Atype, Dim)));
            if Dim = 1 then
               Res := Dim_Length;
            else
               Res := New_Dyadic_Op (ON_Mul_Ov, Res, Dim_Length);
            end if;
         end loop;
         return Res;
      end Get_Array_Type_Length;

      function Get_Array_Length (Arr : Mnode; Atype : Iir) return O_Enode
      is
         Index_List : Iir_List;
         Index_Type : Iir;
         Nbr_Dim : Natural;
         Dim_Length : O_Enode;
         Res : O_Enode;
         Type_Info : Type_Info_Acc;
         Binfo : Type_Info_Acc;
         Index_Info : Type_Info_Acc;
         Bounds : Mnode;
         B : Mnode;
      begin
         Index_List := Get_Index_List (Atype);
         Nbr_Dim := Get_Nbr_Elements (Index_List);

         --  Handle thin array case.
         Type_Info := Get_Info (Atype);
         case Type_Info.Type_Mode is
            when Type_Mode_Ptr_Array
              | Type_Mode_Array =>
               return Get_Array_Type_Length (Atype);
            when Type_Mode_Fat_Array =>
               null;
            when others =>
               raise Internal_Error;
         end case;

         Binfo := Get_Info (Get_Base_Type (Atype));
         for Dim in 1 .. Nbr_Dim loop
            Index_Type := Get_Nth_Element (Index_List, Dim - 1);
            Index_Info := Get_Info (Get_Base_Type (Index_Type));
            B := Get_Array_Bounds (Arr);
            Dim_Length :=
              M2E (Range_To_Length (Bounds_To_Range (B, Atype, Dim)));
            if Dim = 1 then
               Res := Dim_Length;
            else
               Res := New_Dyadic_Op (ON_Mul_Ov, Res, Dim_Length);
            end if;
         end loop;
         return Res;
      end Get_Array_Length;

      function Get_Array_Base (Arr : Mnode) return Mnode
      is
         Info : Type_Info_Acc;
      begin
         Info := Get_Type_Info (Arr);
         case Info.Type_Mode is
            when Type_Mode_Fat_Array
              | Type_Mode_Fat_Acc =>
               declare
                  F : O_Fnode;
                  Kind : Object_Kind_Type;
               begin
                  Kind := Get_Object_Kind (Arr);
                  F := Info.T.Base_Field (Get_Object_Kind (Arr));
                  return Lp2M
                    (New_Selected_Element (M2Lv (Arr),
                                           Info.T.Base_Field (Kind)),
                     Info,
                     Get_Object_Kind (Arr),
                     Info.T.Base_Type (Kind),
                     Info.T.Base_Ptr_Type (Kind));
               end;
            when Type_Mode_Array
              | Type_Mode_Ptr_Array =>
               return Arr;
            when others =>
               raise Internal_Error;
         end case;
      end Get_Array_Base;

      function Index_Base (Base : Mnode; Atype : Iir; Index : O_Enode)
        return Mnode
      is
         El_Type : Iir;
      begin
         El_Type := Get_Element_Subtype (Atype);
         return Lo2M (New_Indexed_Element (M2Lv (Base), Index),
                      Get_Info (El_Type), Get_Object_Kind (Base));
      end Index_Base;

      function Get_Array_Ptr_Base_Ptr
        (Ptr : O_Lnode; Atype : Iir; Is_Sig : Object_Kind_Type)
        return O_Lnode
      is
         Tinfo : Type_Info_Acc;
      begin
         Tinfo := Get_Info (Atype);
         case Tinfo.Type_Mode is
            when Type_Mode_Fat_Array =>
               return New_Selected_Element
                 (New_Access_Element (New_Value (Ptr)),
                  Tinfo.T.Base_Field (Is_Sig));
            when Type_Mode_Array
              | Type_Mode_Ptr_Array =>
               return Ptr;
            when others =>
               raise Internal_Error;
         end case;
      end Get_Array_Ptr_Base_Ptr;

      function Get_Array_Ptr_Range_Ptr (Ptr : O_Lnode;
                                        Array_Type : Iir;
                                        Dim : Natural;
                                        Is_Sig : Object_Kind_Type)
        return O_Enode
      is
         Array_Info : Type_Info_Acc;
         Res : O_Lnode;
         Index_Type : Iir;
         Index_Info : Type_Info_Acc;
      begin
         Array_Info := Get_Info (Array_Type);
         Index_Type := Get_Nth_Element (Get_Index_List (Array_Type), Dim - 1);
         Index_Info := Get_Info (Get_Base_Type (Index_Type));
         case Array_Info.Type_Mode is
            when Type_Mode_Array
              | Type_Mode_Ptr_Array =>
               --  Extract bound variable.
               Res := Get_Var (Array_Info.T.Array_Bounds);
            when Type_Mode_Fat_Array =>
               --  From fat record, extract bounds field.
               Res := New_Acc_Value
                 (New_Selected_Acc_Value
                  (Ptr, Array_Info.T.Bounds_Field (Is_Sig)));
            when others =>
               raise Internal_Error;
         end case;
         --  Extract the range for the dimension.
         Res := New_Selected_Element (Res, Array_Info.T.Bounds_Vector (Dim));
         return New_Address (Res, Index_Info.T.Range_Ptr_Type);
      end Get_Array_Ptr_Range_Ptr;

      function Get_Array_Ptr_Bounds_Ptr
        (Ptr : O_Lnode; Atype : Iir; Is_Sig : Object_Kind_Type)
        return O_Enode
      is
         Info : Type_Info_Acc;
      begin
         Info := Get_Info (Atype);
         case Info.Type_Mode is
            when Type_Mode_Fat_Array =>
               return New_Value
                 (New_Selected_Element (New_Acc_Value (Ptr),
                                        Info.T.Bounds_Field (Is_Sig)));
            when Type_Mode_Array
              | Type_Mode_Ptr_Array =>
               return New_Address (Get_Var (Info.T.Array_Bounds),
                                   Info.T.Bounds_Ptr_Type);
            when others =>
               raise Internal_Error;
         end case;
      end Get_Array_Ptr_Bounds_Ptr;

      function Get_Array_Bounds_Ptr
        (Arr : O_Lnode; Arr_Type : Iir; Is_Sig : Object_Kind_Type)
        return O_Enode
      is
         Type_Info : Type_Info_Acc;
      begin
         Type_Info := Get_Info (Arr_Type);
         case Type_Info.Type_Mode is
            when Type_Mode_Fat_Array =>
               return New_Value (New_Selected_Element
                                 (Arr, Type_Info.T.Bounds_Field (Is_Sig)));
            when Type_Mode_Array
              | Type_Mode_Ptr_Array =>
               return New_Address (Get_Var (Type_Info.T.Array_Bounds),
                                   Type_Info.T.Bounds_Ptr_Type);
            when others =>
               --  Not an array!
               raise Internal_Error;
         end case;
      end Get_Array_Bounds_Ptr;

      procedure Allocate_Fat_Array_Base (Alloc_Kind : Allocation_Kind;
                                         Res : Mnode;
                                         Arr_Type : Iir;
                                         Is_Sig : Object_Kind_Type)
      is
         Dinfo : Type_Info_Acc;
         Length : O_Enode;
      begin
         Dinfo := Get_Info (Get_Base_Type (Arr_Type));
         --  Compute array size.
         Length := Get_Object_Size (Res, Arr_Type);
         --  Allocate the storage for the elements.
         New_Assign_Stmt
           (M2Lp (Chap3.Get_Array_Base (Res)),
            Gen_Alloc (Alloc_Kind, Length, Dinfo.T.Base_Ptr_Type (Is_Sig)));

         Open_Temp;
         if Dinfo.C /= null and then Dinfo.C.Builder_Need_Func then
            --  Build the type.
            Chap3.Gen_Call_Type_Builder (Res, Arr_Type);
         end if;
         Close_Temp;
      end Allocate_Fat_Array_Base;

      procedure Create_Array_Subtype (Sub_Type : Iir)
      is
         Mark : Id_Mark_Type;
      begin
         Push_Identifier_Prefix_Uniq (Mark);
         if Get_Info (Sub_Type) = null then
            --  Minimal subtype creation.
            Translate_Type_Definition (Sub_Type, False);
         end if;
         --  Force creation of variables.
         Chap3.Create_Array_Subtype_Bounds_Var (Sub_Type, True);
         Chap3.Create_Type_Definition_Size_Var (Sub_Type);
         Pop_Identifier_Prefix (Mark);
      end Create_Array_Subtype;

      function Get_Memory_Complex_1
        (Ptr : O_Lnode; Obj_Type : Iir; Kind : Object_Kind_Type)
        return O_Enode
      is
         Info : Type_Info_Acc;
      begin
         Info := Get_Info (Obj_Type);
         case Info.Type_Mode is
            when Type_Mode_Ptr_Array =>
               return New_Value (Ptr);
            when Type_Mode_Array =>
               return Get_Memory_Complex_1
                 (New_Indexed_Element
                  (Ptr, New_Unsigned_Literal (Ghdl_Index_Type, 0)),
                  Get_Element_Subtype (Obj_Type),
                  Kind);
            when Type_Mode_Record =>
               declare
                  El_List : Iir_List;
                  El : Iir_Element_Declaration;
                  El_Type : Iir;
                  El_Info : Type_Info_Acc;
               begin
                  El_List := Get_Element_Declaration_List
                    (Get_Base_Type (Obj_Type));
                  for I in Natural loop
                     El := Get_Nth_Element (El_List, I);
                     exit when El = Null_Iir;
                     El_Type := Get_Type (El);
                     El_Info := Get_Info (El_Type);
                     if El_Info.C /= null then
                        return Get_Memory_Complex_1
                          (New_Selected_Element
                           (Ptr, Get_Info (El).Field_Node (Kind)),
                           El_Type,
                           Kind);
                     end if;
                  end loop;
                  --  Record is known to be complex but has no complex
                  --  element.
                  raise Internal_Error;
               end;
            when Type_Mode_Scalar
              | Type_Mode_Unknown
              | Type_Mode_File
              | Type_Mode_Acc
              | Type_Mode_Fat_Acc
              | Type_Mode_Fat_Array =>
               raise Internal_Error;
         end case;
      end Get_Memory_Complex_1;

      --  VAR_PTR is a pointer to an object of type OBJ_TYPE (and kind KIND).
      --  This object is known to be of a complex type.
      --  Return the value of the first internal pointer of the object.
      function Get_Memory_Complex
        (Var_Ptr : O_Lnode; Obj_Type : Iir; Kind : Object_Kind_Type)
        return O_Enode
      is
         Info : Type_Info_Acc;
         Res : O_Enode;
      begin
         Info := Get_Info (Obj_Type);
         case Info.Type_Mode is
            when Type_Mode_Fat_Array
              | Type_Mode_Array
              | Type_Mode_Ptr_Array =>
               Res := Get_Memory_Complex_1
                 (New_Indexed_Element
                  (New_Acc_Value (Get_Array_Ptr_Base_Ptr
                                  (Var_Ptr, Obj_Type, Kind)),
                   New_Unsigned_Literal (Ghdl_Index_Type, 0)),
                  Get_Element_Subtype (Obj_Type),
                  Kind);
            when Type_Mode_Record =>
               Res := Get_Memory_Complex_1
                 (New_Acc_Value (Var_Ptr), Obj_Type, Kind);
            when Type_Mode_Non_Composite
              | Type_Mode_Unknown =>
               --  Cannot be a complex type.
               raise Internal_Error;
         end case;
         return New_Convert_Ov (Res, Char_Ptr_Type_Node);
      end Get_Memory_Complex;

      --  Copy SRC to DEST.
      --  Both have the same type, OTYPE.
      procedure Translate_Object_Copy (Dest : Mnode;
                                       Src : O_Enode;
                                       Obj_Type : Iir;
                                       Kind : Object_Kind_Type)
      is
         Info : Type_Info_Acc;
         D : Mnode;
      begin
         Info := Get_Info (Obj_Type);
         if Info.C /= null and then Info.C.Builder_Need_Func then
            D := Stabilize (Dest);
            --  A complex type that must be rebuilt.
            --  Save destinaton.
            --  Do the copy.
            case Info.Type_Mode is
               when Type_Mode_Fat_Array =>
                  --  a fat array.
                  Gen_Memcpy
                    (M2Addr (Chap3.Get_Array_Base (D)),
                     New_Value
                     (New_Selected_Element (New_Access_Element (Src),
                                            Info.T.Base_Field (Kind))),
                     Get_Object_Size (Dest, Obj_Type));
               when Type_Mode_Record
                 | Type_Mode_Array
                 | Type_Mode_Ptr_Array =>
                  Gen_Memcpy (M2Addr (D),
                              Src,
                              Get_Object_Size (Dest, Obj_Type));
               when Type_Mode_Unknown
                 | Type_Mode_File
                 | Type_Mode_Scalar
                 | Type_Mode_Acc
                 | Type_Mode_Fat_Acc =>
                  raise Internal_Error;
            end case;
            --  Rebuilt the type.
            Gen_Call_Type_Builder (D, Obj_Type);
         else
            case Info.Type_Mode is
               when Type_Mode_Scalar
                 | Type_Mode_Acc
                 | Type_Mode_File =>
                  --  Scalar or thin pointer.
                  New_Assign_Stmt (M2Lv (Dest), Src);
               when Type_Mode_Fat_Acc =>
                  --  a fat pointer.
                  declare
                     Var_S : O_Lnode;
                     Var_D : O_Lnode;
                  begin
                     Var_S := Create_Temp_Init (Info.Ortho_Ptr_Type (Kind),
                                                Src);
                     Var_D := Create_Temp_Init (Info.Ortho_Ptr_Type (Kind),
                                                M2Addr (Dest));
                     Copy_Fat_Access (Var_D, Var_S, Obj_Type);
                  end;
               when Type_Mode_Fat_Array =>
                  --  a fat array.
                  Gen_Memcpy
                    (M2Addr (Get_Array_Base (Dest)),
                     M2Addr (Get_Array_Base (E2M (Src, Info, Kind))),
                     Get_Object_Size (Dest, Obj_Type));
               when Type_Mode_Record
                 | Type_Mode_Ptr_Array
                 | Type_Mode_Array =>
                  Gen_Memcpy
                    (M2Addr (Dest), Src, Get_Object_Size (Dest, Obj_Type));
               when Type_Mode_Unknown =>
                  raise Internal_Error;
            end case;
         end if;
      end Translate_Object_Copy;

      function Get_Object_Size (Obj : Mnode; Obj_Type : Iir)
        return O_Enode
      is
         Type_Info : Type_Info_Acc;
         Kind : Object_Kind_Type;
      begin
         Type_Info := Get_Type_Info (Obj);
         Kind := Get_Object_Kind (Obj);
         if Type_Info.C /= null
           and then Type_Info.C.Size_Var (Kind) /= null
         then
            return New_Value (Get_Var (Type_Info.C.Size_Var (Kind)));
         end if;
         case Type_Info.Type_Mode is
            when Type_Mode_Non_Composite
              | Type_Mode_Array
              | Type_Mode_Record =>
               return New_Sizeof (Type_Info.Ortho_Type (Kind),
                                  Ghdl_Index_Type);
            when Type_Mode_Fat_Array =>
               declare
                  El_Type : Iir;
               begin
                  El_Type := Get_Element_Subtype (Obj_Type);
                  return New_Dyadic_Op
                    (ON_Mul_Ov,
                     Chap3.Get_Array_Length (Obj, Obj_Type),
                     Get_Object_Size (T2M (El_Type, Kind), El_Type));
               end;
            when others =>
               raise Internal_Error;
         end case;
      end Get_Object_Size;

      function Translate_Object_Allocation
        (Alloc_Kind : Allocation_Kind;
         Obj_Type : Iir;
         Bounds : O_Enode;
         Kind : Object_Kind_Type)
        return Mnode
      is
         Dinfo : Type_Info_Acc;
         L : O_Lnode;
         Res : Mnode;
      begin
         Dinfo := Get_Info (Obj_Type);
         if Dinfo.Type_Mode = Type_Mode_Fat_Array then
            --  Fat pointer to an unconstrained array type.
            L := Create_Temp (Dinfo.Ortho_Type (Kind));
            Res := Vv2M (L, Dinfo, Kind);

            --  Allocate memory for bounds.
            New_Assign_Stmt
              (M2Lp (Chap3.Get_Array_Bounds (Res)),
               Gen_Alloc (Alloc_Kind,
                          New_Sizeof (Dinfo.T.Bounds_Type, Ghdl_Index_Type),
                          Dinfo.T.Bounds_Ptr_Type));

            --  Copy bounds to the allocated area.
            Gen_Memcpy
              (M2Addr (Chap3.Get_Array_Bounds (Res)),
               Bounds,
               New_Sizeof (Dinfo.T.Bounds_Type, Ghdl_Index_Type));

            --  Allocate base.
            Allocate_Fat_Array_Base (Alloc_Kind, Res, Obj_Type, Kind);

            return Res;
         else
            --  FIXME: E2M doesn't work here for scalar types.
            return E2M (Gen_Alloc
                        (Alloc_Kind,
                         Chap3.Get_Object_Size (T2M (Obj_Type, Kind),
                                                Obj_Type),
                         Dinfo.Ortho_Ptr_Type (Kind)),
                        Dinfo, Kind);
         end if;
      end Translate_Object_Allocation;

      --  Performs deallocation of PARAM (the parameter of a deallocate call).
      procedure Translate_Object_Deallocation (Param : Iir)
      is
         --  Performs deallocation of field FIELD of type FTYPE of PTR.
         --  If FIELD is O_FNODE_NULL, deallocate PTR (of type FTYPE).
         --  Here, deallocate means freeing memory and clearing to null.
         procedure Deallocate_1
           (Ptr : Mnode; Field : O_Fnode; Ftype : O_Tnode)
         is
            Assocs : O_Assoc_List;
            L : O_Lnode;
         begin
            for I in 0 .. 1 loop
               L := M2Lv (Ptr);
               if Field /= O_Fnode_Null then
                  L := New_Selected_Element (L, Field);
               end if;
               case I is
                  when 0 =>
                     --  Call deallocator.
                     Start_Association (Assocs, Ghdl_Deallocate);
                     New_Association
                       (Assocs, New_Convert_Ov (New_Value (L), Ghdl_Ptr_Type));
                     New_Procedure_Call (Assocs);
                  when 1 =>
                     --  set the value to 0.
                     New_Assign_Stmt (L, New_Null_Access (Ftype));
               end case;
            end loop;
         end Deallocate_1;

         Param_Type : Iir;
         Val : Mnode;
         Info : Type_Info_Acc;
         Binfo : Type_Info_Acc;
      begin
         --  Compute parameter
         Val := Chap6.Translate_Name (Param);
         if Get_Object_Kind (Val) = Mode_Signal then
            raise Internal_Error;
         end if;
         Stabilize (Val);
         Param_Type := Get_Type (Param);
         Info := Get_Info (Param_Type);
         case Info.Type_Mode is
            when Type_Mode_Fat_Acc =>
               --  This is a fat pointer.
               --  Deallocate base and bounds.
               Binfo := Get_Info (Get_Designated_Type (Param_Type));
               Deallocate_1 (Val, Binfo.T.Base_Field (Mode_Value),
                             Binfo.T.Base_Ptr_Type (Mode_Value));
               Deallocate_1 (Val, Binfo.T.Bounds_Field (Mode_Value),
                             Binfo.T.Bounds_Ptr_Type);
            when Type_Mode_Acc =>
               --  This is a thin pointer.
               Deallocate_1 (Val, O_Fnode_Null,
                             Info.Ortho_Type (Mode_Value));
            when others =>
               raise Internal_Error;
         end case;
      end Translate_Object_Deallocation;

      function Not_In_Range (Value : O_Lnode; Atype : Iir) return O_Enode
      is
         Constr : Iir;

         function Gen_Compare (Left_Op, Right_Op : ON_Op_Kind) return O_Enode
         is
            L, R : O_Enode;
         begin
            L := New_Compare_Op
              (Left_Op,
               New_Value (Value),
               Chap14.Translate_Left_Type_Attribute (Atype),
               Ghdl_Bool_Type_Node);
            R := New_Compare_Op
              (Right_Op,
               New_Value (Value),
               Chap14.Translate_Right_Type_Attribute (Atype),
               Ghdl_Bool_Type_Node);
            return New_Dyadic_Op (ON_Or, L, R);
         end Gen_Compare;

         --Low, High : Iir;
         Var_Res : O_Lnode;
         If_Blk : O_If_Block;
      begin
         Constr := Get_Range_Constraint (Atype);

         if Get_Kind (Constr) = Iir_Kind_Range_Expression then
            --  Constraint is a range expression, therefore, direction is
            --  known.
            if Get_Expr_Staticness (Constr) = Locally then
               --  Range constraint is locally static
               --  FIXME: check low and high if they are not limits...
               --Low := Get_Low_Limit (Constr);
               --High := Get_High_Limit (Constr);
               null;
            end if;
            case Get_Direction (Constr) is
               when Iir_To =>
                  return Gen_Compare (ON_Lt, ON_Gt);
               when Iir_Downto =>
                  return Gen_Compare (ON_Gt, ON_Lt);
            end case;
         end if;

         --  Range constraint is not static
         --    full check (lot's of code ?).
         Var_Res := Create_Temp (Ghdl_Bool_Type_Node);
         Start_If_Stmt
           (If_Blk,
            New_Compare_Op (ON_Eq,
                            Chap14.Translate_Dir_Type_Attribute (Constr),
                            Ghdl_Dir_To_Node,
                            Ghdl_Bool_Type_Node));
         --  To.
         New_Assign_Stmt (Var_Res, Gen_Compare (ON_Lt, ON_Gt));
         New_Else_Stmt (If_Blk);
         --  Downto
         New_Assign_Stmt (Var_Res, Gen_Compare (ON_Gt, ON_Lt));
         Finish_If_Stmt (If_Blk);
         return New_Value (Var_Res);
      end Not_In_Range;

      function Need_Range_Check (Atype : Iir) return Boolean is
      begin
         case Get_Kind (Atype) is
            when Iir_Kind_Integer_Type_Definition
              | Iir_Kind_Enumeration_Type_Definition =>
               return False;
            when others =>
               return True;
         end case;
      end Need_Range_Check;

      procedure Check_Range (Value : O_Lnode; Expr : Iir; Atype : Iir)
      is
         If_Blk : O_If_Block;
      begin
         if not Need_Range_Check (Atype) then
            return;
         end if;

         if Expr /= Null_Iir
           and then Get_Expr_Staticness (Expr) = Locally
           and then Get_Type_Staticness (Atype) = Locally
         then
            if not Eval_Is_In_Bound (Expr, Atype) then
               Chap6.Gen_Bound_Error (Expr);
            end if;
         else
            Open_Temp;
            Start_If_Stmt (If_Blk, Not_In_Range (Value, Atype));
            Chap6.Gen_Bound_Error (Null_Iir);
            Finish_If_Stmt (If_Blk);
            Close_Temp;
         end if;
      end Check_Range;

      procedure Check_Array_Match (L_Type : Iir;
                                   L_Node : O_Lnode;
                                   L_Mode : Object_Kind_Type;
                                   R_Type : Iir;
                                   R_Node : O_Lnode;
                                   R_Mode : Object_Kind_Type;
                                   Loc : Iir)
      is
         L_Tinfo, R_Tinfo : Type_Info_Acc;
      begin
         L_Tinfo := Get_Info (L_Type);
         R_Tinfo := Get_Info (R_Type);
         if L_Tinfo.Type_Mode = Type_Mode_Array
           and R_Tinfo.Type_Mode = Type_Mode_Array
         then
            --  Both left and right are thin array.
            --  Check here the length are the same.
            declare
               L_Indexes : Iir_List;
               R_Indexes : Iir_List;
               L_El : Iir;
               R_El : Iir;
               Err : Boolean;
            begin
               L_Indexes := Get_Index_List (L_Type);
               R_Indexes := Get_Index_List (R_Type);
               Err := False;
               for I in Natural loop
                  L_El := Get_Nth_Element (L_Indexes, I);
                  R_El := Get_Nth_Element (R_Indexes, I);
                  exit when L_El = Null_Iir and R_El = Null_Iir;
                  if Eval_Discrete_Type_Length (L_El)
                    /= Eval_Discrete_Type_Length (R_El)
                  then
                     if not Err then
                        Chap6.Gen_Bound_Error (Loc);
                        Err := True;
                     end if;
                  end if;
               end loop;
            end;
         else
            --  Check length match.
            declare
               Index_List : Iir_List;
               Index : Iir;
               Cond : O_Enode;
               Sub_Cond : O_Enode;
            begin
               Index_List := Get_Index_List (L_Type);
               for I in Natural loop
                  Index := Get_Nth_Element (Index_List, I);
                  exit when Index = Null_Iir;
                  Sub_Cond := New_Compare_Op
                    (ON_Neq,
                     Chap6.Get_Array_Ptr_Bound_Length (L_Node, L_Type,
                                                       I + 1, L_Mode),
                     Chap6.Get_Array_Ptr_Bound_Length (R_Node, R_Type,
                                                       I + 1, R_Mode),
                     Ghdl_Bool_Type_Node);
                  if I = 0 then
                     Cond := Sub_Cond;
                  else
                     Cond := New_Dyadic_Op (ON_Or, Cond, Sub_Cond);
                  end if;
               end loop;
               Chap6.Check_Bound_Error (Cond, Loc, 0);
            end;
         end if;
      end Check_Array_Match;

      procedure Create_Range_From_Length
        (Index_Type : Iir; Length : O_Lnode; Range_Ptr : O_Lnode)
      is
         Iinfo : Type_Info_Acc;
         Op : ON_Op_Kind;
         Diff : O_Enode;
         Left_Bound : O_Enode;
         Var_Right : O_Lnode;
         If_Blk : O_If_Block;
         Range_Constr : Iir;
      begin
         Iinfo := Get_Info (Index_Type);
         Range_Constr := Get_Range_Constraint (Index_Type);
         Start_Declare_Stmt;
         New_Var_Decl (Var_Right, Get_Identifier ("right_bound"),
                       O_Storage_Local, Iinfo.Ortho_Type (Mode_Value));
         New_Assign_Stmt
           (New_Selected_Acc_Value (Range_Ptr, Iinfo.T.Range_Length),
            New_Value (Length));
         New_Assign_Stmt
           (New_Selected_Acc_Value (Range_Ptr, Iinfo.T.Range_Dir),
            Chap7.Translate_Static_Range_Dir (Range_Constr));
         case Get_Direction (Range_Constr) is
            when Iir_To =>
               Op := ON_Add_Ov;
            when Iir_Downto =>
               Op := ON_Sub_Ov;
         end case;

         Start_If_Stmt
           (If_Blk,
            New_Compare_Op (ON_Eq,
                            New_Value (Length),
                            New_Unsigned_Literal (Ghdl_Index_Type, 0),
                            Ghdl_Bool_Type_Node));
         --  Null range.
         New_Assign_Stmt
           (New_Selected_Acc_Value (Range_Ptr, Iinfo.T.Range_Left),
            Chap7.Translate_Range_Expression_Right (Range_Constr, Index_Type));
         New_Assign_Stmt
           (New_Selected_Acc_Value (Range_Ptr, Iinfo.T.Range_Right),
            Chap7.Translate_Range_Expression_Left (Range_Constr, Index_Type));

         New_Else_Stmt (If_Blk);
         New_Assign_Stmt
           (New_Selected_Acc_Value (Range_Ptr, Iinfo.T.Range_Left),
            Chap7.Translate_Range_Expression_Left (Range_Constr, Index_Type));
         Left_Bound := Chap7.Translate_Range_Expression_Left
           (Range_Constr, Index_Type);
         Diff := New_Convert_Ov
           (New_Dyadic_Op (ON_Sub_Ov,
                           New_Value (Length),
                           New_Unsigned_Literal (Ghdl_Index_Type, 1)),
            Iinfo.Ortho_Type (Mode_Value));
         New_Assign_Stmt (Var_Right,
                          New_Dyadic_Op (Op, Left_Bound, Diff));

         --   Check the right bounds is inside the bounds of the index type.
         Chap3.Check_Range (Var_Right, Null_Iir, Index_Type);
         New_Assign_Stmt
           (New_Selected_Acc_Value (Range_Ptr, Iinfo.T.Range_Right),
            New_Value (Var_Right));
         Finish_If_Stmt (If_Blk);
         Finish_Declare_Stmt;
      end Create_Range_From_Length;
   end Chap3;

   package body Chap4 is
      --  Get the ortho type for an object of mode MODE.
      function Get_Object_Type (Tinfo : Type_Info_Acc; Kind : Object_Kind_Type)
        return O_Tnode is
      begin
         if Tinfo.C /= null then
            case Tinfo.Type_Mode is
               when Type_Mode_Fat_Array =>
                  return Tinfo.Ortho_Type (Kind);
               when Type_Mode_Record
                 | Type_Mode_Array
                 | Type_Mode_Ptr_Array =>
                  --  For a complex type, use a pointer.
                  return Tinfo.Ortho_Ptr_Type (Kind);
               when others =>
                  raise Internal_Error;
            end case;
         else
            return Tinfo.Ortho_Type (Kind);
         end if;
      end Get_Object_Type;

      --  Get the ortho type for an object of mode MODE.
      function Get_Element_Type (Tinfo : Type_Info_Acc;
                                 Kind : Object_Kind_Type)
        return O_Tnode is
      begin
         if Tinfo.C /= null then
            --  Always use a pointer for a complex type.
            return Tinfo.Ortho_Ptr_Type (Kind);
         else
            return Tinfo.Ortho_Type (Kind);
         end if;
      end Get_Element_Type;

      procedure Create_Object (El : Iir)
      is
         Obj_Type : O_Tnode;
         Info : Object_Info_Acc;
         Tinfo : Type_Info_Acc;
         Def : Iir;
         Val : Iir;
         Storage : O_Storage;
      begin
         Def := Get_Type (El);
         Val := Get_Default_Value (El);

         --  Be sure the object type was translated.
         if Get_Kind (El) = Iir_Kind_Constant_Declaration
           and then Get_Deferred_Declaration_Flag (El) = False
           and then Get_Deferred_Declaration (El) /= Null_Iir
         then
            --  Do not create the subtype of a full constant declaration,
            --  since it was already created by the deferred declaration.
            --  Use the type of the deferred declaration.
            Def := Get_Type (Get_Deferred_Declaration (El));
         else
            Chap3.Translate_Object_Subtype (El);
         end if;

         Tinfo := Get_Info (Def);
         Obj_Type := Get_Object_Type (Tinfo, Mode_Value);

         Info := Add_Info (El, Kind_Object);

         case Get_Kind (El) is
            when Iir_Kind_Variable_Declaration
              | Iir_Kind_Constant_Interface_Declaration =>
               Info.Object_Var := Create_Var (Create_Var_Identifier (El),
                                              Obj_Type);
            when Iir_Kind_Constant_Declaration =>
               if Get_Deferred_Declaration (El) /= Null_Iir then
                  --  This is a full constant declaration (in a body) of a
                  --  deferred constant declaration (in a package).
                  Storage := O_Storage_Public;
               else
                  Storage := Global_Storage;
               end if;
               if Get_Expr_Staticness (El) = Locally then
                  declare
                     V : O_Enode;
                  begin
                     if Storage = O_Storage_External then
                        V := O_Enode_Null;
                     else
                        V := Chap7.Translate_Static_Expression (Val, Def);
                     end if;
                     Info.Object_Var := Create_Global_Const
                       (Create_Identifier (El), Obj_Type, Storage, V);
                  end;
               else
                  --  Globally static.
                  if Get_Deferred_Declaration_Flag (El) then
                     --  The deferred constant.  Present only in package
                     --  declaration.  Declare it as an external constant, even
                     --  if it may be later declared as a variable!
                     Info.Object_Var := Create_Global_Const
                       (Create_Identifier (El), Obj_Type,
                        O_Storage_External, O_Enode_Null);
                  else
                     Info.Object_Var := Create_Var
                       (Create_Var_Identifier (El), Obj_Type, Storage);
                  end if;
               end if;
            when others =>
               Error_Kind ("create_objet", El);
         end case;
      end Create_Object;

      procedure Create_Signal (Decl : Iir)
      is
         Sig_Type : O_Tnode;
         Type_Info : Type_Info_Acc;
         Info : Ortho_Info_Acc;
         Sig_Type_Def : Iir;
      begin
         Sig_Type_Def := Get_Type (Decl);
         Chap3.Translate_Object_Subtype (Decl);
         Type_Info := Get_Info (Sig_Type_Def);
         Sig_Type := Get_Object_Type (Type_Info, Mode_Signal);
         if Sig_Type = O_Tnode_Null then
            raise Internal_Error;
         end if;

         Info := Add_Info (Decl, Kind_Object);

         Info.Object_Var := Create_Var
           (Create_Var_Identifier (Decl), Sig_Type);

         case Get_Kind (Decl) is
            when Iir_Kind_Signal_Declaration
              | Iir_Kind_Signal_Interface_Declaration =>
               if Type_Info.Type_Mode = Type_Mode_Fat_Array then
                  Info.Object_Type_Desc_Var := Create_Var
                    (Create_Var_Identifier (Decl, "_STB"),
                     Ghdl_Desc_Subarray_Type_Node);
               end if;
               declare
                  Val : O_Enode;
               begin
                  if Global_Storage = O_Storage_External then
                     Val := O_Enode_Null;
                  else
                     Val := Create_String_Len
                       (Get_Identifier (Decl),
                        Create_Identifier (Decl, "_NAMESTR"));
                  end if;
                  Info.Object_Name := Create_Global_Const
                    (Create_Identifier (Decl, "_NAME"), Ghdl_Str_Len_Type_Node,
                     Global_Storage, Val);
               end;
            when Iir_Kind_Guard_Signal_Declaration =>
               --  No name created for guard signal.
               null;
            when others =>
               Error_Kind ("create_signal", Decl);
         end case;
      end Create_Signal;

      procedure Create_Implicit_Signal (Decl : Iir)
      is
         Sig_Type : O_Tnode;
         Type_Info : Type_Info_Acc;
         Info : Ortho_Info_Acc;
         Sig_Type_Def : Iir;
      begin
         Sig_Type_Def := Get_Type (Decl);
         Chap3.Translate_Object_Subtype (Decl);
         Type_Info := Get_Info (Sig_Type_Def);
         Sig_Type := Type_Info.Ortho_Type (Mode_Signal);
         if Sig_Type = O_Tnode_Null then
            raise Internal_Error;
         end if;

         Info := Add_Info (Decl, Kind_Object);

         Info.Object_Var := Create_Var (Create_Uniq_Identifier, Sig_Type);

         Info.Object_Name := null;
      end Create_Implicit_Signal;

      procedure Create_File_Object (El : Iir_File_Declaration)
      is
         Obj_Type : O_Tnode;
         Info : Ortho_Info_Acc;
         Obj_Type_Def : Iir;
      begin
         Obj_Type_Def := Get_Type (El);
         Obj_Type := Get_Ortho_Type (Obj_Type_Def, Mode_Value);

         Info := Add_Info (El, Kind_Object);

         Info.Object_Var := Create_Var (Create_Var_Identifier (El), Obj_Type);
      end Create_File_Object;

      --  Note : OBJ can be a tree.
      procedure Init_Object (Obj : Mnode; Obj_Type : Iir);

      procedure Allocate_Complex_Object (Obj_Type : Iir;
                                         Alloc_Kind : Allocation_Kind;
                                         Var : in out Mnode)
      is
         Type_Info : Type_Info_Acc;
         Kind : Object_Kind_Type;
         Targ : Mnode;
      begin
         Type_Info := Get_Type_Info (Var);
         if Type_Info.Type_Mode = Type_Mode_Fat_Array then
            --  Cannot allocate unconstrained object (since size is unknown).
            raise Internal_Error;
         end if;
         Kind := Get_Object_Kind (Var);

         if Type_Info.C = null then
            --  Object is not complex.
            if Alloc_Kind = Alloc_Stack then
               return;
            else
               raise Internal_Error;
            end if;
         end if;

         if Type_Info.C.Builder_Need_Func and then not Is_Stable (Var) then
            Targ := Create_Temp (Type_Info, Kind);
         else
            Targ := Var;
         end if;

         --  Allocate variable.
         New_Assign_Stmt
           (M2Lp (Targ),
            Gen_Alloc (Alloc_Kind,
                       Chap3.Get_Object_Size (Var, Obj_Type),
                       Type_Info.Ortho_Ptr_Type (Kind)));

         if Type_Info.C.Builder_Need_Func then
            --  Build the type.
            Chap3.Gen_Call_Type_Builder (Targ, Obj_Type);
            if not Is_Stable (Var) then
               New_Assign_Stmt (M2Lp (Var), M2Addr (Targ));
               Var := Targ;
            end if;
         end if;
      end Allocate_Complex_Object;

      --  Note : OBJ can be a tree.
      --  FIXME: should use translate_aggregate_others.
      procedure Init_Array_Object (Obj : Mnode; Obj_Type : Iir)
      is
         Sobj : Mnode;

         --  Type of the object.
         Type_Info : Type_Info_Acc;

         --  Iterator for the elements.
         Index : O_Lnode;

         Upper_Limit : O_Enode;
         Upper_Var : O_Lnode;

         Loop_Blk : O_Loop_Block;
         Label : O_Snode;
         If_Blk : O_If_Block;
      begin
         Type_Info := Get_Info (Obj_Type);

         --  Iterate on all elements of the object.
         Open_Temp;

         if Type_Info.Type_Mode = Type_Mode_Fat_Array then
            Sobj := Stabilize (Obj);
         else
            Sobj := Obj;
         end if;
         Upper_Limit := Chap3.Get_Array_Length (Sobj, Obj_Type);

         if Type_Info.Type_Mode /= Type_Mode_Array then
            Upper_Var := Create_Temp_Init (Ghdl_Index_Type, Upper_Limit);
         else
            Upper_Var := O_Lnode_Null;
         end if;

         Index := Create_Temp (Ghdl_Index_Type);
         Init_Var (Index);
         Start_Loop_Stmt (Loop_Blk, Label);
         if Upper_Var /= O_Lnode_Null then
            Upper_Limit := New_Value (Upper_Var);
         end if;
         Gen_Exit_When (Label,
                        New_Compare_Op (ON_Eq, New_Value (Index), Upper_Limit,
                                        Ghdl_Bool_Type_Node));
         Init_Object (Chap3.Index_Base (Chap3.Get_Array_Base (Sobj),
                                        Obj_Type,
                                        New_Value (Index)),
                      Get_Element_Subtype (Obj_Type));
         Inc_Var (Index);
         Finish_Loop_Stmt (Loop_Blk);

         Close_Temp;
      end Init_Array_Object;

      procedure Init_Object (Obj : Mnode; Obj_Type : Iir)
      is
         Tinfo : Type_Info_Acc;
      begin
         Tinfo := Get_Type_Info (Obj);
         case Tinfo.Type_Mode is
            when Type_Mode_Scalar =>
               New_Assign_Stmt
                 (M2Lv (Obj), Chap14.Translate_Left_Type_Attribute (Obj_Type));
            when Type_Mode_Acc =>
               New_Assign_Stmt
                 (M2Lv (Obj), New_Null_Access (Tinfo.Ortho_Type (Mode_Value)));
            when Type_Mode_Fat_Acc =>
               declare
                  Dinfo : Type_Info_Acc;
                  Sobj : Mnode;
               begin
                  Open_Temp;
                  Sobj := Stabilize (Obj);
                  Dinfo := Get_Info (Get_Designated_Type (Obj_Type));
                  New_Assign_Stmt
                    (New_Selected_Element (M2Lv (Sobj),
                                           Dinfo.T.Bounds_Field (Mode_Value)),
                     New_Null_Access (Dinfo.T.Bounds_Ptr_Type));
                  New_Assign_Stmt
                    (New_Selected_Element (M2Lv (Sobj),
                                           Dinfo.T.Base_Field (Mode_Value)),
                     New_Null_Access
                     (Dinfo.T.Base_Ptr_Type (Mode_Value)));
                  Close_Temp;
               end;
            when Type_Mode_Arrays =>
               Init_Array_Object (Obj, Obj_Type);
            when Type_Mode_Record =>
               declare
                  Sobj : Mnode;
                  List : Iir_List;
                  El : Iir_Element_Declaration;
               begin
                  Open_Temp;
                  Sobj := Stabilize (Obj);
                  List := Get_Element_Declaration_List
                    (Get_Base_Type (Obj_Type));
                  for I in Natural loop
                     El := Get_Nth_Element (List, I);
                     exit when El = Null_Iir;
                     Init_Object (Chap6.Translate_Selected_Element (Sobj, El),
                                  Get_Type (El));
                  end loop;
                  Close_Temp;
               end;
            when Type_Mode_Unknown
              | Type_Mode_File =>
               raise Internal_Error;
         end case;
      end Init_Object;

      --  Generate code to create object OBJ and initialize it with value VAL.
      procedure Elab_Object_Value (Obj : Iir; Value : Iir)
      is
         Obj_Info : Object_Info_Acc;

         Name_Node : Mnode;
         Value_Node : O_Enode;
         Obj_Type : Iir;

         Type_Info : Type_Info_Acc;
         Alloc_Kind : Allocation_Kind;
      begin
         --  Elaborate subtype.
         Obj_Type := Get_Type (Obj);
         Chap3.Elab_Object_Subtype (Obj_Type);
         Type_Info := Get_Info (Obj_Type);
         Obj_Info := Get_Info (Obj);

         Alloc_Kind := Get_Alloc_Kind_For_Var (Obj_Info.Object_Var);

         --  FIXME: the object type may be a fat array!
         --  FIXME: fat array + aggregate ?

         Name_Node := Get_Var (Obj_Info.Object_Var, Type_Info, Mode_Value);

         if Type_Info.C /= null
           and then Type_Info.Type_Mode /= Type_Mode_Fat_Array
         then
            --  FIXME: avoid allocation if the value is a string and
            --  the object is a constant
            Allocate_Complex_Object (Obj_Type, Alloc_Kind, Name_Node);
         end if;

         Open_Temp;
         if Value = Null_Iir then
            --  Performs default initialization.
            Init_Object (Name_Node, Obj_Type);
         elsif Get_Kind (Value) = Iir_Kind_Aggregate then
            if Type_Info.Type_Mode = Type_Mode_Fat_Array then
               --  Allocate.
               declare
                  Aggr_Type : Iir;
               begin
                  Aggr_Type := Get_Type (Value);
                  Chap3.Create_Array_Subtype (Aggr_Type);
                  Stabilize (Name_Node);
                  New_Assign_Stmt
                    (New_Selected_Element
                     (M2Lv (Name_Node), Type_Info.T.Bounds_Field (Mode_Value)),
                     Chap3.Get_Array_Ptr_Bounds_Ptr (O_Lnode_Null, Aggr_Type,
                                                     Mode_Value));
                  Chap3.Allocate_Fat_Array_Base
                    (Alloc_Kind, Name_Node,
                     Get_Base_Type (Aggr_Type), Mode_Value);
               end;
            end if;
            Chap7.Translate_Aggregate (Name_Node, Obj_Type, Value);
         else
            Value_Node := Chap7.Translate_Expression (Value, Obj_Type);

            if Type_Info.Type_Mode = Type_Mode_Fat_Array then
               declare
                  S : O_Lnode;
                  A : Mnode;
               begin
                  Name_Node := Stabilize (Name_Node);
                  S := Create_Temp_Init
                    (Type_Info.Ortho_Ptr_Type (Mode_Value), Value_Node);

                  if Get_Kind (Value) = Iir_Kind_String_Literal
                    and then Get_Kind (Obj) = Iir_Kind_Constant_Declaration
                  then
                     --  No need to allocate space for the object.
                     Copy_Fat_Pointer (Name_Node,
                                       Vp2M (S, Type_Info, Mode_Value));
                  else
                     A := Chap3.Translate_Object_Allocation
                       (Alloc_Kind, Obj_Type,
                        Chap3.Get_Array_Ptr_Bounds_Ptr (S, Get_Type (Value),
                                                        Mode_Value),
                        Mode_Value);
                     Copy_Fat_Pointer (Name_Node, A);
                     Chap3.Translate_Object_Copy
                       (Name_Node, New_Value (S), Obj_Type, Mode_Value);
                  end if;
               end;
            else
               Chap3.Translate_Object_Copy
                 (Name_Node, Value_Node, Obj_Type, Mode_Value);
            end if;
         end if;

         Close_Temp;
      end Elab_Object_Value;

      --  Create code to elaborate OBJ.
      procedure Elab_Object (Obj : Iir)
      is
         Value : Iir;
         Obj1 : Iir;
      begin
         --  A locally static constant is pre-elaborated.
         if Get_Expr_Staticness (Obj) = Locally then
            return;
         end if;

         New_Debug_Line_Stmt (Get_Line_Number (Obj));

         --  Set default value.
         if Get_Kind (Obj) = Iir_Kind_Constant_Declaration then
            if Get_Deferred_Declaration_Flag (Obj) = True then
               --  No code generation for a deferred constant.
               return;
            end if;
            Obj1 := Get_Deferred_Declaration (Obj);
            if Obj1 = Null_Iir then
               Obj1 := Obj;
            end if;
         else
            Obj1 := Obj;
         end if;

         --  Still use the default value of the not deferred constant.
         --  FIXME: what about composite types.
         Value := Get_Default_Value (Obj);
         Elab_Object_Value (Obj1, Value);
      end Elab_Object;

      procedure Add_Associations_For_Resolver
        (Assoc : in out O_Assoc_List; Func : Iir)
      is
         Func_Info : Subprg_Info_Acc;
         Resolv_Info : Subprg_Resolv_Info_Acc;
      begin
         Func_Info := Get_Info (Func);
         Resolv_Info := Func_Info.Subprg_Resolv;
         New_Association (Assoc,
                          New_Subprogram_Address (Resolv_Info.Resolv_Func,
                                                  Ghdl_Ptr_Type));
         if Resolv_Info.Resolv_Block /= Null_Iir then
            New_Association
              (Assoc,
               New_Convert_Ov (Get_Instance_Access (Resolv_Info.Resolv_Block),
                               Ghdl_Ptr_Type));
         else
            New_Association (Assoc, New_Null_Access (Ghdl_Ptr_Type));
         end if;
      end Add_Associations_For_Resolver;

      procedure Elab_Signal_Non_Composite (Targ : Mnode;
                                           Targ_Type : Iir;
                                           Val : Mnode)
      is
         Type_Info : Type_Info_Acc;
         Init_Subprg : O_Dnode;
         Conv : O_Tnode;
         Res : O_Enode;
         Assoc : O_Assoc_List;
         Init_Val : O_Enode;
         --  For the resolution function (if any).
         Func : Iir;
      begin
         Type_Info := Get_Info (Targ_Type);
         case Type_Info.Type_Mode is
            when Type_Mode_B2 =>
               Init_Subprg := Ghdl_Create_Signal_B2;
               Conv := Ghdl_Bool_Type_Node;
            when Type_Mode_E8 =>
               Init_Subprg := Ghdl_Create_Signal_E8;
               Conv := Ghdl_Integer_Type;
            when Type_Mode_I32 =>
               Init_Subprg := Ghdl_Create_Signal_I32;
               Conv := Ghdl_Integer_Type;
            when Type_Mode_F64 =>
               Init_Subprg := Ghdl_Create_Signal_F64;
               Conv := Ghdl_Real_Type;
            when Type_Mode_P64 =>
               Init_Subprg := Ghdl_Create_Signal_P64;
               Conv := Ghdl_Phys_Type;
            when others =>
               Error_Kind ("elab_signal_non_composite", Targ_Type);
         end case;

         if Is_Null (Val) then
            Init_Val := Chap14.Translate_Left_Type_Attribute (Targ_Type);
         else
            Init_Val := M2E (Val);
         end if;
         Start_Association (Assoc, Init_Subprg);
         New_Association (Assoc, New_Convert_Ov (Init_Val, Conv));
         New_Association (Assoc, Chap3.Get_Type_Desc_Addr (Targ_Type));

         if Get_Kind (Targ_Type) in Iir_Kinds_Subtype_Definition then
            Func := Get_Resolution_Function (Targ_Type);
         else
            Func := Null_Iir;
         end if;
         if Func /= Null_Iir then
            Add_Associations_For_Resolver (Assoc, Func);
         else
            New_Association (Assoc, New_Null_Access (Ghdl_Ptr_Type));
            New_Association (Assoc, New_Null_Access (Ghdl_Ptr_Type));
         end if;

         Res := New_Convert_Ov (New_Function_Call (Assoc),
                                Type_Info.Ortho_Type (Mode_Signal));
         New_Assign_Stmt (M2Lv (Targ), Res);
      end Elab_Signal_Non_Composite;

      function Elab_Signal_Prepare_Composite (Val : Mnode; Targ_Type : Iir)
        return Mnode
      is
         Assoc : O_Assoc_List;
         Func : Iir;
      begin
         if Get_Kind (Targ_Type) in Iir_Kinds_Subtype_Definition then
            Func := Get_Resolution_Function (Targ_Type);
            if Func /= Null_Iir then
               --  Add resolver.
               Start_Association (Assoc, Ghdl_Signal_Create_Resolution);
               New_Association (Assoc, Chap3.Get_Type_Desc_Addr (Targ_Type));
               Add_Associations_For_Resolver (Assoc, Func);
               New_Procedure_Call (Assoc);
            end if;
         end if;
         if Is_Null (Val) then
            return Val;
         end if;
         if Get_Type_Info (Val).Type_Mode = Type_Mode_Record then
            return Stabilize (Val);
         else
            return Chap3.Get_Array_Base (Val);
         end if;
      end Elab_Signal_Prepare_Composite;

      function Elab_Signal_Update_Array (Val : Mnode;
                                         Targ_Type : Iir;
                                         Index : O_Lnode)
        return Mnode
      is
      begin
         if Is_Null (Val) then
            --  No default value.
            return Val;
         else
            return Chap3.Index_Base (Val, Targ_Type, New_Value (Index));
         end if;
      end Elab_Signal_Update_Array;

      function Elab_Signal_Update_Record (Val : Mnode;
                                          Targ_Type : Iir;
                                          El : Iir_Element_Declaration)
        return Mnode
      is
      begin
         if Is_Null (Val) then
            --  No default value.
            return Val;
         else
            return Chap6.Translate_Selected_Element (Val, El);
         end if;
      end Elab_Signal_Update_Record;

      procedure Elab_Signal is new Foreach_Non_Composite
        (Data_Type => Mnode,
         Composite_Data_Type => Mnode,
         Do_Non_Composite => Elab_Signal_Non_Composite,
         Prepare_Data_Array => Elab_Signal_Prepare_Composite,
         Update_Data_Array => Elab_Signal_Update_Array,
         Prepare_Data_Record => Elab_Signal_Prepare_Composite,
         Update_Data_Record => Elab_Signal_Update_Record);

      procedure Elab_Signal_Declaration (Decl : Iir; Parent : Iir)
      is
         Sig_Type : Iir;
         Type_Info : Type_Info_Acc;
         Name_Node : Mnode;
         Val : Iir;
         Val_Node : Mnode;
         Mode : Natural;
      begin
         New_Debug_Line_Stmt (Get_Line_Number (Decl));

         Open_Temp;

         Sig_Type := Get_Type (Decl);
         Chap3.Elab_Object_Subtype (Sig_Type);
         Type_Info := Get_Info (Sig_Type);

         Name_Node := Chap6.Translate_Name (Decl);
         if Get_Object_Kind (Name_Node) /= Mode_Signal then
            raise Internal_Error;
         end if;

         if Type_Info.Type_Mode = Type_Mode_Fat_Array then
            Stabilize (Name_Node);
            Chap3.Allocate_Fat_Array_Base
              (Alloc_Memory, Name_Node, Sig_Type, Mode_Signal);
            Chap3.Set_Array_Subtype_Desc
              (Get_Info (Decl).Object_Type_Desc_Var,
               Sig_Type,
               New_Acc_Value
               (New_Selected_Element
                (M2Lv (Name_Node), Type_Info.T.Bounds_Field (Mode_Signal))));
         elsif Type_Info.C /= null then
            Allocate_Complex_Object (Sig_Type, Alloc_Memory, Name_Node);
         end if;

         --  FIXME: build the object.

         case Get_Kind (Decl) is
            when Iir_Kind_Signal_Declaration =>
               Mode := 0;
            when Iir_Kind_Signal_Interface_Declaration =>
               Mode := Iir_Mode'Pos (Get_Mode (Decl));
            when others =>
               Error_Kind ("elab_signal_declaration", Decl);
         end case;

         --  Set the name of the signal.
         declare
            Assoc : O_Assoc_List;
            Desc : O_Enode;
         begin
            Start_Association (Assoc, Ghdl_Signal_Name);
            New_Association
              (Assoc, New_Address (Get_Var (Get_Info (Decl).Object_Name),
                                   Ghdl_Str_Len_Ptr_Node));
            New_Association (Assoc, Chap6.Get_Instance_Name_Ref (Parent));
            New_Association
              (Assoc,
               New_Signed_Literal (Ghdl_Integer_Type, Integer_64 (Mode)));
            New_Association
              (Assoc, New_Signed_Literal
               (Ghdl_Integer_Type,
                Integer_64 (Iir_Signal_Kind'Pos (Get_Signal_Kind (Decl)))));
            if Type_Info.Type_Mode = Type_Mode_Fat_Array then
               Desc := New_Unchecked_Address
                 (Get_Var (Get_Info (Decl).Object_Type_Desc_Var),
                  Ghdl_Desc_Ptr_Type);
            else
               Desc := Chap3.Get_Type_Desc_Addr (Sig_Type);
            end if;
            New_Association (Assoc, Desc);
            New_Procedure_Call (Assoc);
         end;

         Val := Get_Default_Value (Decl);
         if Val = Null_Iir then
            Val_Node := Mnode_Null;
         else
            Val_Node := E2M (Chap7.Translate_Expression (Val, Sig_Type),
                             Get_Info (Sig_Type),
                             Mode_Value);
         end if;
         Elab_Signal (Name_Node, Sig_Type, Val_Node);

         Close_Temp;
      end Elab_Signal_Declaration;

      procedure Elab_Stable_Signal (Decl : Iir)
      is
         Assoc : O_Assoc_List;
         Dtype : Iir;
         Type_Info : Type_Info_Acc;
         Info : Object_Info_Acc;
         Prefix : Iir;
         Prefix_Node : Mnode;
         Res : O_Enode;
         Val : O_Enode;
         Param : Iir;
      begin
         New_Debug_Line_Stmt (Get_Line_Number (Decl));

         Info := Get_Info (Decl);
         Dtype := Get_Type (Decl);
         Type_Info := Get_Info (Dtype);
         --  Create the signal (with the time)
         Start_Association (Assoc, Ghdl_Create_Stable_Signal);
         Param := Get_Parameter (Decl);
         if Param = Null_Iir then
            Val := New_Signed_Literal (Std_Time_Type, 0);
         else
            Val := Chap7.Translate_Expression (Param);
         end if;
         New_Association (Assoc, Val);
         Res := New_Convert_Ov (New_Function_Call (Assoc),
                                Type_Info.Ortho_Type (Mode_Signal));
         New_Assign_Stmt (Get_Var (Info.Object_Var), Res);

         --  Register all signals this depends on.
         Prefix := Get_Prefix (Decl);
         Prefix_Node := Chap6.Translate_Name (Prefix);
         Register_Signal (Prefix_Node, Get_Type (Prefix),
                          Ghdl_Signal_Register_Stable_Prefix);
      end Elab_Stable_Signal;

      procedure Elab_File_Declaration (Decl : Iir_File_Declaration)
      is
         Constr : O_Assoc_List;
         Name : Mnode;
         File_Name : Iir;
         Open_Kind : Iir;
         Mode_Val : O_Enode;
         Str : O_Enode;
         Is_Text : Boolean;
      begin
         --  Elaborate the file.
         --  FIXME: non-text files.
         Name := Chap6.Translate_Name (Decl);
         if Get_Object_Kind (Name) /= Mode_Value then
            raise Internal_Error;
         end if;
         Is_Text := Get_Text_File_Flag (Get_Type (Decl));
         if Is_Text then
            Start_Association (Constr, Ghdl_Text_File_Elaborate);
         else
            Start_Association (Constr, Ghdl_File_Elaborate);
         end if;
         New_Assign_Stmt (M2Lv (Name), New_Function_Call (Constr));

         --  If file_open_information is present, open the file.
         File_Name := Get_File_Logical_Name (Decl);
         if File_Name = Null_Iir then
            return;
         end if;
         Open_Temp;
         Name := Chap6.Translate_Name (Decl);
         Open_Kind := Get_File_Open_Kind (Decl);
         if Open_Kind /= Null_Iir then
            Mode_Val := New_Convert_Ov (Chap7.Translate_Expression (Open_Kind),
                                        Ghdl_Integer_Type);
         else
            case Get_Mode (Decl) is
               when Iir_In_Mode =>
                  Mode_Val := New_Signed_Literal (Ghdl_Integer_Type, 0);
               when Iir_Out_Mode =>
                  Mode_Val := New_Signed_Literal (Ghdl_Integer_Type, 1);
               when others =>
                  raise Internal_Error;
            end case;
         end if;
         Str := Chap7.Translate_Expression (File_Name, String_Type_Definition);

         if Is_Text then
            Start_Association (Constr, Ghdl_Text_File_Open);
         else
            Start_Association (Constr, Ghdl_File_Open);
         end if;
         New_Association (Constr, M2E (Name));
         New_Association (Constr, Mode_Val);
         New_Association (Constr, Str);
         New_Procedure_Call (Constr);
         Close_Temp;
      end Elab_File_Declaration;

      procedure Translate_Type_Declaration (Decl : Iir_Type_Declaration)
      is
      begin
         Chap3.Translate_Named_Type_Definition (Get_Type (Decl),
                                                Get_Identifier (Decl));
      end Translate_Type_Declaration;

      procedure Translate_Subtype_Declaration (Decl : Iir_Subtype_Declaration)
      is
      begin
         Chap3.Translate_Named_Type_Definition (Get_Type (Decl),
                                                Get_Identifier (Decl));
      end Translate_Subtype_Declaration;

      procedure Translate_Bool_Type_Declaration (Decl : Iir_Type_Declaration)
      is
         Mark : Id_Mark_Type;
      begin
         Push_Identifier_Prefix (Mark, Get_Identifier (Decl));
         Chap3.Translate_Bool_Type_Definition (Get_Type (Decl));
         Pop_Identifier_Prefix (Mark);
      end Translate_Bool_Type_Declaration;

      procedure Translate_Alias_Declaration (Decl : Iir_Alias_Declaration)
      is
         Decl_Type : Iir;
         Info : Alias_Info_Acc;
         Tinfo : Type_Info_Acc;
         Atype : O_Tnode;
      begin
         Decl_Type := Get_Type (Decl);
         if Decl_Type = Null_Iir then
            --  FIXME : todo.
            raise Internal_Error;
         end if;

         Chap3.Translate_Named_Type_Definition
           (Decl_Type, Get_Identifier (Decl));

         Info := Add_Info (Decl, Kind_Alias);
         case Get_Kind (Get_Base_Name (Decl)) is
            when Iir_Kind_Signal_Declaration
              | Iir_Kind_Signal_Interface_Declaration =>
               Info.Alias_Kind := Mode_Signal;
            when others =>
               Info.Alias_Kind := Mode_Value;
         end case;

         Tinfo := Get_Info (Decl_Type);
         case Tinfo.Type_Mode is
            when Type_Mode_Fat_Array =>
               --  create an object.
               --  At elaboration: copy base from name, copy bounds from type,
               --   check for matching bounds.
               Atype := Get_Ortho_Type (Decl_Type, Info.Alias_Kind);
            when Type_Mode_Array
              | Type_Mode_Ptr_Array =>
               --  Create an object pointer.
               --  At elaboration: copy base from name.
               Atype := Tinfo.Ortho_Ptr_Type (Info.Alias_Kind);
            when Type_Mode_Scalar =>
               case Info.Alias_Kind is
                  when Mode_Signal =>
                     Atype := Tinfo.Ortho_Type (Mode_Signal);
                  when Mode_Value =>
                     Atype := Tinfo.Ortho_Ptr_Type (Mode_Signal);
               end case;
            when others =>
               raise Internal_Error;
         end case;
         Info.Alias_Var := Create_Var (Create_Var_Identifier (Decl), Atype);
      end Translate_Alias_Declaration;

      procedure Elab_Alias_Declaration (Decl : Iir_Alias_Declaration)
      is
         Decl_Type : Iir;
         Name : Iir;
         Name_Node : Mnode;
         N_Info : Type_Info_Acc;
         Alias_Info : Alias_Info_Acc;
         Name_Type : Iir;
         Tinfo : Type_Info_Acc;
         Kind : Object_Kind_Type;
      begin
         Decl_Type := Get_Type (Decl);
         Tinfo := Get_Info (Decl_Type);

         Alias_Info := Get_Info (Decl);
         Chap3.Elab_Object_Subtype (Decl_Type);
         Name := Get_Name (Decl);
         Name_Type := Get_Type (Name);
         --  Evaluate names.
         Name_Node := Chap6.Translate_Name (Name);
         Kind := Get_Object_Kind (Name_Node);
         N_Info := Get_Info (Name_Type);
         --Chap6.Translate_Name (Decl, Decl_Node, Sig);

         case Tinfo.Type_Mode is
            when Type_Mode_Fat_Array =>
               --  FIXME: to be done.
               raise Internal_Error;
--                Open_Temp;
--                --  Copy base.
--                B := Create_Temp_Ptr
--                  (Decl_Type, Get_Var (Alias_Info.Alias_Var), Kind);
--                S := Create_Temp_Ptr (Name_Type, Name_Node, Kind);
--                New_Assign_Stmt
--                  (Chap3.Get_Array_Ptr_Base_Ptr (B, Decl_Type, Kind),
--                   New_Value
--                   (Chap3.Get_Array_Ptr_Base_Ptr (S, Name_Type, Kind)));
--                --  copy bounds from type.
--                New_Assign_Stmt
--                  (New_Selected_Element (New_Acc_Value (B),
--                                         Tinfo.T.Bounds_Field (Kind)),
--                   New_Address (Get_Var (Tinfo.T.Array_Bounds),
--                                Tinfo.T.Bounds_Ptr_Type));
--                --  check length
--                Chap3.Check_Array_Match (Decl_Type, B, Kind,
--                                         Name_Type, S, Kind,
--                                         Decl);
--                Close_Temp;
            when Type_Mode_Array
              | Type_Mode_Ptr_Array =>
               Open_Temp;
               Stabilize (Name_Node);
               New_Assign_Stmt
                 (Get_Var (Alias_Info.Alias_Var),
                  New_Value (M2Lp (Chap3.Get_Array_Base (Name_Node))));
               Chap3.Check_Array_Match (Decl_Type, O_Lnode_Null, Kind,
                                        Name_Type, M2Lp (Name_Node), Kind,
                                        Decl);
               Close_Temp;
            when Type_Mode_Scalar =>
               case Alias_Info.Alias_Kind is
                  when Mode_Value =>
                     New_Assign_Stmt (Get_Var (Alias_Info.Alias_Var),
                                      M2Addr (Name_Node));
                  when Mode_Signal =>
                     New_Assign_Stmt (Get_Var (Alias_Info.Alias_Var),
                                      M2E (Name_Node));
               end case;
            when others =>
               raise Internal_Error;
         end case;
      end Elab_Alias_Declaration;

      procedure Translate_Port_List (List : Iir_List)
      is
         Port : Iir;
      begin
         if List = null then
            return;
         end if;
         for I in Natural loop
            Port := Get_Nth_Element (List, I);
            exit when Port = Null_Iir;
            Create_Signal (Port);
         end loop;
      end Translate_Port_List;

      procedure Translate_Generic_List (List : Iir_List)
      is
         Decl : Iir;
      begin
         if List = null then
            return;
         end if;
         for I in Natural loop
            Decl := Get_Nth_Element (List, I);
            exit when Decl = Null_Iir;
            Create_Object (Decl);
         end loop;
      end Translate_Generic_List;

      --  Create instance record for a component.
      procedure Translate_Component_Declaration (Decl : Iir)
      is
         Mark : Id_Mark_Type;
         Info : Ortho_Info_Acc;
      begin
         Info := Add_Info (Decl, Kind_Component);
         Push_Identifier_Prefix (Mark, Get_Identifier (Decl));
         Push_Instance_Factory (O_Tnode_Null);

         --  Entity name.
         Info.Comp_Name := Add_Instance_Factory_Field
           (Wki_Name, Ghdl_Instance_Name_Type);

         --  Generic and ports.
         Translate_Generic_List (Get_Generic_List (Decl));
         Translate_Port_List (Get_Port_List (Decl));

         Pop_Instance_Factory (Info.Comp_Type);
         New_Type_Decl (Create_Identifier ("_COMPTYPE"), Info.Comp_Type);
         Info.Comp_Ptr_Type := New_Access_Type (Info.Comp_Type);
         New_Type_Decl (Create_Identifier ("_COMPPTR"), Info.Comp_Ptr_Type);
         Pop_Identifier_Prefix (Mark);
      end Translate_Component_Declaration;

      procedure Translate_Declaration (Decl : Iir)
      is
      begin
         case Get_Kind (Decl) is
            when Iir_Kind_Use_Clause =>
               null;
            when Iir_Kind_Configuration_Specification =>
               null;
            when Iir_Kind_Disconnection_Specification =>
               null;

            when Iir_Kind_Component_Declaration =>
               Chap4.Translate_Component_Declaration (Decl);
            when Iir_Kind_Type_Declaration =>
               Chap4.Translate_Type_Declaration (Decl);
            when Iir_Kind_Subtype_Declaration =>
               Chap4.Translate_Subtype_Declaration (Decl);

            when Iir_Kind_Function_Specification
              | Iir_Kind_Procedure_Specification
              | Iir_Kind_Function_Spec_Body
              | Iir_Kind_Procedure_Spec_Body =>
               raise Internal_Error;
            when Iir_Kind_Function_Body
              | Iir_Kind_Procedure_Body =>
               null;

            --when Iir_Kind_Implicit_Function_Declaration =>
            --when Iir_Kind_Signal_Declaration
            --  | Iir_Kind_Signal_Interface_Declaration =>
               --   Chap4.Create_Object (Decl);

            when Iir_Kind_Variable_Declaration
              | Iir_Kind_Constant_Declaration =>
               Create_Object (Decl);

            when Iir_Kind_Signal_Declaration =>
               Create_Signal (Decl);

            when Iir_Kind_Alias_Declaration =>
               Translate_Alias_Declaration (Decl);

            when Iir_Kind_File_Declaration =>
               Create_File_Object (Decl);

            when Iir_Kind_Attribute_Declaration =>
               Chap3.Translate_Object_Subtype (Decl);

            when Iir_Kind_Attribute_Specification =>
               Chap5.Translate_Attribute_Specification (Decl);

            when Iir_Kind_Implicit_Signal_Declaration =>
               declare
                  Chain : Iir;
               begin
                  Chain := Get_Implicit_Signal_Chain (Decl);
                  while Chain /= Null_Iir loop
                     Chap4.Create_Implicit_Signal (Chain);
                     Chain := Get_Implicit_Signal_Chain (Chain);
                  end loop;
               end;

            when Iir_Kind_Guard_Signal_Declaration =>
               Create_Signal (Decl);

            when Iir_Kind_Group_Template_Declaration =>
               null;
            when Iir_Kind_Group_Declaration =>
               null;

            when others =>
               Error_Kind ("translate_declaration", Decl);
         end case;
      end Translate_Declaration;

      --  Mark FUNC (by adding the subprg_resolv info) iif it can be a
      --  resolution function.
      procedure Check_Resolution_Function (Func : Iir)
      is
         Params_List : Iir_List;
         Param : Iir;
         Res_Type : Iir;
         Info : Subprg_Info_Acc;
      begin
         Params_List := Get_Interface_Declaration (Func);
         if Get_Nbr_Elements (Params_List) /= 1 then
            return;
         end if;
         Param := Get_Type (Get_First_Element (Params_List));
         case Get_Kind (Param) is
            when Iir_Kind_Array_Type_Definition
              | Iir_Kind_Unconstrained_Array_Subtype_Definition =>
               null;
            when others =>
               return;
         end case;
         Res_Type := Get_Return_Type (Func);
         if Get_Base_Type (Get_Element_Subtype (Param))
           /= Get_Base_Type (Res_Type)
         then
            return;
         end if;
         --  FUNC can be a resolution function.
         Info := Get_Info (Func);
         Info.Subprg_Resolv := new Subprg_Resolv_Info;
      end Check_Resolution_Function;

      procedure Translate_Resolution_Function (Func : Iir; Block : Iir)
      is
         --  Type of the resolution function parameter.
         Arr_Type : Iir;
         Base_Type, El_Type : Iir;
         El_Info : Type_Info_Acc;
         Base_Info : Type_Info_Acc;
         Finfo : Subprg_Info_Acc;
         Interface_List : O_Inter_List;
         Rinfo : Subprg_Resolv_Info_Acc;
         Block_Info : Block_Info_Acc;
         Id : O_Ident;
      begin
         Finfo := Get_Info (Func);
         Rinfo := Finfo.Subprg_Resolv;
         if Rinfo = null then
            return;
         end if;
         if Block /= Null_Iir and then Get_Pure_Flag (Func) = False then
            Block_Info := Get_Info (Block);
            Rinfo.Resolv_Block := Block;
         else
            Block_Info := null;
            Rinfo.Resolv_Block := Null_Iir;
         end if;
         Arr_Type := Get_Type
           (Get_First_Element (Get_Interface_Declaration (Func)));
         Base_Type := Get_Base_Type (Arr_Type);
         El_Type := Get_Element_Subtype (Base_Type);
         Base_Info := Get_Info (Base_Type);
         El_Info := Get_Info (El_Type);
         --Allow any types
         --if El_Info.Type_Mode not in Type_Mode_Scalar then
         --   Error_Kind ("translate_resolution_function", Func);
         --end if;
         Id := Create_Identifier (Func, Finfo.Subprg_Overload, "_RESOLV");

         if El_Info.Type_Mode in Type_Mode_Scalar then
            Start_Function_Decl (Interface_List, Id, Global_Storage,
                                 El_Info.Ortho_Type (Mode_Value));
            Rinfo.Var_Res := O_Lnode_Null;
         else
            Start_Procedure_Decl (Interface_List, Id, Global_Storage);
            New_Interface_Decl
              (Interface_List, Rinfo.Var_Res, Get_Identifier ("RESULT"),
               El_Info.Ortho_Ptr_Type (Mode_Value));
         end if;
         New_Interface_Decl
           (Interface_List, Rinfo.Var_Length, Wki_Length, Ghdl_Index_Type);
         New_Interface_Decl
           (Interface_List, Rinfo.Var_Vals, Get_Identifier ("VALS"),
            Base_Info.T.Base_Ptr_Type (Mode_Value));
         if Block_Info /= null then
            New_Interface_Decl (Interface_List, Rinfo.Var_Instance,
                                Wki_Instance, Block_Info.Block_Decls_Ptr_Type);
         else
            Rinfo.Var_Instance := O_Lnode_Null;
         end if;
         Finish_Subprogram_Decl (Interface_List,
                                 Finfo.Subprg_Resolv.Resolv_Func);
      end Translate_Resolution_Function;

      procedure Translate_Resolution_Function_Body (Func : Iir; Block : Iir)
      is
         --  Type of the resolution function parameter.
         Arr_Type : Iir;
         Base_Type, El_Type : Iir;
         El_Info : Type_Info_Acc;
         Base_Info : Type_Info_Acc;

         --  Type and info of the array index.
         Index_Type : Iir;
         Index_Tinfo : Type_Info_Acc;

         Var_Bound : O_Lnode;
         Var_Range_Ptr : O_Lnode;
         Var_Array : O_Lnode;
         Finfo : Subprg_Info_Acc;
         Assoc : O_Assoc_List;
         Rinfo : Subprg_Resolv_Info_Acc;
         Block_Info : Block_Info_Acc;
      begin
         Finfo := Get_Info (Func);
         Rinfo := Finfo.Subprg_Resolv;
         if Rinfo = null then
            return;
         end if;
         Arr_Type := Get_Type
           (Get_First_Element (Get_Interface_Declaration (Func)));
         Base_Type := Get_Base_Type (Arr_Type);
         El_Type := Get_Element_Subtype (Base_Type);
         Base_Info := Get_Info (Base_Type);
         El_Info := Get_Info (El_Type);

         Start_Subprogram_Body (Finfo.Subprg_Resolv.Resolv_Func);
         if Rinfo.Var_Instance /= O_Lnode_Null then
            Block_Info := Get_Info (Block);
            Push_Scope (Block_Info.Block_Decls_Type, Rinfo.Var_Instance);
         end if;
         New_Var_Decl (Var_Bound, Get_Identifier ("BOUND"), O_Storage_Local,
                       Base_Info.T.Bounds_Type);
         New_Var_Decl (Var_Array, Get_Identifier ("ARRAY"), O_Storage_Local,
                       Base_Info.Ortho_Type (Mode_Value));
         Index_Type := Get_First_Element (Get_Index_List (Arr_Type));
         Index_Tinfo := Get_Info (Index_Type);
         New_Var_Decl (Var_Range_Ptr, Get_Identifier ("RANGE_PTR"),
                       O_Storage_Local, Index_Tinfo.T.Range_Ptr_Type);
         New_Assign_Stmt
           (Var_Range_Ptr,
            New_Address (New_Selected_Element (Var_Bound,
                                               Base_Info.T.Bounds_Vector (1)),
                         Index_Tinfo.T.Range_Ptr_Type));

         --  Create range from length
         Chap3.Create_Range_From_Length
           (Index_Type, Rinfo.Var_Length, Var_Range_Ptr);
         --  Fill the array.
         New_Assign_Stmt
           (New_Selected_Element (Var_Array,
                                  Base_Info.T.Bounds_Field (Mode_Value)),
            New_Address (Var_Bound, Base_Info.T.Bounds_Ptr_Type));
         New_Assign_Stmt
           (New_Selected_Element (Var_Array,
                                  Base_Info.T.Base_Field (Mode_Value)),
            New_Value (Rinfo.Var_Vals));
         --  Call the resolution function.
         Start_Association (Assoc, Get_Info (Func).Ortho_Func);
         if Rinfo.Var_Res /= O_Lnode_Null then
            New_Association (Assoc, New_Value (Rinfo.Var_Res));
         end if;
         if Block /= Null_Iir then
            if Rinfo.Var_Instance /= O_Lnode_Null then
               New_Association (Assoc, New_Value (Rinfo.Var_Instance));
            else
               New_Association
                 (Assoc,
                  New_Null_Access (Get_Info (Block).Block_Decls_Ptr_Type));
            end if;
         end if;

         New_Association
           (Assoc, New_Address (Var_Array,
                                Base_Info.Ortho_Ptr_Type (Mode_Value)));
         if Rinfo.Var_Res = O_Lnode_Null then
            New_Return_Stmt (New_Function_Call (Assoc));
         else
            New_Procedure_Call (Assoc);
         end if;
         if Rinfo.Var_Instance /= O_Lnode_Null then
            Pop_Scope (Block_Info.Block_Decls_Type);
         end if;
         Finish_Subprogram_Body;
      end Translate_Resolution_Function_Body;

      procedure Translate_Declaration_List
        (Decl_List : Iir_List; First_List : Iir_List)
      is
         El : Iir;
         Info : Subprg_Info_Acc;
         Id : Name_Id;
      begin
         --  Nothing to do if there is no declarations.
         if Decl_List = null then
            return;
         end if;

         --  Check.
         pragma Debug (Name_Table.Assert_No_Infos);

         --  Set overload number from first list.
         if First_List /= Null_Iir_List then
            for I in Natural loop
               El := Get_Nth_Element (First_List, I);
               exit when El = Null_Iir;
               case Get_Kind (El) is
                  when Iir_Kind_Procedure_Specification
                    | Iir_Kind_Function_Specification =>
                     Info := Get_Info (El);
                     Id := Get_Identifier (El);
                     Name_Table.Set_Info (Id, Info.Subprg_Overload);
                  when Iir_Kind_Procedure_Spec_Body
                    | Iir_Kind_Function_Spec_Body
                    | Iir_Kind_Procedure_Body
                    | Iir_Kind_Function_Body =>
                     --  The first list cannot contains body, since it is
                     --  from a package declaration.
                     Error_Kind ("translate_declaration_list", El);
                  when others =>
                     null;
               end case;
            end loop;
         end if;

         --  First loop: set overload number on subprograms.
         for I in Natural loop
            El := Get_Nth_Element (Decl_List, I);
            exit when El = null;
            case Get_Kind (El) is
               when Iir_Kind_Procedure_Spec_Body
                 | Iir_Kind_Function_Spec_Body
                 | Iir_Kind_Procedure_Specification
                 | Iir_Kind_Function_Specification =>
                  Info := Add_Info (El, Kind_Subprg);
                  Id := Get_Identifier (El);
                  Info.Subprg_Overload := Name_Table.Get_Info (Id) + 1;
                  Name_Table.Set_Info (Id, Info.Subprg_Overload);
                  if Get_Kind (El) in Iir_Kinds_Function_Specification then
                     Check_Resolution_Function (El);
                  end if;
               when others =>
                  null;
            end case;
         end loop;

         --  Second loop: reset the name table info and translate other
         --  declarations.
         for I in Natural loop
            El := Get_Nth_Element (Decl_List, I);
            exit when El = null;
            case Get_Kind (El) is
               when Iir_Kind_Procedure_Spec_Body
                 | Iir_Kind_Function_Spec_Body
                 | Iir_Kind_Procedure_Specification
                 | Iir_Kind_Function_Specification =>
                  Id := Get_Identifier (El);
                  --  If the name is not overloaded, set the overload counter
                  --  to 0.
                  if Name_Table.Get_Info (Id) = 1 then
                     Get_Info (El).Subprg_Overload := 0;
                  end if;
                  Name_Table.Set_Info (Id, 0);
                  --  Note: the specification is *not* translated.
               when Iir_Kind_Function_Body
                 | Iir_Kind_Procedure_Body =>
                  null;
               when others =>
                  Translate_Declaration (El);
            end case;
         end loop;

         --  Clear overload number from first_list.
         if First_List /= Null_Iir_List then
            for I in Natural loop
               El := Get_Nth_Element (First_List, I);
               exit when El = Null_Iir;
               case Get_Kind (El) is
                  when Iir_Kind_Procedure_Specification
                    | Iir_Kind_Function_Specification =>
                     Id := Get_Identifier (El);
                     Name_Table.Set_Info (Id, 0);
                  when others =>
                     null;
               end case;
            end loop;
         end if;
      end Translate_Declaration_List;

      procedure Translate_Declaration_List_Subprograms
        (Decl_List : Iir_List; Block : Iir)
      is
         El : Iir;
      begin
         --  Nothing to do if there is no declarations.
         if Decl_List = Null_Iir_List then
            return;
         end if;

         for I in Natural loop
            El := Get_Nth_Element (Decl_List, I);
            exit when El = Null_Iir_List;
            case Get_Kind (El) is
               when Iir_Kind_Procedure_Specification
                 | Iir_Kind_Function_Specification =>
                  Chap2.Translate_Subprogram_Specification (El, Block);
                  Translate_Resolution_Function (El, Block);
               when Iir_Kind_Function_Spec_Body
                 | Iir_Kind_Procedure_Spec_Body =>
                  Chap2.Translate_Subprogram_Specification (El, Block);
                  Translate_Resolution_Function (El, Block);
                  Chap2.Translate_Subprogram_Body (El);
                  Translate_Resolution_Function_Body (El, Block);
               when Iir_Kind_Function_Body
                 | Iir_Kind_Procedure_Body =>
                  Chap2.Translate_Subprogram_Body (El);
                  Translate_Resolution_Function_Body
                    (Get_Subprogram_Specification (El), Block);
               when Iir_Kind_Type_Declaration =>
                  Chap3.Translate_Type_Subprograms (El, Block);
                  Chap7.Translate_Implicit_Functions_List
                    (Get_Implicit_Functions_List (El), Block);
               when others =>
                  null;
            end case;
         end loop;
      end Translate_Declaration_List_Subprograms;

      procedure Elab_Declaration_List (Decl_List : Iir_List; Parent : Iir)
      is
         Decl : Iir;
      begin
         if Decl_List = Null_Iir_List then
            return;
         end if;
         for I in Natural loop
            Decl := Get_Nth_Element (Decl_List, I);
            exit when Decl = null;

            Open_Temp;

            case Get_Kind (Decl) is
               when Iir_Kind_Use_Clause =>
                  null;
               when Iir_Kind_Component_Declaration =>
                  null;
               when Iir_Kind_Configuration_Specification =>
                  null;
               when Iir_Kind_Disconnection_Specification =>
                  Chap5.Elab_Disconnection_Specification (Decl);

               when Iir_Kind_Type_Declaration =>
                  Chap3.Elab_Type_Declaration (Decl);
               when Iir_Kind_Subtype_Declaration =>
                  Chap3.Elab_Subtype_Declaration (Decl);

               --when Iir_Kind_Signal_Declaration =>
               --   Chap1.Elab_Signal (Decl);
               when Iir_Kind_Variable_Declaration
                 | Iir_Kind_Constant_Declaration =>
                  Elab_Object (Decl);

               when Iir_Kind_Signal_Declaration =>
                  Elab_Signal_Declaration (Decl, Parent);

               when Iir_Kind_Alias_Declaration =>
                  Elab_Alias_Declaration (Decl);

               when Iir_Kind_File_Declaration =>
                  Elab_File_Declaration (Decl);

               when Iir_Kind_Attribute_Declaration =>
                  Chap3.Elab_Object_Subtype (Get_Type (Decl));

               when Iir_Kind_Attribute_Specification =>
                  Chap5.Elab_Attribute_Specification (Decl);

               when Iir_Kind_Function_Specification
                 | Iir_Kind_Function_Spec_Body =>
                  null;
               when Iir_Kind_Function_Body =>
                  null;
               when Iir_Kind_Procedure_Specification
                 | Iir_Kind_Procedure_Spec_Body =>
                  null;
               when Iir_Kind_Procedure_Body =>
                  null;

               --when Iir_Kind_Implicit_Function_Declaration =>
               --   null;

               when Iir_Kind_Implicit_Signal_Declaration =>
                  declare
                     Chain : Iir;
                  begin
                     Chain := Get_Implicit_Signal_Chain (Decl);
                     while Chain /= Null_Iir loop
                        case Get_Kind (Chain) is
                           when Iir_Kind_Stable_Attribute =>
                              Elab_Stable_Signal (Chain);
                           when others =>
                              Error_Kind ("elab_implicit_signal", Chain);
                        end case;
                        --Elab_Implicit_Signal (Chain);
                        Chain := Get_Implicit_Signal_Chain (Chain);
                     end loop;
                  end;

               when Iir_Kind_Group_Template_Declaration
                 | Iir_Kind_Group_Declaration =>
                  null;

               when others =>
                  Error_Kind ("elab_declaration_list", Decl);
            end case;

            Close_Temp;

         end loop;
      end Elab_Declaration_List;

      procedure Elab_Generic_List (List : Iir_List)
      is
      begin
         --  Nothing to do, all the work is done by the generic association.
         return;
      end Elab_Generic_List;

      procedure Elab_Port_List (List : Iir_List; Parent : Iir)
      is
         Port : Iir;
      begin
         if List = Null_Iir_List then
            return;
         end if;
         for I in Natural loop
            Port := Get_Nth_Element (List, I);
            exit when Port = Null_Iir;
            Elab_Signal_Declaration (Port, Parent);
         end loop;
      end Elab_Port_List;

      --  Create a record that describe thes location of an IIR node and
      --  returns the address of it.
      function Get_Location (N : Iir) return O_Enode
      is
         Constr : O_Record_Aggr_List;
         Aggr : O_Enode;
         Name : Name_Id;
         Line : Natural;
         Col : Natural;
         C : O_Lnode;
      begin
         Files_Map.Location_To_Position (Get_Location (N), Name, Line, Col);

         Start_Record_Aggr (Constr, Ghdl_Location_Type_Node);
         New_Record_Aggr_El
           (Constr,
            New_Address (Current_Filename_Node, Ghdl_Str_Len_Ptr_Node));
         New_Record_Aggr_El (Constr, New_Signed_Literal (Ghdl_Integer_Type,
                                                         Integer_64 (Line)));
         New_Record_Aggr_El (Constr, New_Signed_Literal (Ghdl_Integer_Type,
                                                         Integer_64 (Col)));
         Finish_Record_Aggr (Constr, Aggr);
         New_Const_Decl (C, Create_Uniq_Identifier, O_Storage_Private,
                         Ghdl_Location_Type_Node, Aggr);

         return New_Address (C, Ghdl_Location_Ptr_Node);
      end Get_Location;
   end Chap4;

   package body Chap5 is
      procedure Translate_Attribute_Specification
        (Spec : Iir_Attribute_Specification)
      is
         Attr : Iir_Attribute_Declaration;
         Mark : Id_Mark_Type;
         Info : Object_Info_Acc;
         Atinfo : Type_Info_Acc;
      begin
         Attr := Get_Attribute_Designator (Spec);
         Atinfo := Get_Info (Get_Type (Attr));
         Push_Identifier_Prefix (Mark, Get_Identifier (Attr));
         Info := Add_Info (Spec, Kind_Object);
         Info.Object_Var := Create_Var
           (Create_Uniq_Identifier,
            Atinfo.Ortho_Type (Mode_Value),
            Global_Storage);
         Pop_Identifier_Prefix (Mark);
      end Translate_Attribute_Specification;

      procedure Elab_Attribute_Specification
        (Spec : Iir_Attribute_Specification)
      is
         Attr : Iir_Attribute_Declaration;
      begin
         Attr := Get_Attribute_Designator (Spec);
         --  Kludge
         Set_Info (Attr, Get_Info (Spec));
         Chap4.Elab_Object_Value (Attr, Get_Expression (Spec));
         Set_Back_End_Info (Attr, System.Null_Address);
      end Elab_Attribute_Specification;

      procedure Elab_Disconnection_Specification
        (Spec : Iir_Disconnection_Specification)
      is
      begin
         null;
      end Elab_Disconnection_Specification;

      type Connect_Data is record
         Actual_Node : Mnode;
         Actual_Type : Iir;
         Actual : Iir;
         Formal : Iir;
      end record;

      procedure Connect_Scalar (Formal_Node : Mnode;
                                Formal_Type : Iir;
                                Data : Connect_Data)
      is
         Formal_Mode : Iir_Mode;
         Rule1, Rule2 : Boolean;
         Act_Node, Form_Node : O_Lnode;
      begin
         Formal_Mode := Get_Mode (Data.Formal);

         --  Two rules:

         --  LRM93 4.3.1.2
         --  For a signal of a scalar type, each source is either a driver or
         --  an OUT, INOUT, BUFFER or LINKAGE port of a component instance
         --  or of a block statement with which the signal is associated.
         Rule1 := Formal_Mode in Iir_Out_Mode .. Iir_Linkage_Mode;

         --  LRM93 12.6.2
         --  For a scalar signal S, the effective value of S is determined in
         --  the following manner:
         --  *  If S is [...] a port of mode BUFFER or [...], then the
         --     effective value of S is the same as the driving value of S.
         --  *  If S is a connected port of mode IN or INOUT, then the
         --     effective value of S is the same as the effective value of the
         --     actual part of the association element that associates an
         --     actual with S.
         --  *  [...]
         case Formal_Mode is
            when Iir_In_Mode
              | Iir_Inout_Mode =>
               Rule2 := True;
            when others =>
               Rule2 := False;
         end case;

         if Rule1 and Rule2 then
            Open_Temp;
            Act_Node := Create_Temp_Init
              (Ghdl_Signal_Ptr,
               New_Convert_Ov (New_Value (M2Lv (Data.Actual_Node)),
                               Ghdl_Signal_Ptr));
            Form_Node := Create_Temp_Init
              (Ghdl_Signal_Ptr,
               New_Convert_Ov (New_Value (M2Lv (Formal_Node)),
                               Ghdl_Signal_Ptr));
         else
            Act_Node := M2Lv (Data.Actual_Node);
            Form_Node := M2Lv (Formal_Node);
         end if;

         if Rule1 then
            --  According to rule 1, formal is a driver to actual.
            declare
               Constr : O_Assoc_List;
            begin
               Start_Association (Constr, Ghdl_Signal_Add_Source);
               New_Association (Constr, New_Convert_Ov (New_Value (Act_Node),
                                                        Ghdl_Signal_Ptr));
               New_Association (Constr, New_Convert_Ov (New_Value (Form_Node),
                                                        Ghdl_Signal_Ptr));
               New_Procedure_Call (Constr);
            end;
         end if;

         if Rule2 then
            --  According to rule 2, the effective value of formal is the
            --  effective value of actual.
            declare
               Constr : O_Assoc_List;
            begin
               Start_Association (Constr, Ghdl_Signal_Effective_Value);
               New_Association (Constr, New_Convert_Ov (New_Value (Form_Node),
                                                        Ghdl_Signal_Ptr));
               New_Association (Constr, New_Convert_Ov (New_Value (Act_Node),
                                                        Ghdl_Signal_Ptr));
               New_Procedure_Call (Constr);
            end;
         end if;

         if Rule1 and Rule2 then
            Close_Temp;
         end if;
      end Connect_Scalar;

      function Connect_Prepare_Data_Composite (Data : Connect_Data;
                                               Formal_Type : Iir)
        return Connect_Data
      is
         Res : Connect_Data;
         Atype : Iir;
      begin
         Atype := Get_Base_Type (Data.Actual_Type);
         if Get_Kind (Atype) = Iir_Kind_Record_Type_Definition then
            Res := Data;
            Stabilize (Res.Actual_Node);
            return Res;
         else
            return Data;
         end if;
      end Connect_Prepare_Data_Composite;

      function Connect_Update_Data_Array (Data : Connect_Data;
                                          Formal_Type : Iir;
                                          Index : O_Lnode)
        return Connect_Data
      is
         Res : Connect_Data;
      begin
         --  FIXME: should check matching elements!
         Res := (Actual_Node =>
                   Chap3.Index_Base (Chap3.Get_Array_Base (Data.Actual_Node),
                                     Data.Actual_Type, New_Value (Index)),
                 Actual_Type => Get_Element_Subtype (Data.Actual_Type),
                 Actual => Data.Actual,
                 Formal => Data.Formal);
         return Res;
      end Connect_Update_Data_Array;

      function Connect_Update_Data_Record (Data : Connect_Data;
                                           Formal_Type : Iir;
                                           El : Iir_Element_Declaration)
        return Connect_Data
      is
         Res : Connect_Data;
      begin
         Res := (Actual_Node =>
                   Chap6.Translate_Selected_Element (Data.Actual_Node, El),
                 Actual_Type => Get_Type (El),
                 Actual => Data.Actual,
                 Formal => Data.Formal);
         return Res;
      end Connect_Update_Data_Record;

      procedure Connect is new Foreach_Non_Composite
        (Data_Type => Connect_Data,
         Composite_Data_Type => Connect_Data,
         Do_Non_Composite => Connect_Scalar,
         Prepare_Data_Array => Connect_Prepare_Data_Composite,
         Update_Data_Array => Connect_Update_Data_Array,
         Prepare_Data_Record => Connect_Prepare_Data_Composite,
         Update_Data_Record => Connect_Update_Data_Record);

      procedure Elab_Unconstrained_Port (Port : Iir; Actual : Iir)
      is
         Act_Node : Mnode;
         Bounds : Mnode;
         Tinfo : Type_Info_Acc;
         Bound_Var : O_Lnode;
         Actual_Type : Iir;
      begin
         Actual_Type := Get_Type (Actual);
         Open_Temp;
         if Get_Kind (Actual_Type)
           not in Iir_Kinds_Unconstrained_Array_Type_Definition
         then
            Chap3.Create_Array_Subtype (Actual_Type);
            Tinfo := Get_Info (Actual_Type);
            Bounds := Chap3.Get_Array_Type_Bounds (Actual_Type);
            if Get_Alloc_Kind_For_Var (Tinfo.T.Array_Bounds) = Alloc_Stack then
               --  We need a copy.
               Bound_Var := Create_Temp (Tinfo.T.Bounds_Ptr_Type);
               New_Assign_Stmt
                 (Bound_Var,
                  Gen_Alloc (Alloc_Memory,
                             New_Sizeof (Tinfo.T.Bounds_Type, Ghdl_Index_Type),
                             Tinfo.T.Bounds_Ptr_Type));
               Gen_Memcpy (New_Value (Bound_Var),
                           M2Addr (Bounds),
                           New_Sizeof (Tinfo.T.Bounds_Type, Ghdl_Index_Type));
               Bounds := Lp2M (Bound_Var, Tinfo, Mode_Value,
                               Tinfo.T.Bounds_Type,
                               Tinfo.T.Bounds_Ptr_Type);
            end if;
         else
            Bounds := Chap3.Get_Array_Bounds (Chap6.Translate_Name (Actual));
         end if;
         Act_Node := Chap6.Translate_Name (Port);
         New_Assign_Stmt
           (-- FIXME: this works only because it is not stabilized,
            -- and therefore the bounds field is returned and not
            -- a pointer to the bounds.
            M2Lp (Chap3.Get_Array_Bounds (Act_Node)),
            M2Addr (Bounds));
         Close_Temp;
      end Elab_Unconstrained_Port;

      procedure Elab_Unconstrained_Ports_Of_Port_List (List : Iir_List)
      is
         El : Iir;
         Formal, Actual : Iir;
         Formal_Type : Iir;
         Ftinfo : Type_Info_Acc;
         Formal_Node, Actual_Node : Mnode;
         Data : Connect_Data;
      begin
         if List = Null_Iir_List then
            return;
         end if;
         for I in Natural loop
            El := Get_Nth_Element (List, I);
            exit when El = Null_Iir;
            Formal := Get_Associated_Formal (El);
            Formal_Type := Get_Type (Formal);
            Ftinfo := Get_Info (Formal_Type);
            Actual := Null_Iir;
            if Ftinfo.Type_Mode = Type_Mode_Fat_Array then
               case Get_Kind (El) is
                  when Iir_Kind_Association_Element_By_Expression =>
                     if Get_Whole_Association_Flag (El) then
                        Actual := Get_Actual (El);
                     end if;
               when others =>
                  Error_Kind ("elab_unconstrained_ports_of_port_list", El);
               end case;
            end if;
            if Actual /= Null_Iir then
               Elab_Unconstrained_Port (Formal, Actual);
            end if;
         end loop;
      end Elab_Unconstrained_Ports_Of_Port_List;

      procedure Elab_Port_Map_Aspect_List (List : Iir_List)
      is
         El : Iir;
         Formal, Actual : Iir;
         Formal_Node, Actual_Node : Mnode;
         Data : Connect_Data;
      begin
         if List = Null_Iir_List then
            return;
         end if;
         for I in Natural loop
            El := Get_Nth_Element (List, I);
            exit when El = Null_Iir;
            case Get_Kind (El) is
               when Iir_Kind_Association_Element_By_Expression =>
                  Open_Temp;
                  Formal := Get_Formal (El);
                  if Get_In_Conversion (El) /= Null_Iir then
                     --  TODO.
                     raise Internal_Error;
                  end if;
                  if Get_Out_Conversion (El) /= Null_Iir then
                     --  TODO.
                     raise Internal_Error;
                  end if;
                  --  translate formal
                  Formal_Node := Chap6.Translate_Name (Formal);
                  if Get_Object_Kind (Formal_Node) /= Mode_Signal then
                     raise Internal_Error;
                  end if;
                  --  translate actual (abort if not a signal).
                  Actual := Get_Actual (El);
                  Actual_Node := Chap6.Translate_Name (Actual);
                  if Get_Object_Kind (Actual_Node) /= Mode_Signal then
                     --  FIXME: actual may not be a signal, or even a name!
                     raise Internal_Error;
                  end if;
                  --  one to one connection.
                  Data := (Actual_Node => Actual_Node,
                           Actual_Type => Get_Type (Actual),
                           Actual => Actual,
                           Formal => Get_Associated_Formal (El));
                  Connect (Formal_Node, Get_Type (Formal), Data);
                  Close_Temp;
               when Iir_Kind_Association_Element_Open =>
                  null;
               when Iir_Kind_Association_Element_By_Individual =>
                  null;
               when others =>
                  Error_Kind ("elab_port_map_aspect_list", El);
            end case;
         end loop;
      end Elab_Port_Map_Aspect_List;

      procedure Elab_Generic_Map_Aspect_List (List : Iir_List)
      is
         El : Iir;
         Formal : Iir;
      begin
         if List = Null_Iir_List then
            return;
         end if;
         for I in Natural loop
            El := Get_Nth_Element (List, I);
            exit when El = Null_Iir;
            Formal := Get_Associated_Formal (El);
            case Get_Kind (El) is
               when Iir_Kind_Association_Element_By_Expression =>
                  Chap4.Elab_Object_Value (Formal, Get_Actual (El));
               when Iir_Kind_Association_Element_Open =>
                  Chap4.Elab_Object_Value (Formal, Get_Default_Value (Formal));
               when others =>
                  Error_Kind ("elab_generic_map_aspect_list", El);
            end case;
         end loop;
      end Elab_Generic_Map_Aspect_List;
   end Chap5;

   package body Chap6 is
      --  Extract from fat array FAT_ARRAY the range corresponding to dimension
      --  DIM.
      function Fat_Array_To_Base (Fat_Array : O_Lnode;
                                  Array_Info : Type_Info_Acc;
                                  Is_Sig : Object_Kind_Type)
        return O_Lnode
      is
         Lval : O_Lnode;
      begin
         --  From fat record, extract base pointer field.
         Lval := New_Selected_Element
           (Fat_Array, Array_Info.T.Base_Field (Is_Sig));
         --  Dereference it.
         Lval := New_Access_Element (New_Value (Lval));
         return Lval;
      end Fat_Array_To_Base;

      --  Extract from fat array FAT_ARRAY the range corresponding to dimension
      --  DIM.
      function Fat_Array_To_Range (Fat_Array : O_Lnode;
                                   Array_Type : Iir;
                                   Dim : Natural;
                                   Is_Sig : Object_Kind_Type)
        return O_Lnode
      is
         Lval : O_Lnode;
         Array_Info : Type_Info_Acc;
      begin
         Array_Info := Get_Info (Array_Type);
         case Array_Info.Type_Mode is
            when Type_Mode_Fat_Array =>
               --  From fat record, extract bounds field.
               Lval := New_Selected_Element
                 (Fat_Array, Array_Info.T.Bounds_Field (Is_Sig));
               --  Dereference it.
               Lval := New_Access_Element (New_Value (Lval));
            when Type_Mode_Ptr_Array =>
               Lval := Get_Var (Array_Info.T.Array_Bounds);
            when others =>
               raise Internal_Error;
         end case;
         --  Extract the range for the dimension.
         return New_Selected_Element (Lval, Array_Info.T.Bounds_Vector (Dim));
      end Fat_Array_To_Range;

      function Get_Array_Bound_Dir (Arr : O_Lnode;
                                    Arr_Type : Iir;
                                    Dim : Natural;
                                    Is_Sig : Object_Kind_Type)
        return O_Enode
      is
         Index_Type : Iir;
         Tinfo : Type_Info_Acc;
         Rinfo : Type_Info_Acc;
         Constraint : Iir;
      begin
         Index_Type := Get_Nth_Element (Get_Index_List (Arr_Type), Dim - 1);
         Tinfo := Get_Info (Arr_Type);

         case Tinfo.Type_Mode is
            when Type_Mode_Fat_Array
              | Type_Mode_Ptr_Array =>
               Rinfo := Get_Info (Get_Base_Type (Index_Type));
               return New_Value
                 (New_Selected_Element
                  (Fat_Array_To_Range (Arr, Arr_Type, Dim, Is_Sig),
                   Rinfo.T.Range_Dir));
            when Type_Mode_Array =>
               Constraint := Get_Range_Constraint (Index_Type);
               return Chap7.Translate_Static_Range_Dir (Constraint);
            when others =>
               raise Internal_Error;
         end case;
      end Get_Array_Bound_Dir;

      function Get_Array_Bound_Length (Arr : O_Lnode;
                                       Arr_Type : Iir;
                                       Dim : Natural;
                                       Is_Sig : Object_Kind_Type)
        return O_Enode
      is
         Tinfo : Type_Info_Acc;
         Index_Type : Iir;
         Rinfo : Type_Info_Acc;
         Constraint : Iir;
      begin
         Index_Type := Get_Nth_Element (Get_Index_List (Arr_Type), Dim - 1);

         Tinfo := Get_Info (Arr_Type);
         case Tinfo.Type_Mode is
            when Type_Mode_Fat_Array
              | Type_Mode_Ptr_Array =>
               Rinfo := Get_Info (Get_Base_Type (Index_Type));
               return New_Value
                 (New_Selected_Element
                  (Fat_Array_To_Range (Arr, Arr_Type, Dim, Is_Sig),
                   Rinfo.T.Range_Length));
            when Type_Mode_Array =>
               Constraint := Get_Range_Constraint (Index_Type);
               return Chap7.Translate_Static_Range_Length (Constraint);
            when others =>
               raise Internal_Error;
         end case;
      end Get_Array_Bound_Length;

      function Get_Array_Ptr_Bound_Length (Ptr : O_Lnode;
                                           Arr_Type : Iir;
                                           Dim : Natural;
                                           Is_Sig : Object_Kind_Type)
        return O_Enode
      is
         Tinfo : Type_Info_Acc;
      begin
         Tinfo := Get_Info (Arr_Type);
         case Tinfo.Type_Mode is
            when Type_Mode_Fat_Array =>
               return Get_Array_Bound_Length
                 (New_Acc_Value (Ptr), Arr_Type, Dim, Is_Sig);
            when Type_Mode_Array
              | Type_Mode_Ptr_Array =>
               return Get_Array_Bound_Length
                 (O_Lnode_Null, Arr_Type, Dim, Is_Sig);
            when others =>
               raise Internal_Error;
         end case;
      end Get_Array_Ptr_Bound_Length;

      function Get_Array_Bound_Left (Arr : O_Lnode;
                                     Arr_Type : Iir;
                                     Dim : Natural;
                                     Is_Sig : Object_Kind_Type)
        return O_Enode
      is
         Type_Info : Type_Info_Acc;
         Index_Type : Iir;
         Rinfo : Type_Info_Acc;
         Constraint : Iir;
      begin
         Index_Type := Get_Nth_Element (Get_Index_List (Arr_Type), Dim - 1);

         Type_Info := Get_Info (Arr_Type);
         case Type_Info.Type_Mode is
            when Type_Mode_Ptr_Array
              | Type_Mode_Fat_Array =>
               Rinfo := Get_Info (Get_Base_Type (Index_Type));
               return New_Value
                 (New_Selected_Element
                  (Fat_Array_To_Range (Arr, Arr_Type, Dim, Is_Sig),
                   Rinfo.T.Range_Left));
            when Type_Mode_Array =>
               Constraint := Get_Range_Constraint (Index_Type);
               return Chap7.Translate_Static_Range_Left (Constraint);
            when others =>
               raise Internal_Error;
         end case;
      end Get_Array_Bound_Left;

      function Get_Array_Bound_Right (Arr : O_Lnode;
                                      Arr_Type : Iir;
                                      Dim : Natural;
                                      Is_Sig : Object_Kind_Type)
        return O_Enode
      is
         Type_Info : Type_Info_Acc;
         Index_Type : Iir;
         Rinfo : Type_Info_Acc;
         Constraint : Iir;
      begin
         Index_Type := Get_Nth_Element (Get_Index_List (Arr_Type), Dim - 1);

         Type_Info := Get_Info (Arr_Type);

         case Type_Info.Type_Mode is
            when Type_Mode_Fat_Array
              | Type_Mode_Ptr_Array =>
               Rinfo := Get_Info (Get_Base_Type (Index_Type));
               return New_Value
                 (New_Selected_Element
                  (Fat_Array_To_Range (Arr, Arr_Type, Dim, Is_Sig),
                   Rinfo.T.Range_Right));
            when Type_Mode_Array =>
               Constraint := Get_Range_Constraint (Index_Type);
               return Chap7.Translate_Static_Range_Right (Constraint);
            when others =>
               raise Internal_Error;
         end case;
      end Get_Array_Bound_Right;

      --  There is a uniq number associated which each error.
      Bound_Error_Number : Unsigned_64 := 0;

      procedure Gen_Bound_Error (Loc : Iir)
      is
         Constr : O_Assoc_List;
      begin
         Start_Association (Constr, Ghdl_Bound_Check_Failed_L0);
         New_Association (Constr, New_Unsigned_Literal (Ghdl_Index_Type,
                                                        Bound_Error_Number));
         New_Procedure_Call (Constr);
         Bound_Error_Number := Bound_Error_Number + 1;
      end Gen_Bound_Error;

      procedure Gen_Program_Error (Loc : Iir)
      is
         Constr : O_Assoc_List;
      begin
         Start_Association (Constr, Ghdl_Program_Error);
         New_Procedure_Call (Constr);
      end Gen_Program_Error;

      --  Generate code to emit a failure if COND is TRUE, indicating an
      --  index violation for dimension DIM of an array.  LOC is usually
      --  the expression which has computed the index and is used only for
      --  its location.
      procedure Check_Bound_Error (Cond : O_Enode; Loc : Iir; Dim : Natural)
      is
         If_Blk : O_If_Block;
      begin
         Start_If_Stmt (If_Blk, Cond);
         Gen_Bound_Error (Loc);
         Finish_If_Stmt (If_Blk);
      end Check_Bound_Error;

      function Translate_Index_To_Offset (Range_Ptr : O_Lnode;
                                          Index : O_Enode;
                                          Index_Type : Iir;
                                          Loc : Iir)
        return O_Enode
      is
         Dir : O_Enode;
         If_Blk : O_If_Block;
         Res : O_Lnode;
         Off : O_Lnode;
         Cond1, Cond2: O_Enode;
         Index_Node : O_Lnode;
         Bound_Node : O_Lnode;
         Index_Info : Type_Info_Acc;
      begin
         Index_Info := Get_Info (Index_Type);

         Res := Create_Temp (Ghdl_Index_Type);

         Open_Temp;

         Index_Node := Create_Temp_Init
           (Index_Info.Ortho_Type (Mode_Value), Index);
         Bound_Node := Create_Temp_Init
           (Index_Info.Ortho_Type (Mode_Value),
            New_Value_Selected_Acc_Value (Range_Ptr, Index_Info.T.Range_Left));
         Off := Create_Temp (Index_Info.Ortho_Type (Mode_Value));

         Dir := New_Value_Selected_Acc_Value (Range_Ptr,
                                              Index_Info.T.Range_Dir);

         Start_If_Stmt (If_Blk,
                        New_Compare_Op (ON_Eq, Dir, Ghdl_Dir_To_Node,
                                        Ghdl_Bool_Type_Node));
         --  Direction TO:  INDEX - LEFT.
         New_Assign_Stmt (Off, New_Dyadic_Op (ON_Sub_Ov,
                                              New_Value (Index_Node),
                                              New_Value (Bound_Node)));
         New_Else_Stmt (If_Blk);
         --  Direction DOWNTO: LEFT - INDEX.
         New_Assign_Stmt (Off, New_Dyadic_Op (ON_Sub_Ov,
                                              New_Value (Bound_Node),
                                              New_Value (Index_Node)));
         Finish_If_Stmt (If_Blk);

         --  Get the offset.
         New_Assign_Stmt
           (Res, New_Convert_Ov (New_Value (Off), Ghdl_Index_Type));

         --  Check bounds.
         Cond1 := New_Compare_Op
           (ON_Lt,
            New_Value (Off),
            New_Signed_Literal (Index_Info.Ortho_Type (Mode_Value), 0),
            Ghdl_Bool_Type_Node);

         Cond2 := New_Compare_Op
           (ON_Ge,
            New_Value (Res),
            New_Value_Selected_Acc_Value (Range_Ptr,
                                          Index_Info.T.Range_Length),
            Ghdl_Bool_Type_Node);
         Check_Bound_Error (New_Dyadic_Op (ON_Or, Cond1, Cond2), Loc, 0);

         Close_Temp;

         return New_Value (Res);
      end Translate_Index_To_Offset;

      --  Translate index EXPR in dimension DIM of thin array into an
      --  offset.
      --  This checks bounds.
      function Translate_Thin_Index_Offset (Index_Type : Iir;
                                            Index_Range : Iir;
                                            Dim : Natural;
                                            Expr : Iir)
        return O_Enode
      is
         Obound : O_Enode;
         Res : O_Lnode;
         Off : O_Lnode;
         Cond2: O_Enode;
         Index : O_Enode;
         Index_Info : Type_Info_Acc;
         V : Iir_Int64;
         B : Iir_Int64;
      begin
         if Get_Expr_Staticness (Expr) = Locally then
            V := Eval_Pos (Expr);
            B := Eval_Pos (Get_Left_Limit (Index_Range));
            if Get_Direction (Index_Range) = Iir_To then
               B := V - B;
            else
               B := B - V;
            end if;
            return New_Unsigned_Literal (Ghdl_Index_Type, Unsigned_64 (B));
         else
            Index_Info := Get_Info (Get_Base_Type (Index_Type));
            Res := Create_Temp (Ghdl_Index_Type);
            Off := Create_Temp (Index_Info.Ortho_Type (Mode_Value));

            Index := Chap7.Translate_Expression (Expr, Index_Type);

            Obound := Chap7.Translate_Static_Range_Left
                 (Index_Range, Index_Type);

            if Get_Direction (Index_Range) = Iir_To then
               --  Direction TO:  INDEX - LEFT.
               Index := New_Dyadic_Op (ON_Sub_Ov, Index, Obound);
            else
               --  Direction DOWNTO:  LEFT - INDEX.
               Index := New_Dyadic_Op (ON_Sub_Ov, Obound, Index);
            end if;

            New_Assign_Stmt (Off, Index);

            --  Get the offset.
            New_Assign_Stmt
              (Res, New_Convert_Ov (New_Value (Off), Ghdl_Index_Type));

            --  Since the value is unsigned, both left and right bounds are
            --  checked in the same time.
            Cond2 := New_Compare_Op
              (ON_Ge, New_Value (Res),
               Chap7.Translate_Static_Range_Length (Index_Range),
               Ghdl_Bool_Type_Node);
            Check_Bound_Error (Cond2, Expr, Dim);

            return New_Value (Res);
         end if;
      end Translate_Thin_Index_Offset;

      --  Translate an indexed name.
      function Translate_Indexed_Name (Expr : Iir) return Mnode
      is
         Prefix : Iir;
         Name_Node : Mnode;
         Prefix_Type : Iir;
         Index : Iir;
         Index_List : Iir_List;
         Type_List : Iir_List;
         Offset : O_Lnode;
         R : O_Enode;
         Length : O_Enode;
         Itype : Iir;
         Ibasetype : Iir;
         Prefix_Info : Type_Info_Acc;
         Nbr_Dim : Natural;
         Fat_Ptr : O_Lnode;
         Range_Ptr : Mnode;
      begin
         Prefix := Get_Prefix (Expr);
         Prefix_Type := Get_Type (Prefix);
         Prefix_Info := Get_Info (Prefix_Type);
         case Prefix_Info.Type_Mode is
            when Type_Mode_Fat_Array =>
               Name_Node := Stabilize (Translate_Name (Prefix));
            when Type_Mode_Array =>
               null;
            when Type_Mode_Ptr_Array =>
               --  FIXME: should save the bounds address ?
               Fat_Ptr := O_Lnode_Null;
            when others =>
               raise Internal_Error;
         end case;
         Index_List := Get_Index_List (Expr);
         Type_List := Get_Index_List (Prefix_Type);
         Nbr_Dim := Get_Nbr_Elements (Index_List);
         Offset := Create_Temp (Ghdl_Index_Type);
         for Dim in 1 .. Nbr_Dim loop
            Index := Get_Nth_Element (Index_List, Dim - 1);
            Itype := Get_Nth_Element (Type_List, Dim - 1);
            Ibasetype := Get_Base_Type (Itype);
            Open_Temp;
            --  Compute index for the current dimension.
            case Prefix_Info.Type_Mode is
               when Type_Mode_Fat_Array =>
                  Range_Ptr := Chap3.Get_Array_Range (Name_Node, Prefix_Type,
                                                      Dim);
               when Type_Mode_Ptr_Array =>
                  Range_Ptr := Chap3.Bounds_To_Range
                    (Chap3.Get_Array_Type_Bounds (Prefix_Type),
                     Prefix_Type, Dim);
               when Type_Mode_Array =>
                  null;
               when others =>
                  raise Internal_Error;
            end case;
            case Prefix_Info.Type_Mode is
               when Type_Mode_Fat_Array
                 | Type_Mode_Ptr_Array =>
                  Range_Ptr := Stabilize (Range_Ptr);
                  R := Translate_Index_To_Offset
                    (M2Vp (Range_Ptr),
                     Chap7.Translate_Expression (Index, Ibasetype),
                     Ibasetype,
                     Index);
               when Type_Mode_Array =>
                  --  BASE is a thin array.
                  R := Translate_Thin_Index_Offset
                    (Ibasetype, Get_Range_Constraint (Itype), Dim, Index);
               when others =>
                  raise Internal_Error;
            end case;
            if Dim = 1 then
               --  First dimension.
               New_Assign_Stmt (Offset, R);
            else
               --  If there are more dimension(s) to follow, then multiply
               --  the current offset by the length of the current dimension.
               case Prefix_Info.Type_Mode is
                  when Type_Mode_Fat_Array
                    | Type_Mode_Ptr_Array =>
                     Length := M2E (Chap3.Range_To_Length (Range_Ptr));
                  when Type_Mode_Array =>
                     Length := Chap7.Translate_Static_Range_Length
                       (Get_Range_Constraint (Itype));
                  when others =>
                     raise Internal_Error;
               end case;
               New_Assign_Stmt
                 (Offset,
                  New_Dyadic_Op (ON_Add_Ov,
                                 New_Dyadic_Op (ON_Mul_Ov,
                                                New_Value (Offset), Length),
                                 R));
            end if;
            Close_Temp;
         end loop;

         case Prefix_Info.Type_Mode is
            when Type_Mode_Fat_Array =>
               null;
            when Type_Mode_Array
              | Type_Mode_Ptr_Array =>
               Name_Node := Chap6.Translate_Name (Prefix);
            when others =>
               raise Internal_Error;
         end case;
         return Chap3.Index_Base
           (Chap3.Get_Array_Base (Name_Node), Prefix_Type, New_Value (Offset));
      end Translate_Indexed_Name;

      function Translate_Slice_Name (Expr : Iir_Slice_Name) return Mnode
      is
         --  Prefix of the slice.
         Prefix : Iir;

         --  Type of the prefix.
         Prefix_Type : Iir;

         --  Type info of the prefix.
         Prefix_Info : Type_Info_Acc;

         --  Type of the slice.
         Slice_Type : Iir;
         Slice_Info : Type_Info_Acc;
         Slice_Binfo : Type_Info_Acc;

         --  Type of the first (and only) index of the prefix array type.
         Index_Type : Iir;

         --  True iff the direction of the slice is known at compile time.
         Static_Range : Boolean;

         --  Suffix of the slice (discrete range).
         Expr_Range : Iir;

         --  Ortho node for the prefix.
         Prefix_Node : Mnode;

         --  Object kind of the prefix.
         Kind : Object_Kind_Type;

         --  Variable pointing to the prefix.
         Prefix_Var : Mnode;

         --  Type info of the slice base type.
         --Info : Type_Info_Acc;

         --  Type info of the range base type.
         Index_Info : Type_Info_Acc;

         --  Variables pointing to slice and prefix ranges.
         Slice_Range : Mnode;
         Prefix_Range : Mnode;

         Res : O_Lnode;

         Diff : O_Lnode;
         Unsigned_Diff : O_Lnode;
         If_Blk : O_If_Block;
         If_Blk1 : O_If_Block;
      begin
         --  Evaluate the prefix.
         Prefix := Get_Prefix (Expr);
         Slice_Type := Get_Type (Expr);
         Expr_Range := Get_Suffix (Expr);
         Prefix_Type := Get_Type (Prefix);
         Index_Type := Get_Nth_Element (Get_Index_List (Prefix_Type), 0);

         --  Evaluate slice bounds.
         Chap3.Create_Array_Subtype (Slice_Type);

         Prefix_Node := Chap6.Translate_Name (Prefix);
         Kind := Get_Object_Kind (Prefix_Node);

         Prefix_Info := Get_Info (Prefix_Type);
         Slice_Info := Get_Info (Slice_Type);

         if Slice_Info.Type_Mode = Type_Mode_Array
           and then Prefix_Info.Type_Mode = Type_Mode_Array
         then
            --  Both prefix and result are constrained array.
            declare
               Prefix_Left, Slice_Left : Iir_Int64;
               Off : Iir_Int64;
               Slice_Index_Type : Iir;
               Slice_Range : Iir;
               Slice_Length : Iir_Int64;
               Index_Range : Iir;
            begin
               Index_Range := Get_Range_Constraint (Index_Type);
               Prefix_Left := Eval_Pos (Get_Left_Limit (Index_Range));
               Slice_Index_Type := Get_First_Element
                 (Get_Index_List (Slice_Type));
               Slice_Range := Get_Range_Constraint (Slice_Index_Type);
               Slice_Left := Eval_Pos (Get_Left_Limit (Slice_Range));
               Slice_Length := Eval_Discrete_Range_Length (Slice_Range);
               if Slice_Length = 0 then
                  --  Null slice.
                  return Prefix_Node;
               end if;
               if Get_Direction (Index_Range) /= Get_Direction (Slice_Range)
               then
                  --  Must have been caught by sem.
                  --  FIXME: this is allowed with vhdl87
                  raise Internal_Error;
               end if;
               --  Both prefix and slice are thin array.
               case Get_Direction (Index_Range) is
                  when Iir_To =>
                     Off := Slice_Left - Prefix_Left;
                  when Iir_Downto =>
                     Off := Prefix_Left - Slice_Left;
               end case;
               if Off < 0 then
                  --  Must have been caught by sem.
                  raise Internal_Error;
               end if;
               if Off + Slice_Length > Eval_Discrete_Range_Length (Index_Range)
               then
                  --  Must have been caught by sem.
                  raise Internal_Error;
               end if;
               return Lv2M
                 (New_Slice (M2Lv (Prefix_Node),
                             Slice_Info.Ortho_Type (Kind),
                             New_Unsigned_Literal (Ghdl_Index_Type,
                                                   Unsigned_64 (Off))),
                  Slice_Info,
                  Kind);
            end;
         end if;

         Slice_Binfo := Get_Info (Get_Base_Type (Slice_Type));

         --  Save prefix.
         Prefix_Var := Stabilize (Prefix_Node);

         Index_Info := Get_Info (Get_Base_Type (Index_Type));

         --  Save prefix bounds.
         Prefix_Range := Stabilize
           (Chap3.Get_Array_Range (Prefix_Var, Prefix_Type, 1));

         --  Save slice bounds.
         Slice_Range := Stabilize
           (Chap3.Bounds_To_Range (Chap3.Get_Array_Type_Bounds (Slice_Type),
                                   Slice_Type, 1));

         --  TRUE if the direction of the slice is known.
         Static_Range := Get_Kind (Expr_Range) = Iir_Kind_Range_Expression;

         --  Check direction against same direction, error if different.
         --  FIXME: what about v87 -> if different then null slice
         if not Static_Range
           or else Get_Kind (Prefix_Type) /= Iir_Kind_Array_Subtype_Definition
         then
            --  Check same direction.
            Check_Bound_Error
              (New_Compare_Op
               (ON_Neq,
                M2E (Chap3.Range_To_Dir (Prefix_Range)),
                M2E (Chap3.Range_To_Dir (Slice_Range)),
                Ghdl_Bool_Type_Node),
               Expr, 1);
         end if;

         --  Check if not a null slice.
         --  FIXME: why ?
         --Start_If_Stmt
         --  (If_Blk,
         --   New_Compare_Op
         --   (ON_Neq,
         --    Get_Array_Bound_Length (Res, Prefix_Type, 1, Sig),
         --    New_Unsigned_Literal (Ghdl_Index_Type, 0),
         --    Ghdl_Bool_Type_Node));

         Diff := Create_Temp (Index_Info.Ortho_Type (Mode_Value));

         --  Compute the offset in the prefix.
         if not Static_Range then
            Start_If_Stmt
              (If_Blk1,
               New_Compare_Op
               (ON_Eq,
                M2E (Chap3.Range_To_Dir (Slice_Range)),
                Ghdl_Dir_To_Node,
                Ghdl_Bool_Type_Node));
         end if;
         if not Static_Range or else Get_Direction (Expr_Range) = Iir_To then
            --  Diff = slice - bounds.
            New_Assign_Stmt
              (Diff, New_Dyadic_Op (ON_Sub_Ov,
                                    M2E (Chap3.Range_To_Left (Slice_Range)),
                                    M2E (Chap3.Range_To_Left (Prefix_Range))));
         end if;
         if not Static_Range then
            New_Else_Stmt (If_Blk1);
         end if;
         if not Static_Range or else Get_Direction (Expr_Range) = Iir_Downto
         then
            --  Diff = bounds - slice.
            New_Assign_Stmt
              (Diff, New_Dyadic_Op (ON_Sub_Ov,
                                    M2E (Chap3.Range_To_Left (Prefix_Range)),
                                    M2E (Chap3.Range_To_Left (Slice_Range))));
         end if;
         if not Static_Range then
            Finish_If_Stmt (If_Blk1);
         end if;

         --  Note: this also check for overflow.
         Unsigned_Diff := Create_Temp (Ghdl_Index_Type);
         New_Assign_Stmt (Unsigned_Diff,
                          New_Convert_Ov (New_Value (Diff), Ghdl_Index_Type));

         --  Check bounds.
         declare
            Err_1 : O_Enode;
            Err_2 : O_Enode;
         begin
            --  Bounds error if left of slice is before left of prefix.
            Err_1 := New_Compare_Op
              (ON_Lt,
               New_Value (Diff),
               New_Signed_Literal (Index_Info.Ortho_Type (Mode_Value), 0),
               Ghdl_Bool_Type_Node);
            --  Bounds error if right of slice is after right of prefix.
            Err_2 := New_Compare_Op
              (ON_Gt,
               New_Dyadic_Op (ON_Add_Ov,
                              New_Value (Unsigned_Diff),
                              M2E (Chap3.Range_To_Length (Slice_Range))),
               M2E (Chap3.Range_To_Length (Prefix_Range)),
               Ghdl_Bool_Type_Node);
            Check_Bound_Error (New_Dyadic_Op (ON_Or, Err_1, Err_2), Expr, 1);
         end;

         --  Create the result (fat array) and assign the bounds field.
         case Slice_Info.Type_Mode is
            when Type_Mode_Fat_Array =>
               Res := Create_Temp (Slice_Info.Ortho_Type (Kind));
               New_Assign_Stmt
                 (New_Selected_Element (Res,
                                        Slice_Info.T.Bounds_Field (Kind)),
                  New_Value (M2Lp (Slice_Range)));
               New_Assign_Stmt
                 (New_Selected_Element (Res, Slice_Info.T.Base_Field (Kind)),
                  New_Address
                  (New_Slice (M2Lv (Chap3.Get_Array_Base (Prefix_Var)),
                              Slice_Info.T.Base_Type (Kind),
                              New_Value (Unsigned_Diff)),
                   Slice_Info.T.Base_Ptr_Type (Kind)));
               return Vv2M (Res, Slice_Info, Kind);
            when Type_Mode_Array
              | Type_Mode_Ptr_Array =>
               Res := New_Slice
                 (M2Lv (Chap3.Get_Array_Base (Prefix_Var)),
                  Slice_Info.T.Base_Type (Kind),
                  New_Value (Unsigned_Diff));
               return Lv2M (Res,
                            True,
                            Slice_Info.T.Base_Type (Kind),
                            Slice_Info.T.Base_Ptr_Type (Kind),
                            Slice_Info, Kind);
            when others =>
               raise Internal_Error;
         end case;

         --Finish_If_Stmt (If_Blk);

      end Translate_Slice_Name;

      function Translate_Interface_Name
        (Inter : Iir; Info : Ortho_Info_Acc; Kind : Object_Kind_Type)
        return Mnode
      is
         Type_Info : Type_Info_Acc;
      begin
         Type_Info := Get_Info (Get_Type (Inter));
         case Info.Kind is
            when Kind_Object =>
               --  For a generic or a port.
               return Get_Var (Info.Object_Var, Type_Info, Kind);
            when Kind_Interface =>
               --  For a parameter.
               case Type_Info.Type_Mode is
                  when Type_Mode_Unknown =>
                     raise Internal_Error;
                  when Type_Mode_By_Value =>
                     --  Parameter is passed by value.
                     if Info.Interface_Field /= O_Fnode_Null then
                        --  And by copy.
                        return Lv2M
                          (New_Selected_Acc_Value
                           (Info.Interface_Node, Info.Interface_Field),
                           Type_Info, Kind);
                     else
                        return Lv2M (Info.Interface_Node, Type_Info, Kind);
                     end if;
                  when Type_Mode_By_Ref
                    | Type_Mode_By_Copy =>
                     --  Parameter is passed by reference, dereference it.
                     return Lp2M (Info.Interface_Node, Type_Info, Kind);
               end case;
            when others =>
               raise Internal_Error;
         end case;
      end Translate_Interface_Name;

      function Translate_Selected_Element (Prefix : Mnode;
                                           El : Iir_Element_Declaration)
        return Mnode
      is
         El_Info : Field_Info_Acc;
         Kind : Object_Kind_Type;
      begin
         El_Info := Get_Info (El);
         Kind := Get_Object_Kind (Prefix);
         return Lo2M (New_Selected_Element (M2Lv (Prefix),
                                            El_Info.Field_Node (Kind)),
                      Get_Info (Get_Type (El)), Kind);
      end Translate_Selected_Element;

      function Translate_Name (Name : Iir) return Mnode
      is
         Name_Type : Iir;
         Name_Info : Ortho_Info_Acc;
         Type_Info : Type_Info_Acc;
      begin
         Name_Type := Get_Type (Name);
         Name_Info := Get_Info (Name);
         Type_Info := Get_Info (Name_Type);
         case Get_Kind (Name) is
            when Iir_Kind_Constant_Declaration
              | Iir_Kind_Variable_Declaration
              | Iir_Kind_File_Declaration =>
               return Get_Var (Name_Info.Object_Var, Type_Info, Mode_Value);

            when Iir_Kind_Attribute_Value =>
               return Get_Var
                 (Get_Info (Get_Attribute_Specification (Name)).Object_Var,
                  Type_Info, Mode_Value);

            when Iir_Kind_Alias_Declaration =>
               --  FIXME: check for type_mode_array, which is in fact a
               --  pointer.
               case Type_Info.Type_Mode is
                  when Type_Mode_Fat_Array =>
                     --  Is such a mode possible ??
                     raise Internal_Error;
                  when Type_Mode_Ptr_Array
                    | Type_Mode_Array
                    | Type_Mode_Scalar =>
                     return Lp2M (Get_Var (Name_Info.Alias_Var),
                                  Type_Info, Name_Info.Alias_Kind);
                  when others =>
                     raise Internal_Error;
               end case;

            when Iir_Kind_Signal_Declaration
              | Iir_Kind_Stable_Attribute
              | Iir_Kind_Guard_Signal_Declaration =>
               return Get_Var (Name_Info.Object_Var, Type_Info, Mode_Signal);

            when Iir_Kind_Constant_Interface_Declaration =>
               return Translate_Interface_Name (Name, Name_Info, Mode_Value);

            when Iir_Kind_File_Interface_Declaration =>
               return Translate_Interface_Name (Name, Name_Info, Mode_Value);

            when Iir_Kind_Variable_Interface_Declaration =>
               if Name_Info.Interface_Field /= O_Fnode_Null then
                  --  Passed via the result record.
                  return Lv2M
                    (New_Selected_Element
                     (New_Acc_Value (Name_Info.Interface_Node),
                      Name_Info.Interface_Field),
                     Type_Info, Mode_Value);
               else
                  return Translate_Interface_Name
                    (Name, Name_Info, Mode_Value);
               end if;

            when Iir_Kind_Signal_Interface_Declaration =>
               return Translate_Interface_Name (Name, Name_Info, Mode_Signal);

            when Iir_Kind_Indexed_Name =>
               return Translate_Indexed_Name (Name);

            when Iir_Kind_Slice_Name =>
               return Translate_Slice_Name (Name);

            when Iir_Kind_Selected_By_All_Name
              | Iir_Kind_Implicit_Dereference =>
               declare
                  Pfx : O_Enode;
               begin
                  Pfx := Chap7.Translate_Expression (Get_Prefix (Name));
                  --  FIXME: what about fat pointer ??
                  return Lv2M (New_Access_Element (Pfx),
                               Type_Info, Mode_Value);
               end;

            when Iir_Kind_Selected_Element =>
               return Translate_Selected_Element
                 (Translate_Name (Get_Prefix (Name)),
                  Get_Selected_Element (Name));

            when Iir_Kind_Function_Call =>
               --  This can appear as a prefix of a name, therefore, the
               --  result is always a composite type.
               declare
                  Imp : Iir;
               begin
                  Imp := Get_Implementation (Name);
                  if Get_Kind (Imp) = Iir_Kind_Implicit_Function_Declaration
                  then
                     --  FIXME : to be done
                     raise Internal_Error;
                  else
                     return E2M (Chap7.Translate_Function_Call
                                 (Imp, Get_Parameter_Association_List (Name)),
                                 Type_Info, Mode_Value);
                  end if;
               end;

            when others =>
               Error_Kind ("translate_name", Name);
         end case;
      end Translate_Name;

      type Inst_Type (Is_Pkg : Boolean := False) is record
         case Is_Pkg is
            when True =>
               Lfield : O_Lnode;
            when False =>
               Itype : O_Tnode;
               Ifield : O_Fnode;
         end case;
      end record;

      function Get_Instance_Name (Inst : Inst_Type) return O_Lnode is
      begin
         if Inst.Is_Pkg then
            return Inst.Lfield;
         else
            return New_Selected_Element
              (Get_Instance_Ref (Inst.Itype), Inst.Ifield);
         end if;
      end Get_Instance_Name;

      function Get_Instance_Ref (Inst : Inst_Type) return O_Enode is
      begin
         return New_Address
           (Get_Instance_Name (Inst), Ghdl_Instance_Name_Acc);
      end Get_Instance_Ref;

      function Get_Field (Inst : Inst_Type; F : O_Fnode) return O_Lnode is
      begin
         return New_Selected_Element (Get_Instance_Name (Inst), F);
      end Get_Field;

      function Get_Inst_Type_From_Block (Blk : Iir) return Inst_Type
      is
         Info : Ortho_Info_Acc;
      begin
         Info := Get_Info (Blk);
         case Get_Kind (Blk) is
            when Iir_Kind_Entity_Declaration
              | Iir_Kind_Architecture_Declaration
              | Iir_Kind_Component_Instantiation_Statement
              | Iir_Kind_Block_Statement
              | Iir_Kind_Generate_Statement =>
               return (Is_Pkg => False,
                       Itype => Info.Block_Decls_Type,
                       Ifield => Info.Block_Name);
            when Iir_Kind_Component_Declaration =>
               return (Is_Pkg => False,
                       Itype => Info.Comp_Type,
                       Ifield => Info.Comp_Name);
            when Iir_Kind_Process_Statement
              | Iir_Kind_Sensitized_Process_Statement =>
               return (Is_Pkg => False,
                       Itype => Info.Process_Decls_Type,
                       Ifield => Info.Process_Name);
            when Iir_Kind_Package_Declaration =>
               return (Is_Pkg => True,
                       Lfield => Info.Package_Name);
            when others =>
               Error_Kind ("get_inst_type_from_block", Blk);
         end case;
      end Get_Inst_Type_From_Block;

      procedure Link_With_Parent (Parent_Inst, Child_Inst : Inst_Type)
      is
      begin
         --  set this.parent to parent.
         New_Assign_Stmt
           (Get_Field (Child_Inst, Ghdl_Instance_Name_Parent),
            Get_Instance_Ref (Parent_Inst));
         --  set brother to parent.child
         New_Assign_Stmt
           (Get_Field (Child_Inst, Ghdl_Instance_Name_Brother),
            New_Value (Get_Field (Parent_Inst, Ghdl_Instance_Name_Child)));
         --  set parent.child to this.
         New_Assign_Stmt
           (Get_Field (Parent_Inst, Ghdl_Instance_Name_Child),
            Get_Instance_Ref (Child_Inst));
      end Link_With_Parent;

      procedure Set_Instance_Name (Decl : Iir)
      is
         N : O_Lnode;
         Decl_Inst : Inst_Type;
         Parent_Inst : Inst_Type;
         Parent : Iir;
         V : O_Enode;
         Id : Name_Id;
      begin
         case Get_Kind (Decl) is
            when Iir_Kind_Entity_Declaration =>
               V := Ghdl_Name_Entity;
            when Iir_Kind_Architecture_Declaration =>
               V := Ghdl_Name_Architecture;
            when Iir_Kind_Component_Instantiation_Statement =>
               V := Ghdl_Name_Component;
            when Iir_Kind_Block_Statement =>
               V := Ghdl_Name_Block;
            when Iir_Kind_Generate_Statement =>
               declare
                  Scheme : Iir;
               begin
                  Scheme := Get_Generation_Scheme (Decl);
                  if Get_Kind (Scheme) = Iir_Kind_Iterator_Declaration then
                     V := Ghdl_Name_Generate_Iterative;
                  else
                     V := Ghdl_Name_Generate_Conditional;
                  end if;
               end;
            when Iir_Kind_Process_Statement
              | Iir_Kind_Sensitized_Process_Statement =>
               V := Ghdl_Name_Process;
            when Iir_Kind_Package_Declaration =>
               V := Ghdl_Name_Package;
            when others =>
               Error_Kind ("set_instance_name(2)", Decl);
         end case;

         case Get_Kind (Decl) is
            when Iir_Kind_Package_Declaration =>
               Parent := Null_Iir;
               Id := Null_Identifier;
            when Iir_Kind_Entity_Declaration =>
               Parent := Null_Iir;
               Id := Null_Identifier;
            when Iir_Kind_Architecture_Declaration =>
               Parent := Get_Entity (Decl);
               Id := Null_Identifier;
            when Iir_Kind_Component_Instantiation_Statement
              | Iir_Kind_Block_Statement
              | Iir_Kind_Generate_Statement
              | Iir_Kind_Process_Statement
              | Iir_Kind_Sensitized_Process_Statement =>
               Parent := Get_Parent (Decl);
               Id := Get_Identifier (Decl);
            when others =>
               Error_Kind ("set_instance_name(3)", Decl);
         end case;

         Decl_Inst := Get_Inst_Type_From_Block (Decl);
         declare
            Ident : O_Ident;
            Identstr : O_Ident;
         begin
            if Id = Null_Identifier then
               Ident := Create_Identifier ("_NAME");
               Identstr := Create_Identifier ("_NAMESTR");
            else
               Ident := Create_Identifier (Id, "_NAME");
               Identstr := Create_Identifier (Id, "_NAMESTR");
            end if;
            New_Const_Decl
              (N, Ident,
               O_Storage_Private, Ghdl_Str_Len_Type_Node,
               Create_String_Len (Get_Identifier (Decl), Identstr));
         end;

         New_Assign_Stmt (Get_Field (Decl_Inst, Ghdl_Instance_Name_Kind), V);

         New_Assign_Stmt (Get_Field (Decl_Inst, Ghdl_Instance_Name_Name),
                          New_Address (N, Ghdl_Str_Len_Ptr_Node));

         New_Assign_Stmt (Get_Field (Decl_Inst, Ghdl_Instance_Name_Child),
                          New_Null_Access (Ghdl_Instance_Name_Acc));

         if Parent /= Null_Iir then
            Parent_Inst := Get_Inst_Type_From_Block (Parent);
         end if;

         --  set parent
         if Parent = Null_Iir then
            New_Assign_Stmt (Get_Field (Decl_Inst, Ghdl_Instance_Name_Parent),
                             New_Null_Access (Ghdl_Instance_Name_Acc));
            New_Assign_Stmt (Get_Field (Decl_Inst, Ghdl_Instance_Name_Brother),
                             New_Null_Access (Ghdl_Instance_Name_Acc));
         else
            Link_With_Parent (Parent_Inst, Decl_Inst);
         end if;
      end Set_Instance_Name;

      procedure Link_Instance_Name (Parent : Iir; Child : Iir)
      is
         Decl_Inst : Inst_Type;
         Parent_Inst : Inst_Type;
      begin
         if Parent = Null_Iir then
            Parent_Inst := (Is_Pkg => True,
                            Lfield => Ghdl_Hierarchy_Top);
         else
            Parent_Inst := Get_Inst_Type_From_Block (Parent);
         end if;
         Decl_Inst := Get_Inst_Type_From_Block (Child);
         Link_With_Parent (Parent_Inst, Decl_Inst);
      end Link_Instance_Name;

      function Get_Instance_Name_Ref (Blk : Iir) return O_Enode is
      begin
         return Get_Instance_Ref (Get_Inst_Type_From_Block (Blk));
      end Get_Instance_Name_Ref;

   end Chap6;

   package body Chap7 is
      function Translate_Static_String_Literal (Str : Iir)
        return O_Enode
      is
         use Name_Table;

         Lit_Type : Iir;
         Lit_List : Iir_List;
         Element_Type : Iir;
         Element_Otype : O_Tnode;
         Literal_List : Iir_List;
         Lit : Iir;
         List : O_Array_Aggr_List;
         Res : O_Enode;
      begin
         Lit_Type := Get_Type (Str);
         Chap3.Translate_Anonymous_Type_Definition (Lit_Type);
         Lit_List := Get_String_Literals_List (Str);

         Start_Array_Aggr (List, Get_Ortho_Type (Lit_Type, Mode_Value));

         if Lit_List /= Null_Iir_List then
            for I in Natural loop
               Lit := Get_Nth_Element (Lit_List, I);
               exit when Lit = Null_Iir;
               New_Array_Aggr_El (List, Get_Ortho_Expr (Lit));
            end loop;
         else
            Element_Type := Get_Element_Subtype (Lit_Type);
            Element_Otype := Get_Ortho_Type (Element_Type, Mode_Value);
            Literal_List :=
              Get_Enumeration_Literal_List (Get_Base_Type (Element_Type));
            Image (Get_Identifier (Str));
            for I in 1 .. Name_Length loop
               Lit := Find_Name_In_List
                 (Literal_List,
                  Name_Table.Get_Identifier (Name_Buffer (I)));
               New_Array_Aggr_El (List, Get_Ortho_Expr (Lit));
            end loop;
         end if;

         Finish_Array_Aggr (List, Res);
         return Res;
      end Translate_Static_String_Literal;

      function Translate_Static_Bit_String_Literal
        (Lit : Iir_Bit_String_Literal)
        return O_Enode
      is
         use Name_Table;
         type Enode_Array is array (Natural range 0 .. 1) of O_Enode;
         L : Enode_Array;
         Lit_Type : Iir;
         List : O_Array_Aggr_List;
         Res : O_Enode;
         N : Natural;
         C : Character;
         B : Natural;
      begin
         Lit_Type := Get_Type (Lit);
         L (0) := Get_Ortho_Expr (Get_Bit_String_0 (Lit));
         L (1) := Get_Ortho_Expr (Get_Bit_String_1 (Lit));
         Chap3.Translate_Anonymous_Type_Definition (Lit_Type);
         Start_Array_Aggr (List, Get_Ortho_Type (Lit_Type, Mode_Value));
         B := Get_Bit_String_Base (Lit);
         Image (Get_Identifier (Lit));
         for I in 1 .. Name_Length loop
            C := Name_Buffer (I);
            if C /= '_' then
               if C in '0' .. '9' then
                  N := Character'Pos (C) - Character'Pos ('0');
               elsif C in 'A' .. 'F' then
                  N := Character'Pos (C) - Character'Pos ('A') + 10;
               elsif C in 'a' .. 'f' then
                  N := Character'Pos (C) - Character'Pos ('a') + 10;
               else
                  raise Internal_Error;
               end if;
               if B >= 4 then
                  New_Array_Aggr_El (List, L ((N / 8) mod 2));
               end if;
               if B >= 3 then
                  New_Array_Aggr_El (List, L ((N / 4) mod 2));
               end if;
               if B >= 2 then
                  New_Array_Aggr_El (List, L ((N / 2) mod 2));
               end if;
               New_Array_Aggr_El (List, L (N mod 2));
            end if;
         end loop;
         Finish_Array_Aggr (List, Res);
         return Res;
      end Translate_Static_Bit_String_Literal;

      function Translate_String_Literal (Str : Iir) return O_Enode
      is
         Var : Var_Acc;
         Info : Type_Info_Acc;
         Str_Type : Iir;
         Res : O_Enode;
      begin
         Str_Type := Get_Type (Str);
         Chap3.Translate_Anonymous_Type_Definition (Str_Type);
         Info := Get_Info (Str_Type);
         case Get_Kind (Str) is
            when Iir_Kind_String_Literal =>
               Res := Translate_Static_String_Literal (Str);
            when Iir_Kind_Bit_String_Literal =>
               Res := Translate_Static_Bit_String_Literal (Str);
            when others =>
               raise Internal_Error;
         end case;
         Var := Create_Global_Const
           (Create_Uniq_Identifier, Info.Ortho_Type (Mode_Value),
            O_Storage_Private, Res);
         return New_Address (Get_Var (Var), Info.Ortho_Ptr_Type (Mode_Value));
      end Translate_String_Literal;

      function Convert_Static_Constrained_To_Unconstrained
        (Expr : O_Enode; Expr_Type : Iir; Atype : Iir_Array_Type_Definition)
        return O_Enode
      is
         Res : O_Enode;
         Constr : O_Record_Aggr_List;
         Arr_Base : O_Lnode;
         Tinfo : Type_Info_Acc;
      begin
         raise Internal_Error;
         return O_Enode_Null;
         --New_Const_Decl (Arr_Base, Create_Uniq_Identifier, O_Storage_Private,
         --                Get_Ortho_Type (Expr_Type), Expr);
--          Tinfo := Get_Info (Atype);
--          Start_Record_Aggr (Constr, Tinfo.Ortho_Type (Mode_Value));
--          New_Record_Aggr_El
--            (Constr,
--             New_Address (Arr_Base, Tinfo.T.Base_Ptr_Type (Mode_Value)));
--          Chap3.Create_Array_Subtype (Expr_Type);
--          New_Record_Aggr_El
--            (Constr,
--           Chap3.Get_Array_Ptr_Bounds_Ptr (O_Lnode_Null, Atype, Mode_Value));
--          Finish_Record_Aggr (Constr, Res);
--          return Res;
      end Convert_Static_Constrained_To_Unconstrained;

      function Translate_Static_Expression (Expr : Iir; Rtype : Iir)
        return O_Enode
      is
         Expr_Type : Iir;
         Expr_Otype : O_Tnode;
         Tinfo : Type_Info_Acc;
         Res : O_Enode;
      begin
         Expr_Type := Get_Type (Expr);
         Tinfo := Get_Info (Expr_Type);
         if Rtype /= Null_Iir then
            Expr_Otype := Get_Ortho_Type (Rtype, Mode_Value);
         else
            if Tinfo = null then
               --  FIXME: this is a working kludge, in the case where EXPR_TYPE
               --  is a subtype which was not yet translated.
               --  (eg: evaluared array attribute)
               Tinfo := Get_Info (Get_Base_Type (Expr_Type));
            end if;
            Expr_Otype := Tinfo.Ortho_Type (Mode_Value);
         end if;
         case Get_Kind (Expr) is
            when Iir_Kind_Integer_Literal =>
               return New_Signed_Literal
                 (Expr_Otype, Integer_64 (Get_Value (Expr)));
            when Iir_Kind_Enumeration_Literal =>
               return Get_Ortho_Expr (Get_Enumeration_Decl (Expr));
            when Iir_Kind_Floating_Point_Literal =>
               return New_Float_Literal
                 (Expr_Otype, IEEE_Float_64 (Get_Fp_Value (Expr)));
            when Iir_Kind_Physical_Literal =>
               return New_Signed_Literal
                 (Expr_Otype,
                  Integer_64 (Get_Abstract_Literal (Expr))
                  * Integer_64 (Get_Value (Get_Name (Expr))));
            when Iir_Kind_String_Literal =>
               Res := Translate_Static_String_Literal (Expr);
               if Get_Kind (Rtype)
                 in Iir_Kinds_Unconstrained_Array_Type_Definition
               then
                  Res := Convert_Static_Constrained_To_Unconstrained
                    (Res, Expr_Type, Rtype);
               end if;
               return Res;
            when Iir_Kind_Bit_String_Literal =>
               Res := Translate_Static_Bit_String_Literal (Expr);
               if Get_Kind (Rtype)
                 in Iir_Kinds_Unconstrained_Array_Type_Definition
               then
                  Res := Convert_Static_Constrained_To_Unconstrained
                    (Res, Expr_Type, Rtype);
               end if;
               return Res;

            when others =>
               Error_Kind ("translate_static_expression", Expr);
         end case;
      end Translate_Static_Expression;

      function Translate_Static_Range_Left
        (Expr : Iir; Range_Type : Iir := Null_Iir)
        return O_Enode
      is
         Left : O_Enode;
         Bound : Iir;
      begin
         Bound := Get_Left_Limit (Expr);
         Left := Chap7.Translate_Static_Expression (Bound, Range_Type);
         if Range_Type /= Null_Iir and then Get_Type (Bound) /= Range_Type then
            Left := New_Convert_Ov
              (Left, Get_Ortho_Type (Range_Type, Mode_Value));
         end if;
         return Left;
      end Translate_Static_Range_Left;

      function Translate_Static_Range_Right
        (Expr : Iir; Range_Type : Iir := Null_Iir)
        return O_Enode
      is
         Right : O_Enode;
      begin
         Right := Chap7.Translate_Static_Expression (Get_Right_Limit (Expr),
                                                     Range_Type);
         if Range_Type /= Null_Iir then
            Right := New_Convert_Ov (Right,
                                     Get_Ortho_Type (Range_Type, Mode_Value));
         end if;
         return Right;
      end Translate_Static_Range_Right;

      function Translate_Static_Range_Dir (Expr : Iir) return O_Enode
      is
      begin
         case Get_Direction (Expr) is
            when Iir_To =>
               return Ghdl_Dir_To_Node;
            when Iir_Downto =>
               return Ghdl_Dir_Downto_Node;
         end case;
      end Translate_Static_Range_Dir;

      function Translate_Static_Range_Length (Expr : Iir) return O_Enode
      is
      begin
         return New_Unsigned_Literal
           (Ghdl_Index_Type,
            Unsigned_64 (Eval_Discrete_Range_Length (Expr)));
      end Translate_Static_Range_Length;

      function Translate_Range_Expression_Left (Expr : Iir;
                                                Range_Type : Iir := Null_Iir)
        return O_Enode
      is
         Left : O_Enode;
      begin
         Left := Chap7.Translate_Expression (Get_Left_Limit (Expr));
         if Range_Type /= Null_Iir then
            Left := New_Convert_Ov (Left,
                                    Get_Ortho_Type (Range_Type, Mode_Value));
         end if;
         return Left;
      end Translate_Range_Expression_Left;

      function Translate_Range_Expression_Right (Expr : Iir;
                                                 Range_Type : Iir := Null_Iir)
        return O_Enode
      is
         Right : O_Enode;
      begin
         Right := Chap7.Translate_Expression (Get_Right_Limit (Expr));
         if Range_Type /= Null_Iir then
            Right := New_Convert_Ov (Right,
                                     Get_Ortho_Type (Range_Type, Mode_Value));
         end if;
         return Right;
      end Translate_Range_Expression_Right;

      --  Compute the length of LEFT DIR (to/downto) RIGHT.
      function Compute_Range_Length
        (Left : O_Enode; Right : O_Enode; Dir : Iir_Direction)
        return O_Enode
      is
         L : O_Enode;
         R : O_Enode;
         Val : O_Enode;
         Tmp : O_Lnode;
         Res : O_Lnode;
         If_Blk : O_If_Block;
      begin
         L := New_Convert_Ov (Left, Ghdl_Integer_Type);
         R := New_Convert_Ov (Right, Ghdl_Integer_Type);

         case Dir is
            when Iir_To =>
               Val := New_Dyadic_Op (ON_Sub_Ov, R, L);
            when Iir_Downto =>
               Val := New_Dyadic_Op (ON_Sub_Ov, L, R);
         end case;

         Res := Create_Temp (Ghdl_Index_Type);
         Open_Temp;
         Tmp := Create_Temp (Ghdl_Integer_Type);
         New_Assign_Stmt (Tmp, Val);
         Start_If_Stmt
           (If_Blk,
            New_Compare_Op (ON_Lt, New_Value (Tmp),
                            New_Signed_Literal (Ghdl_Integer_Type, 0),
                            Ghdl_Bool_Type_Node));
         Init_Var (Res);
         New_Else_Stmt (If_Blk);
         Val := New_Convert_Ov (New_Value (Tmp), Ghdl_Index_Type);
         Val := New_Dyadic_Op (ON_Add_Ov, Val,
                               New_Unsigned_Literal (Ghdl_Index_Type, 1));
         New_Assign_Stmt (Res, Val);
         Finish_If_Stmt (If_Blk);
         Close_Temp;
         return New_Value (Res);
      end Compute_Range_Length;

      function Translate_Range_Expression_Length (Expr : Iir) return O_Enode
      is
         Left, Right : O_Enode;
      begin
         if Get_Expr_Staticness (Expr) = Locally then
            return Translate_Static_Range_Length (Expr);
         else
            Left := Chap7.Translate_Expression (Get_Left_Limit (Expr));
            Right := Chap7.Translate_Expression (Get_Right_Limit (Expr));

            return Compute_Range_Length (Left, Right, Get_Direction (Expr));
         end if;
      end Translate_Range_Expression_Length;

      function Translate_Range_Length (Expr : Iir) return O_Enode is
      begin
         case Get_Kind (Expr) is
            when Iir_Kind_Range_Expression =>
               return Translate_Range_Expression_Length (Expr);
            when Iir_Kind_Range_Array_Attribute =>
               return Chap14.Translate_Length_Array_Attribute (Expr, Null_Iir);
            when others =>
               Error_Kind ("translate_range_length", Expr);
         end case;
      end Translate_Range_Length;

      function Translate_Association (Assoc : Iir)
        return O_Enode
      is
         Actual, Formal : Iir;
         Res : O_Enode;
      begin
         Formal := Get_Associated_Formal (Assoc);
         case Get_Kind (Assoc) is
            when Iir_Kind_Association_Element_By_Expression =>
               Actual := Get_Actual (Assoc);
            when Iir_Kind_Association_Element_Open =>
               Actual := Get_Default_Value (Formal);
            when others =>
               Error_Kind ("translate_association", Assoc);
         end case;

         case Get_Kind (Formal) is
            when Iir_Kind_Constant_Interface_Declaration =>
               Res := Translate_Expression (Actual, Get_Type (Formal));
               return Res;
            when Iir_Kind_Signal_Interface_Declaration =>
               return M2E (Chap6.Translate_Name (Actual));
            when others =>
               Error_Kind ("translate_association", Formal);
         end case;
      end Translate_Association;

      function Translate_Function_Call (Imp : Iir; Assocs : Iir_List)
        return O_Enode
      is
         Constr : O_Assoc_List;
         Assoc : Iir;
         Info : Subprg_Info_Acc;
         Res : Mnode;
      begin
         Info := Get_Info (Imp);

         if Info.Use_Stack2 then
            Create_Temp_Stack2_Mark;
         end if;

         if Info.Res_Interface /= O_Lnode_Null then
            --  Composite result.
            --  If we need to allocate, do it before starting the call!
            declare
               Res_Type : Iir;
               Res_Info : Type_Info_Acc;
            begin
               Res_Type := Get_Return_Type (Imp);
               Res_Info := Get_Info (Res_Type);
               Res := Create_Temp (Res_Info);
               if Res_Info.Type_Mode /= Type_Mode_Fat_Array then
                  Chap4.Allocate_Complex_Object (Res_Type, Alloc_Stack, Res);
               end if;
            end;
         end if;

         Start_Association (Constr, Info.Ortho_Func);

         if Info.Res_Interface /= O_Lnode_Null then
            --  Composite result.
            New_Association (Constr, M2E (Res));
         end if;

         Chap2.Add_Subprg_Instance_Assoc (Constr, Imp);

         if Assocs /= Null_Iir then
            for I in Natural loop
               Assoc := Get_Nth_Element (Assocs, I);
               exit when Assoc = Null_Iir;
               --  FIXME: evaluate expression before, because we
               --  may allocate objects.
               New_Association (Constr, Translate_Association (Assoc));
            end loop;
         end if;

         if Info.Res_Interface /= O_Lnode_Null then
            --  Composite result.
            New_Procedure_Call (Constr);
            return M2E (Res);
         else
            return New_Function_Call (Constr);
         end if;
      end Translate_Function_Call;

      function Translate_Operator_Function_Call
        (Imp : Iir; Left : Iir;  Right : Iir; Res_Type : Iir)
        return O_Enode
      is
         function Add_Assoc (Actual : Iir; Formal : Iir; List : Iir_List)
           return Iir
         is
            R : Iir;
         begin
            R := Create_Iir (Iir_Kind_Association_Element_By_Expression);
            Location_Copy (R, Actual);
            Set_Actual (R, Actual);
            Set_Formal (R, Formal);
            Set_Associated_Formal (R, Formal);
            Append_Element (List, R);
            return R;
         end Add_Assoc;

         Assocs : Iir_List;
         Inters : Iir_List;
         El_L : Iir;
         El_R : Iir;
         Res : O_Enode;
      begin
         Assocs := Create_Iir (Iir_Kind_Association_List);
         Inters := Get_Interface_Declaration (Imp);

         El_L := Add_Assoc (Left, Get_Nth_Element (Inters, 0), Assocs);

         if Right /= Null_Iir then
            El_R := Add_Assoc (Right, Get_Nth_Element (Inters, 1), Assocs);
         end if;

         Res := Translate_Function_Call (Imp, Assocs);

         Free_Iir (El_L);
         if Right /= Null_Iir then
            Free_Iir (El_R);
         end if;
         Free_Iir (Assocs);

         return Translate_Implicit_Conv
           (Res, Get_Return_Type (Imp), Res_Type, Mode_Value);
      end Translate_Operator_Function_Call;

      function Convert_Constrained_To_Unconstrained
        (Expr : O_Enode; Expr_Type : Iir; Atype : Iir; Kind : Object_Kind_Type)
        return O_Enode
      is
         Res : O_Lnode;
         Type_Info : Type_Info_Acc;
         Expr_Type_Info : Type_Info_Acc;
      begin
         Expr_Type_Info := Get_Info (Expr_Type);
         Type_Info := Get_Info (Atype);
         Res := Create_Temp (Type_Info.Ortho_Type (Kind));
         New_Assign_Stmt
           (New_Selected_Element (Res, Type_Info.T.Base_Field (Kind)),
            New_Convert_Ov (Expr, Type_Info.T.Base_Ptr_Type (Kind)));

         -- FIXME: to do.
         --  Be sure the bounds variable was created.
         --  This may be necessary for on-the-fly types, such as strings.
         Chap3.Create_Array_Subtype (Expr_Type);
         New_Assign_Stmt
           (New_Selected_Element (Res, Type_Info.T.Bounds_Field (Kind)),
            Chap3.Get_Array_Ptr_Bounds_Ptr (O_Lnode_Null, Expr_Type, Kind));
         return L2e_Node (Res, Type_Info, Kind);
      end Convert_Constrained_To_Unconstrained;

      function Convert_Array_To_Thin_Array
        (Expr : O_Enode;
         Expr_Type : Iir;
         Atype : Iir;
         Is_Sig : Object_Kind_Type)
        return O_Enode
      is
         Ptr : O_Lnode;
         Expr_Type_Info : Type_Info_Acc;
         Expr_Indexes: Iir_List;
         Loop_Succ_Blk, Loop_Fail_Blk : O_Loop_Block;
         Success_Label, Failure_Label : O_Snode;
         If_Blk : O_If_Block;
      begin
         Expr_Type_Info := Get_Info (Expr_Type);
         Ptr := Create_Temp_Init
           (Expr_Type_Info.Ortho_Ptr_Type (Is_Sig), Expr);

         Open_Temp;
         --  Check each dimension.
         Expr_Indexes := Get_Index_List (Expr_Type);
         Start_Loop_Stmt (Loop_Succ_Blk, Success_Label);
         Start_Loop_Stmt (Loop_Fail_Blk, Failure_Label);
         for I in 1 .. Get_Nbr_Elements (Expr_Indexes) loop
            Gen_Exit_When
              (Failure_Label,
               New_Compare_Op
               (ON_Neq,
                Chap6.Get_Array_Ptr_Bound_Length (Ptr, Expr_Type, I, Is_Sig),
                Chap6.Get_Array_Bound_Length (O_Lnode_Null, Atype, I, Is_Sig),
                Ghdl_Bool_Type_Node));
         end loop;
         New_Exit_Stmt (Success_Label);
         Finish_Loop_Stmt (Loop_Fail_Blk);
         Chap6.Gen_Bound_Error (Expr_Type);   --  FIXME: location.
         Finish_Loop_Stmt (Loop_Succ_Blk);
         Close_Temp;

         return New_Value
           (Chap3.Get_Array_Ptr_Base_Ptr (Ptr, Expr_Type, Is_Sig));
      end Convert_Array_To_Thin_Array;

      --  Convert (if necessary) EXPR translated from EXPR_ORIG to type ATYPE.
      function Translate_Implicit_Conv (Expr : O_Enode;
                                        Expr_Type : Iir;
                                        Atype : Iir;
                                        Is_Sig : Object_Kind_Type)
        return O_Enode
      is
         Ainfo : Type_Info_Acc;
         Einfo : Type_Info_Acc;
      begin
         if Atype = Expr_Type then
            return Expr;
         end if;
         if Expr_Type = Universal_Integer_Type_Definition then
            return New_Convert_Ov (Expr, Get_Ortho_Type (Atype, Mode_Value));
         elsif Expr_Type = Universal_Real_Type_Definition then
            return New_Convert_Ov (Expr, Get_Ortho_Type (Atype, Mode_Value));
         elsif Get_Kind (Expr_Type) in Iir_Kinds_Array_Type_Definition then
            Ainfo := Get_Info (Atype);
            Einfo := Get_Info (Expr_Type);
            case Ainfo.Type_Mode is
               when Type_Mode_Fat_Array =>
                  --  X to unconstrained.
                  case Einfo.Type_Mode is
                     when Type_Mode_Fat_Array =>
                        --  unconstrained to unconstrained.
                        return Expr;
                     when Type_Mode_Array
                       | Type_Mode_Ptr_Array =>
                        --  constrained to unconstrained.
                        return Convert_Constrained_To_Unconstrained
                          (Expr, Expr_Type, Atype, Is_Sig);
                     when others =>
                        raise Internal_Error;
                  end case;
               when Type_Mode_Array =>
                  --  X to constrained.
                  case Einfo.Type_Mode is
                     when Type_Mode_Fat_Array
                       | Type_Mode_Ptr_Array =>
                        --  unconstrained to constrained.
                        return Convert_Array_To_Thin_Array (Expr, Expr_Type,
                                                            Atype, Is_Sig);
                     when Type_Mode_Array =>
                        --  constrained to constrained.
                        declare
                           E_List, A_List : Iir_List;
                           E_El, A_El : Iir;
                        begin
                           E_List := Get_Index_List (Expr_Type);
                           A_List := Get_Index_List (Atype);
                           for I in Natural loop
                              E_El := Get_Nth_Element (E_List, I);
                              A_El := Get_Nth_Element (A_List, I);
                              exit when E_El = Null_Iir
                                and then A_El = Null_Iir;
                              if Eval_Discrete_Type_Length (E_El)
                                /= Eval_Discrete_Type_Length (A_El)
                              then
                                 --  FIXME: generate a bound error ?
                                 --  Even if this is caught at compile-time,
                                 --  the code is not required to run.
                                 raise Internal_Error;
                              end if;
                           end loop;
                        end;
                        return Expr;
                     when others =>
                        raise Internal_Error;
                  end case;
               when Type_Mode_Ptr_Array =>
                  case Einfo.Type_Mode is
                     when Type_Mode_Fat_Array
                       | Type_Mode_Array
                       | Type_Mode_Ptr_Array =>
                        return Convert_Array_To_Thin_Array
                          (Expr, Expr_Type, Atype, Is_Sig);
                     when others =>
                        raise Internal_Error;
                  end case;
               when others =>
                  raise Internal_Error;
            end case;
         else
            return Expr;
         end if;
      end Translate_Implicit_Conv;

      type Predefined_To_Onop_Type is array (Iir_Predefined_Functions)
        of ON_Op_Kind;
      Predefined_To_Onop : constant Predefined_To_Onop_Type :=
        (Iir_Predefined_Boolean_Or => ON_Or,
         Iir_Predefined_Boolean_Not => ON_Not,
         Iir_Predefined_Boolean_And => ON_And,
         Iir_Predefined_Boolean_Xor => ON_Xor,

         Iir_Predefined_Bit_Not => ON_Not,
         Iir_Predefined_Bit_And => ON_And,
         Iir_Predefined_Bit_Or => ON_Or,
         Iir_Predefined_Bit_Xor => ON_Xor,

         Iir_Predefined_Integer_Equality => ON_Eq,
         Iir_Predefined_Integer_Inequality => ON_Neq,
         Iir_Predefined_Integer_Less_Equal => ON_Le,
         Iir_Predefined_Integer_Less => ON_Lt,
         Iir_Predefined_Integer_Greater => ON_Gt,
         Iir_Predefined_Integer_Greater_Equal => ON_Ge,
         Iir_Predefined_Integer_Plus => ON_Add_Ov,
         Iir_Predefined_Integer_Minus => ON_Sub_Ov,
         Iir_Predefined_Integer_Mul => ON_Mul_Ov,
         Iir_Predefined_Integer_Rem => ON_Rem_Ov,
         Iir_Predefined_Integer_Mod => ON_Mod_Ov,
         Iir_Predefined_Integer_Div => ON_Div_Ov,
         Iir_Predefined_Integer_Absolute => ON_Abs_Ov,
         Iir_Predefined_Integer_Negation => ON_Neg_Ov,

         Iir_Predefined_Enum_Equality => ON_Eq,
         Iir_Predefined_Enum_Inequality => ON_Neq,
         Iir_Predefined_Enum_Greater_Equal => ON_Ge,
         Iir_Predefined_Enum_Greater => ON_Gt,
         Iir_Predefined_Enum_Less => ON_Lt,
         Iir_Predefined_Enum_Less_Equal => ON_Le,

         Iir_Predefined_Physical_Equality => ON_Eq,
         Iir_Predefined_Physical_Inequality => ON_Neq,
         Iir_Predefined_Physical_Less => ON_Lt,
         Iir_Predefined_Physical_Less_Equal => ON_Le,
         Iir_Predefined_Physical_Greater => ON_Gt,
         Iir_Predefined_Physical_Greater_Equal => ON_Ge,
         Iir_Predefined_Physical_Negation => ON_Neg_Ov,
         Iir_Predefined_Physical_Absolute => ON_Abs_Ov,
         Iir_Predefined_Physical_Minus => ON_Sub_Ov,
         Iir_Predefined_Physical_Plus => ON_Add_Ov,

         Iir_Predefined_Floating_Greater => ON_Gt,
         Iir_Predefined_Floating_Greater_Equal => ON_Ge,
         Iir_Predefined_Floating_Less => ON_Lt,
         Iir_Predefined_Floating_Less_Equal => ON_Le,
         Iir_Predefined_Floating_Equality => ON_Eq,
         Iir_Predefined_Floating_Inequality => ON_Neq,
         Iir_Predefined_Floating_Minus => ON_Sub_Ov,
         Iir_Predefined_Floating_Plus => ON_Add_Ov,
         Iir_Predefined_Floating_Mul => ON_Mul_Ov,
         Iir_Predefined_Floating_Div => ON_Div_Ov,
         Iir_Predefined_Floating_Negation => ON_Neg_Ov,
         Iir_Predefined_Floating_Absolute => ON_Abs_Ov,

         others => ON_Nil);

      function Translate_Shortcut_Operator
        (Imp : Iir_Implicit_Function_Declaration; Left, Right : Iir)
        return O_Enode
      is
         Rtype : Iir;
         Res : O_Lnode;
         Res_Type : O_Tnode;
         If_Blk : O_If_Block;
         Op : ON_Op_Kind;
         Val : Integer;
         V : O_Enode;
         Kind : Iir_Predefined_Functions;
         Invert : Boolean;
      begin
         Rtype := Get_Return_Type (Imp);
         Res_Type := Get_Ortho_Type (Rtype, Mode_Value);
         Res := Create_Temp_Init
           (Res_Type, Chap7.Translate_Expression (Left));
         Open_Temp;
         Kind := Get_Implicit_Definition (Imp);

         --  Short cut: RIGHT is the result (and must be evaluated) iff
         --  LEFT is equal to VAL (ie '0' or false for 0, '1' or true for 1).
         case Kind is
            when Iir_Predefined_Bit_And
              | Iir_Predefined_Boolean_And =>
               Op := ON_And;
               Invert := False;
               Val := 1;
            when Iir_Predefined_Bit_Nand
              | Iir_Predefined_Boolean_Nand =>
               Op := ON_And;
               Invert := True;
               Val := 1;
            when Iir_Predefined_Bit_Or
              | Iir_Predefined_Boolean_Or =>
               Op := ON_Or;
               Invert := False;
               Val := 0;
            when Iir_Predefined_Bit_Nor
              | Iir_Predefined_Boolean_Nor =>
               Op := ON_Or;
               Invert := True;
               Val := 0;
            when others =>
               Ada.Text_IO.Put_Line
                 ("translate_shortcut_operator: cannot handle "
                  & Iir_Predefined_Functions'Image (Kind));
               raise Internal_Error;
         end case;

         V := Get_Ortho_Expr
           (Get_Nth_Element (Get_Enumeration_Literal_List (Rtype), Val));
         Start_If_Stmt (If_Blk,
                        New_Compare_Op (ON_Eq, New_Value (Res), V,
                                        Ghdl_Bool_Type_Node));
         Open_Temp;
         New_Assign_Stmt (Res, Chap7.Translate_Expression (Right));
         Close_Temp;
         Finish_If_Stmt (If_Blk);
         Close_Temp;
         if Invert then
            return New_Monadic_Op (ON_Not, New_Value (Res));
         else
            return New_Value (Res);
         end if;
      end Translate_Shortcut_Operator;

      function Translate_Lib_Operator (Left, Right : O_Enode; Func : O_Dnode)
        return O_Enode
      is
         Constr : O_Assoc_List;
      begin
         Start_Association (Constr, Func);
         New_Association (Constr, Left);
         if Right /= O_Enode_Null then
            New_Association (Constr, Right);
         end if;
         return New_Function_Call (Constr);
      end Translate_Lib_Operator;

      function Translate_Predefined_Lib_Operator
        (Left, Right : O_Enode; Func : Iir_Implicit_Function_Declaration)
        return O_Enode
      is
         Info : Subprg_Info_Acc;
         Constr : O_Assoc_List;
      begin
         Info := Get_Info (Func);
         Start_Association (Constr, Info.Ortho_Func);
         Chap2.Add_Subprg_Instance_Assoc (Constr, Func);
         New_Association (Constr, Left);
         if Right /= O_Enode_Null then
            New_Association (Constr, Right);
         end if;
         return New_Function_Call (Constr);
      end Translate_Predefined_Lib_Operator;

      function Translate_Predefined_Array_Operator
        (Left, Right : O_Enode; Func : Iir)
        return O_Enode
      is
         Res : O_Lnode;
         Constr : O_Assoc_List;
         Info : Type_Info_Acc;
      begin
         Create_Temp_Stack2_Mark;
         Info := Get_Info (Get_Return_Type (Func));
         Res := Create_Temp (Info.Ortho_Type (Mode_Value));
         Start_Association (Constr, Get_Info (Func).Ortho_Func);
         Chap2.Add_Subprg_Instance_Assoc (Constr, Func);
         New_Association (Constr,
                          New_Address (Res, Info.Ortho_Ptr_Type (Mode_Value)));
         New_Association (Constr, Left);
         if Right /= O_Enode_Null then
            New_Association (Constr, Right);
         end if;
         New_Procedure_Call (Constr);
         return New_Address (Res, Info.Ortho_Ptr_Type (Mode_Value));
      end Translate_Predefined_Array_Operator;

      --  Create an array aggregate containing one element, EL.
      function Translate_Element_To_Array (El : O_Enode; Arr_Type : Iir)
        return O_Enode
      is
         Res : O_Lnode;
         Ainfo : Type_Info_Acc;
         Einfo : Type_Info_Acc;
         V : O_Lnode;
      begin
         Ainfo := Get_Info (Arr_Type);
         Einfo := Get_Info (Get_Element_Subtype (Arr_Type));
         Res := Create_Temp (Ainfo.Ortho_Type (Mode_Value));
         if Is_Composite (Einfo) then
            New_Assign_Stmt
              (New_Selected_Element (Res, Ainfo.T.Base_Field (Mode_Value)),
               New_Convert_Ov (El, Ainfo.T.Base_Ptr_Type (Mode_Value)));
         else
            V := Create_Temp_Init (Einfo.Ortho_Type (Mode_Value), El);
            New_Assign_Stmt
              (New_Selected_Element (Res, Ainfo.T.Base_Field (Mode_Value)),
               New_Convert_Ov (New_Address (V,
                                            Einfo.Ortho_Ptr_Type (Mode_Value)),
                               Ainfo.T.Base_Ptr_Type (Mode_Value)));
         end if;
         New_Assign_Stmt
           (New_Selected_Element (Res, Ainfo.T.Bounds_Field (Mode_Value)),
            New_Address (Get_Var (Ainfo.T.Array_1bound),
                         Ainfo.T.Bounds_Ptr_Type));
         return New_Address (Res, Ainfo.Ortho_Ptr_Type (Mode_Value));
      end Translate_Element_To_Array;

      function Translate_Concat_Operator
        (Left_Tree, Right_Tree : O_Enode;
         Imp : Iir_Implicit_Function_Declaration;
         Res_Type : Iir)
        return O_Enode
      is
         Arr_El1 : O_Enode;
         Arr_El2 : O_Enode;
         Ret_Type : Iir;
         Res : O_Enode;
         Kind : Iir_Predefined_Functions;
      begin
         Ret_Type := Get_Return_Type (Imp);
         Kind := Get_Implicit_Definition (Imp);
         case Kind is
            when Iir_Predefined_Element_Array_Concat
              | Iir_Predefined_Element_Element_Concat =>
               Arr_El1 := Translate_Element_To_Array (Left_Tree, Ret_Type);
            when others =>
               Arr_El1 := Left_Tree;
         end case;
         case Kind is
            when Iir_Predefined_Array_Element_Concat
              | Iir_Predefined_Element_Element_Concat =>
               Arr_El2 := Translate_Element_To_Array (Right_Tree, Ret_Type);
            when others =>
               Arr_El2 := Right_Tree;
         end case;
         Res := Translate_Predefined_Array_Operator (Arr_El1, Arr_El2, Imp);
         return Translate_Implicit_Conv (Res, Ret_Type, Res_Type, Mode_Value);
      end Translate_Concat_Operator;

      function Translate_Predefined_Operator
        (Imp : Iir_Implicit_Function_Declaration;
         Left, Right : Iir;
         Res_Type : Iir)
        return O_Enode
      is
         Left_Tree : O_Enode;
         Right_Tree : O_Enode;
         Kind : Iir_Predefined_Functions;
         Left_Type : Iir;
         Right_Type : Iir;
         Res_Otype : O_Tnode;
         Op : ON_Op_Kind;
         Interface_List : Iir_List;
         Res : O_Enode;
      begin
         Kind := Get_Implicit_Definition (Imp);
         if Iir_Predefined_Shortcut_P (Kind) then
            return Translate_Shortcut_Operator (Imp, Left, Right);
         end if;

         Res_Otype := Get_Ortho_Type (Res_Type, Mode_Value);
         Interface_List := Get_Interface_Declaration (Imp);
         if Left = Null_Iir then
            Left_Tree := O_Enode_Null;
         else
            Left_Type := Get_Type (Get_Nth_Element (Interface_List, 0));
            Left_Tree := Translate_Expression (Left, Left_Type);
         end if;

         if Right = Null_Iir then
            Right_Tree := O_Enode_Null;
         else
            Right_Type := Get_Type (Get_Nth_Element (Interface_List, 1));
            Right_Tree := Translate_Expression (Right, Right_Type);
         end if;

         Op := Predefined_To_Onop (Kind);
         if Op /= ON_Nil then
            case Op is
               when ON_Eq
                 | ON_Neq
                 | ON_Ge
                 | ON_Gt
                 | ON_Le
                 | ON_Lt =>
                  Res := New_Compare_Op (Op, Left_Tree, Right_Tree,
                                         Std_Boolean_Type_Node);
               when ON_Add_Ov
                 | ON_Sub_Ov
                 | ON_Mul_Ov
                 | ON_Div_Ov
                 | ON_Rem_Ov
                 | ON_Mod_Ov
                 | ON_Xor =>
                  Res := New_Dyadic_Op (Op, Left_Tree, Right_Tree);
               when ON_Abs_Ov
                 | ON_Neg_Ov
                 | ON_Not =>
                  Res := New_Monadic_Op (Op, Left_Tree);
               when others =>
                  Ada.Text_IO.Put_Line
                    ("translate_predefined_operator: cannot handle "
                     & ON_Op_Kind'Image (Op));
                  raise Internal_Error;
            end case;
            Res := Translate_Implicit_Conv
              (Res, Get_Return_Type (Imp), Res_Type, Mode_Value);
            return Res;
         end if;

         case Kind is
            when Iir_Predefined_Integer_Identity
              | Iir_Predefined_Floating_Identity
              | Iir_Predefined_Physical_Identity =>
               return Translate_Implicit_Conv
                 (Left_Tree, Left_Type, Res_Type, Mode_Value);

            when Iir_Predefined_Access_Equality
              | Iir_Predefined_Access_Inequality =>
               if Is_Composite (Get_Info (Left_Type)) then
                  --  a fat pointer.
                  declare
                     T : Type_Info_Acc;
                     B : Type_Info_Acc;
                     L, R : O_Lnode;
                     V1, V2 : O_Enode;
                     Op1, Op2 : ON_Op_Kind;
                  begin
                     if Kind = Iir_Predefined_Access_Equality then
                        Op1 := ON_Eq;
                        Op2 := ON_And;
                     else
                        Op1 := ON_Neq;
                        Op2 := ON_Or;
                     end if;
                     T := Get_Info (Left_Type);
                     B := Get_Info (Get_Designated_Type (Left_Type));
                     L := Create_Temp (T.Ortho_Ptr_Type (Mode_Value));
                     R := Create_Temp (T.Ortho_Ptr_Type (Mode_Value));
                     New_Assign_Stmt (L, Left_Tree);
                     New_Assign_Stmt (R, Right_Tree);
                     V1 := New_Compare_Op
                       (Op1,
                        New_Value_Selected_Acc_Value
                        (L, B.T.Base_Field (Mode_Value)),
                        New_Value_Selected_Acc_Value
                        (R, B.T.Base_Field (Mode_Value)),
                        Std_Boolean_Type_Node);
                     V2 := New_Compare_Op
                       (Op1,
                        New_Value_Selected_Acc_Value
                        (L, B.T.Bounds_Field (Mode_Value)),
                        New_Value_Selected_Acc_Value
                        (R, B.T.Bounds_Field (Mode_Value)),
                        Std_Boolean_Type_Node);
                     return New_Dyadic_Op (Op2, V1, V2);
                  end;
               else
                  --  a thin pointer.
                  if Kind = Iir_Predefined_Access_Equality then
                     return New_Compare_Op
                       (ON_Eq, Left_Tree, Right_Tree, Std_Boolean_Type_Node);
                  else
                     return New_Compare_Op
                       (ON_Neq, Left_Tree, Right_Tree, Std_Boolean_Type_Node);
                  end if;
               end if;

            when Iir_Predefined_Physical_Integer_Div =>
               return New_Dyadic_Op (ON_Div_Ov, Left_Tree,
                                     New_Convert_Ov (Right_Tree, Res_Otype));
            when Iir_Predefined_Physical_Physical_Div =>
               return New_Convert_Ov
                 (New_Dyadic_Op (ON_Div_Ov, Left_Tree, Right_Tree), Res_Otype);

            --  LRM 7.2.6
            --  Multiplication of a value P of a physical type Tp by a
            --  value I of type INTEGER is equivalent to the following
            --  computation: Tp'Val (Tp'Pos (P) * I)
            --  FIXME: this is not what is really done...
            when Iir_Predefined_Integer_Physical_Mul =>
               return New_Dyadic_Op (ON_Mul_Ov,
                                     New_Convert_Ov (Left_Tree, Res_Otype),
                                     Right_Tree);
            when Iir_Predefined_Physical_Integer_Mul =>
               return New_Dyadic_Op (ON_Mul_Ov, Left_Tree,
                                     New_Convert_Ov (Right_Tree, Res_Otype));

            --  LRM 7.2.6
            --  Multiplication of a value P of a physical type Tp by a
            --  value F of type REAL is equivalten to the following
            --  computation: Tp'Val (INTEGER (REAL (Tp'Pos (P)) * F))
            --  FIXME: we do not restrict with INTEGER.
            when Iir_Predefined_Physical_Real_Mul =>
               declare
                  Right_Otype : O_Tnode;
               begin
                  Right_Otype := Get_Ortho_Type (Right_Type, Mode_Value);
                  return New_Convert_Ov
                    (New_Dyadic_Op (ON_Mul_Ov,
                                    New_Convert_Ov (Left_Tree, Right_Otype),
                                    Right_Tree),
                     Res_Otype);
               end;
            when Iir_Predefined_Physical_Real_Div =>
               declare
                  Right_Otype : O_Tnode;
               begin
                  Right_Otype := Get_Ortho_Type (Right_Type, Mode_Value);
                  return New_Convert_Ov
                    (New_Dyadic_Op (ON_Div_Ov,
                                    New_Convert_Ov (Left_Tree, Right_Otype),
                                    Right_Tree),
                     Res_Otype);
               end;
            when Iir_Predefined_Real_Physical_Mul =>
               declare
                  Left_Otype : O_Tnode;
               begin
                  Left_Otype := Get_Ortho_Type (Left_Type, Mode_Value);
                  return New_Convert_Ov
                    (New_Dyadic_Op (ON_Mul_Ov,
                                    Left_Tree,
                                    New_Convert_Ov (Right_Tree, Left_Otype)),
                     Res_Otype);
               end;

            when Iir_Predefined_Universal_R_I_Mul =>
               return New_Dyadic_Op (ON_Mul_Ov,
                                     Left_Tree,
                                     New_Convert_Ov (Right_Tree, Res_Otype));

            when Iir_Predefined_Floating_Exp =>
               return Translate_Lib_Operator
                 (Left_Tree, Right_Tree, Ghdl_Real_Exp);
            when Iir_Predefined_Integer_Exp =>
               return Translate_Lib_Operator
                 (Left_Tree, Right_Tree, Ghdl_Integer_Exp);

            when Iir_Predefined_Array_Inequality
              | Iir_Predefined_Record_Inequality =>
               return New_Monadic_Op
                 (ON_Not, Translate_Predefined_Lib_Operator
                  (Left_Tree, Right_Tree, Imp));
            when Iir_Predefined_Array_Equality
              | Iir_Predefined_Record_Equality =>
               return Translate_Predefined_Lib_Operator
                 (Left_Tree, Right_Tree, Imp);

            when Iir_Predefined_Array_Greater =>
               return New_Compare_Op
                 (ON_Eq,
                  Translate_Predefined_Lib_Operator (Left_Tree, Right_Tree,
                                                     Imp),
                  Ghdl_Compare_Gt,
                  Std_Boolean_Type_Node);
            when Iir_Predefined_Array_Greater_Equal =>
               return New_Compare_Op
                 (ON_Ge,
                  Translate_Predefined_Lib_Operator (Left_Tree, Right_Tree,
                                                     Imp),
                  Ghdl_Compare_Eq,
                  Std_Boolean_Type_Node);
            when Iir_Predefined_Array_Less =>
               return New_Compare_Op
                 (ON_Eq,
                  Translate_Predefined_Lib_Operator (Left_Tree, Right_Tree,
                                                     Imp),
                  Ghdl_Compare_Lt,
                  Std_Boolean_Type_Node);
            when Iir_Predefined_Array_Less_Equal =>
               return New_Compare_Op
                 (ON_Le,
                  Translate_Predefined_Lib_Operator (Left_Tree, Right_Tree,
                                                     Imp),
                  Ghdl_Compare_Eq,
                  Std_Boolean_Type_Node);

            when Iir_Predefined_Bit_Array_And
              | Iir_Predefined_Bit_Array_Or
              | Iir_Predefined_Bit_Array_Nand
              | Iir_Predefined_Bit_Array_Nor
              | Iir_Predefined_Bit_Array_Xor
              | Iir_Predefined_Bit_Array_Xnor
              | Iir_Predefined_Bit_Array_Not
              | Iir_Predefined_Boolean_Array_And
              | Iir_Predefined_Boolean_Array_Or
              | Iir_Predefined_Boolean_Array_Nand
              | Iir_Predefined_Boolean_Array_Nor
              | Iir_Predefined_Boolean_Array_Xor
              | Iir_Predefined_Boolean_Array_Xnor
              | Iir_Predefined_Boolean_Array_Not =>
               declare
                  Res : O_Enode;
                  Ret_Type : Iir;
               begin
                  Ret_Type := Get_Return_Type (Imp);
                  Res := Translate_Predefined_Array_Operator
                    (Left_Tree, Right_Tree, Imp);
                  return Translate_Implicit_Conv
                    (Res, Ret_Type, Res_Type, Mode_Value);
               end;

            when Iir_Predefined_Array_Array_Concat
              | Iir_Predefined_Element_Array_Concat
              | Iir_Predefined_Array_Element_Concat
              | Iir_Predefined_Element_Element_Concat =>
               return Translate_Concat_Operator
                 (Left_Tree, Right_Tree, Imp, Res_Type);

            when Iir_Predefined_Endfile =>
               return Translate_Lib_Operator
                 (Left_Tree, O_Enode_Null, Ghdl_File_Endfile);

            when Iir_Predefined_Now_Function =>
               return New_Value (Ghdl_Now);

            when others =>
               Ada.Text_IO.Put_Line
                 ("translate_predefined_operator(2): cannot handle "
                  & Iir_Predefined_Functions'Image (Kind));
               raise Internal_Error;
               return O_Enode_Null;
         end case;
      end Translate_Predefined_Operator;

      --  Assign EXPR to TARGET.
      procedure Translate_Assign
        (Target : Mnode; Expr : Iir; Target_Type : Iir)
      is
         T_Info : Type_Info_Acc;
         Val : O_Enode;
      begin
         if Get_Kind (Expr) = Iir_Kind_Aggregate then
            Translate_Aggregate (Target, Target_Type, Expr);
            return;
         end if;

         Open_Temp;
         Val := Chap7.Translate_Expression (Expr, Target_Type);

         T_Info := Get_Info (Target_Type);
         case T_Info.Type_Mode is
            when Type_Mode_Scalar =>
               if not Chap3.Need_Range_Check (Target_Type) then
                  New_Assign_Stmt (M2Lv (Target), Val);
               else
                  declare
                     V : O_Lnode;
                  begin
                     Open_Temp;
                     V := Create_Temp_Init (T_Info.Ortho_Type (Mode_Value),
                                            Val);
                     Chap3.Check_Range (V, Expr, Target_Type);
                     New_Assign_Stmt (M2Lv (Target), New_Value (V));
                     Close_Temp;
                  end;
               end if;
            when Type_Mode_Acc =>
               New_Assign_Stmt (M2Lv (Target), Val);
            when Type_Mode_Fat_Acc =>
               declare
                  T, E : O_Lnode;
               begin
                  T := Create_Temp_Ptr
                    (Target_Type, M2Lv (Target), Mode_Value);
                  E := Create_Temp_Init
                    (T_Info.Ortho_Ptr_Type (Mode_Value), Val);
                  Copy_Fat_Access (T, E, Target_Type);
               end;
            when Type_Mode_Fat_Array =>
               declare
                  T : Mnode;
                  E : O_Lnode;
               begin
                  T := Stabilize (Target);
                  E := Create_Temp_Init
                    (T_Info.Ortho_Ptr_Type (Mode_Value), Val);
                  Chap3.Check_Array_Match (Target_Type, M2Lp (T), Mode_Value,
                                           Get_Type (Expr), E, Mode_Value,
                                           Null_Iir);
                  Chap3.Translate_Object_Copy
                    (T, New_Value (E), Target_Type, Mode_Value);
               end;
            when Type_Mode_Array
              | Type_Mode_Ptr_Array =>
               --  Source is of type TARGET_TYPE, so no length check is
               --  necessary.
               Chap3.Translate_Object_Copy
                 (Target, Val, Target_Type, Mode_Value);
            when Type_Mode_Record =>
               Chap3.Translate_Object_Copy
                 (Target, Val, Target_Type, Mode_Value);
            when Type_Mode_Unknown
              | Type_Mode_File =>
               raise Internal_Error;
         end case;

         Close_Temp;
      end Translate_Assign;

      --  If AGGR is of the form (others => (others => EXPR)) (where the
      --   number of (others => ) sub-aggregate is at least 1, return EXPR
      --   otherwise return NULL_IIR.
      function Is_Aggregate_Others (Aggr : Iir_Aggregate) return Iir
      is
         List : Iir_List;
         Aggr1 : Iir;
      begin
         Aggr1 := Aggr;
         loop
            List := Get_Association_Choices_List (Aggr1);
            if Get_Nbr_Elements (List) /= 1 then
               return Null_Iir;
            end if;
            Aggr1 := Get_First_Element (List);
            if Get_Kind (Aggr1) /= Iir_Kind_Association_Choice_By_Others then
               return Null_Iir;
            end if;
            Aggr1 := Get_Associated (Aggr1);
            case Get_Kind (Aggr1) is
               when Iir_Kind_Aggregate =>
                  if Get_Type (Aggr1) /= Null_Iir then
                     --  Stop when a sub-aggregate is in fact an aggregate.
                     return Aggr1;
                  end if;
               when Iir_Kind_String_Literal
                 | Iir_Kind_Bit_String_Literal =>
                  return Null_Iir;
                  --Error_Kind ("is_aggregate_others", Aggr1);
               when others =>
                  return Aggr1;
            end case;
         end loop;
      end Is_Aggregate_Others;

      --  Generate code for (others => EL).
      procedure Translate_Aggregate_Others
        (Target : Mnode; Target_Type : Iir; Aggr : Iir; El : Iir)
      is
         Base_Ptr : Mnode;
         Info : Type_Info_Acc;
         It : O_Lnode;
         Len : O_Lnode;
         Len_Val : O_Enode;
         If_Blk : O_If_Block;
         Loop_Blk : O_Loop_Block;
         Label : O_Snode;
         Arr_Var : Mnode;
         El_Node : Mnode;
      begin
         Open_Temp;

         Info := Get_Info (Target_Type);
         case Info.Type_Mode is
            when Type_Mode_Fat_Array =>
               Arr_Var := Stabilize (Target);
               Base_Ptr := Stabilize (Chap3.Get_Array_Base (Arr_Var));
               Len_Val := Chap3.Get_Array_Length (Arr_Var, Target_Type);
            when Type_Mode_Ptr_Array
              | Type_Mode_Array =>
               Base_Ptr := Stabilize (Chap3.Get_Array_Base (Target));
               Len_Val := Chap3.Get_Array_Type_Length (Target_Type);
            when others =>
               raise Internal_Error;
         end case;
         --  FIXME: use this (since this use one variable instead of two):
         --  I := length;
         --  loop
         --    exit when I = 0;
         --    I := I - 1;
         --    A[I] := xxx;
         --  end loop;
         Len := Create_Temp_Init (Ghdl_Index_Type, Len_Val);
         if True then
            It := Create_Temp (Ghdl_Index_Type);
         else
            New_Var_Decl (It, Wki_I, O_Storage_Local, Ghdl_Index_Type);
         end if;
         Init_Var (It);
         Start_Loop_Stmt (Loop_Blk, Label);
         Gen_Exit_When (Label,
                        New_Compare_Op (ON_Eq, New_Value (It), New_Value (Len),
                                        Ghdl_Bool_Type_Node));
         El_Node := Chap3.Index_Base (Base_Ptr, Target_Type, New_Value (It));
         --New_Assign_Stmt (El_Node, Chap7.Translate_Expression (El));
         Translate_Assign (El_Node, El, Get_Element_Subtype (Target_Type));
         Inc_Var (It);
         Finish_Loop_Stmt (Loop_Blk);

         Close_Temp;
      end Translate_Aggregate_Others;

      procedure Translate_Array_Aggregate_Gen
        (Base_Ptr : Mnode;
         Bounds_Ptr : Mnode;
         Aggr : Iir;
         Aggr_Type : Iir;
         Dim : Natural;
         Var_Index : O_Lnode)
      is
         Index_List : Iir_List;
         List : Iir_List;
         Expr_Type : Iir;
         Final : Boolean;

         procedure Do_Assign (Expr : Iir)
         is
         begin
            if Final then
               Translate_Assign (Chap3.Index_Base (Base_Ptr, Aggr_Type,
                                                   New_Value (Var_Index)),
                                 Expr, Expr_Type);
               Inc_Var (Var_Index);
            else
               Translate_Array_Aggregate_Gen
                 (Base_Ptr, Bounds_Ptr, Expr, Aggr_Type, Dim + 1, Var_Index);
            end if;
         end Do_Assign;

         P : Natural;
         El : Iir;
      begin
         Index_List := Get_Index_List (Aggr_Type);
         List := Get_Association_Choices_List (Aggr);

         --  FINAL is true if the elements of the aggregate are elements of
         --  the array.
         if Get_Nbr_Elements (Index_List) = Dim then
            Expr_Type := Get_Element_Subtype (Aggr_Type);
            Final:= True;
         else
            Final := False;
         end if;

         --  First, assign positionnal association.
         --  FIXME: count the number of positionnal association and generate
         --   an error if there is more positionnal association than elements
         --   in the array.
         P := 0;
         loop
            El := Get_Nth_Element (List, P);
            if El = Null_Iir then
               --  There is only positionnal associations.
               return;
            end if;
            exit when Get_Kind (El) /= Iir_Kind_Association_Choice_By_None;
            Do_Assign (Get_Associated (El));
            P := P + 1;
         end loop;

         --  Then, assign named or others association.
         if P + 1 = Get_Nbr_Elements (List) then
            --  There is only one choice
            case Get_Kind (El) is
               when Iir_Kind_Association_Choice_By_List
                 | Iir_Kind_Association_Choice_By_Others =>
                  --  falltrough...
                  null;
               when Iir_Kind_Association_Choice_By_Expression =>
                  Do_Assign (Get_Associated (El));
                  return;
               when Iir_Kind_Association_Choice_By_Range =>
                  declare
                     Var_Length : O_Lnode;
                     Var_I : O_Lnode;
                     Loop_Blk : O_Loop_Block;
                     Label : O_Snode;
                  begin
                     Open_Temp;
                     Var_Length := Create_Temp_Init
                       (Ghdl_Index_Type,
                        Chap7.Translate_Range_Length (Get_Expression (El)));
                     Var_I := Create_Temp (Ghdl_Index_Type);
                     Init_Var (Var_I);
                     Start_Loop_Stmt (Loop_Blk, Label);
                     Gen_Exit_When (Label,
                                    New_Compare_Op (ON_Eq,
                                                    New_Value (Var_I),
                                                    New_Value (Var_Length),
                                                    Ghdl_Bool_Type_Node));
                     Do_Assign (Get_Associated (El));
                     Inc_Var (Var_I);
                     Finish_Loop_Stmt (Loop_Blk);
                     Close_Temp;
                  end;
                  return;
               when others =>
                  Error_Kind ("translate_array_aggregate_gen", El);
            end case;
         end if;

         --  Several choices..
         declare
            Range_Type : Iir;
            Var_Pos : O_Lnode;
            Var_Len : O_Lnode;
            Range_Ptr : Mnode;
            Rtinfo : Type_Info_Acc;
            If_Blk1, If_Blk : O_If_Block;
            Case_Blk : O_Case_Block;
            Choice : Iir;
            Loop_Blk : O_Loop_Block;
            Label : O_Snode;
         begin
            Open_Temp;
            --  Create a loop from left +- number of positionnals associations
            --   to/downto right.
            Range_Type :=
              Get_Base_Type (Get_Nth_Element (Index_List, Dim - 1));
            Rtinfo := Get_Info (Range_Type);
            Var_Pos := Create_Temp (Rtinfo.Ortho_Type (Mode_Value));
            Range_Ptr := Stabilize
              (Chap3.Bounds_To_Range (Bounds_Ptr, Aggr_Type, Dim));
            New_Assign_Stmt (Var_Pos, M2E (Chap3.Range_To_Left (Range_Ptr)));
            Var_Len := Create_Temp (Ghdl_Index_Type);
            New_Assign_Stmt (Var_Len, M2E (Chap3.Range_To_Length (Range_Ptr)));
            if P /= 0 then
               Start_If_Stmt
                 (If_Blk,
                  New_Compare_Op (ON_Eq,
                                  M2E (Chap3.Range_To_Dir (Range_Ptr)),
                                  Ghdl_Dir_To_Node,
                                  Ghdl_Bool_Type_Node));
               Chap8.Gen_Update_Iterator (Var_Pos, Iir_To, Unsigned_64 (P),
                                          Range_Type);
               New_Else_Stmt (If_Blk);
               Chap8.Gen_Update_Iterator (Var_Pos, Iir_Downto, Unsigned_64 (P),
                                          Range_Type);
               Finish_If_Stmt (If_Blk);
            end if;

            --  Start loop.
            Start_Loop_Stmt (Loop_Blk, Label);
            --  Check if end of loop.
            Gen_Exit_When
              (Label,
               New_Compare_Op (ON_Eq,
                               New_Value (Var_Len),
                               New_Unsigned_Literal (Ghdl_Index_Type, 0),
                               Ghdl_Bool_Type_Node));

            --  convert aggr into a case statement.
            Start_Case_Stmt (Case_Blk, New_Value (Var_Pos));
            loop
               Choice := Get_Nth_Element (List, P);
               exit when Choice = Null_Iir;
               Start_Branch (Case_Blk);
               Chap8.Translate_Case_Choice (Choice, Range_Type, Case_Blk);
               Do_Assign (Get_Associated (Choice));
               Finish_Branch (Case_Blk, False);
               P := P + 1;
            end loop;
            Finish_Case_Stmt (Case_Blk);
            --  Update var_pos
            Start_If_Stmt
              (If_Blk,
               New_Compare_Op (ON_Eq,
                               M2E (Chap3.Range_To_Dir (Range_Ptr)),
                               Ghdl_Dir_To_Node,
                               Ghdl_Bool_Type_Node));
            Chap8.Gen_Update_Iterator (Var_Pos, Iir_To, Unsigned_64 (1),
                                       Range_Type);
            New_Else_Stmt (If_Blk);
            Chap8.Gen_Update_Iterator (Var_Pos, Iir_Downto, Unsigned_64 (1),
                                       Range_Type);
            Finish_If_Stmt (If_Blk);
            New_Assign_Stmt
              (Var_Len,
               New_Dyadic_Op (ON_Sub_Ov,
                              New_Value (Var_Len),
                              New_Unsigned_Literal (Ghdl_Index_Type, 1)));
            Finish_Loop_Stmt (Loop_Blk);
            Close_Temp;
         end;
      end Translate_Array_Aggregate_Gen;

      procedure Translate_Record_Aggregate (Target : Mnode; Aggr : Iir)
      is
         Targ : Mnode;
         Aggr_Type : Iir := Get_Type (Aggr);
         El_List : Iir_List := Get_Element_Declaration_List
           (Get_Base_Type (Aggr_Type));
         Nbr_El : Natural := Get_Nbr_Elements (El_List);

         --  Record which elements of the record have been set.  The 'others'
         --  clause applies to all elements not already set.
         type Bool_Array_Type is array (0 .. Nbr_El - 1) of Boolean;
         pragma Pack (Bool_Array_Type);
         Set_Array : Bool_Array_Type := (others => False);

         --  The expression associated.
         El_Expr : Iir;

         --  Set an elements.
         procedure Set_El (El : Iir_Element_Declaration) is
         begin
            Translate_Assign (Chap6.Translate_Selected_Element (Targ, El),
                              El_Expr, Get_Type (El));
            Set_Array (Natural (Get_Element_Position (El))) := True;
         end Set_El;

         Assoc_List : Iir_List;
         Assoc : Iir;
         El : Iir;
      begin
         Open_Temp;
         Targ := Stabilize (Target);
         Assoc_List := Get_Association_Choices_List (Aggr);
         for I in Natural loop
            Assoc := Get_Nth_Element (Assoc_List, I);
            exit when Assoc = Null_Iir;
            El_Expr := Get_Associated (Assoc);
            case Get_Kind (Assoc) is
               when Iir_Kind_Association_Choice_By_None =>
                  El := Get_Nth_Element (El_List, I);
                  Set_El (El);
               when Iir_Kind_Association_Choice_By_Name =>
                  El := Get_Name (Assoc);
                  Set_El (El);
               when Iir_Kind_Association_Choice_By_List =>
                  declare
                     Sub_List : Iir_List;
                     Sub_El : Iir;
                  begin
                     Sub_List := Get_Choice_List (Assoc);
                     for J in Natural loop
                        Sub_El := Get_Nth_Element (Sub_List, J);
                        exit when Sub_El = Null_Iir;
                        Set_El (Get_Name (Sub_El));
                     end loop;
                  end;
               when Iir_Kind_Association_Choice_By_Others =>
                  for J in Set_Array'Range loop
                     if not Set_Array (J) then
                        Set_El (Get_Nth_Element (El_List, J));
                     end if;
                  end loop;
               when others =>
                  Error_Kind ("translate_record_aggregate", Assoc);
            end case;
         end loop;
         Close_Temp;
      end Translate_Record_Aggregate;

      procedure Translate_Array_Aggregate
        (Target : Mnode; Target_Type : Iir; Aggr : Iir)
      is
         Aggr_Type : Iir;
         Base : Mnode;
         Bounds : Mnode;
         Var_Index : O_Lnode;
         Targ : Mnode;
         Tinfo : Type_Info_Acc;

         Range_Ptr : Mnode;
         Rinfo : Type_Info_Acc;
         Bt : Iir;

         function Check_Value
           (Lval : Iir; Lop : ON_Op_Kind; Rval : Iir; Rop : ON_Op_Kind)
            return O_Enode
         is
            L, R : O_Enode;
         begin
            L := New_Compare_Op (Lop,
                                 Translate_Static_Expression (Lval, Bt),
                                 M2E (Chap3.Range_To_Left (Range_Ptr)),
                                 Ghdl_Bool_Type_Node);
            R := New_Compare_Op (Rop,
                                 Translate_Static_Expression (Rval, Bt),
                                 M2E (Chap3.Range_To_Right (Range_Ptr)),
                                 Ghdl_Bool_Type_Node);
            return New_Dyadic_Op (ON_Or, L, R);
         end Check_Value;

         Index_List : Iir_List;
         Targ_Index_List : Iir_List;
         Subtarg_Type : Iir;
         Subaggr_Type : Iir;
         L, H : Iir;
         Max : Natural;
         Has_Others : Boolean;

         Aggr_Info : Iir_Aggregate_Info;

         Var_Err : O_Lnode;
         E : O_Enode;
         If_Blk : O_If_Block;
         Op : ON_Op_Kind;
      begin
         Tinfo := Get_Info (Target_Type);
         Open_Temp;
         --  FIXME: to be removed ?
         --Chap3.Translate_Type_Definition (Aggr_Type);
         Targ := Stabilize (Target);
         Base := Stabilize (Chap3.Get_Array_Base (Targ));
         Bounds := Stabilize (Chap3.Get_Array_Bounds (Targ));

         --  Check type
         Aggr_Type := Get_Type (Aggr);
         Index_List := Get_Index_List (Aggr_Type);
         Targ_Index_List := Get_Index_List (Target_Type);
         Aggr_Info := Get_Aggregate_Info (Aggr);
         for I in Natural loop
            Subaggr_Type := Get_Nth_Element (Index_List, I);
            exit when Subaggr_Type = Null_Iir;
            Subtarg_Type := Get_Nth_Element (Targ_Index_List, I);

            Bt := Get_Base_Type (Subaggr_Type);
            Rinfo := Get_Info (Bt);

            if Get_Type_Staticness (Subaggr_Type) /= Locally then
               --  Aggregate has dynamic bounds.
               if Subaggr_Type /= Subtarg_Type then
                  --  And it is not the same as the target.
                  --  Must be checked.

                  Open_Temp;
                  declare
                     A_Range : O_Lnode;
                     Rng_Ptr : O_Lnode;
                  begin
                     --  Evaluate the range.
                     Chap3.Translate_Anonymous_Type_Definition (Subaggr_Type);

                     A_Range := Create_Temp (Rinfo.T.Range_Type);
                     Rng_Ptr := Create_Temp_Ptr
                       (Rinfo.T.Range_Ptr_Type, A_Range);
                     Chap7.Translate_Range_Ptr
                       (Rng_Ptr,
                        Get_Range_Constraint (Subaggr_Type),
                        Subaggr_Type);

                     --  Check range length VS target length.
                     Chap6.Check_Bound_Error
                       (New_Compare_Op
                        (ON_Neq,
                         M2E (Chap3.Range_To_Length
                              (Vv2M (A_Range,
                                     Rinfo,
                                     Mode_Value,
                                     Rinfo.T.Range_Type,
                                     Rinfo.T.Range_Ptr_Type))),
                         M2E (Chap3.Range_To_Length
                              (Chap3.Bounds_To_Range
                               (Bounds, Target_Type, I + 1))),
                         Ghdl_Bool_Type_Node),
                        Aggr, I);
                  end;
                  Close_Temp;
               end if;
            else
               --  Note: if the aggregate has no others, then the bounds
               --  must be the same, otherwise, aggregate bounds must be
               --  inside type bounds.
               Has_Others := Get_Aggr_Others_Flag (Aggr_Info);

               Open_Temp;

               --  Pointer to the range.
               Range_Ptr := Stabilize
                 (Chap3.Bounds_To_Range (Bounds, Target_Type, I + 1));
               Var_Err := Create_Temp (Ghdl_Bool_Type_Node);
               Max := Get_Aggr_Max_Length (Aggr_Info);
               L := Get_Aggr_Low_Limit (Aggr_Info);
               H := Get_Aggr_High_Limit (Aggr_Info);

               if L /= Null_Iir then
                  Start_If_Stmt
                    (If_Blk,
                     New_Compare_Op (ON_Eq,
                                     M2E (Chap3.Range_To_Dir (Range_Ptr)),
                                     Ghdl_Dir_To_Node,
                                     Ghdl_Bool_Type_Node));
                  if Has_Others then
                     E := Check_Value (L, ON_Lt, H, ON_Gt);
                  else
                     E := Check_Value (L, ON_Neq, H, ON_Neq);
                  end if;
                  New_Assign_Stmt (Var_Err, E);
                  New_Else_Stmt (If_Blk);
                  if Has_Others then
                     E := Check_Value (H, ON_Gt, L, ON_Lt);
                  else
                     E := Check_Value (H, ON_Neq, L, ON_Neq);
                  end if;
                  New_Assign_Stmt (Var_Err, E);
                  Finish_If_Stmt (If_Blk);
                  -- If L and H are greather than the maximum length,
                  -- then there is no need to check with max.
                  if Natural (Eval_Pos (H) - Eval_Pos (L) + 1) >= Max then
                     Max := 0;
                  end if;
               end if;
               if Max > 0 then
                  if Has_Others then
                     Op := ON_Lt;
                  else
                     Op := ON_Neq;
                  end if;
                  E := New_Compare_Op
                    (Op,
                     M2E (Chap3.Range_To_Length (Range_Ptr)),
                     New_Unsigned_Literal (Ghdl_Index_Type,
                                           Unsigned_64 (Max)),
                     Ghdl_Bool_Type_Node);
                  if L /= Null_Iir then
                     E := New_Dyadic_Op (ON_Or, E, New_Value (Var_Err));
                  end if;
                  New_Assign_Stmt (Var_Err, E);
               end if;
               Chap6.Check_Bound_Error (New_Value (Var_Err), Aggr, I);
               Close_Temp;
            end if;

            --  Next dimension.
            Aggr_Info := Get_Sub_Aggregate_Info (Aggr_Info);
         end loop;

         Var_Index := Create_Temp_Init
           (Ghdl_Index_Type,
            New_Unsigned_Literal (Ghdl_Index_Type, 0));
         Translate_Array_Aggregate_Gen
           (Base, Bounds, Aggr, Aggr_Type, 1, Var_Index);
         Close_Temp;
      end Translate_Array_Aggregate;

      procedure Translate_Aggregate
        (Target : Mnode; Target_Type : Iir; Aggr : Iir)
      is
         Aggr_Type : Iir;
         El : Iir;
      begin
         Aggr_Type := Get_Type (Aggr);
         case Get_Kind (Aggr_Type) is
            when Iir_Kind_Array_Subtype_Definition
              | Iir_Kind_Array_Type_Definition =>
               El := Is_Aggregate_Others (Aggr);
               if El /= Null_Iir then
                  Translate_Aggregate_Others (Target, Target_Type, Aggr, El);
               else
                  Translate_Array_Aggregate (Target, Target_Type, Aggr);
               end if;
            when Iir_Kind_Record_Type_Definition
              | Iir_Kind_Record_Subtype_Definition =>
               Translate_Record_Aggregate (Target, Aggr);
            when others =>
               Error_Kind ("translate_aggregate", Aggr_Type);
         end case;
      end Translate_Aggregate;

      function Translate_Allocator_By_Expression (Expr : Iir)
        return O_Enode
      is
         Val : O_Enode;
         Val_M : Mnode;
         P_Type : Iir;
         P_Info : Type_Info_Acc;
         D_Type : Iir;
         D_Info : Type_Info_Acc;
         R : Mnode;
         Rtype : O_Tnode;
      begin
         P_Type := Get_Type (Expr);
         P_Info := Get_Info (P_Type);
         D_Type := Get_Designated_Type (P_Type);
         D_Info := Get_Info (D_Type);
         --  Compute the expression.
         Val := Translate_Expression (Get_Expression (Expr), D_Type);
         --  Allocate memory for the object.
         case P_Info.Type_Mode is
            when Type_Mode_Fat_Acc =>
               Val_M := Stabilize (E2M (Val, D_Info, Mode_Value));
               R := Chap3.Translate_Object_Allocation
                 (Alloc_Heap, D_Type,
                  M2Addr (Chap3.Get_Array_Bounds (Val_M)),
                  Mode_Value);
               Val := M2E (Val_M);
               Rtype := P_Info.Ortho_Ptr_Type (Mode_Value);
            when Type_Mode_Acc =>
               R := Chap3.Translate_Object_Allocation
                 (Alloc_Heap, D_Type, O_Enode_Null, Mode_Value);
               Rtype := P_Info.Ortho_Type (Mode_Value);
            when others =>
               raise Internal_Error;
         end case;
         Stabilize (R);
         Chap3.Translate_Object_Copy (R, Val, D_Type, Mode_Value);
         return New_Convert_Ov (M2Addr (R), Rtype);
      end Translate_Allocator_By_Expression;

      function Translate_Allocator_By_Subtype (Expr : Iir)
        return O_Enode
      is
         Sub_Type : Iir;
         Bounds : O_Enode;
         Res : Mnode;
         Rtype : O_Tnode;
         P_Type : Iir;
         P_Info : Type_Info_Acc;
         D_Type : Iir;
         D_Info : Type_Info_Acc;
      begin
         P_Type := Get_Type (Expr);
         P_Info := Get_Info (P_Type);
         D_Type := Get_Designated_Type (P_Type);
         D_Info := Get_Info (D_Type);
         case P_Info.Type_Mode is
            when Type_Mode_Fat_Acc =>
               --  FIXME: should allocate bounds, and directly set bounds
               --  from the range.
               Sub_Type := Get_Expression (Expr);
               Chap3.Create_Array_Subtype (Sub_Type);
               Bounds := M2E (Chap3.Get_Array_Type_Bounds (Sub_Type));
               Rtype := P_Info.Ortho_Ptr_Type (Mode_Value);
            when Type_Mode_Acc =>
               Bounds := O_Enode_Null;
               Rtype := P_Info.Ortho_Type (Mode_Value);
            when others =>
               raise Internal_Error;
         end case;
         Res := Chap3.Translate_Object_Allocation
           (Alloc_Heap, D_Type, Bounds, Mode_Value);
         --  FIXME: should initialize RES.
         return New_Convert_Ov (M2Addr (Res), Rtype);
      end Translate_Allocator_By_Subtype;

      function Translate_Fat_Array_Type_Conversion
        (Expr : O_Enode; Expr_Type : Iir; Res_Type : Iir; Loc : Iir)
        return O_Enode;

      function Translate_Array_Subtype_Conversion
        (Expr : O_Enode; Expr_Type : Iir; Res_Type : Iir; Loc : Iir)
        return O_Enode
      is
         Res_Info : Type_Info_Acc;
         Expr_Info : Type_Info_Acc;
      begin
         Res_Info := Get_Info (Res_Type);
         Expr_Info := Get_Info (Expr_Type);
         case Res_Info.Type_Mode is
            when Type_Mode_Array =>
               declare
                  E : O_Lnode;
               begin
                  E := Create_Temp_Init (Expr_Info.Ortho_Ptr_Type (Mode_Value),
                                         Expr);
                  Chap3.Check_Array_Match (Res_Type, O_Lnode_Null, Mode_Value,
                                           Expr_Type, E, Mode_Value,
                                           Loc);
                  return New_Convert_Ov
                    (New_Value (Chap3.Get_Array_Ptr_Base_Ptr
                                (E, Expr_Type, Mode_Value)),
                     Res_Info.Ortho_Ptr_Type (Mode_Value));
               end;
            when Type_Mode_Fat_Array =>
               declare
                  Res : O_Lnode;
                  E : O_Lnode;
               begin
                  Res := Create_Temp (Res_Info.Ortho_Type (Mode_Value));
                  Open_Temp;
                  E := Create_Temp_Init (Expr_Info.Ortho_Ptr_Type (Mode_Value),
                                         Expr);
                  --  Convert base.
                  New_Assign_Stmt
                    (New_Selected_Element
                     (Res, Res_Info.T.Base_Field (Mode_Value)),
                     New_Value (Chap3.Get_Array_Ptr_Base_Ptr
                                (E, Expr_Type, Mode_Value)));
                  --  Copy subtype bounds.
                  New_Assign_Stmt
                    (New_Selected_Element
                     (Res, Res_Info.T.Bounds_Field (Mode_Value)),
                     Chap3.Get_Array_Bounds_Ptr (O_Lnode_Null, Expr_Type,
                                                 Mode_Value));
                  --  Check array match.
                  Chap3.Check_Array_Match (Res_Type, Res, Mode_Value,
                                           Expr_Type, E, Mode_Value,
                                           Loc);
                  Close_Temp;
                  return New_Address
                    (Res, Res_Info.Ortho_Ptr_Type (Mode_Value));
               end;
            when others =>
               Error_Kind ("translate_array_subtype_conversion", Res_Type);
         end case;
      end Translate_Array_Subtype_Conversion;

      function Translate_Type_Conversion
        (Expr : O_Enode; Expr_Type : Iir; Res_Type : Iir; Loc : Iir)
        return O_Enode
      is
         Res_Info : Type_Info_Acc;
      begin
         Res_Info := Get_Info (Res_Type);
         case Get_Kind (Res_Type) is
            when Iir_Kinds_Scalar_Type_Definition =>
               --  If res_type = expr_type, do not convert.
               --  FIXME: range check ?
               return New_Convert_Ov (Expr, Res_Info.Ortho_Type (Mode_Value));
            when Iir_Kind_Array_Subtype_Definition =>
               return Translate_Array_Subtype_Conversion
                 (Expr, Expr_Type, Res_Type, Loc);
            when Iir_Kind_Array_Type_Definition
              | Iir_Kind_Unconstrained_Array_Subtype_Definition =>
               return Translate_Fat_Array_Type_Conversion
                 (Expr, Expr_Type, Res_Type, Loc);
            when others =>
               Error_Kind ("translate_type_conversion", Res_Type);
         end case;
      end Translate_Type_Conversion;

      function Translate_Fat_Array_Type_Conversion
        (Expr : O_Enode; Expr_Type : Iir; Res_Type : Iir; Loc : Iir)
        return O_Enode
      is
         Res : O_Lnode;
         Res_Ptr : O_Lnode;
         E : O_Lnode;
         Bounds : O_Lnode;
         Res_Indexes : Iir;
         Expr_Indexes : Iir;
         R_El : Iir;
         E_El : Iir;
         Res_Info : Type_Info_Acc;
         Expr_Info : Type_Info_Acc;
      begin
         Res_Info := Get_Info (Res_Type);
         Expr_Info := Get_Info (Expr_Type);
         Res := Create_Temp (Res_Info.Ortho_Type (Mode_Value));
         Bounds := Create_Temp (Res_Info.T.Bounds_Type);
         Open_Temp;
         Res_Ptr := Create_Temp_Ptr
           (Res_Info.Ortho_Ptr_Type (Mode_Value), Res);
         E := Create_Temp_Init (Expr_Info.Ortho_Ptr_Type (Mode_Value), Expr);
         --  Set base.
         New_Assign_Stmt
           (New_Selected_Element (Res, Res_Info.T.Base_Field (Mode_Value)),
            New_Convert_Ov (New_Value (Chap3.Get_Array_Ptr_Base_Ptr
                                       (E, Expr_Type, Mode_Value)),
                            Res_Info.T.Base_Ptr_Type (Mode_Value)));
         --  Set bounds.
         New_Assign_Stmt
           (New_Selected_Element (Res, Res_Info.T.Bounds_Field (Mode_Value)),
            New_Address (Bounds, Res_Info.T.Bounds_Ptr_Type));
         --  Convert bounds.
         Res_Indexes := Get_Index_List (Res_Type);
         Expr_Indexes := Get_Index_List (Expr_Type);
         for I in Natural loop
            R_El := Get_Nth_Element (Res_Indexes, I);
            E_El := Get_Nth_Element (Expr_Indexes, I);
            exit when R_El = Null_Iir;
            declare
               Rb_Ptr : O_Lnode;
               Eb_Ptr : O_Lnode;
               Rr_Info : Type_Info_Acc;
               Er_Info : Type_Info_Acc;
            begin
               Open_Temp;
               Rr_Info := Get_Info (R_El);
               Rb_Ptr := Create_Temp_Init
                 (Rr_Info.T.Range_Ptr_Type,
                  Chap3.Get_Array_Ptr_Range_Ptr (Res_Ptr, Res_Type, I + 1,
                                                 Mode_Value));
               Er_Info := Get_Info (Get_Base_Type (E_El));
               Eb_Ptr := Create_Temp_Init
                 (Er_Info.T.Range_Ptr_Type,
                  Chap3.Get_Array_Ptr_Range_Ptr (E, Expr_Type, I + 1,
                                                 Mode_Value));
               --  Convert left and right.
               New_Assign_Stmt
                 (New_Selected_Acc_Value (Rb_Ptr, Rr_Info.T.Range_Left),
                  Translate_Type_Conversion
                  (New_Value_Selected_Acc_Value (Eb_Ptr, Er_Info.T.Range_Left),
                   E_El, R_El, Loc));
               New_Assign_Stmt
                 (New_Selected_Acc_Value (Rb_Ptr, Rr_Info.T.Range_Right),
                  Translate_Type_Conversion
                  (New_Value_Selected_Acc_Value (Eb_Ptr,
                                                 Er_Info.T.Range_Right),
                   E_El, R_El, Loc));
               --  Copy Dir and Length.
               New_Assign_Stmt
                 (New_Selected_Acc_Value (Rb_Ptr, Rr_Info.T.Range_Dir),
                  New_Value_Selected_Acc_Value (Eb_Ptr, Er_Info.T.Range_Dir));
               New_Assign_Stmt
                 (New_Selected_Acc_Value (Rb_Ptr, Rr_Info.T.Range_Length),
                  New_Value_Selected_Acc_Value (Eb_Ptr,
                                                Er_Info.T.Range_Length));
               Close_Temp;
            end;
         end loop;
         Close_Temp;
         return New_Address (Res, Res_Info.Ortho_Ptr_Type (Mode_Value));
      end Translate_Fat_Array_Type_Conversion;


      function Translate_Signal_Value (Sig : O_Enode; Sig_Type : Iir)
        return O_Enode
      is
         procedure Translate_Signal_Non_Composite
           (Targ : Mnode;
            Targ_Type : Iir;
            Data : Mnode)
         is
         begin
            New_Assign_Stmt (M2Lv (Targ),
                             Read_Value (M2E (Data), Targ_Type));
         end Translate_Signal_Non_Composite;

         function Sig2val_Prepare_Composite (Data : Mnode; Targ_Type : Iir)
           return Mnode
         is
         begin
            if Get_Type_Info (Data).Type_Mode = Type_Mode_Fat_Array then
               return Stabilize (Chap3.Get_Array_Base (Data));
            else
               return Stabilize (Data);
            end if;
         end Sig2val_Prepare_Composite;

         function Sig2val_Update_Data_Array
           (Val : Mnode; Targ_Type : Iir; Index : O_Lnode) return Mnode
         is
         begin
            return Chap3.Index_Base (Val, Targ_Type, New_Value (Index));
         end Sig2val_Update_Data_Array;

         function Sig2val_Update_Data_Record
           (Val : Mnode; Targ_Type : Iir; El : Iir_Element_Declaration)
           return Mnode
         is
         begin
            return Chap6.Translate_Selected_Element (Val, El);
         end Sig2val_Update_Data_Record;

         procedure Translate_Signal_Target is new Foreach_Non_Composite
           (Data_Type => Mnode,
            Composite_Data_Type => Mnode,
            Do_Non_Composite => Translate_Signal_Non_Composite,
            Prepare_Data_Array => Sig2val_Prepare_Composite,
            Update_Data_Array => Sig2val_Update_Data_Array,
            Prepare_Data_Record => Sig2val_Prepare_Composite,
            Update_Data_Record => Sig2val_Update_Data_Record);

         Tinfo : Type_Info_Acc;
      begin
         Tinfo := Get_Info (Sig_Type);
         if Tinfo.Type_Mode in Type_Mode_Scalar then
            return Read_Value (Sig, Sig_Type);
         else
            declare
               Lres : O_Lnode;
               Res : Mnode;
               Var_Val : Mnode;
            begin
               --  allocate result array
               if Tinfo.Type_Mode = Type_Mode_Fat_Array then
                  Lres := Create_Temp (Tinfo.Ortho_Type (Mode_Value));
                  Res := Lo2M (Lres, Tinfo, Mode_Value);

                  Var_Val := Stabilize (E2M (Sig, Tinfo, Mode_Signal));

                  --  Copy bounds.
                  New_Assign_Stmt
                    (M2Lp (Chap3.Get_Array_Bounds (Res)),
                     M2Addr (Chap3.Get_Array_Bounds (Var_Val)));

                  --  Allocate base.
                  Chap3.Allocate_Fat_Array_Base
                    (Alloc_Stack, Res, Sig_Type, Mode_Value);
               elsif Tinfo.C /= null then
                  Res := Create_Temp (Tinfo);
                  Chap4.Allocate_Complex_Object (Sig_Type, Alloc_Stack, Res);
               else
                  Res := Create_Temp (Tinfo);
               end if;

               Open_Temp;

               if Tinfo.Type_Mode /= Type_Mode_Fat_Array then
                  Var_Val := Stabilize (E2M (Sig, Tinfo, Mode_Signal));
               end if;

               Translate_Signal_Target (Res, Sig_Type, Var_Val);
               Close_Temp;
               return M2Addr (Res);
            end;
         end if;

--          case Tinfo.Type_Mode is
--             when Type_Mode_Scalar =>
--                return Read_Value (Sig, Sig_Type);
--             when Type_Mode_Array =>
--                declare
--                   Res : Mnode;
--                   Sig_Var : Mnode;
--                   Index_Var : O_Lnode;
--                   Loop_Blk : O_Loop_Block;
--                   Label : O_Snode;
--                begin
--                   Res := Vv2M (Create_Temp (Tinfo.Ortho_Type (Mode_Value)),
--                                Tinfo, Mode_Value);
--                   Open_Temp;
--                   Sig_Var := Stabilize (E2M (Sig, Tinfo, Mode_Signal));
--                   Index_Var := Create_Temp_Init
--                     (Ghdl_Index_Type,
--                      New_Unsigned_Literal (Ghdl_Index_Type, 0));
--                   Start_Loop_Stmt (Loop_Blk, Label);
--                   Gen_Exit_When
--                     (Label,
--                      New_Compare_Op (ON_Eq,
--                                      New_Value (Index_Var),
--                                      Chap3.Get_Thin_Array_Length (Sig_Type),
--                                      Ghdl_Bool_Type_Node));
--                   Translate_Signal_Target
--                    (Chap3.Index_Base (Res, Sig_Type, New_Value (Index_Var)),
--                      Get_Element_Subtype (Sig_Type),
--                      Chap3.Index_Base (Sig_Var, Sig_Type,
--                                        New_Value (Index_Var)));
--                   Inc_Var (Index_Var);
--                   Finish_Loop_Stmt (Loop_Blk);
--                   Close_Temp;
--                   return New_Address (M2Lv (Res),
--                                       Tinfo.Ortho_Ptr_Type (Mode_Value));
--                end;
--             when Type_Mode_Fat_Array =>
--                declare
--                   Var_Res : O_Lnode;
--                   Var_Length : O_Lnode;
--                   Var_I : O_Lnode;
--                   Var_Val : O_Lnode;
--                   Var_Base : Mnode;
--                   El_Type : Iir;
--                   Loop_Blk : O_Loop_Block;
--                   Label : O_Snode;
--                begin
--                   El_Type := Get_Element_Subtype (Sig_Type);
--                   Var_Res := Create_Temp (Tinfo.Ortho_Type (Mode_Value));
--                   Open_Temp;
--                   Var_Val := Create_Temp_Init
--                     (Tinfo.Ortho_Ptr_Type (Mode_Signal), Sig);
--                   --  Copy bounds.
--                   New_Assign_Stmt
--                     (New_Selected_Element (Var_Res,
--                                          Tinfo.T.Bounds_Field (Mode_Value)),
--                      Chap3.Get_Array_Ptr_Bounds_Ptr (Var_Val, Sig_Type,
--                                                      Mode_Signal));
--                   --  Compute length of array
--                   Var_Length := Create_Temp_Init
--                     (Ghdl_Index_Type,
--                      Chap3.Get_Array_Ptr_Length
--                      (Var_Val, Sig_Type, Mode_Signal));
--                   --  allocate result array
--                   New_Assign_Stmt
--                     (New_Selected_Element (Var_Res,
--                                            Tinfo.T.Base_Field (Mode_Value)),
--                      Gen_Alloc
--                      (Alloc_Stack,
--                       New_Dyadic_Op (ON_Mul_Ov,
--                                      Chap3.Get_Object_Size
--                                      (El_Type, O_Lnode_Null, Mode_Value),
--                                      New_Value (Var_Length)),
--                       Tinfo.T.Base_Ptr_Type (Mode_Value)));
--                   Var_Base := Stabilize (Chap3.Get_Array_Base
--                                       (Vp2M (Var_Val, Tinfo, Mode_Signal)));
--                   --  Convert.
--                   Var_I := Create_Temp (Ghdl_Index_Type);
--                   Init_Var (Var_I);
--                   Start_Loop_Stmt (Loop_Blk, Label);
--                   Gen_Exit_When
--                     (Label, New_Compare_Op (ON_Ge,
--                                             New_Value (Var_I),
--                                             New_Value (Var_Length),
--                                             Ghdl_Bool_Type_Node));
--                   --  Translate element.
--                   Translate_Signal_Target
--                     (Chap3.Index_Base (Chap3.Get_Array_Base
--                                        (Vv2M (Var_Res, Tinfo, Mode_Value)),
--                                        Sig_Type,
--                                        New_Value (Var_I)),
--                      El_Type,
--                   Chap3.Index_Base (Var_Base, Sig_Type, New_Value (Var_I)));
--                   Inc_Var (Var_I);
--                   Finish_Loop_Stmt (Loop_Blk);
--                   Close_Temp;
--                   return New_Address
--                     (Var_Res, Tinfo.Ortho_Ptr_Type (Mode_Value));
--                end;
--             when Type_Mode_Ptr_Array =>
--                declare
--                   Var_Res : O_Lnode;
--                   Var_Length : O_Lnode;
--                   Var_I : O_Lnode;
--                   Var_Val : Mnode;
--                   El_Type : Iir;
--                   Loop_Blk : O_Loop_Block;
--                   Label : O_Snode;
--                begin
--                   --  allocate result array
--                   Var_Res := Chap4.Allocate_Complex_Object
--                     (Sig_Type, Alloc_Stack, Mode_Value, O_Lnode_Null);

--                   El_Type := Get_Element_Subtype (Sig_Type);
--                   Open_Temp;
--                   Var_Val := Stabilize (E2M (Sig, Tinfo, Mode_Signal));
--                   --  Compute length of array
--                   Var_Length := Create_Temp_Init
--                   (Ghdl_Index_Type, Chap3.Get_Array_Type_Length (Sig_Type));
--                   --  Convert.
--                   Var_I := Create_Temp (Ghdl_Index_Type);
--                   Init_Var (Var_I);
--                   Start_Loop_Stmt (Loop_Blk, Label);
--                   Gen_Exit_When
--                     (Label, New_Compare_Op (ON_Ge,
--                                             New_Value (Var_I),
--                                             New_Value (Var_Length),
--                                             Ghdl_Bool_Type_Node));
--                   --  Translate element.
--                   Translate_Signal_Target
--                     (Chap3.Index_Base (Chap3.Get_Array_Base
--                                        (Vp2M (Var_Res, Tinfo, Mode_Value)),
--                                        Sig_Type,
--                                        New_Value (Var_I)),
--                      El_Type,
--                    Chap3.Index_Base (Var_Val, Sig_Type, New_Value (Var_I)));
--                   Inc_Var (Var_I);
--                   Finish_Loop_Stmt (Loop_Blk);
--                   Close_Temp;
--                   return New_Value (Var_Res);
--                end;
--             when Type_Mode_Record =>
--                declare
--                   Res : O_Lnode;
--                   Sig_Var : Mnode;
--                   Elements : Iir_List;
--                   El : Iir_Element_Declaration;
--                begin
--                   Res := Create_Temp (Tinfo.Ortho_Type (Mode_Value));
--                   Open_Temp;
--                   Sig_Var := Stabilize (E2M (Sig, Tinfo, Mode_Signal));
--                   Elements := Get_Element_Declaration_List
--                     (Get_Base_Type (Sig_Type));
--                   for I in Natural loop
--                      El := Get_Nth_Element (Elements, I);
--                      exit when El = Null_Iir;
--                      Translate_Signal_Target
--                        (Chap6.Translate_Selected_Element
--                         (Vv2M (Res, Tinfo, Mode_Value), El),
--                         Get_Type (El),
--                         Chap6.Translate_Selected_Element (Sig_Var, El));
--                   end loop;
--                   Close_Temp;
--                 return New_Address (Res, Tinfo.Ortho_Ptr_Type (Mode_Value));
--                end;
--             when others =>
--                Error_Kind ("translate_signal", Sig_Type);
--          end case;
      end Translate_Signal_Value;

      --  Get the effective value of a simple signal SIG.
      function Read_Signal_Value (Sig : O_Enode; Sig_Type : Iir)
        return O_Enode is
      begin
         return New_Value (New_Access_Element (Sig));
      end Read_Signal_Value;

      --  Get the value of signal SIG.
      function Translate_Signal is new Translate_Signal_Value
        (Read_Value => Read_Signal_Value);

      function Translate_Expression (Expr : Iir; Rtype : Iir := Null_Iir)
        return O_Enode
      is
         Imp : Iir;
         Expr_Type : Iir;
         Res_Type : Iir;
         Res : O_Enode;
      begin
         Expr_Type := Get_Type (Expr);
         if Rtype = Null_Iir then
            Res_Type := Expr_Type;
         else
            Res_Type := Rtype;
         end if;
         case Get_Kind (Expr) is
            when Iir_Kind_Integer_Literal
              | Iir_Kind_Enumeration_Literal
              | Iir_Kind_Floating_Point_Literal
              | Iir_Kind_Physical_Literal =>
               return Translate_Static_Expression (Expr, Rtype);

            when Iir_Kind_String_Literal
              | Iir_Kind_Bit_String_Literal =>
               Res := Translate_String_Literal (Expr);
               Res := Translate_Implicit_Conv
                 (Res, Expr_Type, Rtype, Mode_Value);
               return Res;

            when Iir_Kind_Aggregate =>
               declare
                  Aggr_Type : Iir;
                  Tinfo : Type_Info_Acc;
                  Mres : Mnode;
               begin
                  if Rtype = Null_Iir then
                     raise Internal_Error;
                  end if;

                  --  Extract the type of the aggregate.
                  if Get_Kind (Rtype) /= Iir_Kind_Array_Type_Definition then
                     Aggr_Type := Rtype;
                  else
                     Aggr_Type := Expr_Type;
                     if Get_Kind (Expr_Type)
                       = Iir_Kind_Array_Subtype_Definition
                     then
                        Chap3.Create_Array_Subtype (Expr_Type);
                     end if;
                  end if;

                  --  FIXME: this may be not necessary
                  Tinfo := Get_Info (Aggr_Type);

                  --  The result area has to be created
                  if Tinfo.C /= null then
                     Mres := Create_Temp (Tinfo);
                     Chap4.Allocate_Complex_Object
                       (Aggr_Type, Alloc_Stack, Mres);
                  else
                     --  if thin array/record:
                     --    create result
                     Mres := Create_Temp (Tinfo);
                  end if;

                  Translate_Aggregate (Mres, Aggr_Type, Expr);
                  Res := M2E (Mres);

                  if Aggr_Type /= Rtype then
                     Res := Translate_Implicit_Conv
                       (Res, Aggr_Type, Rtype, Mode_Value);
                  end if;
                  return Res;
               end;

            when Iir_Kind_Null_Literal =>
               declare
                  L : O_Lnode;
                  Otype : O_Tnode;
                  B : Type_Info_Acc;
                  Tinfo : Type_Info_Acc;
               begin
                  Tinfo := Get_Info (Expr_Type);
                  Otype := Tinfo.Ortho_Type (Mode_Value);
                  if Tinfo.Type_Mode = Type_Mode_Fat_Acc then
                     --  Create a fat null pointer.
                     --  FIXME: should be optimized!!
                     L := Create_Temp (Otype);
                     B := Get_Info (Get_Designated_Type (Expr_Type));
                     New_Assign_Stmt
                       (New_Selected_Element (L, B.T.Base_Field (Mode_Value)),
                        New_Null_Access (B.T.Base_Ptr_Type (Mode_Value)));
                     New_Assign_Stmt
                       (New_Selected_Element
                        (L, B.T.Bounds_Field (Mode_Value)),
                        New_Null_Access (B.T.Bounds_Ptr_Type));
                     return New_Address (L, Tinfo.Ortho_Ptr_Type (Mode_Value));
                  else
                     return New_Null_Access (Otype);
                  end if;
               end;

            when Iir_Kind_Allocator_By_Expression =>
               return Translate_Allocator_By_Expression (Expr);
            when Iir_Kind_Allocator_By_Subtype =>
               return Translate_Allocator_By_Subtype (Expr);

            when Iir_Kind_Qualified_Expression =>
               --  FIXME: check type.
               Res := Translate_Expression (Get_Expression (Expr), Expr_Type);
               return Translate_Implicit_Conv
                 (Res, Expr_Type, Rtype, Mode_Value);

            when Iir_Kind_Constant_Declaration
              | Iir_Kind_Variable_Declaration
              | Iir_Kind_Signal_Declaration
              | Iir_Kind_File_Declaration
              | Iir_Kind_Alias_Declaration
              | Iir_Kind_Constant_Interface_Declaration
              | Iir_Kind_Variable_Interface_Declaration
              | Iir_Kind_Signal_Interface_Declaration
              | Iir_Kind_File_Interface_Declaration
              | Iir_Kind_Indexed_Name
              | Iir_Kind_Slice_Name
              | Iir_Kind_Selected_By_All_Name
              | Iir_Kind_Selected_Element
              | Iir_Kind_Implicit_Dereference
              | Iir_Kind_Stable_Attribute
              | Iir_Kind_Guard_Signal_Declaration
              | Iir_Kind_Attribute_Value =>
               declare
                  L : Mnode;
                  Expr_Type_Info : Type_Info_Acc;
               begin
                  L := Chap6.Translate_Name (Expr);

                  Expr_Type_Info := Get_Info (Expr_Type);
                  Res := M2E (L);
                  if Get_Object_Kind (L) = Mode_Signal then
                     Res := Translate_Signal (Res, Expr_Type);
                  end if;
               end;
               if Rtype /= Null_Iir then
                  Res := Translate_Implicit_Conv
                    (Res, Expr_Type, Rtype, Mode_Value);
               end if;
               return Res;

            when Iir_Kind_Iterator_Declaration =>
               declare
                  Expr_Info : Ortho_Info_Acc;
               begin
                  Expr_Info := Get_Info (Expr);
                  Res := New_Value (Get_Var (Expr_Info.Iterator_Var));
                  if Rtype /= Null_Iir then
                     Res := New_Convert_Ov
                       (Res, Get_Ortho_Type (Rtype, Mode_Value));
                  end if;
                  return Res;
               end;

            when Iir_Kinds_Dyadic_Operator =>
               Imp := Get_Implementation (Expr);
               if Get_Kind (Imp) = Iir_Kind_Implicit_Function_Declaration then
                  return Translate_Predefined_Operator
                    (Imp, Get_Left (Expr), Get_Right (Expr), Res_Type);
               else
                  return Translate_Operator_Function_Call
                    (Imp, Get_Left (Expr), Get_Right (Expr), Res_Type);
               end if;
            when Iir_Kinds_Monadic_Operator =>
               Imp := Get_Implementation (Expr);
               if Get_Kind (Imp) = Iir_Kind_Implicit_Function_Declaration then
                  return Translate_Predefined_Operator
                    (Imp, Get_Operand (Expr), Null_Iir, Res_Type);
               else
                  return Translate_Operator_Function_Call
                    (Imp, Get_Operand (Expr), Null_Iir, Res_Type);
               end if;
            when Iir_Kind_Function_Call =>
               Imp := Get_Implementation (Expr);
               if Get_Kind (Imp) = Iir_Kind_Implicit_Function_Declaration then
                  declare
                     List : Iir_List;
                     Left, Right : Iir;
                  begin
                     List := Get_Parameter_Association_List (Expr);
                     if List = Null_Iir_List then
                        Left := Null_Iir;
                        Right := Null_Iir;
                     else
                        Left := Get_Actual (Get_Nth_Element (List, 0));
                        if Get_Nbr_Elements (List) = 1 then
                           Right := Null_Iir;
                        else
                           Right := Get_Actual (Get_Nth_Element (List, 1));
                        end if;
                     end if;
                     return Translate_Predefined_Operator
                       (Imp, Left, Right, Res_Type);
                  end;
               else
                  Res := Translate_Function_Call
                    (Imp, Get_Parameter_Association_List (Expr));
                  return Translate_Implicit_Conv
                    (Res, Get_Return_Type (Imp), Res_Type, Mode_Value);
               end if;

            when Iir_Kind_Type_Conversion =>
               declare
                  Conv_Expr : Iir;
               begin
                  Conv_Expr := Get_Expression (Expr);
                  Res := Translate_Type_Conversion
                    (Translate_Expression (Conv_Expr), Get_Type (Conv_Expr),
                     Expr_Type, Expr);
                  return Translate_Implicit_Conv
                    (Res, Expr_Type, Res_Type, Mode_Value);
               end;

            when Iir_Kind_Length_Array_Attribute =>
               return Chap14.Translate_Length_Array_Attribute
                 (Expr, Res_Type);
            when Iir_Kind_Low_Array_Attribute =>
               return Chap14.Translate_Low_Array_Attribute (Expr);
            when Iir_Kind_High_Array_Attribute =>
               return Chap14.Translate_High_Array_Attribute (Expr);
            when Iir_Kind_Left_Array_Attribute =>
               return Chap14.Translate_Left_Array_Attribute (Expr);
            when Iir_Kind_Right_Array_Attribute =>
               return Chap14.Translate_Right_Array_Attribute (Expr);

            when Iir_Kind_Val_Attribute =>
               return Chap14.Translate_Val_Attribute (Expr);
            when Iir_Kind_Pos_Attribute =>
               return Chap14.Translate_Pos_Attribute (Expr, Res_Type);

            when Iir_Kind_Succ_Attribute
               | Iir_Kind_Pred_Attribute =>
               return Chap14.Translate_Succ_Pred_Attribute (Expr);

            when Iir_Kind_Event_Attribute =>
               return Chap14.Translate_Event_Attribute (Expr);
            when Iir_Kind_Active_Attribute =>
               return Chap14.Translate_Active_Attribute (Expr);
            when Iir_Kind_Last_Value_Attribute =>
               return Chap14.Translate_Last_Value_Attribute (Expr);
            when Iir_Kind_High_Type_Attribute =>
               return Chap14.Translate_High_Type_Attribute (Get_Prefix (Expr));
            when Iir_Kind_Low_Type_Attribute =>
               return Chap14.Translate_Low_Type_Attribute (Get_Prefix (Expr));
            when Iir_Kind_Left_Type_Attribute =>
               return Chap14.Translate_Left_Type_Attribute (Get_Prefix (Expr));
            when Iir_Kind_Right_Type_Attribute =>
               return Chap14.Translate_Right_Type_Attribute
                 (Get_Prefix (Expr));

            when Iir_Kind_Last_Event_Attribute =>
               return Chap14.Translate_Last_Time_Attribute
                 (Get_Prefix (Expr), Ghdl_Signal_Last_Event_Node);
            --when Iir_Kind_Last_Active_Attribute =>
            --   return Chap14.Translate_Last_Time_Attribute
            --     (Get_Prefix (Expr), Ghdl_Signal_Last_Active_Attribute);

            when others =>
               Error_Kind ("translate_expression", Expr);
         end case;
      end Translate_Expression;

--       procedure Translate_Range_Expression
--         (Res : O_Lnode; Expr : Iir; Range_Type : Iir)
--       is
--          T_Info : Type_Info_Acc;
--       begin
--          T_Info := Get_Info (Range_Type);
--          Open_Temp;
--          New_Assign_Stmt
--            (New_Selected_Element (Res, T_Info.T.Range_Left),
--             Chap7.Translate_Range_Expression_Left (Expr, Range_Type));
--          New_Assign_Stmt
--            (New_Selected_Element (Res, T_Info.T.Range_Right),
--             Chap7.Translate_Range_Expression_Right (Expr, Range_Type));
--          New_Assign_Stmt (New_Selected_Element (Res, T_Info.T.Range_Dir),
--                           Chap7.Translate_Static_Range_Dir (Expr));
--          if T_Info.T.Range_Length /= O_Fnode_Null then
--             Open_Temp;
--          New_Assign_Stmt (New_Selected_Element (Res, T_Info.T.Range_Length),
--                            Chap7.Translate_Range_Expression_Length (Expr));
--             Close_Temp;
--          end if;
--          Close_Temp;
--       end Translate_Range_Expression;

      procedure Translate_Range_Expression_Ptr
        (Res_Ptr : O_Lnode; Expr : Iir; Range_Type : Iir)
      is
         T_Info : Type_Info_Acc;
      begin
         T_Info := Get_Info (Range_Type);
         Open_Temp;
         New_Assign_Stmt
           (New_Selected_Acc_Value (Res_Ptr, T_Info.T.Range_Left),
            Chap7.Translate_Range_Expression_Left (Expr, Range_Type));
         New_Assign_Stmt
           (New_Selected_Acc_Value (Res_Ptr, T_Info.T.Range_Right),
            Chap7.Translate_Range_Expression_Right (Expr, Range_Type));
         New_Assign_Stmt
           (New_Selected_Acc_Value (Res_Ptr, T_Info.T.Range_Dir),
            Chap7.Translate_Static_Range_Dir (Expr));
         if T_Info.T.Range_Length /= O_Fnode_Null then
            if Get_Expr_Staticness (Expr) = Locally then
               New_Assign_Stmt
                 (New_Selected_Acc_Value (Res_Ptr, T_Info.T.Range_Length),
                  Translate_Static_Range_Length (Expr));
            else
               Open_Temp;
               New_Assign_Stmt
                 (New_Selected_Acc_Value (Res_Ptr, T_Info.T.Range_Length),
                  Compute_Range_Length
                  (New_Value_Selected_Acc_Value (Res_Ptr, T_Info.T.Range_Left),
                   New_Value_Selected_Acc_Value (Res_Ptr,
                                                 T_Info.T.Range_Right),
                   Get_Direction (Expr)));
               Close_Temp;
            end if;
         end if;
         Close_Temp;
      end Translate_Range_Expression_Ptr;

      --  Reverse range ARANGE.
      procedure Translate_Reverse_Range_Ptr
        (Res_Ptr : O_Lnode; Arange : O_Lnode; Range_Type : Iir)
      is
         Rinfo : Type_Info_Acc;
         Ptr : O_Lnode;
         If_Blk : O_If_Block;
      begin
         Rinfo := Get_Info (Range_Type);
         Open_Temp;
         Ptr := Create_Temp_Ptr (Rinfo.T.Range_Ptr_Type, Arange);
         New_Assign_Stmt
           (New_Selected_Acc_Value (Res_Ptr, Rinfo.T.Range_Left),
            New_Value_Selected_Acc_Value (Ptr, Rinfo.T.Range_Right));
         New_Assign_Stmt
           (New_Selected_Acc_Value (Res_Ptr, Rinfo.T.Range_Right),
            New_Value_Selected_Acc_Value (Ptr, Rinfo.T.Range_Left));
         New_Assign_Stmt
           (New_Selected_Acc_Value (Res_Ptr, Rinfo.T.Range_Length),
            New_Value_Selected_Acc_Value (Ptr, Rinfo.T.Range_Length));
         Start_If_Stmt
           (If_Blk,
            New_Compare_Op
            (ON_Eq,
             New_Value_Selected_Acc_Value (Ptr, Rinfo.T.Range_Dir),
             Ghdl_Dir_To_Node,
             Ghdl_Bool_Type_Node));
         New_Assign_Stmt
           (New_Selected_Acc_Value (Res_Ptr, Rinfo.T.Range_Dir),
            Ghdl_Dir_Downto_Node);
         New_Else_Stmt (If_Blk);
         New_Assign_Stmt
           (New_Selected_Acc_Value (Res_Ptr, Rinfo.T.Range_Dir),
            Ghdl_Dir_To_Node);
         Finish_If_Stmt (If_Blk);
         Close_Temp;
      end Translate_Reverse_Range_Ptr;

      procedure Copy_Range (Dest_Ptr : O_Lnode;
                            Src_Ptr : O_Lnode;
                            Info : Type_Info_Acc)
      is
      begin
         New_Assign_Stmt
           (New_Selected_Acc_Value (Dest_Ptr, Info.T.Range_Left),
            New_Value_Selected_Acc_Value (Src_Ptr, Info.T.Range_Left));
         New_Assign_Stmt
           (New_Selected_Acc_Value (Dest_Ptr, Info.T.Range_Right),
            New_Value_Selected_Acc_Value (Src_Ptr, Info.T.Range_Right));
         New_Assign_Stmt
           (New_Selected_Acc_Value (Dest_Ptr, Info.T.Range_Dir),
            New_Value_Selected_Acc_Value (Src_Ptr, Info.T.Range_Dir));
         if Info.T.Range_Length /= O_Fnode_Null then
            New_Assign_Stmt
              (New_Selected_Acc_Value (Dest_Ptr, Info.T.Range_Length),
               New_Value_Selected_Acc_Value (Src_Ptr, Info.T.Range_Length));
         end if;
      end Copy_Range;

      procedure Translate_Range_Ptr
        (Res_Ptr : O_Lnode; Arange : Iir; Range_Type : Iir)
      is
      begin
         case Get_Kind (Arange) is
            when Iir_Kind_Range_Array_Attribute =>
               declare
                  Ptr : O_Lnode;
                  Rinfo : Type_Info_Acc;
               begin
                  Rinfo := Get_Info (Range_Type);
                  Open_Temp;
                  Ptr := Create_Temp_Ptr
                    (Rinfo.T.Range_Ptr_Type,
                     Chap14.Translate_Range_Array_Attribute (Arange));
                  Copy_Range (Res_Ptr, Ptr, Rinfo);
                  Close_Temp;
               end;
            when Iir_Kind_Reverse_Range_Array_Attribute =>
               Translate_Reverse_Range_Ptr
                 (Res_Ptr,
                  Chap14.Translate_Range_Array_Attribute (Arange),
                  Range_Type);
            when Iir_Kind_Range_Expression =>
               Translate_Range_Expression_Ptr (Res_Ptr, Arange, Range_Type);
            when others =>
               Error_Kind ("translate_range_ptr", Arange);
         end case;
      end Translate_Range_Ptr;

      procedure Translate_Discrete_Range_Ptr (Res_Ptr : O_Lnode; Arange : Iir)
      is
      begin
         case Get_Kind (Arange) is
            when Iir_Kind_Integer_Subtype_Definition
              | Iir_Kind_Enumeration_Subtype_Definition =>
               if not Is_Anonymous_Type_Definition (Arange) then
                  declare
                     Ptr : O_Lnode;
                     Rinfo : Type_Info_Acc;
                  begin
                     Rinfo := Get_Info (Arange);
                     Open_Temp;
                     Ptr := Create_Temp_Ptr
                       (Rinfo.T.Range_Ptr_Type, Get_Var (Rinfo.T.Range_Var));
                     Copy_Range (Res_Ptr, Ptr, Rinfo);
                     Close_Temp;
                  end;
               else
                  Translate_Range_Ptr (Res_Ptr,
                                       Get_Range_Constraint (Arange),
                                       Get_Base_Type (Arange));
               end if;
            when Iir_Kind_Range_Array_Attribute
              | Iir_Kind_Reverse_Range_Array_Attribute
              | Iir_Kind_Range_Expression =>
               Translate_Range_Ptr (Res_Ptr, Arange, Get_Type (Arange));
            when others =>
               Error_Kind ("translate_discrete_range_ptr", Arange);
         end case;
      end Translate_Discrete_Range_Ptr;

      function Translate_Range (Arange : Iir; Range_Type : Iir)
        return O_Lnode is
      begin
         case Get_Kind (Arange) is
            when Iir_Kind_Range_Array_Attribute =>
               return Chap14.Translate_Range_Array_Attribute (Arange);
            when Iir_Kind_Reverse_Range_Array_Attribute =>
               declare
                  Res : O_Lnode;
                  Res_Ptr : O_Lnode;
                  Rinfo : Type_Info_Acc;
               begin
                  Rinfo := Get_Info (Range_Type);
                  Res := Create_Temp (Rinfo.T.Range_Type);
                  Open_Temp;
                  Res_Ptr := Create_Temp_Ptr (Rinfo.T.Range_Ptr_Type, Res);
                  Translate_Reverse_Range_Ptr
                    (Res_Ptr,
                     Chap14.Translate_Range_Array_Attribute (Arange),
                     Range_Type);
                  Close_Temp;
                  return Res;
               end;
            when Iir_Kind_Range_Expression =>
               declare
                  Res : O_Lnode;
                  Ptr : O_Lnode;
                  T_Info : Type_Info_Acc;
               begin
                  T_Info := Get_Info (Range_Type);
                  Res := Create_Temp (T_Info.T.Range_Type);
                  Open_Temp;
                  Ptr := Create_Temp_Ptr (T_Info.T.Range_Ptr_Type, Res);
                  Translate_Range_Expression_Ptr (Ptr, Arange, Range_Type);
                  Close_Temp;
                  return Res;
               end;
            when others =>
               Error_Kind ("translate_range", Arange);
         end case;
         return O_Lnode_Null;
      end Translate_Range;

      function Translate_Static_Range (Arange : Iir; Range_Type : Iir)
        return O_Enode
      is
         Constr : O_Record_Aggr_List;
         Res : O_Enode;
         T_Info : Type_Info_Acc;
      begin
         T_Info := Get_Info (Range_Type);
         Start_Record_Aggr (Constr, T_Info.T.Range_Type);
         New_Record_Aggr_El
           (Constr, Chap7.Translate_Static_Range_Left (Arange, Range_Type));
         New_Record_Aggr_El
           (Constr, Chap7.Translate_Static_Range_Right (Arange, Range_Type));
         New_Record_Aggr_El
           (Constr, Chap7.Translate_Static_Range_Dir (Arange));
         if T_Info.T.Range_Length /= O_Fnode_Null then
            New_Record_Aggr_El
              (Constr, Chap7.Translate_Static_Range_Length (Arange));
         end if;
         Finish_Record_Aggr (Constr, Res);
         return Res;
      end Translate_Static_Range;

      procedure Translate_Predefined_Array_Compare (Subprg : Iir)
      is
         procedure Gen_Compare (L, R : O_Lnode)
         is
            If_Blk1, If_Blk2 : O_If_Block;
         begin
            Start_If_Stmt
              (If_Blk1, New_Compare_Op (ON_Neq, New_Value (L), New_Value (R),
                                       Ghdl_Bool_Type_Node));
            Start_If_Stmt
              (If_Blk2, New_Compare_Op (ON_Gt, New_Value (L), New_Value (R),
                                        Ghdl_Bool_Type_Node));
            New_Return_Stmt (Ghdl_Compare_Gt);
            New_Else_Stmt (If_Blk2);
            New_Return_Stmt (Ghdl_Compare_Lt);
            Finish_If_Stmt (If_Blk2);
            Finish_If_Stmt (If_Blk1);
         end Gen_Compare;

         F_Info : Subprg_Info_Acc;
         Arr_Type : Iir_Array_Type_Definition;
         Arr_Ptr_Type : O_Tnode;
         Info : Type_Info_Acc;
         Id : Name_Id;
         L, R : O_Lnode;
         Interface_List : O_Inter_List;
         If_Blk : O_If_Block;
         Var_L_Len, Var_R_Len : O_Lnode;
         Var_L_El, Var_R_El : O_Lnode;
         Var_I, Var_Len : O_Lnode;
         Loop_Blk : O_Loop_Block;
         Label : O_Snode;
         El_Otype : O_Tnode;
      begin
         Arr_Type := Get_Type
           (Get_First_Element (Get_Interface_Declaration (Subprg)));
         Info := Get_Info (Arr_Type);
         Id := Get_Identifier (Get_Type_Declarator (Arr_Type));
         Arr_Ptr_Type := Info.Ortho_Ptr_Type (Mode_Value);

         F_Info := Add_Info (Subprg, Kind_Subprg);
         Chap2.Clear_Instance_Data (F_Info.Subprg_Instance);
         F_Info.Subprg_Overload := 0;

         --  Create function.
         Start_Function_Decl (Interface_List, Create_Identifier (Id, "_CMP"),
                              Global_Storage, Ghdl_Compare_Type);
         New_Interface_Decl (Interface_List, L, Wki_Left, Arr_Ptr_Type);
         New_Interface_Decl (Interface_List, R, Wki_Right, Arr_Ptr_Type);
         Finish_Subprogram_Decl (Interface_List, F_Info.Ortho_Func);

         if Global_Storage = O_Storage_External then
            return;
         end if;

         El_Otype := Get_Ortho_Type
           (Get_Element_Subtype (Arr_Type), Mode_Value);
         Start_Subprogram_Body (F_Info.Ortho_Func);
         --  Compute length of L and R.
         New_Var_Decl (Var_L_Len, Get_Identifier ("l_len"),
                       O_Storage_Local, Ghdl_Index_Type);
         New_Var_Decl (Var_R_Len, Get_Identifier ("r_len"),
                       O_Storage_Local, Ghdl_Index_Type);
         New_Var_Decl (Var_Len, Wki_Length, O_Storage_Local, Ghdl_Index_Type);
         New_Var_Decl (Var_I, Wki_I, O_Storage_Local, Ghdl_Index_Type);
         New_Assign_Stmt
           (Var_L_Len,
            Chap6.Get_Array_Ptr_Bound_Length (L, Arr_Type, 1, Mode_Value));
         New_Assign_Stmt
           (Var_R_Len,
            Chap6.Get_Array_Ptr_Bound_Length (R, Arr_Type, 1, Mode_Value));
         --  Find the minimum length.
         Start_If_Stmt (If_Blk,
                        New_Compare_Op (ON_Ge,
                                        New_Value (Var_L_Len),
                                        New_Value (Var_R_Len),
                                        Ghdl_Bool_Type_Node));
         New_Assign_Stmt (Var_Len, New_Value (Var_R_Len));
         New_Else_Stmt (If_Blk);
         New_Assign_Stmt (Var_Len, New_Value (Var_L_Len));
         Finish_If_Stmt (If_Blk);

         --  for each element, compare elements; if not equal return the
         --       comparaison result.
         Init_Var (Var_I);
         Start_Loop_Stmt (Loop_Blk, Label);
         Start_If_Stmt (If_Blk, New_Compare_Op (ON_Ge,
                                                New_Value (Var_I),
                                                New_Value (Var_Len),
                                                Ghdl_Bool_Type_Node));
         --  Compare the length and return the result.
         Gen_Compare (Var_L_Len, Var_R_Len);
         New_Return_Stmt (Ghdl_Compare_Eq);
         Finish_If_Stmt (If_Blk);
         Start_Declare_Stmt;
         New_Var_Decl (Var_L_El, Get_Identifier ("l_el"), O_Storage_Local,
                       El_Otype);
         New_Var_Decl (Var_R_El, Get_Identifier ("r_el"), O_Storage_Local,
                       El_Otype);
         New_Assign_Stmt
           (Var_L_El,
            New_Value
            (New_Indexed_Element
             (New_Acc_Value (Chap3.Get_Array_Ptr_Base_Ptr
                             (L, Arr_Type, Mode_Value)),
              New_Value (Var_I))));
         New_Assign_Stmt
           (Var_R_El,
            New_Value
            (New_Indexed_Element
             (New_Acc_Value (Chap3.Get_Array_Ptr_Base_Ptr
                             (R, Arr_Type, Mode_Value)),
              New_Value (Var_I))));
         Gen_Compare (Var_L_El, Var_R_El);
         Finish_Declare_Stmt;
         Inc_Var (Var_I);
         Finish_Loop_Stmt (Loop_Blk);
         Finish_Subprogram_Body;
      end Translate_Predefined_Array_Compare;

      --  Find the declaration of the predefined function IMP in type
      --  definition BASE_TYPE.
      function Find_Predefined_Function
        (Base_Type : Iir; Imp : Iir_Predefined_Functions)
        return Iir
      is
         El : Iir;
         List : Iir_List;
      begin
         List := Get_Implicit_Functions_List (Get_Type_Declarator (Base_Type));
         for I in Natural loop
            El := Get_Nth_Element (List, I);
            exit when El = Null_Iir;
            if Get_Implicit_Definition (El) = Imp then
               return El;
            end if;
         end loop;
         raise Internal_Error;
      end Find_Predefined_Function;

      function Translate_Equality (L, R : Mnode; Etype : Iir)
        return O_Enode
      is
         Tinfo : Type_Info_Acc;
      begin
         Tinfo := Get_Type_Info (L);
         case Tinfo.Type_Mode is
            when Type_Mode_Scalar
              | Type_Mode_Acc =>
               return New_Compare_Op (ON_Eq, M2E (L), M2E (R),
                                      Ghdl_Bool_Type_Node);
            when Type_Mode_Fat_Acc =>
               --  a fat pointer.
               declare
                  B : Type_Info_Acc;
                  Ln, Rn : Mnode;
                  V1, V2 : O_Enode;
               begin
                  B := Get_Info (Get_Designated_Type (Etype));
                  Ln := Stabilize (L);
                  Rn := Stabilize (R);
                  V1 := New_Compare_Op
                    (ON_Eq,
                     New_Value (New_Selected_Element
                                (M2Lv (Ln), B.T.Base_Field (Mode_Value))),
                     New_Value (New_Selected_Element
                                (M2Lv (Rn), B.T.Base_Field (Mode_Value))),
                     Std_Boolean_Type_Node);
                  V2 := New_Compare_Op
                    (ON_Eq,
                     New_Value (New_Selected_Element
                                (M2Lv (Ln), B.T.Bounds_Field (Mode_Value))),
                     New_Value (New_Selected_Element
                                (M2Lv (Rn), B.T.Bounds_Field (Mode_Value))),
                     Std_Boolean_Type_Node);
                  return New_Dyadic_Op (ON_And, V1, V2);
               end;

            when Type_Mode_Array
              | Type_Mode_Ptr_Array =>
               declare
                  Lc, Rc : O_Enode;
                  Base_Type : Iir_Array_Type_Definition;
                  Func : Iir;
               begin
                  Base_Type := Get_Base_Type (Etype);
                  Lc := Translate_Implicit_Conv
                    (M2E (L), Etype, Base_Type, Mode_Value);
                  Rc := Translate_Implicit_Conv
                    (M2E (R), Etype, Base_Type, Mode_Value);
                  Func := Find_Predefined_Function
                    (Base_Type, Iir_Predefined_Array_Equality);
                  return Translate_Predefined_Lib_Operator (Lc, Rc, Func);
               end;

            when Type_Mode_Record =>
               declare
                  Func : Iir;
               begin
                  Func := Find_Predefined_Function
                    (Etype, Iir_Predefined_Record_Equality);
                  return Translate_Predefined_Lib_Operator
                    (M2E (L), M2E (R), Func);
               end;

            when Type_Mode_Unknown
              | Type_Mode_File
              | Type_Mode_Fat_Array =>
               raise Internal_Error;
         end case;
      end Translate_Equality;

      procedure Translate_Predefined_Array_Equality (Subprg : Iir; Block : Iir)
      is
         F_Info : Subprg_Info_Acc;
         Arr_Type : Iir_Array_Type_Definition;
         Arr_Ptr_Type : O_Tnode;
         Info : Type_Info_Acc;
         Id : Name_Id;
         Var_L, Var_R : O_Lnode;
         L, R : Mnode;
         Interface_List : O_Inter_List;
         Indexes : Iir_List;
         Nbr_Indexes : Natural;
         If_Blk : O_If_Block;
         Var_I : O_Lnode;
         Var_Len : O_Lnode;
         Loop_Blk : O_Loop_Block;
         Label : O_Snode;
         Le, Re : Mnode;
         El_Type : Iir;
      begin
         Arr_Type := Get_Type
           (Get_First_Element (Get_Interface_Declaration (Subprg)));
         El_Type := Get_Element_Subtype (Arr_Type);
         Info := Get_Info (Arr_Type);
         Id := Get_Identifier (Get_Type_Declarator (Arr_Type));
         Arr_Ptr_Type := Info.Ortho_Ptr_Type (Mode_Value);

         F_Info := Add_Info (Subprg, Kind_Subprg);
         F_Info.Subprg_Overload := 0;

         --  Create function.
         Start_Function_Decl (Interface_List, Create_Identifier (Id, "_EQ"),
                              Global_Storage, Std_Boolean_Type_Node);
         Chap2.Create_Subprg_Instance (Interface_List, Subprg, Block);
         New_Interface_Decl (Interface_List, Var_L, Wki_Left, Arr_Ptr_Type);
         New_Interface_Decl (Interface_List, Var_R, Wki_Right, Arr_Ptr_Type);
         Finish_Subprogram_Decl (Interface_List, F_Info.Ortho_Func);

         if Global_Storage = O_Storage_External then
            return;
         end if;

         L := Vp2M (Var_L, Info, Mode_Value);
         R := Vp2M (Var_R, Info, Mode_Value);

         Indexes := Get_Index_List (Arr_Type);
         Nbr_Indexes := Get_Nbr_Elements (Indexes);

         Start_Subprogram_Body (F_Info.Ortho_Func);
         Chap2.Start_Subprg_Instance_Use (Subprg);
         --  for each dimension:  if length mismatch: return false
         for I in 1 .. Nbr_Indexes loop
            Start_If_Stmt
              (If_Blk,
               New_Compare_Op
               (ON_Neq,
                M2E (Chap3.Range_To_Length
                     (Chap3.Get_Array_Range (L, Arr_Type, I))),
                M2E (Chap3.Range_To_Length
                     (Chap3.Get_Array_Range (R, Arr_Type, I))),
                Std_Boolean_Type_Node));
            New_Return_Stmt (Std_Boolean_False_Node);
            Finish_If_Stmt (If_Blk);
         end loop;

         --  for each element: if element is not equal, return false
         New_Var_Decl (Var_I, Wki_I, O_Storage_Local, Ghdl_Index_Type);
         New_Var_Decl (Var_Len, Wki_Length, O_Storage_Local, Ghdl_Index_Type);
         New_Assign_Stmt (Var_Len, Chap3.Get_Array_Length (L, Arr_Type));
         Init_Var (Var_I);
         Start_Loop_Stmt (Loop_Blk, Label);
         --  If the end of the array is reached, return TRUE.
         Start_If_Stmt (If_Blk,
                        New_Compare_Op (ON_Ge,
                                        New_Value (Var_I),
                                        New_Value (Var_Len),
                                        Ghdl_Bool_Type_Node));
         New_Return_Stmt (Std_Boolean_True_Node);
         Finish_If_Stmt (If_Blk);
         Open_Temp;
         Le := Chap3.Index_Base (Chap3.Get_Array_Base (L), Arr_Type,
                                 New_Value (Var_I));
         Re := Chap3.Index_Base (Chap3.Get_Array_Base (R), Arr_Type,
                                 New_Value (Var_I));
         Start_If_Stmt
           (If_Blk,
            New_Monadic_Op (ON_Not, Translate_Equality (Le, Re, El_Type)));
         New_Return_Stmt (Std_Boolean_False_Node);
         Finish_If_Stmt (If_Blk);
         Close_Temp;
         Inc_Var (Var_I);
         Finish_Loop_Stmt (Loop_Blk);
         Chap2.Finish_Subprg_Instance_Use (Subprg);
         Finish_Subprogram_Body;
      end Translate_Predefined_Array_Equality;

      procedure Translate_Predefined_Record_Equality
        (Subprg : Iir; Block : Iir)
      is
         F_Info : Subprg_Info_Acc;
         Rec_Type : Iir_Record_Type_Definition;
         Rec_Ptr_Type : O_Tnode;
         Info : Type_Info_Acc;
         Id : Name_Id;
         Var_L, Var_R : O_Lnode;
         L, R : Mnode;
         Interface_List : O_Inter_List;
         If_Blk : O_If_Block;
         Le, Re : Mnode;
         List : Iir_List;

         El : Iir_Element_Declaration;
      begin
         Rec_Type := Get_Type
           (Get_First_Element (Get_Interface_Declaration (Subprg)));
         Info := Get_Info (Rec_Type);
         Id := Get_Identifier (Get_Type_Declarator (Rec_Type));
         Rec_Ptr_Type := Info.Ortho_Ptr_Type (Mode_Value);

         F_Info := Add_Info (Subprg, Kind_Subprg);
         Chap2.Clear_Instance_Data (F_Info.Subprg_Instance);
         F_Info.Subprg_Overload := 0;

         --  Create function.
         Start_Function_Decl (Interface_List, Create_Identifier (Id, "_EQ"),
                              Global_Storage, Std_Boolean_Type_Node);
         Chap2.Create_Subprg_Instance (Interface_List, Subprg, Block);
         New_Interface_Decl (Interface_List, Var_L, Wki_Left, Rec_Ptr_Type);
         New_Interface_Decl (Interface_List, Var_R, Wki_Right, Rec_Ptr_Type);
         Finish_Subprogram_Decl (Interface_List, F_Info.Ortho_Func);

         if Global_Storage = O_Storage_External then
            return;
         end if;

         Start_Subprogram_Body (F_Info.Ortho_Func);
         Chap2.Start_Subprg_Instance_Use (Subprg);

         L := Vp2M (Var_L, Info, Mode_Value);
         R := Vp2M (Var_R, Info, Mode_Value);

         --   Compare each element.
         List := Get_Element_Declaration_List (Rec_Type);
         for I in Natural loop
            El := Get_Nth_Element (List, I);
            exit when El = Null_Iir;

            Le := Chap6.Translate_Selected_Element (L, El);
            Re := Chap6.Translate_Selected_Element (R, El);

            Open_Temp;
            Start_If_Stmt
              (If_Blk,
               New_Monadic_Op (ON_Not,
                               Translate_Equality (Le, Re, Get_Type (El))));
            New_Return_Stmt (Std_Boolean_False_Node);
            Finish_If_Stmt (If_Blk);
            Close_Temp;
         end loop;
         New_Return_Stmt (Std_Boolean_True_Node);
         Chap2.Finish_Subprg_Instance_Use (Subprg);
         Finish_Subprogram_Body;
      end Translate_Predefined_Record_Equality;

      procedure Translate_Predefined_Array_Array_Concat
        (Subprg : Iir; Block : Iir)
      is
         F_Info : Subprg_Info_Acc;
         Arr_Type : Iir_Array_Type_Definition;
         Arr_Ptr_Type : O_Tnode;

         --  Info for the array type.
         Info : Type_Info_Acc;

         --  Info for the index type.
         Iinfo : Type_Info_Acc;
         Index_Type : Iir;

         Index_Otype : O_Tnode;
         Id : Name_Id;
         Interface_List : O_Inter_List;
         Var_Res, Var_L, Var_R : O_Lnode;
         Res, L, R : Mnode;
         Var_Length, Var_L_Len, Var_R_Len : O_Lnode;
         Var_Bounds, Var_Right : O_Lnode;
         V_Bounds : Mnode;
         If_Blk : O_If_Block;
      begin
         Arr_Type := Get_Return_Type (Subprg);
         Info := Get_Info (Arr_Type);
         Id := Get_Identifier (Get_Type_Declarator (Arr_Type));
         Arr_Ptr_Type := Info.Ortho_Ptr_Type (Mode_Value);

         F_Info := Add_Info (Subprg, Kind_Subprg);
         F_Info.Subprg_Overload := 0;
         F_Info.Use_Stack2 := True;

         --  Create function.
         Start_Procedure_Decl
           (Interface_List, Create_Identifier (Id, "_CONCAT"), Global_Storage);
         --  Note: contrary to user function which returns composite value
         --  via a result record, a concatenation returns its value without
         --  the use of the record.
         Chap2.Create_Subprg_Instance (Interface_List, Subprg, Block);
         New_Interface_Decl (Interface_List, Var_Res, Wki_Res, Arr_Ptr_Type);
         New_Interface_Decl (Interface_List, Var_L, Wki_Left, Arr_Ptr_Type);
         New_Interface_Decl (Interface_List, Var_R, Wki_Right, Arr_Ptr_Type);
         Finish_Subprogram_Decl (Interface_List, F_Info.Ortho_Func);

         if Global_Storage = O_Storage_External then
            return;
         end if;

         Index_Type := Get_First_Element (Get_Index_List (Arr_Type));
         Iinfo := Get_Info (Index_Type);
         Index_Otype := Iinfo.Ortho_Type (Mode_Value);

         Start_Subprogram_Body (F_Info.Ortho_Func);
         Chap2.Start_Subprg_Instance_Use (Subprg);
         New_Var_Decl (Var_Length, Wki_Length, O_Storage_Local,
                       Ghdl_Index_Type);
         New_Var_Decl (Var_L_Len, Get_Identifier ("l_len"), O_Storage_Local,
                       Ghdl_Index_Type);
         New_Var_Decl (Var_R_Len, Get_Identifier ("r_len"), O_Storage_Local,
                       Ghdl_Index_Type);
         New_Var_Decl (Var_Bounds, Get_Identifier ("bounds"), O_Storage_Local,
                       Info.T.Bounds_Ptr_Type);

         L := Vp2M (Var_L, Info, Mode_Value);
         R := Vp2M (Var_R, Info, Mode_Value);
         Res := Vp2M (Var_Res, Info, Mode_Value);
         V_Bounds := Vp2M (Var_Bounds, Info, Mode_Value,
                           Info.T.Bounds_Type, Info.T.Bounds_Ptr_Type);

         --  Compute length.
         New_Assign_Stmt
           (Var_L_Len, Chap3.Get_Array_Length (L, Arr_Type));
         New_Assign_Stmt
           (Var_R_Len, Chap3.Get_Array_Length (R, Arr_Type));
         New_Assign_Stmt
           (Var_Length,
            New_Dyadic_Op (ON_Add_Ov,
                           New_Value (Var_L_Len), New_Value (Var_R_Len)));

         --  Check case where the result is the right operand.
         declare
            Len : O_Enode;
         begin
            if Flags.Vhdl_Std = Vhdl_87 then
               --  LRM87 7.2.4
               --  [...], unless the left operand is a null array, in which
               --  case the result of the concatenation is the right operand.
               Len := New_Value (Var_L_Len);

            else
               --  LRM93 7.2.4
               --  If both operands are null arrays, then the result of the
               --  concatenation is the right operand.
               --  GHDL: since the length type is unsigned, then both operands
               --   are null arrays iff the result is a null array.
               Len := New_Value (Var_Length);
            end if;

            Start_If_Stmt
              (If_Blk,
               New_Compare_Op (ON_Eq,
                               Len,
                               New_Unsigned_Literal (Ghdl_Index_Type, 0),
                               Ghdl_Bool_Type_Node));
            Copy_Fat_Pointer (Res, R);
            New_Return_Stmt;
            Finish_If_Stmt (If_Blk);
         end;

         --  Allocate bounds.
         New_Assign_Stmt
           (Var_Bounds,
            Gen_Alloc (Alloc_Return,
                       New_Sizeof (Info.T.Bounds_Type, Ghdl_Index_Type),
                       Info.T.Bounds_Ptr_Type));
         New_Assign_Stmt
           (M2Lp (Chap3.Get_Array_Bounds (Res)), New_Value (Var_Bounds));

         --  Set bound.
         if Flags.Vhdl_Std = Vhdl_87 then
            --  Set length.
            New_Assign_Stmt
              (M2Lv (Chap3.Range_To_Length
                     (Chap3.Bounds_To_Range (V_Bounds, Arr_Type, 1))),
               New_Value (Var_Length));

            --  Set direction, left bound and right bound.
            --  LRM87 7.2.4
            --  The left bound of this result is the left bound of the left
            --  operand, unless the left operand is a null array, in which
            --  case the result of the concatenation is the right operand.
            --  The direction of the result is the direction of the left
            --  operand, unless the left operand is a null array, in which
            --  case the direction of the result is that of the right operand.
            declare
               Var_Dir, Var_Left : O_Lnode;
               Var_Length1 : O_Lnode;
            begin
               Start_Declare_Stmt;
               New_Var_Decl (Var_Right, Get_Identifier ("right_bound"),
                             O_Storage_Local, Index_Otype);
               New_Var_Decl (Var_Dir, Wki_Dir, O_Storage_Local,
                             Ghdl_Dir_Type_Node);
               New_Var_Decl (Var_Left, Get_Identifier ("left_bound"),
                             O_Storage_Local, Iinfo.Ortho_Type (Mode_Value));
               New_Var_Decl (Var_Length1, Get_Identifier ("length_1"),
                             O_Storage_Local, Ghdl_Index_Type);
               New_Assign_Stmt
                 (Var_Dir,
                  M2E (Chap3.Range_To_Dir
                       (Chap3.Get_Array_Range (L, Arr_Type, 1))));
               New_Assign_Stmt
                 (M2Lv (Chap3.Range_To_Dir
                        (Chap3.Bounds_To_Range (V_Bounds, Arr_Type, 1))),
                  New_Value (Var_Dir));
               New_Assign_Stmt
                 (Var_Left,
                  M2E (Chap3.Range_To_Left
                       (Chap3.Get_Array_Range (L, Arr_Type, 1))));
               --  Note this substraction cannot overflow, since LENGTH >= 1.
               New_Assign_Stmt
                 (Var_Length1,
                  New_Dyadic_Op (ON_Sub_Ov,
                                 New_Value (Var_Length),
                                 New_Unsigned_Literal (Ghdl_Index_Type, 1)));
               New_Assign_Stmt
                 (M2Lv (Chap3.Range_To_Left
                        (Chap3.Bounds_To_Range (V_Bounds, Arr_Type, 1))),
                  New_Value (Var_Left));
               Start_If_Stmt
                 (If_Blk,
                  New_Compare_Op (ON_Eq, New_Value (Var_Dir),
                                  Ghdl_Dir_To_Node, Ghdl_Bool_Type_Node));
               New_Assign_Stmt
                 (Var_Right,
                  New_Dyadic_Op (ON_Add_Ov,
                                 New_Value (Var_Left),
                                 New_Convert_Ov (New_Value (Var_Length1),
                                                 Index_Otype)));
               New_Else_Stmt (If_Blk);
               New_Assign_Stmt
                 (Var_Right,
                  New_Dyadic_Op (ON_Sub_Ov,
                                 New_Value (Var_Left),
                                 New_Convert_Ov (New_Value (Var_Length1),
                                                 Index_Otype)));
               Finish_If_Stmt (If_Blk);
               --   Check the right bounds is inside the bounds of the
               --   index type.
               Chap3.Check_Range (Var_Right, Null_Iir, Index_Type);
               New_Assign_Stmt
                 (M2Lv (Chap3.Range_To_Right
                        (Chap3.Bounds_To_Range (V_Bounds, Arr_Type, 1))),
                  New_Value (Var_Right));
               Finish_Declare_Stmt;
            end;
         else
            --  LRM93 7.2.4
            --  [...], the direction and bounds of the result are determined
            --  as follows: Let S be the index subtype of the base type of the
            --  result.  The direction of the result of the concatenation is
            --  the direction of S, and the left bound of the result is
            --  S'LEFT.
            declare
               Var_Range_Ptr : O_Lnode;
            begin
               Start_Declare_Stmt;
               New_Var_Decl (Var_Range_Ptr, Get_Identifier ("range_ptr"),
                             O_Storage_Local, Iinfo.T.Range_Ptr_Type);
               New_Assign_Stmt
                 (Var_Range_Ptr,
                  M2Addr (Chap3.Bounds_To_Range (V_Bounds, Arr_Type, 1)));
               Chap3.Create_Range_From_Length
                 (Index_Type, Var_Length, Var_Range_Ptr);
               Finish_Declare_Stmt;
            end;
         end if;

         --  Allocate array base.
         Chap3.Allocate_Fat_Array_Base
           (Alloc_Return, Res, Arr_Type, Mode_Value);

         --  Copy left.
         declare
            V_Arr : O_Lnode;
            Var_Arr : Mnode;
         begin
            Open_Temp;
            V_Arr := Create_Temp (Info.Ortho_Type (Mode_Value));
            Var_Arr := Vv2M (V_Arr, Info, Mode_Value);
            New_Assign_Stmt
              (M2Lp (Chap3.Get_Array_Bounds (Var_Arr)),
               M2Addr (Chap3.Get_Array_Bounds (L)));
            New_Assign_Stmt
              (M2Lp (Chap3.Get_Array_Base (Var_Arr)),
               M2Addr (Chap3.Get_Array_Base (Res)));
            Chap3.Translate_Object_Copy
              (Var_Arr, New_Value (Var_L), Arr_Type, Mode_Value);
            Close_Temp;
         end;

         --  Copy right.
         declare
            V_Arr : O_Lnode;
            Var_Arr : Mnode;
         begin
            Open_Temp;
            V_Arr := Create_Temp (Info.Ortho_Type (Mode_Value));
            Var_Arr := Vv2M (V_Arr, Info, Mode_Value);
            New_Assign_Stmt
              (M2Lp (Chap3.Get_Array_Bounds (Var_Arr)),
               M2Addr (Chap3.Get_Array_Bounds (R)));
            New_Assign_Stmt
              (M2Lp (Chap3.Get_Array_Base (Var_Arr)),
               New_Address (New_Slice (M2Lv (Chap3.Get_Array_Base (Res)),
                                       Info.T.Base_Type (Mode_Value),
                                       New_Value (Var_L_Len)),
                            Info.T.Base_Ptr_Type (Mode_Value)));
            Chap3.Translate_Object_Copy
              (Var_Arr, New_Value (Var_R), Arr_Type, Mode_Value);
            Close_Temp;
         end;
         Chap2.Finish_Subprg_Instance_Use (Subprg);
         Finish_Subprogram_Body;
      end Translate_Predefined_Array_Array_Concat;

      procedure Translate_Predefined_Array_Logical (Subprg : Iir)
      is
         F_Info : Subprg_Info_Acc;
         Arr_Type : Iir_Array_Type_Definition;
         Arr_Ptr_Type : O_Tnode;
         --  Info for the array type.
         Info : Type_Info_Acc;
         Id : Name_Id;
         Interface_List : O_Inter_List;
         Var_Res : O_Lnode;
         Res : Mnode;
         L, R : O_Lnode;
         Var_Length, Var_I : O_Lnode;
         Var_Base, Var_L_Base, Var_R_Base : O_Lnode;
         If_Blk : O_If_Block;
         Loop_Blk : O_Loop_Block;
         Label : O_Snode;
         Name : O_Ident;
         Is_Monadic : Boolean;
         El, L_El : O_Enode;
         Op : ON_Op_Kind;
         Do_Invert : Boolean;
      begin
         Arr_Type := Get_Type
           (Get_First_Element (Get_Interface_Declaration (Subprg)));
         Info := Get_Info (Arr_Type);
         Id := Get_Identifier (Get_Type_Declarator (Arr_Type));
         Arr_Ptr_Type := Info.Ortho_Ptr_Type (Mode_Value);

         F_Info := Add_Info (Subprg, Kind_Subprg);
         Chap2.Clear_Instance_Data (F_Info.Subprg_Instance);
         F_Info.Subprg_Overload := 0;
         F_Info.Use_Stack2 := True;

         Is_Monadic := False;
         case Get_Implicit_Definition (Subprg) is
            when Iir_Predefined_Bit_Array_And
              | Iir_Predefined_Boolean_Array_And =>
               Name := Create_Identifier (Id, "_AND");
               Op := ON_And;
               Do_Invert := False;
            when Iir_Predefined_Bit_Array_Or
              | Iir_Predefined_Boolean_Array_Or =>
               Name := Create_Identifier (Id, "_OR");
               Op := ON_Or;
               Do_Invert := False;
            when Iir_Predefined_Bit_Array_Nand
              | Iir_Predefined_Boolean_Array_Nand =>
               Name := Create_Identifier (Id, "_NAND");
               Op := ON_And;
               Do_Invert := True;
            when Iir_Predefined_Bit_Array_Nor
              | Iir_Predefined_Boolean_Array_Nor =>
               Name := Create_Identifier (Id, "_NOR");
               Op := ON_Or;
               Do_Invert := True;
            when Iir_Predefined_Bit_Array_Xor
              | Iir_Predefined_Boolean_Array_Xor =>
               Name := Create_Identifier (Id, "_XOR");
               Op := ON_Xor;
               Do_Invert := False;
            when Iir_Predefined_Bit_Array_Xnor
              | Iir_Predefined_Boolean_Array_Xnor =>
               Name := Create_Identifier (Id, "_XNOR");
               Op := ON_Xor;
               Do_Invert := True;
            when Iir_Predefined_Bit_Array_Not
              | Iir_Predefined_Boolean_Array_Not =>
               Name := Create_Identifier (Id, "_NOT");
               Is_Monadic := True;
               Op := ON_Not;
               Do_Invert := False;
            when others =>
               raise Internal_Error;
         end case;

         --  Create function.
         Start_Procedure_Decl (Interface_List, Name, Global_Storage);
         --  Note: contrary to user function which returns composite value
         --  via a result record, a concatenation returns its value without
         --  the use of the record.
         New_Interface_Decl (Interface_List, Var_Res, Wki_Res, Arr_Ptr_Type);
         New_Interface_Decl (Interface_List, L, Wki_Left, Arr_Ptr_Type);
         if not Is_Monadic then
            New_Interface_Decl (Interface_List, R, Wki_Right, Arr_Ptr_Type);
         end if;
         Finish_Subprogram_Decl (Interface_List, F_Info.Ortho_Func);

         if Global_Storage = O_Storage_External then
            return;
         end if;

         Start_Subprogram_Body (F_Info.Ortho_Func);
         New_Var_Decl (Var_Length, Wki_Length, O_Storage_Local,
                       Ghdl_Index_Type);
         New_Var_Decl (Var_I, Wki_I, O_Storage_Local, Ghdl_Index_Type);
         New_Var_Decl (Var_Base, Get_Identifier ("base"), O_Storage_Local,
                       Info.T.Base_Ptr_Type (Mode_Value));
         New_Var_Decl (Var_L_Base, Get_Identifier ("l_base"), O_Storage_Local,
                       Info.T.Base_Ptr_Type (Mode_Value));
         if not Is_Monadic then
            New_Var_Decl
              (Var_R_Base, Get_Identifier ("r_base"), O_Storage_Local,
               Info.T.Base_Ptr_Type (Mode_Value));
         end if;
         Open_Temp;
         --  Get length of LEFT.
         New_Assign_Stmt
           (Var_Length,
            Chap6.Get_Array_Ptr_Bound_Length (L, Arr_Type, 1, Mode_Value));
         --  If dyadic, check RIGHT has the same length.
         if not Is_Monadic then
            Chap6.Check_Bound_Error
              (New_Compare_Op
               (ON_Neq,
                New_Value (Var_Length),
                Chap6.Get_Array_Ptr_Bound_Length (R, Arr_Type, 1, Mode_Value),
                Ghdl_Bool_Type_Node),
               Subprg, 0);
         end if;

         --  Create the result from LEFT bound.
         Res := Chap3.Translate_Object_Allocation
           (Alloc_Return, Arr_Type,
            Chap3.Get_Array_Ptr_Bounds_Ptr (L, Arr_Type, Mode_Value),
            Mode_Value);
         Copy_Fat_Pointer (Vp2M (Var_Res, Info, Mode_Value), Res);
         New_Assign_Stmt
           (Var_Base, M2Addr (Chap3.Get_Array_Base (Res)));
         New_Assign_Stmt
           (Var_L_Base,
            New_Value (Chap3.Get_Array_Ptr_Base_Ptr
                       (L, Arr_Type, Mode_Value)));
         if not Is_Monadic then
            New_Assign_Stmt
              (Var_R_Base,
               New_Value (Chap3.Get_Array_Ptr_Base_Ptr
                          (R, Arr_Type, Mode_Value)));
         end if;

         --  Do the logical operation on each element.
         Init_Var (Var_I);
         Start_Loop_Stmt (Loop_Blk, Label);
         Start_If_Stmt (If_Blk,
                        New_Compare_Op (ON_Ge,
                                        New_Value (Var_I),
                                        New_Value (Var_Length),
                                        Ghdl_Bool_Type_Node));
         New_Return_Stmt;
         Finish_If_Stmt (If_Blk);
         L_El := New_Value (New_Indexed_Element (New_Acc_Value (Var_L_Base),
                                                 New_Value (Var_I)));
         if Is_Monadic then
            El := New_Monadic_Op (Op, L_El);
         else
            El := New_Dyadic_Op
              (Op, L_El,
               New_Value (New_Indexed_Element (New_Acc_Value (Var_R_Base),
                                               New_Value (Var_I))));
         end if;
         if Do_Invert then
            El := New_Monadic_Op (ON_Not, El);
         end if;

         New_Assign_Stmt (New_Indexed_Element (New_Acc_Value (Var_Base),
                                               New_Value (Var_I)),
                          El);
         Inc_Var (Var_I);
         Finish_Loop_Stmt (Loop_Blk);
         Close_Temp;
         Finish_Subprogram_Body;
      end Translate_Predefined_Array_Logical;

      procedure Translate_File_Subprogram (Ftype : Iir; Block : Iir)
      is
         Etype : Iir;
         Tinfo : Type_Info_Acc;
      begin
         Etype := Get_Type_Mark (Ftype);
         Tinfo := Get_Info (Etype);
         case Tinfo.Type_Mode is
            when Type_Mode_Scalar =>
               --  Intrinsic.
               null;
            when others =>
               Error_Kind ("translate_file_subprogram", Etype);
         end case;
      end Translate_File_Subprogram;

      procedure Translate_Implicit_Functions_List (List : Iir_List;
                                                   Block : Iir)
      is
         Subprg : Iir;
         Kind : Iir_Predefined_Functions;
         Arr_Eq_Info : Subprg_Info_Acc;
         Rec_Eq_Info : Subprg_Info_Acc;
         Arr_Cmp_Info : Subprg_Info_Acc;
         Arr_Concat_Info : Subprg_Info_Acc;
      begin
         if List = Null_Iir_List then
            return;
         end if;

         --  Be independant of declaration order since the same subprogram
         --  may be used for several implicit operators (eg. array comparaison)
         Arr_Eq_Info := null;
         Arr_Cmp_Info := null;
         Arr_Concat_Info := null;
         Rec_Eq_Info := null;

         for I in Natural loop
            Subprg := Get_Nth_Element (List, I);
            exit when Subprg = Null_Iir;

            Kind := Get_Implicit_Definition (Subprg);
            if Predefined_To_Onop (Kind) = ON_Nil then
               --  Maybe not intrinsic.

               case Kind is
                  when Iir_Predefined_Access_Equality
                    | Iir_Predefined_Access_Inequality =>
                     --  Intrinsic.
                     null;
                  when Iir_Predefined_Deallocate =>
                     --  Intrinsic.
                     null;
                  when Iir_Predefined_Integer_Identity
                    | Iir_Predefined_Integer_Exp =>
                     --  Intrinsic.
                     null;

                  when Iir_Predefined_Record_Equality
                    | Iir_Predefined_Record_Inequality =>
                     if Rec_Eq_Info = null then
                        Translate_Predefined_Record_Equality (Subprg, Block);
                        Rec_Eq_Info := Get_Info (Subprg);
                     else
                        Set_Info (Subprg, Rec_Eq_Info);
                     end if;

                  when Iir_Predefined_Array_Equality
                    | Iir_Predefined_Array_Inequality =>
                     if Arr_Eq_Info = null then
                        Translate_Predefined_Array_Equality (Subprg, Block);
                        Arr_Eq_Info := Get_Info (Subprg);
                     else
                        Set_Info (Subprg, Arr_Eq_Info);
                     end if;

                  when Iir_Predefined_Array_Greater
                    | Iir_Predefined_Array_Greater_Equal
                    | Iir_Predefined_Array_Less
                    | Iir_Predefined_Array_Less_Equal =>
                     if Arr_Cmp_Info = null then
                        Translate_Predefined_Array_Compare (Subprg);
                        Arr_Cmp_Info := Get_Info (Subprg);
                     else
                        Set_Info (Subprg, Arr_Cmp_Info);
                     end if;

                  when Iir_Predefined_Array_Array_Concat
                    | Iir_Predefined_Array_Element_Concat
                    | Iir_Predefined_Element_Array_Concat
                    | Iir_Predefined_Element_Element_Concat =>
                     if Arr_Concat_Info = null then
                        Translate_Predefined_Array_Array_Concat
                          (Subprg, Block);
                        Arr_Concat_Info := Get_Info (Subprg);
                     else
                        Set_Info (Subprg, Arr_Concat_Info);
                     end if;

                  when Iir_Predefined_Bit_Array_And
                    | Iir_Predefined_Bit_Array_Or
                    | Iir_Predefined_Bit_Array_Nand
                    | Iir_Predefined_Bit_Array_Nor
                    | Iir_Predefined_Bit_Array_Xor
                    | Iir_Predefined_Bit_Array_Xnor
                    | Iir_Predefined_Bit_Array_Not
                    | Iir_Predefined_Boolean_Array_And
                    | Iir_Predefined_Boolean_Array_Or
                    | Iir_Predefined_Boolean_Array_Nand
                    | Iir_Predefined_Boolean_Array_Nor
                    | Iir_Predefined_Boolean_Array_Xor
                    | Iir_Predefined_Boolean_Array_Xnor
                    | Iir_Predefined_Boolean_Array_Not =>
                     Translate_Predefined_Array_Logical (Subprg);

                  when Iir_Predefined_Physical_Identity =>
                     null;

                  when Iir_Predefined_Physical_Integer_Mul
                    | Iir_Predefined_Physical_Integer_Div
                    | Iir_Predefined_Integer_Physical_Mul
                    | Iir_Predefined_Physical_Real_Mul
                    | Iir_Predefined_Physical_Real_Div
                    | Iir_Predefined_Real_Physical_Mul
                    | Iir_Predefined_Physical_Physical_Div =>
                     null;

                  when Iir_Predefined_Floating_Exp
                    | Iir_Predefined_Floating_Identity =>
                     null;

                  when Iir_Predefined_File_Open
                    | Iir_Predefined_File_Open_Status
                    | Iir_Predefined_File_Close
                    | Iir_Predefined_Endfile =>
                     --  All of them have predefined definitions.
                     null;

                  when Iir_Predefined_Write
                    | Iir_Predefined_Read_Length
                    | Iir_Predefined_Read =>
                     declare
                        P_List : Iir_List;
                        File_Type : Iir;
                     begin
                        P_List := Get_Interface_Declaration (Subprg);
                        File_Type := Get_Type (Get_First_Element (P_List));
                        if not Get_Text_File_Flag (File_Type) then
                           Translate_File_Subprogram (File_Type, Block);
                        end if;
                     end;
                  when others =>
                     Error_Kind ("translate_implicit_function ("
                                 & Iir_Predefined_Functions'Image (Kind) & ")",
                                 Subprg);
               end case;
            end if;
         end loop;
      end Translate_Implicit_Functions_List;
   end Chap7;

   package body Chap8 is
      procedure Translate_Return_Statement (Stmt : Iir_Return_Statement)
      is
         Expr : Iir;
         Ret_Type : Iir;
         Ret_Info : Type_Info_Acc;
         Val : O_Lnode;
         Area : Mnode;
         Subprg_Info : Ortho_Info_Acc;
      begin
         Expr := Get_Expression (Stmt);
         if Expr = Null_Iir then
            --  Return in a procedure.
            New_Return_Stmt;
            return;
         end if;

         --  Return in a function.
         Ret_Type := Get_Return_Type (Chap2.Current_Subprogram);
         Ret_Info := Get_Info (Ret_Type);
         Subprg_Info := Get_Info (Chap2.Current_Subprogram);
         case Ret_Info.Type_Mode is
            when Type_Mode_Thin =>
               --  * if the return type is scalar, simply returns.
               declare
                  V : O_Lnode;
               begin
                  V := Create_Temp (Ret_Info.Ortho_Type (Mode_Value));
                  Open_Temp;
                  New_Assign_Stmt
                    (V, Chap7.Translate_Expression (Expr, Ret_Type));
                  Close_Temp;
                  Chap3.Check_Range (V, Expr, Ret_Type);
                  New_Return_Stmt (New_Value (V));
               end;
            when Type_Mode_Fat_Array =>
               --  * if the return type is unconstrained: allocate an area from
               --    the secondary stack, copy it to the area, and fill the fat
               --    pointer.
               --  Evaluate the result.
               Val := Create_Temp_Init
                 (Ret_Info.Ortho_Ptr_Type (Mode_Value),
                  Chap7.Translate_Expression (Expr, Ret_Type));
               Area := Chap3.Translate_Object_Allocation
                 (Alloc_Return, Ret_Type,
                  Chap3.Get_Array_Ptr_Bounds_Ptr (Val, Ret_Type, Mode_Value),
                  Mode_Value);
               Stabilize (Area);
               Chap3.Translate_Object_Copy (Area,
                                            New_Value (Val),
                                            Ret_Type, Mode_Value);
               Copy_Fat_Pointer
                 (Vp2M (Subprg_Info.Res_Interface, Ret_Info, Mode_Value),
                  Area);
               New_Return_Stmt;
            when Type_Mode_Record
              | Type_Mode_Array
              | Type_Mode_Ptr_Array
              | Type_Mode_Fat_Acc =>
               --  * if the return type is a constrained composite type, copy
               --    it to the result area.
               --  Create a temporary area so that if the expression use
               --  stack2, it will be freed before the return (otherwise,
               --  the stack area will be lost).
               declare
                  V : Mnode;
               begin
                  Open_Temp;
                  V := Vp2M (Subprg_Info.Res_Interface, Ret_Info, Mode_Value);
                  Chap3.Translate_Object_Copy
                    (V, Chap7.Translate_Expression (Expr, Ret_Type),
                     Ret_Type, Mode_Value);
                  Close_Temp;
                  New_Return_Stmt;
               end;
            when Type_Mode_Unknown =>
               raise Internal_Error;
         end case;
      end Translate_Return_Statement;

      procedure Translate_If_Statement (Stmt : Iir)
      is
         Blk : O_If_Block;
         Else_Clause : Iir;
      begin
         Start_If_Stmt
           (Blk, Chap7.Translate_Expression (Get_Condition (Stmt)));

         Translate_Statements_List (Get_Then_Sequence_List (Stmt));

         Else_Clause := Get_Else_Clause (Stmt);
         if Else_Clause /= Null_Iir then
            New_Else_Stmt (Blk);
            if Get_Condition (Else_Clause) = Null_Iir then
               Translate_Statements_List
                 (Get_Then_Sequence_List (Else_Clause));
            else
               Open_Temp;
               Translate_If_Statement (Else_Clause);
               Close_Temp;
            end if;
         end if;
         Finish_If_Stmt (Blk);
      end Translate_If_Statement;

      function Get_Range_Ptr_Field_Value (O_Range : O_Lnode; Field : O_Fnode)
        return O_Enode
      is
      begin
         return New_Value (New_Selected_Element
                           (New_Access_Element (New_Value (O_Range)), Field));
      end Get_Range_Ptr_Field_Value;

      --  Inc or dec ITERATOR according to DIR.
      procedure Gen_Update_Iterator (Iterator : O_Lnode;
                                     Dir : Iir_Direction;
                                     Val : Unsigned_64;
                                     Itype : Iir)
      is
         Op : ON_Op_Kind;
         Base_Type : Iir;
         V : O_Enode;
      begin
         case Dir is
            when Iir_To =>
               Op := ON_Add_Ov;
            when Iir_Downto =>
               Op := ON_Sub_Ov;
         end case;
         Base_Type := Get_Base_Type (Itype);
         case Get_Kind (Base_Type) is
            when Iir_Kind_Integer_Type_Definition =>
               V := New_Signed_Literal
                 (Get_Ortho_Type (Base_Type, Mode_Value), Integer_64 (Val));
            when Iir_Kind_Enumeration_Type_Definition =>
               declare
                  List : Iir_List;
               begin
                  List := Get_Enumeration_Literal_List (Base_Type);
                  --  FIXME: what about type E is ('T') ??
                  if Natural (Val) > Get_Nbr_Elements (List) then
                     raise Internal_Error;
                  end if;
                  V := Get_Info
                    (Get_Nth_Element (List, Natural (Val))).Expr_Node;
               end;

            when others =>
               Error_Kind ("gen_update_iterator", Base_Type);
         end case;
         New_Assign_Stmt (Iterator,
                          New_Dyadic_Op (Op, New_Value (Iterator), V));
      end Gen_Update_Iterator;

      type For_Loop_Data is record
         Iterator : Iir_Iterator_Declaration;
         Stmt : Iir_For_Loop_Statement;
         --  If around the loop, to check if the loop must be executed.
         If_Blk : O_If_Block;
         Blk_Next, Blk_Exit : O_Loop_Block;
         Label_Next, Label_Exit : O_Snode;
         --  Right bound of the iterator, used only if the iterator is a
         --  range expression.
         O_Right : O_Lnode;
         --  Range variable of the iterator, used only if the iterator is not
         --  a range expression.
         O_Range : O_Lnode;
      end record;

      procedure Start_For_Loop (Iterator : Iir_Iterator_Declaration;
                                Stmt : Iir_For_Loop_Statement;
                                Data : out For_Loop_Data)
      is
         Iter_Type : Iir;
         Iter_Base_Type : Iir;
         Var_Iter : Var_Acc;
         Constraint : Iir;
         Cond_To, Cond_Downto : O_Enode;
         Cond_Dir : O_Enode;
         Cond : O_Enode;
         Dir : Iir_Direction;
         Iter_Type_Info : Ortho_Info_Acc;
         Op : ON_Op_Kind;
         If_Blk1 : O_If_Block;
      begin
         --  Initialize DATA.
         Data.Iterator := Iterator;
         Data.Stmt := Stmt;

         Iter_Type := Get_Type (Iterator);
         Iter_Base_Type := Get_Base_Type (Iter_Type);
         Iter_Type_Info := Get_Info (Iter_Base_Type);
         Var_Iter := Get_Info (Iterator).Iterator_Var;

         Open_Temp;

         Constraint := Get_Range_Constraint (Iter_Type);
         if Get_Kind (Constraint) = Iir_Kind_Range_Expression then
            New_Assign_Stmt
              (Get_Var (Var_Iter), Chap7.Translate_Range_Expression_Left
               (Constraint, Iter_Base_Type));
            Dir := Get_Direction (Constraint);
            Data.O_Right := Create_Temp
              (Iter_Type_Info.Ortho_Type (Mode_Value));
            New_Assign_Stmt
              (Data.O_Right, Chap7.Translate_Range_Expression_Right
               (Constraint, Iter_Base_Type));
            case Dir is
               when Iir_To =>
                  Op := ON_Le;
               when Iir_Downto =>
                  Op := ON_Ge;
            end case;
            Cond := New_Compare_Op
              (Op, New_Value (Get_Var (Var_Iter)), New_Value (Data.O_Right),
               Ghdl_Bool_Type_Node);
         else
            Data.O_Range := Create_Temp (Iter_Type_Info.T.Range_Ptr_Type);
            Open_Temp;
            New_Assign_Stmt
              (Data.O_Range, New_Address (Chap7.Translate_Range
                                          (Constraint, Iter_Base_Type),
                                          Iter_Type_Info.T.Range_Ptr_Type));
            New_Assign_Stmt (Get_Var (Var_Iter), Get_Range_Ptr_Field_Value
                             (Data.O_Range, Iter_Type_Info.T.Range_Left));
            Close_Temp;
            --  Before starting the loop, check wether there will be at least
            --  one iteration.
            Cond_To := New_Compare_Op
              (ON_Le, New_Value (Get_Var (Var_Iter)),
               Get_Range_Ptr_Field_Value (Data.O_Range,
                                          Iter_Type_Info.T.Range_Right),
               Ghdl_Bool_Type_Node);
            Cond_Dir := New_Compare_Op
              (ON_Eq,
               Get_Range_Ptr_Field_Value (Data.O_Range,
                                          Iter_Type_Info.T.Range_Dir),
               Ghdl_Dir_To_Node,
               Ghdl_Bool_Type_Node);
            Cond_To := New_Dyadic_Op (ON_And, Cond_Dir, Cond_To);
            Cond_Downto := New_Compare_Op
              (ON_Ge, New_Value (Get_Var (Var_Iter)),
               Get_Range_Ptr_Field_Value (Data.O_Range,
                                          Iter_Type_Info.T.Range_Right),
               Ghdl_Bool_Type_Node);
            Cond_Dir := New_Compare_Op
              (ON_Eq,
               Get_Range_Ptr_Field_Value (Data.O_Range,
                                          Iter_Type_Info.T.Range_Dir),
               Ghdl_Dir_Downto_Node,
               Ghdl_Bool_Type_Node);
            Cond_Downto := New_Dyadic_Op (ON_And, Cond_Dir, Cond_Downto);
            Cond := New_Dyadic_Op (ON_Or, Cond_To, Cond_Downto);
         end if;

         Start_If_Stmt (Data.If_Blk, Cond);

         --  Start loop.
         --  There are two blocks: one for the exit, one for the next.
         Start_Loop_Stmt (Data.Blk_Exit, Data.Label_Exit);
         Start_Loop_Stmt (Data.Blk_Next, Data.Label_Next);

         if Stmt /= Null_Iir then
            declare
               Loop_Info : Loop_Info_Acc;
            begin
               Loop_Info := Add_Info (Stmt, Kind_Loop);
               Loop_Info.Label_Exit := Data.Label_Exit;
               Loop_Info.Label_Next := Data.Label_Next;
            end;
         end if;
      end Start_For_Loop;

      procedure Finish_For_Loop (Data : in out For_Loop_Data)
      is
         Cond : O_Enode;
         If_Blk1 : O_If_Block;
         Iter_Type : Iir;
         Iter_Base_Type : Iir;
         Iter_Type_Info : Type_Info_Acc;
         Var_Iter : Var_Acc;
         Constraint : Iir;
      begin
         New_Exit_Stmt (Data.Label_Next);
         Finish_Loop_Stmt (Data.Blk_Next);

         --  Check end of loop.
         --  Equality is necessary and enough.
         Iter_Type := Get_Type (Data.Iterator);
         Iter_Base_Type := Get_Base_Type (Iter_Type);
         Iter_Type_Info := Get_Info (Iter_Base_Type);
         Var_Iter := Get_Info (Data.Iterator).Iterator_Var;

         Constraint := Get_Range_Constraint (Iter_Type);

         if Get_Kind (Constraint) = Iir_Kind_Range_Expression then
            Cond := New_Value (Data.O_Right);
         else
            Cond := Get_Range_Ptr_Field_Value
              (Data.O_Range, Iter_Type_Info.T.Range_Right);
         end if;
         Gen_Exit_When (Data.Label_Exit,
                        New_Compare_Op (ON_Eq, New_Value (Get_Var (Var_Iter)),
                                        Cond, Ghdl_Bool_Type_Node));

         --  Update the iterator.
         if Get_Kind (Constraint) = Iir_Kind_Range_Expression then
            Gen_Update_Iterator
              (Get_Var (Var_Iter), Get_Direction (Constraint),
               1, Iter_Base_Type);
         else
            Start_If_Stmt
              (If_Blk1, New_Compare_Op
               (ON_Eq,
                Get_Range_Ptr_Field_Value (Data.O_Range,
                                           Iter_Type_Info.T.Range_Dir),
                Ghdl_Dir_To_Node,
                Ghdl_Bool_Type_Node));
            Gen_Update_Iterator
              (Get_Var (Var_Iter), Iir_To, 1, Iter_Base_Type);
            New_Else_Stmt (If_Blk1);
            Gen_Update_Iterator
              (Get_Var (Var_Iter), Iir_Downto, 1, Iter_Base_Type);
            Finish_If_Stmt (If_Blk1);
         end if;

         Finish_Loop_Stmt (Data.Blk_Exit);
         Finish_If_Stmt (Data.If_Blk);
         Close_Temp;

         if Data.Stmt /= Null_Iir then
            declare
               Loop_Info : Loop_Info_Acc;
            begin
               Loop_Info := Get_Info (Data.Stmt);
               Free (Loop_Info);
            end;
         end if;
      end Finish_For_Loop;

      procedure Translate_For_Loop_Statement (Stmt : Iir_For_Loop_Statement)
      is
         Iterator : Iir;
         Iter_Type : Iir;
         Iter_Base_Type : Iir;
         Iter_Type_Info : Type_Info_Acc;
         Loop_Info : Loop_Info_Acc;
         It_Info : Ortho_Info_Acc;
         O_Range : O_Lnode;
         O_Right : O_Lnode;
         Blk_Next, Blk_Exit : O_Loop_Block;
         Cond_To, Cond_Downto : O_Enode;
         Cond_Dir : O_Enode;
         Cond : O_Enode;
         Dir : Iir_Direction;
         Op : ON_Op_Kind;
         If_Blk, If_Blk1 : O_If_Block;
         Var_Iter : Var_Acc;
         Constraint : Iir;
      begin
         Start_Declare_Stmt;
         Iterator := Get_Iterator_Scheme (Stmt);
         Iter_Type := Get_Type (Iterator);
         Iter_Base_Type := Get_Base_Type (Iter_Type);
         Iter_Type_Info := Get_Info (Iter_Base_Type);

         Chap3.Translate_Object_Subtype (Iterator, False);

         --  Create info for the iterator.
         It_Info := Add_Info (Iterator, Kind_Iterator);
         Var_Iter := Create_Var
           (Create_Var_Identifier (Iterator),
            Iter_Type_Info.Ortho_Type (Mode_Value),
            O_Storage_Local);
         It_Info.Iterator_Var := Var_Iter;

         Open_Temp;

         Constraint := Get_Range_Constraint (Iter_Type);
         if Get_Kind (Constraint) = Iir_Kind_Range_Expression then
            New_Assign_Stmt
              (Get_Var (Var_Iter), Chap7.Translate_Range_Expression_Left
               (Constraint, Iter_Base_Type));
            Dir := Get_Direction (Constraint);
            O_Right := Create_Temp (Iter_Type_Info.Ortho_Type (Mode_Value));
            New_Assign_Stmt
              (O_Right, Chap7.Translate_Range_Expression_Right
               (Constraint, Iter_Base_Type));
            case Dir is
               when Iir_To =>
                  Op := ON_Le;
               when Iir_Downto =>
                  Op := ON_Ge;
            end case;
            Cond := New_Compare_Op
              (Op, New_Value (Get_Var (Var_Iter)), New_Value (O_Right),
               Ghdl_Bool_Type_Node);
         else
            O_Range := Create_Temp (Iter_Type_Info.T.Range_Ptr_Type);
            Open_Temp;
            New_Assign_Stmt
              (O_Range, New_Address (Chap7.Translate_Range
                                     (Constraint, Iter_Base_Type),
                                     Iter_Type_Info.T.Range_Ptr_Type));
            New_Assign_Stmt (Get_Var (Var_Iter), Get_Range_Ptr_Field_Value
                             (O_Range, Iter_Type_Info.T.Range_Left));
            Close_Temp;
            --  Before starting the loop, check wether there will be at least
            --  one iteration.
            Cond_To := New_Compare_Op
              (ON_Le, New_Value (Get_Var (Var_Iter)),
               Get_Range_Ptr_Field_Value (O_Range,
                                          Iter_Type_Info.T.Range_Right),
               Ghdl_Bool_Type_Node);
            Cond_Dir := New_Compare_Op
              (ON_Eq,
               Get_Range_Ptr_Field_Value (O_Range, Iter_Type_Info.T.Range_Dir),
               Ghdl_Dir_To_Node,
               Ghdl_Bool_Type_Node);
            Cond_To := New_Dyadic_Op (ON_And, Cond_Dir, Cond_To);
            Cond_Downto := New_Compare_Op
              (ON_Ge, New_Value (Get_Var (Var_Iter)),
               Get_Range_Ptr_Field_Value (O_Range,
                                          Iter_Type_Info.T.Range_Right),
               Ghdl_Bool_Type_Node);
            Cond_Dir := New_Compare_Op
              (ON_Eq,
               Get_Range_Ptr_Field_Value (O_Range, Iter_Type_Info.T.Range_Dir),
               Ghdl_Dir_Downto_Node,
               Ghdl_Bool_Type_Node);
            Cond_Downto := New_Dyadic_Op (ON_And, Cond_Dir, Cond_Downto);
            Cond := New_Dyadic_Op (ON_Or, Cond_To, Cond_Downto);
         end if;

         Start_If_Stmt (If_Blk, Cond);

         --  Start loop.
         --  There are two blocks: one for the exit, one for the next.
         Loop_Info := Add_Info (Stmt, Kind_Loop);
         Start_Loop_Stmt (Blk_Exit, Loop_Info.Label_Exit);
         Start_Loop_Stmt (Blk_Next, Loop_Info.Label_Next);

         Translate_Statements_List (Get_Statement_List (Stmt));

         New_Exit_Stmt (Loop_Info.Label_Next);
         Finish_Loop_Stmt (Blk_Next);

         --  Check end of loop.
         --  Equality is necessary and enough.
         if Get_Kind (Constraint) = Iir_Kind_Range_Expression then
            Cond := New_Value (O_Right);
         else
            Cond := Get_Range_Ptr_Field_Value
              (O_Range, Iter_Type_Info.T.Range_Right);
         end if;
         Gen_Exit_When (Loop_Info.Label_Exit,
                        New_Compare_Op (ON_Eq, New_Value (Get_Var (Var_Iter)),
                                        Cond, Ghdl_Bool_Type_Node));

         --  Update the iterator.
         if Get_Kind (Constraint) = Iir_Kind_Range_Expression then
            Gen_Update_Iterator (Get_Var (Var_Iter), Dir, 1, Iter_Base_Type);
         else
            Start_If_Stmt
              (If_Blk1, New_Compare_Op
               (ON_Eq,
                Get_Range_Ptr_Field_Value (O_Range,
                                           Iter_Type_Info.T.Range_Dir),
                Ghdl_Dir_To_Node,
                Ghdl_Bool_Type_Node));
            Gen_Update_Iterator
              (Get_Var (Var_Iter), Iir_To, 1, Iter_Base_Type);
            New_Else_Stmt (If_Blk1);
            Gen_Update_Iterator
              (Get_Var (Var_Iter), Iir_Downto, 1, Iter_Base_Type);
            Finish_If_Stmt (If_Blk1);
         end if;

         Finish_Loop_Stmt (Blk_Exit);
         Finish_If_Stmt (If_Blk);
         Close_Temp;
         Finish_Declare_Stmt;

         Free (Loop_Info);
      end Translate_For_Loop_Statement;

      procedure Translate_While_Loop_Statement
        (Stmt : Iir_While_Loop_Statement)
      is
         Loop_Blk : O_Loop_Block;
         Info : Loop_Info_Acc;
         Cond : Iir;
         If_Blk : O_If_Block;
      begin
         Info := Add_Info (Stmt, Kind_Loop);

         Start_Loop_Stmt (Loop_Blk, Info.Label_Exit);
         Info.Label_Next := O_Snode_Null;

         Open_Temp;
         Cond := Get_Condition (Stmt);
         if Cond /= Null_Iir then
            Gen_Exit_When
              (Info.Label_Exit,
               New_Monadic_Op (ON_Not, Chap7.Translate_Expression (Cond)));
         end if;
         Close_Temp;

         Translate_Statements_List (Get_Statement_List (Stmt));

         Finish_Loop_Stmt (Loop_Blk);
         Free (Info);
      end Translate_While_Loop_Statement;

      procedure Translate_Exit_Next_Statement (Stmt : Iir)
      is
         Cond : Iir;
         If_Blk : O_If_Block;
         Info : Loop_Info_Acc;
      begin
         Cond := Get_Condition (Stmt);
         if Cond /= Null_Iir then
            Start_If_Stmt (If_Blk, Chap7.Translate_Expression (Cond));
         end if;
         Info := Get_Info (Get_Loop (Stmt));
         case Get_Kind (Stmt) is
            when Iir_Kind_Exit_Statement =>
               New_Exit_Stmt (Info.Label_Exit);
            when Iir_Kind_Next_Statement =>
               if Info.Label_Next /= O_Snode_Null then
                  --  For-loop.
                  New_Exit_Stmt (Info.Label_Next);
               else
                  --  While-loop.
                  New_Next_Stmt (Info.Label_Exit);
               end if;
            when others =>
               raise Internal_Error;
         end case;
         if Cond /= Null_Iir then
            Finish_If_Stmt (If_Blk);
         end if;
      end Translate_Exit_Next_Statement;

      procedure Translate_Variable_Assignment_Statement
        (Stmt : Iir_Variable_Assignment_Statement)
      is
         Target : Iir;
         Targ_Type : Iir;
         Expr : Iir;
         Targ_Node : Mnode;
      begin
         Target := Get_Target (Stmt);
         Targ_Type := Get_Type (Target);
         if Get_Kind (Target) = Iir_Kind_Aggregate then
            --  Not implemented.
            raise Internal_Error;
         end if;
         --  Note: must evaluate target before the expression, since the
         --  expression can be unconstrained (eg: (others => ' '))
         Targ_Node := Chap6.Translate_Name (Target);
         Expr := Get_Expression (Stmt);
         Chap7.Translate_Assign (Targ_Node, Expr, Targ_Type);
      end Translate_Variable_Assignment_Statement;

      procedure Translate_Assertion_Statement (Stmt : Iir_Assertion_Statement)
      is
         Expr : Iir;
         If_Blk : O_If_Block;
         No_If : Boolean;
         Msg : O_Enode;
         Severity : O_Enode;
         Assocs : O_Assoc_List;
         Loc : O_Enode;
      begin
         Expr := Get_Assertion_Condition (Stmt);
         if Get_Expr_Staticness (Expr) = Locally then
            if Eval_Pos (Expr) = 1 then
               --  Assert TRUE is a noop.
               --  FIXME: generate a noop.
               return;
            else
               No_If := True;
            end if;
         else
            --  An assertion is reported if the condition is false!
            Start_If_Stmt (If_Blk,
                           New_Monadic_Op (ON_Not,
                                           Chap7.Translate_Expression (Expr)));
            No_If := False;
            --  Note: it is necessary to create a declare block, to avoid bad
            --  order with the if block.
            Open_Temp;
         end if;
         Loc := Chap4.Get_Location (Stmt);
         Expr := Get_Report_Expression (Stmt);
         if Expr = Null_Iir then
            Msg := New_Address (Std_Assert_Default_Report_Node,
                                Std_String_Ptr_Node);
         else
            Msg := Chap7.Translate_Expression (Expr, String_Type_Definition);
         end if;
         Expr := Get_Severity_Expression (Stmt);
         if Expr = Null_Iir then
            Severity := Get_Ortho_Expr (Severity_Level_Error);
         else
            Severity := Chap7.Translate_Expression (Expr);
         end if;
         --  Do call.
         Start_Association (Assocs, Ghdl_Assert_Failed);
         New_Association (Assocs, Msg);
         New_Association (Assocs, Severity);
         New_Association (Assocs, Loc);
         New_Procedure_Call (Assocs);
         if not No_If then
            Close_Temp;
            Finish_If_Stmt (If_Blk);
         end if;
      end Translate_Assertion_Statement;

      function Translate_Simple_String_Choice
        (Expr : O_Lnode;
         Val : Iir;
         Val_Node : O_Lnode;
         Tinfo : Type_Info_Acc;
         Func : O_Dnode)
        return O_Enode
      is
      begin
         New_Assign_Stmt
           (New_Selected_Element (Val_Node, Tinfo.T.Base_Field (Mode_Value)),
            Chap7.Translate_Expression (Val, Get_Type (Val)));
         return Chap7.Translate_Lib_Operator
           (New_Value (Expr),
            New_Address (Val_Node, Tinfo.Ortho_Ptr_Type (Mode_Value)),
            Func);
      end Translate_Simple_String_Choice;

      procedure Translate_String_Choice
        (Expr : O_Lnode;
         Val_Node : O_Lnode;
         Tinfo : Type_Info_Acc;
         Func : O_Dnode;
         Cond_Var : O_Lnode;
         Choice_List : Iir_List;
         Choice_Index : Natural)
      is
         Choice : Iir;
         Cond : O_Enode;
         If_Blk : O_If_Block;
      begin
         Choice := Get_Nth_Element (Choice_List, Choice_Index);
         if Choice = Null_Iir then
            return;
         end if;

         case Get_Kind (Choice) is
            when Iir_Kind_Association_Choice_By_Expression =>
               Cond := Translate_Simple_String_Choice
                 (Expr, Get_Expression (Choice), Val_Node, Tinfo, Func);
            when Iir_Kind_Association_Choice_By_List =>
               declare
                  List : Iir_List;
                  Ch : Iir;
                  V : O_Enode;
               begin
                  List := Get_Choice_List (Choice);
                  for I in Natural loop
                     Ch := Get_Nth_Element (List, I);
                     exit when Ch = Null_Iir;
                     V := Translate_Simple_String_Choice
                       (Expr, Get_Expression (Ch), Val_Node, Tinfo, Func);
                     if I = 0 then
                        New_Assign_Stmt (Cond_Var, V);
                     else
                        New_Assign_Stmt
                          (Cond_Var,
                           New_Dyadic_Op (ON_Or, New_Value (Cond_Var), V));
                     end if;
                  end loop;
                  Cond := New_Value (Cond_Var);
               end;
            when Iir_Kind_Association_Choice_By_Others =>
               Translate_Statements_List (Get_Associated (Choice));
               return;
            when others =>
               Error_Kind ("translate_string_choice", Choice);
         end case;
         Start_If_Stmt (If_Blk, Cond);
         Translate_Statements_List (Get_Associated (Choice));
         New_Else_Stmt (If_Blk);
         Translate_String_Choice (Expr, Val_Node, Tinfo, Func, Cond_Var,
                                  Choice_List, Choice_Index + 1);
         Finish_If_Stmt (If_Blk);
      end Translate_String_Choice;

      --  Case statement whose expression is an unidim array.
      procedure Translate_String_Case_Statement (Stmt : Iir_Case_Statement)
      is
         Expr : Iir;
         Expr_Type : Iir;
         Base_Type : Iir;
         --  Node containing the address of the selector.
         Expr_Node : O_Lnode;
         --  Node containing the current choice.
         C_Node : O_Lnode;
         Tinfo : Type_Info_Acc;
         List : Iir_List;
         If_Blk : O_If_Block;
         Func : Iir;
         Func_Node : O_Dnode;
         Cond_Var : O_Lnode;
      begin
         --  Translate into if/elsif statements.
         --  FIXME: if the number of literals ** length of the array < 256,
         --   use a case statement.
         Expr := Get_Expression (Stmt);
         Expr_Type := Get_Type (Expr);
         Base_Type := Get_Base_Type (Expr_Type);
         Tinfo := Get_Info (Base_Type);

         Expr_Node := Create_Temp_Init
           (Tinfo.Ortho_Ptr_Type (Mode_Value),
            Chap7.Translate_Expression (Expr, Base_Type));
         C_Node := Create_Temp (Tinfo.Ortho_Type (Mode_Value));
         New_Assign_Stmt
           (New_Selected_Element (C_Node, Tinfo.T.Bounds_Field (Mode_Value)),
            New_Value_Selected_Acc_Value
            (Expr_Node, Tinfo.T.Bounds_Field (Mode_Value)));

         Cond_Var := Create_Temp (Std_Boolean_Type_Node);

         Func := Chap7.Find_Predefined_Function
           (Base_Type, Iir_Predefined_Array_Equality);
         Func_Node := Get_Info (Func).Ortho_Func;

         List := Get_Association_Choices_List (Stmt);
         Translate_String_Choice
           (Expr_Node, C_Node, Tinfo, Func_Node, Cond_Var, List, 0);
      end Translate_String_Case_Statement;

      procedure Translate_Case_Choice
        (Choice : Iir; Choice_Type : Iir; Blk : in out O_Case_Block)
      is
         Expr : Iir;
      begin
         case Get_Kind (Choice) is
            when Iir_Kind_Association_Choice_By_Others =>
               New_Default_Choice (Blk);
            when Iir_Kind_Association_Choice_By_Expression =>
               Expr := Get_Expression (Choice);
               New_Expr_Choice
                 (Blk, Chap7.Translate_Static_Expression (Expr, Choice_Type));
            when Iir_Kind_Association_Choice_By_Range =>
               declare
                  H, L : Iir;
               begin
                  Expr := Get_Expression (Choice);
                  Get_Low_High_Limit (Expr, L, H);
                  New_Range_Choice
                    (Blk,
                     Chap7.Translate_Static_Expression (L, Choice_Type),
                     Chap7.Translate_Static_Expression (H, Choice_Type));
               end;
            when Iir_Kind_Association_Choice_By_List =>
               declare
                  List : Iir_List;
                  Ch : Iir;
               begin
                  List := Get_Choice_List (Choice);
                  for I in Natural loop
                     Ch := Get_Nth_Element (List, I);
                     exit when Ch = Null_Iir;
                     Translate_Case_Choice (Ch, Choice_Type, Blk);
                  end loop;
               end;
            when others =>
               Error_Kind ("translate_case_choice", Choice);
         end case;
      end Translate_Case_Choice;

      procedure Translate_Case_Statement (Stmt : Iir_Case_Statement)
      is
         Expr : Iir;
         Expr_Type : Iir;
         Case_Blk : O_Case_Block;
         List : Iir_List;
         Choice : Iir;
      begin
         Expr := Get_Expression (Stmt);
         Expr_Type := Get_Type (Expr);
         if Get_Kind (Expr_Type) = Iir_Kind_Array_Subtype_Definition then
            Translate_String_Case_Statement (Stmt);
            return;
         end if;
         Start_Case_Stmt (Case_Blk, Chap7.Translate_Expression (Expr));
         List := Get_Association_Choices_List (Stmt);
         for I in Natural loop
            Choice := Get_Nth_Element (List, I);
            exit when Choice = Null_Iir;
            Start_Branch (Case_Blk);
            Translate_Case_Choice (Choice, Expr_Type, Case_Blk);
            Translate_Statements_List (Get_Associated (Choice));
            Finish_Branch (Case_Blk, False);
         end loop;
         Finish_Case_Stmt (Case_Blk);
      end Translate_Case_Statement;

      procedure Translate_Write_Procedure_Call (Params : Iir_List)
      is
         F_Assoc : Iir;
         Value_Assoc : Iir;
         Value : O_Lnode;
         Formal_Type : Iir;
         Tinfo : Type_Info_Acc;
         Assocs : O_Assoc_List;
      begin
         F_Assoc := Get_Nth_Element (Params, 0);
         Value_Assoc := Get_Nth_Element (Params, 1);
         Formal_Type := Get_Type (Get_Formal (Value_Assoc));
         Tinfo := Get_Info (Formal_Type);
         case Tinfo.Type_Mode is
            when Type_Mode_Scalar =>
               Open_Temp;
               Start_Association (Assocs, Ghdl_Write_Scalar);
               --    compute file parameter (get an index)
               New_Association
                 (Assocs, Chap7.Translate_Expression (Get_Actual (F_Assoc)));
               --    compute the value.
               Value := Create_Temp (Tinfo.Ortho_Type (Mode_Value));
               New_Assign_Stmt
                 (Value,
                  Chap7.Translate_Expression (Get_Actual (Value_Assoc)));
               New_Association
                 (Assocs, New_Unchecked_Address (Value, Ghdl_Ptr_Type));
               --    length.
               New_Association (Assocs,
                                New_Sizeof (Tinfo.Ortho_Type (Mode_Value),
                                            Ghdl_Index_Type));
               --    call a predefined procedure
               New_Procedure_Call (Assocs);
            when others =>
               raise Internal_Error;
         end case;
      end Translate_Write_Procedure_Call;

      procedure Translate_Read_Procedure_Call (Params : Iir_List)
      is
         F_Assoc : Iir;
         Value_Assoc : Iir;
         Value : Mnode;
         Formal_Type : Iir;
         Tinfo : Type_Info_Acc;
         Assocs : O_Assoc_List;
      begin
         F_Assoc := Get_Nth_Element (Params, 0);
         Value_Assoc := Get_Nth_Element (Params, 1);
         Formal_Type := Get_Type (Get_Formal (Value_Assoc));
         Tinfo := Get_Info (Formal_Type);
         case Tinfo.Type_Mode is
            when Type_Mode_Scalar =>
               Open_Temp;
               Start_Association (Assocs, Ghdl_Read_Scalar);
               --    compute file parameter (get an index)
               New_Association
                 (Assocs, Chap7.Translate_Expression (Get_Actual (F_Assoc)));
               --  value
               Value := Chap6.Translate_Name (Get_Actual (Value_Assoc));
               New_Association
                 (Assocs, New_Convert_Ov (M2Addr (Value), Ghdl_Ptr_Type));
               --    length.
               New_Association (Assocs,
                                New_Sizeof (Tinfo.Ortho_Type (Mode_Value),
                                            Ghdl_Index_Type));
               --    call a predefined procedure
               New_Procedure_Call (Assocs);
            when others =>
               raise Internal_Error;
         end case;
      end Translate_Read_Procedure_Call;

      procedure Translate_Implicit_Procedure_Call (Stmt : Iir_Procedure_Call)
      is
         Kind : Iir_Predefined_Functions;
         Params : Iir_List;
      begin
         Kind := Get_Implicit_Definition (Get_Implementation (Stmt));
         Params := Get_Parameter_Association_List (Stmt);
         case Kind is
            when Iir_Predefined_Write =>
               --  Check wether text or not.
               declare
                  File_Param : Iir;
                  Assocs : O_Assoc_List;
               begin
                  File_Param := Get_Nth_Element (Params, 0);
                  -- FIXME: do the test.
                  if Get_Text_File_Flag (Get_Type (Get_Formal (File_Param)))
                  then
                     --  If text:
                     Start_Association (Assocs, Ghdl_Text_Write);
                     --    compute file parameter (get an index)
                     New_Association
                       (Assocs,
                        Chap7.Translate_Expression (Get_Actual (File_Param)));
                     --    compute string parameter (get a fat array pointer)
                     New_Association
                       (Assocs, Chap7.Translate_Expression
                        (Get_Actual (Get_Nth_Element (Params, 1)),
                         String_Type_Definition));
                     --    call a predefined procedure
                     New_Procedure_Call (Assocs);
                  else
                     Translate_Write_Procedure_Call (Params);
                  end if;
               end;

            when Iir_Predefined_Read_Length =>
               --  FIXME: works only for text read length.
               declare
                  File_Param : Iir;
                  Assocs : O_Assoc_List;
                  Str : O_Enode;
                  Res : Mnode;
               begin
                  File_Param := Get_Nth_Element (Params, 0);
                  Str := Chap7.Translate_Expression
                    (Get_Actual (Get_Nth_Element (Params, 1)),
                     String_Type_Definition);
                  Res := Chap6.Translate_Name
                    (Get_Actual (Get_Nth_Element (Params, 2)));
                  Start_Association (Assocs, Ghdl_Text_Read_Length);
                  --    compute file parameter (get an index)
                  New_Association
                    (Assocs,
                     Chap7.Translate_Expression (Get_Actual (File_Param)));
                  --    compute string parameter (get a fat array pointer)
                  New_Association (Assocs, Str);
                  --    call a predefined procedure
                  New_Assign_Stmt
                    (M2Lv (Res), New_Function_Call (Assocs));
               end;

            when Iir_Predefined_Read =>
               Translate_Read_Procedure_Call (Params);

            when Iir_Predefined_Deallocate =>
               Chap3.Translate_Object_Deallocation
                 (Get_Actual (Get_Nth_Element (Params, 0)));

            when Iir_Predefined_File_Open =>
               declare
                  File_Param : Iir;
                  Name_Param : Iir;
                  Kind_Param : Iir;
                  Constr : O_Assoc_List;
               begin
                  File_Param := Get_Actual (Get_Nth_Element (Params, 0));
                  Name_Param := Get_Actual (Get_Nth_Element (Params, 1));
                  Kind_Param := Get_Actual (Get_Nth_Element (Params, 2));
                  if Get_Text_File_Flag (Get_Type (File_Param)) then
                     Start_Association (Constr, Ghdl_Text_File_Open);
                  else
                     Start_Association (Constr, Ghdl_File_Open);
                  end if;
                  New_Association
                    (Constr, Chap7.Translate_Expression (File_Param));
                  New_Association
                    (Constr, New_Convert_Ov
                     (Chap7.Translate_Expression (Kind_Param),
                      Ghdl_Integer_Type));
                  New_Association
                    (Constr,
                     Chap7.Translate_Expression (Name_Param,
                                                 String_Type_Definition));
                  New_Procedure_Call (Constr);
               end;

            when Iir_Predefined_File_Open_Status =>
               declare
                  Status_Param : Iir;
                  File_Param : Iir;
                  Name_Param : Iir;
                  Kind_Param : Iir;
                  Constr : O_Assoc_List;
                  Status : Mnode;
               begin
                  Status_Param := Get_Actual (Get_Nth_Element (Params, 0));
                  Status := Chap6.Translate_Name (Status_Param);
                  File_Param := Get_Actual (Get_Nth_Element (Params, 1));
                  Name_Param := Get_Actual (Get_Nth_Element (Params, 2));
                  Kind_Param := Get_Actual (Get_Nth_Element (Params, 3));
                  if Get_Text_File_Flag (Get_Type (File_Param)) then
                     Start_Association (Constr, Ghdl_Text_File_Open_Status);
                  else
                     Start_Association (Constr, Ghdl_File_Open_Status);
                  end if;
                  New_Association
                    (Constr, Chap7.Translate_Expression (File_Param));
                  New_Association
                    (Constr, New_Convert_Ov
                     (Chap7.Translate_Expression (Kind_Param),
                      Ghdl_Integer_Type));
                  New_Association
                    (Constr,
                     Chap7.Translate_Expression (Name_Param,
                                                 String_Type_Definition));
                  New_Assign_Stmt
                    (M2Lv (Status),
                     New_Convert_Ov (New_Function_Call (Constr),
                                     Std_File_Open_Status_Type));
               end;

            when Iir_Predefined_File_Close =>
               declare
                  File_Param : Iir;
                  Constr : O_Assoc_List;
               begin
                  File_Param := Get_Actual (Get_Nth_Element (Params, 0));
                  if Get_Text_File_Flag (Get_Type (File_Param)) then
                     Start_Association (Constr, Ghdl_Text_File_Close);
                  else
                     Start_Association (Constr, Ghdl_File_Close);
                  end if;
                  New_Association
                    (Constr, Chap7.Translate_Expression (File_Param));
                  New_Procedure_Call (Constr);
               end;

            when others =>
               Ada.Text_IO.Put_Line
                 ("translate_implicit_procedure_call: cannot handle "
                  & Iir_Predefined_Functions'Image (Kind));
               raise Internal_Error;
         end case;
      end Translate_Implicit_Procedure_Call;

      procedure Translate_Procedure_Call (Stmt : Iir_Procedure_Call)
      is
         type Mnode_Array is array (Natural range <>) of Mnode;
         type O_Enode_Array is array (Natural range <>) of O_Enode;
         List : Iir_List := Get_Parameter_Association_List (Stmt);
         Nbr_Assoc : Natural := Get_Nbr_Elements (List);
         Params : Mnode_Array (0 .. Nbr_Assoc - 1);
         E_Params : O_Enode_Array (0 .. Nbr_Assoc - 1);
         Imp : Iir;
         Info : Subprg_Info_Acc;
         Res : O_Lnode;
         El : Iir;
         Constr : O_Assoc_List;
         Act : Iir;
         Actual_Type : Iir;
         Formal : Iir;
         Formal_Type : Iir;
         Ftype_Info : Type_Info_Acc;
         Atype_Info : Type_Info_Acc;
         Formal_Info : Ortho_Info_Acc;
         Val : O_Enode;
         Param : Mnode;
         Ptr : O_Lnode;
      begin
         Imp := Get_Implementation (Stmt);
         Info := Get_Info (Imp);

         --  Create an in-out result record for in-out arguments passed by
         --  value.
         if Info.Res_Record_Type /= O_Tnode_Null then
            Res := Create_Temp (Info.Res_Record_Type);
         else
            Res := O_Lnode_Null;
         end if;

         --  Evaluate in-out parameters and parameters passed by ref, since
         --  they can add declarations.
         --  Non-composite in-out parameters address are saved in order to
         --  be able to assignate the result.
         for I in Natural loop
            El := Get_Nth_Element (List, I);
            exit when El = Null_Iir;

            Params (I) := Mnode_Null;
            E_Params (I) := O_Enode_Null;

            Formal := Get_Formal (El);

            case Get_Kind (El) is
               when Iir_Kind_Association_Element_Open =>
                  Act := Get_Default_Value (Formal);
               when Iir_Kind_Association_Element_By_Expression =>
                  Act := Get_Actual (El);
               when others =>
                  Error_Kind ("translate_procedure_call", El);
            end case;
            Actual_Type := Get_Type (Act);

            Formal_Type := Get_Type (Formal);
            Ftype_Info := Get_Info (Formal_Type);
            Formal_Info := Get_Info (Formal);

            if Formal_Info.Interface_Field /= O_Fnode_Null then
               --  Copy-out argument.
               --  This is not a composite type.
               Param := Chap6.Translate_Name (Act);
               if Get_Object_Kind (Param) /= Mode_Value then
                  raise Internal_Error;
               end if;
               Params (I) := Stabilize (Param);
               Ptr := New_Selected_Element (Res, Formal_Info.Interface_Field);
               case Ftype_Info.Type_Mode is
                  when Type_Mode_Thin =>
                     New_Assign_Stmt (Ptr, M2E (Params (I)));
                  when Type_Mode_Fat_Acc =>
                     Copy_Fat_Pointer (Lv2M (Ptr, Ftype_Info, Mode_Value),
                                       Params (I));
                  when others =>
                     raise Internal_Error;
               end case;
            elsif Ftype_Info.Type_Mode not in Type_Mode_By_Value then
               --  Passed by reference.
               case Get_Kind (Formal) is
                  when Iir_Kind_Constant_Interface_Declaration
                    | Iir_Kind_File_Interface_Declaration =>
                     E_Params (I) := Chap7.Translate_Expression
                       (Act, Formal_Type);
                  when Iir_Kind_Variable_Interface_Declaration =>
                     Param := Chap6.Translate_Name (Act);
                     if Get_Object_Kind (Param) /= Mode_Value then
                        raise Internal_Error;
                     end if;
                     --  Atype may not have been set (eg: slice).
                     Atype_Info := Get_Info (Actual_Type);
                     E_Params (I) := M2E (Param);
                     if Formal_Type /= Actual_Type then
                        --  Implicit array conversion or subtype check.
                        E_Params (I) := Chap7.Translate_Implicit_Conv
                          (E_Params (I), Actual_Type, Formal_Type, Mode_Value);
                     end if;
                  when Iir_Kind_Signal_Interface_Declaration =>
                     Param := Chap6.Translate_Name (Act);
                     if Get_Object_Kind (Param) /= Mode_Signal then
                        raise Internal_Error;
                     end if;
                     --  Atype may not have been set (eg: slice).
                     Atype_Info := Get_Info (Actual_Type);
                     E_Params (I) := M2E (Param);
                     if Formal_Type /= Actual_Type then
                        --  Implicit array conversion or subtype check.
                        E_Params (I) := Chap7.Translate_Implicit_Conv
                          (E_Params (I), Actual_Type,
                           Formal_Type, Mode_Signal);
                     end if;
                  when others =>
                     Error_Kind ("translate_procedure_call(2)", Formal);
               end case;
            end if;
         end loop;

         Start_Association (Constr, Info.Ortho_Func);
         if Res /= O_Lnode_Null then
            New_Association (Constr, New_Address (Res, Info.Res_Record_Ptr));
         end if;
         Chap2.Add_Subprg_Instance_Assoc (Constr, Imp);

         --  Parameters.
         for I in Natural loop
            El := Get_Nth_Element (List, I);
            exit when El = Null_Iir;
            Formal := Get_Formal (El);
            Formal_Info := Get_Info (Formal);
            Formal_Type := Get_Type (Formal);
            Ftype_Info := Get_Info (Formal_Type);
            --  FIXME: what about fat pointer (not a composite).

            if Formal_Info.Interface_Field = O_Fnode_Null then
               if Ftype_Info.Type_Mode in Type_Mode_By_Value then
                  --  Parameter passed by value.
                  if E_Params (I) /= O_Enode_Null then
                     Val := E_Params (I);
                     raise Internal_Error;
                  else
                     case Get_Kind (El) is
                        when Iir_Kind_Association_Element_Open =>
                           Act := Get_Default_Value (Formal);
                        when Iir_Kind_Association_Element_By_Expression =>
                           Act := Get_Actual (El);
                        when others =>
                           Error_Kind ("translate_procedure_call(2)", El);
                     end case;
                     case Get_Kind (Formal) is
                        when Iir_Kind_Signal_Interface_Declaration =>
                           Param := Chap6.Translate_Name (Act);
                           --  This is a scalar.
                           Val := M2E (Param);
                        when others =>
                           Val := Chap7.Translate_Expression
                             (Act, Formal_Type);
                     end case;
                  end if;
                  New_Association (Constr, Val);
               else
                  --  Parameter passed by ref, which was already computed.
                  New_Association (Constr, E_Params (I));
               end if;
            end if;
         end loop;

         New_Procedure_Call (Constr);

         --  Copy-out non-composite parameters.
         for I in Natural loop
            El := Get_Nth_Element (List, I);
            exit when El = Null_Iir;
            Formal := Get_Formal (El);
            Formal_Type := Get_Type (Formal);
            Ftype_Info := Get_Info (Formal_Type);
            Formal_Info := Get_Info (Formal);
            if Params (I) /= Mnode_Null then
               if Formal_Info.Interface_Field /= O_Fnode_Null then
                  Ptr := New_Selected_Element
                    (Res, Formal_Info.Interface_Field);
                  case Ftype_Info.Type_Mode is
                     when Type_Mode_Thin =>
                        New_Assign_Stmt (M2Lv (Params (I)), New_Value (Ptr));
                     when Type_Mode_Fat_Acc =>
                        Copy_Fat_Pointer
                          (Params (I), Lv2M (Ptr, Ftype_Info, Mode_Value));
                     when others =>
                        raise Internal_Error;
                  end case;
               end if;
            end if;
         end loop;
      end Translate_Procedure_Call;

      procedure Translate_Wait_Statement (Stmt : Iir)
      is
         Sensitivity : Iir_List;
         Cond : Iir;
         Timeout : Iir;
         Constr : O_Assoc_List;
      begin
         Sensitivity := Get_Sensitivity_List (Stmt);
         Cond := Get_Condition_Clause (Stmt);
         Timeout := Get_Timeout_Clause (Stmt);

         --  Check for simple cases.
         if Sensitivity = Null_Iir
           and then Cond = Null_Iir
         then
            if Timeout = Null_Iir then
               --  Process exit.
               Start_Association (Constr, Ghdl_Process_Wait_Exit);
               New_Procedure_Call (Constr);
            else
               --  Wait for a timeout.
               Start_Association (Constr, Ghdl_Process_Wait_Timeout);
               New_Association (Constr, Chap7.Translate_Expression
                                (Timeout, Time_Type_Definition));
               New_Procedure_Call (Constr);
            end if;
            return;
         end if;

         --  Evaluate the timeout (if any) and register it,
         if Timeout /= Null_Iir then
            Start_Association (Constr, Ghdl_Process_Wait_Set_Timeout);
            New_Association (Constr, Chap7.Translate_Expression
                             (Timeout, Time_Type_Definition));
            New_Procedure_Call (Constr);
         end if;

         --  Evaluate the sensitivity list and register it.
         if Sensitivity /= null then
            Register_Signal_List
              (Sensitivity, Ghdl_Process_Wait_Add_Sensitivity);
         end if;

         if Cond = Null_Iir then
            declare
               V : O_Lnode;
            begin
               --  declare
               --     v : __ghdl_bool_type_node;
               --  begin
               --     v := suspend ();
               --  end;
               Open_Temp;
               V := Create_Temp (Ghdl_Bool_Type_Node);
               Start_Association (Constr, Ghdl_Process_Wait_Suspend);
               New_Assign_Stmt (V, New_Function_Call (Constr));
               Close_Temp;
            end;
         else
            declare
               Loop_Blk : O_Loop_Block;
               Label : O_Snode;
               If_Blk : O_If_Block;
            begin
               --  start loop
               Start_Loop_Stmt (Loop_Blk, Label);

               --    if suspend() then        --  return true if timeout.
               --      exit;
               --    end if;
               Start_Association (Constr, Ghdl_Process_Wait_Suspend);
               Gen_Exit_When (Label, New_Function_Call (Constr));

               --    if condition then
               --      exit;
               --    end if;
               Open_Temp;
               Gen_Exit_When
                 (Label,
                  Chap7.Translate_Expression (Cond, Boolean_Type_Definition));
               Close_Temp;

               --  end loop;
               Finish_Loop_Stmt (Loop_Blk);
            end;
         end if;

         --  wait_close;
         Start_Association (Constr, Ghdl_Process_Wait_Close);
         New_Procedure_Call (Constr);
      end Translate_Wait_Statement;

      --  Signal assignment.
      procedure Gen_Simple_Signal_Assign_Non_Composite (Targ : Mnode;
                                                        Targ_Type : Iir;
                                                        Val : O_Enode)
      is
         Type_Info : Type_Info_Acc;
         Subprg : O_Dnode;
         Conv : O_Tnode;
         Assoc : O_Assoc_List;
      begin
         Type_Info := Get_Info (Targ_Type);
         case Type_Info.Type_Mode is
            when Type_Mode_B2 =>
               Subprg := Ghdl_Signal_Simple_Assign_B2;
               Conv := Ghdl_Bool_Type_Node;
            when Type_Mode_E8 =>
               Subprg := Ghdl_Signal_Simple_Assign_E8;
               Conv := Ghdl_Integer_Type;
            when Type_Mode_I32 =>
               Subprg := Ghdl_Signal_Simple_Assign_I32;
               Conv := Ghdl_Integer_Type;
            when Type_Mode_P64 =>
               Subprg := Ghdl_Signal_Simple_Assign_P64;
               Conv := Ghdl_Phys_Type;
            when Type_Mode_F64 =>
               Subprg := Ghdl_Signal_Simple_Assign_F64;
               Conv := Ghdl_Real_Type;
            when Type_Mode_Array =>
               raise Internal_Error;
            when others =>
               Error_Kind ("gen_signal_assign_non_composite", Targ_Type);
         end case;
         if Chap3.Need_Range_Check (Targ_Type) then
            declare
               If_Blk : O_If_Block;
               Val2 : O_Lnode;
               Targ2 : O_Lnode;
            begin
               Open_Temp;
               Val2 := Create_Temp_Init
                 (Type_Info.Ortho_Type (Mode_Value), Val);
               Targ2 := Create_Temp_Init
                 (Ghdl_Signal_Ptr, New_Convert_Ov (New_Value (M2Lv (Targ)),
                                                   Ghdl_Signal_Ptr));
               Start_If_Stmt (If_Blk, Chap3.Not_In_Range (Val2, Targ_Type));
               Start_Association (Assoc, Ghdl_Signal_Simple_Assign_Error);
               New_Association (Assoc, New_Value (Targ2));
               New_Procedure_Call (Assoc);
               New_Else_Stmt (If_Blk);
               Start_Association (Assoc, Subprg);
               New_Association (Assoc, New_Value (Targ2));
               New_Association
                 (Assoc, New_Convert_Ov (New_Value (Val2), Conv));
               New_Procedure_Call (Assoc);
               Finish_If_Stmt (If_Blk);
               Close_Temp;
            end;
         else
            Start_Association (Assoc, Subprg);
            New_Association (Assoc, New_Convert_Ov (New_Value (M2Lv (Targ)),
                                                    Ghdl_Signal_Ptr));
            New_Association (Assoc, New_Convert_Ov (Val, Conv));
            New_Procedure_Call (Assoc);
         end if;
      end Gen_Simple_Signal_Assign_Non_Composite;

--       procedure Gen_Simple_Signal_Prepare_Data_Composite (Val : O_Enode;
--                                                          Targ_Type : Iir) is
--       begin
--          null;
--       end Gen_Simple_Signal_Prepare_Data_Composite;

--       function Gen_Simple_Signal_Update_Data_Array (Val : O_Enode;
--                                                     Targ_Type : Iir;
--                                                     Index : O_Lnode)
--         return O_Enode
--       is
--          Base : O_Lnode;
--       begin
--          Base := Chap3.Get_Array_Base
--            (New_Access_Element (Val), Targ_Type, Mode_Value);
--          return New_Value (New_Indexed_Element (Base, New_Value (Index)));
--       end Gen_Simple_Signal_Update_Data_Array;

--       function Gen_Simple_Signal_Update_Data_Record
--         (Val : O_Enode; Targ_Type : Iir; El : Iir_Element_Declaration)
--         return O_Enode
--       is
--       begin
--          return New_Value (New_Selected_Element
--                            (New_Access_Element (Val),
--                             Get_Info (El).Field_Node (Mode_Value)));
--       end Gen_Simple_Signal_Update_Data_Record;

      procedure Gen_Simple_Signal_Assign is new Foreach_Non_Composite
        (Data_Type => O_Enode,
         Composite_Data_Type => Mnode,
         Do_Non_Composite => Gen_Simple_Signal_Assign_Non_Composite,
         Prepare_Data_Array => Gen_Oenode_Prepare_Data_Composite,
         Update_Data_Array => Gen_Oenode_Update_Data_Array,
         Prepare_Data_Record => Gen_Oenode_Prepare_Data_Composite,
         Update_Data_Record => Gen_Oenode_Update_Data_Record);

      type Signal_Assign_Data is record
         Expr : O_Enode;
         Reject : O_Lnode;
         After : O_Lnode;
      end record;

      procedure Gen_Start_Signal_Assign_Non_Composite
        (Targ : Mnode; Targ_Type : Iir; Data : Signal_Assign_Data)
      is
         Type_Info : Type_Info_Acc;
         Subprg : O_Dnode;
         Conv : O_Tnode;
         Assoc : O_Assoc_List;
      begin
         if Data.Expr = O_Enode_Null then
            --  Null transaction.
            Start_Association (Assoc, Ghdl_Signal_Start_Assign_Null);
            New_Association (Assoc, New_Convert_Ov (New_Value (M2Lv (Targ)),
                                                    Ghdl_Signal_Ptr));
            New_Association (Assoc, New_Value (Data.Reject));
            New_Association (Assoc, New_Value (Data.After));
            New_Procedure_Call (Assoc);
            return;
         end if;

         Type_Info := Get_Info (Targ_Type);
         case Type_Info.Type_Mode is
            when Type_Mode_B2 =>
               Subprg := Ghdl_Signal_Start_Assign_B2;
               Conv := Ghdl_Bool_Type_Node;
            when Type_Mode_E8 =>
               Subprg := Ghdl_Signal_Start_Assign_E8;
               Conv := Ghdl_Integer_Type;
            when Type_Mode_I32 =>
               Subprg := Ghdl_Signal_Start_Assign_I32;
               Conv := Ghdl_Integer_Type;
            when Type_Mode_P64 =>
               Subprg := Ghdl_Signal_Start_Assign_P64;
               Conv := Ghdl_Phys_Type;
            when Type_Mode_F64 =>
               Subprg := Ghdl_Signal_Start_Assign_F64;
               Conv := Ghdl_Real_Type;
            when Type_Mode_Array =>
               raise Internal_Error;
            when others =>
               Error_Kind ("gen_signal_assign_non_composite", Targ_Type);
         end case;
         --  Check range.
         if Chap3.Need_Range_Check (Targ_Type) then
            declare
               If_Blk : O_If_Block;
               V : O_Lnode;
               Starg : O_Lnode;
            begin
               Open_Temp;
               V := Create_Temp_Init
                 (Type_Info.Ortho_Type (Mode_Value), Data.Expr);
               Starg := Create_Temp_Init
                 (Ghdl_Signal_Ptr,
                  New_Convert_Ov (New_Value (M2Lv (Targ)), Ghdl_Signal_Ptr));
               Start_If_Stmt (If_Blk, Chap3.Not_In_Range (V, Targ_Type));
               Start_Association (Assoc, Ghdl_Signal_Start_Assign_Error);
               New_Association (Assoc, New_Value (Starg));
               New_Association (Assoc, New_Value (Data.Reject));
               New_Association (Assoc, New_Value (Data.After));
               New_Procedure_Call (Assoc);
               New_Else_Stmt (If_Blk);
               Start_Association (Assoc, Subprg);
               New_Association (Assoc, New_Value (Starg));
               New_Association (Assoc, New_Value (Data.Reject));
               New_Association (Assoc, New_Convert_Ov (New_Value (V), Conv));
               New_Association (Assoc, New_Value (Data.After));
               New_Procedure_Call (Assoc);
               Finish_If_Stmt (If_Blk);
               Close_Temp;
            end;
         else
            Start_Association (Assoc, Subprg);
            New_Association (Assoc, New_Convert_Ov (New_Value (M2Lv (Targ)),
                                                    Ghdl_Signal_Ptr));
            New_Association (Assoc, New_Value (Data.Reject));
            New_Association (Assoc, New_Convert_Ov (Data.Expr, Conv));
            New_Association (Assoc, New_Value (Data.After));
            New_Procedure_Call (Assoc);
         end if;
      end Gen_Start_Signal_Assign_Non_Composite;

      function Gen_Signal_Prepare_Data_Composite
        (Val : Signal_Assign_Data; Targ_Type : Iir)
        return Signal_Assign_Data is
      begin
         return Val;
      end Gen_Signal_Prepare_Data_Composite;

      function Gen_Signal_Update_Data_Array
        (Val : Signal_Assign_Data;
         Targ_Type : Iir;
         Index : O_Lnode)
        return Signal_Assign_Data
      is
         Res : Signal_Assign_Data;
      begin
         if Val.Expr = O_Enode_Null then
            --  Handle null transaction.
            return Val;
         end if;
         Res := Signal_Assign_Data'
           (Expr => New_Value (New_Indexed_Element
                               (New_Access_Element (Val.Expr),
                                New_Value (Index))),
            Reject => Val.Reject,
            After => Val.After);
         return Res;
      end Gen_Signal_Update_Data_Array;

      function Gen_Signal_Update_Data_Record
        (Val : Signal_Assign_Data;
         Targ_Type : Iir;
         El : Iir_Element_Declaration)
        return Signal_Assign_Data
      is
         Res : Signal_Assign_Data;
      begin
         if Val.Expr = O_Enode_Null then
            --  Handle null transaction.
            return Val;
         end if;
         Res := Signal_Assign_Data'
           (Expr => New_Value (New_Selected_Element
                               (New_Access_Element (Val.Expr),
                                Get_Info (El).Field_Node (Mode_Value))),
            Reject => Val.Reject,
            After => Val.After);
         return Res;
      end Gen_Signal_Update_Data_Record;

      procedure Gen_Start_Signal_Assign is new Foreach_Non_Composite
        (Data_Type => Signal_Assign_Data,
         Composite_Data_Type => Signal_Assign_Data,
         Do_Non_Composite => Gen_Start_Signal_Assign_Non_Composite,
         Prepare_Data_Array => Gen_Signal_Prepare_Data_Composite,
         Update_Data_Array => Gen_Signal_Update_Data_Array,
         Prepare_Data_Record => Gen_Signal_Prepare_Data_Composite,
         Update_Data_Record => Gen_Signal_Update_Data_Record);

      procedure Gen_Next_Signal_Assign_Non_Composite
        (Targ : Mnode; Targ_Type : Iir; Data : Signal_Assign_Data)
      is
         Type_Info : Type_Info_Acc;
         Subprg : O_Dnode;
         Conv : O_Tnode;
         Assoc : O_Assoc_List;
      begin
         if Data.Expr = O_Enode_Null then
            --  Null transaction.
            Start_Association (Assoc, Ghdl_Signal_Next_Assign_Null);
            New_Association (Assoc, New_Convert_Ov (New_Value (M2Lv (Targ)),
                                                    Ghdl_Signal_Ptr));
            New_Association (Assoc, New_Value (Data.After));
            New_Procedure_Call (Assoc);
            return;
         end if;

         Type_Info := Get_Info (Targ_Type);
         case Type_Info.Type_Mode is
            when Type_Mode_B2 =>
               Subprg := Ghdl_Signal_Next_Assign_B2;
               Conv := Ghdl_Bool_Type_Node;
            when Type_Mode_E8 =>
               Subprg := Ghdl_Signal_Next_Assign_E8;
               Conv := Ghdl_Integer_Type;
            when Type_Mode_I32 =>
               Subprg := Ghdl_Signal_Next_Assign_I32;
               Conv := Ghdl_Integer_Type;
            when Type_Mode_P64 =>
               Subprg := Ghdl_Signal_Next_Assign_P64;
               Conv := Ghdl_Phys_Type;
            when Type_Mode_F64 =>
               Subprg := Ghdl_Signal_Next_Assign_F64;
               Conv := Ghdl_Real_Type;
            when Type_Mode_Array =>
               raise Internal_Error;
            when others =>
               Error_Kind ("gen_signal_next_assign_non_composite", Targ_Type);
         end case;
         -- FIXME: check in range.
         Start_Association (Assoc, Subprg);
         New_Association (Assoc, New_Convert_Ov (New_Value (M2Lv (Targ)),
                                                 Ghdl_Signal_Ptr));
         New_Association (Assoc, New_Convert_Ov (Data.Expr, Conv));
         New_Association (Assoc, New_Value (Data.After));
         New_Procedure_Call (Assoc);
      end Gen_Next_Signal_Assign_Non_Composite;

      procedure Gen_Next_Signal_Assign is new Foreach_Non_Composite
        (Data_Type => Signal_Assign_Data,
         Composite_Data_Type => Signal_Assign_Data,
         Do_Non_Composite => Gen_Next_Signal_Assign_Non_Composite,
         Prepare_Data_Array => Gen_Signal_Prepare_Data_Composite,
         Update_Data_Array => Gen_Signal_Update_Data_Array,
         Prepare_Data_Record => Gen_Signal_Prepare_Data_Composite,
         Update_Data_Record => Gen_Signal_Update_Data_Record);

      procedure Translate_Signal_Assignment_Statement (Stmt : Iir)
      is
         Target : Iir;
         Target_Type : Iir;
         Wl : Iir_List;
         We : Iir_Waveform_Element;
         Targ : Mnode;
         Val : O_Enode;
         Value : Iir;
      begin
         Target := Get_Target (Stmt);
         Target_Type := Get_Type (Target);
         if Get_Kind (Target) = Iir_Kind_Aggregate then
            --  FIXME: todo
            Error_Kind ("translate_signal_assignment(1)", Target);
         end if;
         Wl := Get_Waveform_List (Stmt);
         if Wl = Null_Iir then
            --  FIXME: disconnection.
            Error_Kind ("translate_signal_assignment(2)", Stmt);
         end if;
         Targ := Chap6.Translate_Name (Target);
         if Get_Object_Kind (Targ) /= Mode_Signal then
            raise Internal_Error;
         end if;

         --  Handle a simple and common case: only one waveform, inertial,
         --  and no time (eg: sig <= expr).
         We := Get_Nth_Element (Wl, 0);
         Value := Get_We_Value (We);
         if Get_Nbr_Elements (Wl) = 1
           and then Get_Time (We) = Null_Iir
           and then Get_Delay_Mechanism (Stmt) = Iir_Inertial_Delay
           and then Get_Reject_Time_Expression (Stmt) = Null_Iir
           and then Get_Kind (Value) /= Iir_Kind_Null_Literal
         then
            Val := Chap7.Translate_Expression (Get_We_Value (We), Target_Type);
            Gen_Simple_Signal_Assign (Targ, Target_Type, Val);
            return;
         end if;

         --  General case.
         declare
            Var_Targ : Mnode;
            Targ_Tinfo : Type_Info_Acc;
         begin
            Open_Temp;
            Targ_Tinfo := Get_Info (Target_Type);
            Var_Targ := Stabilize (Targ, True);

            --  Translate the first waveform element.
            declare
               Reject_Time : O_Lnode;
               After_Time : O_Lnode;
               Del : Iir;
               Rej : Iir;
               Val : O_Enode;
            begin
               Open_Temp;
               Reject_Time := Create_Temp (Std_Time_Type);
               After_Time := Create_Temp (Std_Time_Type);
               Del := Get_Time (We);
               if Del = Null_Iir then
                  New_Assign_Stmt (After_Time,
                                   New_Signed_Literal (Std_Time_Type, 0));
               else
                  New_Assign_Stmt
                    (After_Time,
                     Chap7.Translate_Expression (Del, Time_Type_Definition));
               end if;
               case Get_Delay_Mechanism (Stmt) is
                  when Iir_Transport_Delay =>
                     New_Assign_Stmt (Reject_Time,
                                      New_Signed_Literal (Std_Time_Type, 0));
                  when Iir_Inertial_Delay =>
                     Rej := Get_Reject_Time_Expression (Stmt);
                     if Rej = Null_Iir then
                        New_Assign_Stmt (Reject_Time, New_Value (After_Time));
                     else
                        New_Assign_Stmt
                          (Reject_Time, Chap7.Translate_Expression
                           (Rej, Time_Type_Definition));
                     end if;
               end case;
               if Get_Kind (Value) = Iir_Kind_Null_Literal then
                  Val := O_Enode_Null;
               else
                  Val := Chap7.Translate_Expression (Value, Target_Type);
               end if;
               Gen_Start_Signal_Assign
                 (Var_Targ, Target_Type,
                  Signal_Assign_Data'(Expr => Val,
                                      Reject => Reject_Time,
                                      After => After_Time));
               Close_Temp;
            end;

            --  Translate other waveform elements.
            for I in 1 .. Natural'Last loop
               We := Get_Nth_Element (Wl, I);
               exit when We = Null_Iir;
               declare
                  After_Time : O_Lnode;
                  Val : O_Enode;
               begin
                  Open_Temp;
                  After_Time := Create_Temp (Std_Time_Type);
                  New_Assign_Stmt
                    (After_Time,
                     Chap7.Translate_Expression (Get_Time (We),
                                                 Time_Type_Definition));
                  Value := Get_We_Value (We);
                  if Get_Kind (Value) = Iir_Kind_Null_Literal then
                     Val := O_Enode_Null;
                  else
                     Val := Chap7.Translate_Expression (Value, Target_Type);
                  end if;
                  Gen_Next_Signal_Assign
                    (Var_Targ, Target_Type,
                     Signal_Assign_Data'(Expr => Val,
                                         Reject => O_Lnode_Null,
                                         After => After_Time));
                  Close_Temp;
               end;
            end loop;

            Close_Temp;
         end;
      end Translate_Signal_Assignment_Statement;

      procedure Translate_Statement (Stmt : Iir)
      is
      begin
         New_Debug_Line_Stmt (Get_Line_Number (Stmt));
         Open_Temp;
         case Get_Kind (Stmt) is
            when Iir_Kind_Return_Statement =>
               Translate_Return_Statement (Stmt);

            when Iir_Kind_If_Statement =>
               Translate_If_Statement (Stmt);
            when Iir_Kind_Assertion_Statement =>
               Translate_Assertion_Statement (Stmt);
            when Iir_Kind_Case_Statement =>
               Translate_Case_Statement (Stmt);

            when Iir_Kind_For_Loop_Statement =>
               Translate_For_Loop_Statement (Stmt);
            when Iir_Kind_While_Loop_Statement =>
               Translate_While_Loop_Statement (Stmt);
            when Iir_Kind_Next_Statement
              | Iir_Kind_Exit_Statement =>
               Translate_Exit_Next_Statement (Stmt);

            when Iir_Kind_Signal_Assignment_Statement =>
               Translate_Signal_Assignment_Statement (Stmt);
            when Iir_Kind_Variable_Assignment_Statement =>
               Translate_Variable_Assignment_Statement (Stmt);

            when Iir_Kind_Null_Statement =>
               --  A null statement is translated to a NOP, so that the
               --  statement generates code (and a breakpoint can be set on
               --  it).
               --  Emit_Nop;
               null;

            when Iir_Kind_Procedure_Call =>
               declare
                  Imp : Iir;
               begin
                  Imp := Get_Implementation (Stmt);
                  if Get_Kind (Imp) = Iir_Kind_Implicit_Procedure_Declaration
                  then
                     Translate_Implicit_Procedure_Call (Stmt);
                  else
                     Translate_Procedure_Call (Stmt);
                  end if;
               end;

            when Iir_Kind_Wait_Statement =>
               Translate_Wait_Statement (Stmt);

            when others =>
               Error_Kind ("translate_statement", Stmt);
         end case;
         Close_Temp;
      end Translate_Statement;

      procedure Translate_Statements_List (List : Iir_List)
      is
         Stmt : Iir;
      begin
         for I in Natural loop
            Stmt := Get_Nth_Element (List, I);
            exit when Stmt = null;
            Translate_Statement (Stmt);
         end loop;
      end Translate_Statements_List;
   end Chap8;

   package body Chap9 is
      procedure Translate_Process_Statement (Proc : Iir; Base : Block_Info_Acc)
      is
         Inter_List : O_Inter_List;
         Instance : O_Lnode;
         Info : Proc_Info_Acc;
      begin
         Info := Get_Info (Proc);
         Start_Procedure_Decl (Inter_List, Create_Identifier ("PROC"),
                               O_Storage_Private);
         New_Interface_Decl (Inter_List, Instance, Wki_Instance,
                             Base.Block_Decls_Ptr_Type);
         Finish_Subprogram_Decl (Inter_List, Info.Process_Subprg);

         Start_Subprogram_Body (Info.Process_Subprg);
         Push_Local_Factory;
         --  Push scope for process declarations
         Push_Scope (Info.Process_Decls_Type, Info.Process_Parent_Field,
                     Base.Block_Decls_Type);
         --  Push scope for architecture declarations.
         Push_Scope (Base.Block_Decls_Type, Instance);
         Chap8.Translate_Statements_List (Get_Process_Statement_List (Proc));
         Pop_Scope (Base.Block_Decls_Type);
         Pop_Scope (Info.Process_Decls_Type);
         Pop_Local_Factory;
         Finish_Subprogram_Body;
      end Translate_Process_Statement;

      procedure Translate_Implicit_Guard_Signal
        (Guard : Iir; Base : Block_Info_Acc)
      is
         Info : Object_Info_Acc;
         Inter_List : O_Inter_List;
         Instance : O_Lnode;
         Guard_Expr : Iir;
      begin
         Guard_Expr := Get_Guard_Expression (Guard);
         --  Create the subprogram to compute the value of GUARD.
         Info := Get_Info (Guard);
         Start_Function_Decl (Inter_List, Create_Identifier ("_GUARD_PROC"),
                              O_Storage_Private, Std_Boolean_Type_Node);
         New_Interface_Decl (Inter_List, Instance, Wki_Instance,
                             Base.Block_Decls_Ptr_Type);
         Finish_Subprogram_Decl (Inter_List, Info.Object_Function);

         Start_Subprogram_Body (Info.Object_Function);
         Push_Local_Factory;
         Push_Scope (Base.Block_Decls_Type, Instance);
         Open_Temp;
         New_Return_Stmt (Chap7.Translate_Expression (Guard_Expr));
         Close_Temp;
         Pop_Scope (Base.Block_Decls_Type);
         Pop_Local_Factory;
         Finish_Subprogram_Body;
      end Translate_Implicit_Guard_Signal;

      --  Create the instance for block BLOCK.
      --  BLOCK can be either an entity, an architecture or a block statement.
      procedure Translate_Block_Declarations (Block : Iir; Origin : Iir)
      is
         Decl_List : Iir_List;
         Stmt_List : Iir_List;
         El : Iir;
      begin
         Decl_List := Get_Declaration_List (Block);
         Chap4.Translate_Declaration_List (Decl_List, Null_Iir_List);

         Stmt_List := Get_Concurrent_Statement_List (Block);
         if Stmt_List = Null_Iir then
            return;
         end if;
         for I in Natural loop
            El := Get_Nth_Element (Stmt_List, I);
            exit when El = Null_Iir;
            case Get_Kind (El) is
               when Iir_Kind_Process_Statement
                 | Iir_Kind_Sensitized_Process_Statement =>
                  declare
                     Mark : Id_Mark_Type;
                     Info : Ortho_Info_Acc;
                     Itype : O_Tnode;
                     Field : O_Fnode;
                  begin
                     Push_Identifier_Prefix (Mark, Get_Identifier (El));
                     --  Start child record.
                     Push_Instance_Factory (O_Tnode_Null);
                     Info := Add_Info (El, Kind_Process);
                     Info.Process_Name := Add_Instance_Factory_Field
                       (Wki_Name, Ghdl_Instance_Name_Type);
                     Chap4.Translate_Declaration_List
                       (Get_Declaration_List (El), Null_Iir_List);
                     Pop_Instance_Factory (Itype);
                     New_Type_Decl (Create_Identifier ("INSTTYPE"), Itype);
                     Pop_Identifier_Prefix (Mark);
                     --  Create a field in the parent record.
                     Field := Add_Instance_Factory_Field
                       (Create_Identifier_Without_Prefix (El), Itype);
                     --  Set info in child record.
                     Info.Process_Decls_Type := Itype;
                     Info.Process_Parent_Field := Field;
                  end;
               when Iir_Kind_Component_Instantiation_Statement =>
                  declare
                     Comp : Iir;
                     Field : O_Fnode;
                     Info : Block_Info_Acc;
                     Comp_Info : Comp_Info_Acc;
                  begin
                     Comp := Get_Instantiated_Unit (El);
                     if Get_Kind (Comp) = Iir_Kind_Component_Declaration then
                        Comp_Info := Get_Info (Comp);
                        Field := Add_Instance_Factory_Field
                          (Create_Identifier_Without_Prefix (El),
                           Comp_Info.Comp_Type);
                        Info := Add_Info (El, Kind_Block);
                        Info.Block_Decls_Type := Comp_Info.Comp_Type;
                        Info.Block_Decls_Ptr_Type := Comp_Info.Comp_Ptr_Type;
                        Info.Block_Name := Comp_Info.Comp_Name;
                        Info.Block_Parent_Field := Field;
                     end if;
                  end;
               when Iir_Kind_Block_Statement =>
                  declare
                     Info : Block_Info_Acc;
                     Hdr : Iir_Block_Header;
                     Guard : Iir;
                     Mark : Id_Mark_Type;
                     Field : O_Fnode;
                  begin
                     Push_Identifier_Prefix (Mark, Get_Identifier (El));

                     Info := Add_Info (El, Kind_Block);
                     Chap1.Start_Block_Decl (El);
                     Push_Instance_Factory (Info.Block_Decls_Type);

                     --  Block name.
                     Info.Block_Name := Add_Instance_Factory_Field
                       (Wki_Name, Ghdl_Instance_Name_Type);

                     Guard := Get_Guard_Decl (El);
                     if Guard /= Null_Iir then
                        Chap4.Translate_Declaration (Guard);
                     end if;

                     --  generics, ports.
                     Hdr := Get_Block_Header (El);
                     if Hdr /= Null_Iir then
                        Chap4.Translate_Generic_List (Get_Generic_List (Hdr));
                        Chap4.Translate_Port_List (Get_Port_List (Hdr));
                     end if;

                     Chap9.Translate_Block_Declarations (El, Origin);

                     Pop_Instance_Factory (Info.Block_Decls_Type);
                     Pop_Identifier_Prefix (Mark);

                     --  Create a field in the parent record.
                     Field := Add_Instance_Factory_Field
                       (Create_Identifier_Without_Prefix (El),
                        Info.Block_Decls_Type);
                     --  Set info in child record.
                     Info.Block_Parent_Field := Field;
                  end;
               when Iir_Kind_Generate_Statement =>
                  declare
                     Info : Block_Info_Acc;
                     Mark : Id_Mark_Type;
                     Scheme : Iir;
                     Iter_Type : Iir;
                     It_Info : Ortho_Info_Acc;
                  begin
                     Scheme := Get_Generation_Scheme (El);
                     if Get_Kind (Scheme) = Iir_Kind_Iterator_Declaration then
                        Iter_Type := Get_Type (Scheme);
                        Chap3.Translate_Object_Subtype (Scheme, False);
                     end if;

                     Push_Identifier_Prefix (Mark, Get_Identifier (El));

                     Info := Add_Info (El, Kind_Block);
                     Chap1.Start_Block_Decl (El);
                     Push_Instance_Factory (Info.Block_Decls_Type);

                     --  Add a parent field in the current instance.
                     Info.Block_Origin_Field := Add_Instance_Factory_Field
                       (Get_Identifier ("ORIGIN"),
                        Get_Info (Origin).Block_Decls_Ptr_Type);

                     --  Block name.
                     Info.Block_Name := Add_Instance_Factory_Field
                       (Wki_Name, Ghdl_Instance_Name_Type);

                     --  Iterator.
                     if Get_Kind (Scheme) = Iir_Kind_Iterator_Declaration then
                        Info.Block_Iterator_TD_Field :=
                          Add_Instance_Factory_Field
                          (Get_Identifier ("ITERATOR_TD"), Ghdl_Desc_Ptr_Type);
                        Info.Block_Iterator_Addr_Field :=
                          Add_Instance_Factory_Field
                          (Get_Identifier ("ITERATOR_ADDR"), Ghdl_Ptr_Type);
                        Info.Block_Configured_Field :=
                          Add_Instance_Factory_Field
                          (Get_Identifier ("CONFIGURED"), Ghdl_Bool_Type_Node);
                        It_Info := Add_Info (Scheme, Kind_Iterator);
                        It_Info.Iterator_Var := Create_Var
                          (Create_Var_Identifier (Scheme),
                           Get_Info (Get_Base_Type (Iter_Type)).Ortho_Type
                           (Mode_Value));
                     end if;

                     Chap9.Translate_Block_Declarations (El, El);

                     Pop_Instance_Factory (Info.Block_Decls_Type);
                     Pop_Identifier_Prefix (Mark);

                     if Get_Kind (Scheme) = Iir_Kind_Iterator_Declaration then
                        --  Create array type of block_decls_type
                        Info.Block_Decls_Array_Type := New_Array_Type
                          (Info.Block_Decls_Type, Ghdl_Index_Type);
                        New_Type_Decl (Create_Identifier ("INSTARRTYPE"),
                                       Info.Block_Decls_Array_Type);
                        --  Create access to the array type.
                        Info.Block_Decls_Array_Ptr_Type := New_Access_Type
                          (Info.Block_Decls_Array_Type);
                        New_Type_Decl (Create_Identifier ("INSTARRPTR"),
                                       Info.Block_Decls_Array_Ptr_Type);
                        --  Iterator range field in the parent record.
                        Info.Block_Range_Field := Add_Instance_Factory_Field
                          (Create_Identifier_Without_Prefix (El, "_RANGE"),
                           Get_Info (Get_Base_Type (Iter_Type)).T.Range_Type);
                        --  Add a field in parent record
                        Info.Block_Parent_Field := Add_Instance_Factory_Field
                          (Create_Identifier_Without_Prefix (El),
                           Info.Block_Decls_Array_Ptr_Type);
                     else
                        --  Create an access field in the parent record.
                        Info.Block_Parent_Field := Add_Instance_Factory_Field
                          (Create_Identifier_Without_Prefix (El),
                           Info.Block_Decls_Ptr_Type);
                     end if;
                  end;
               when others =>
                  Error_Kind ("translate_block_declarations", El);
            end case;
         end loop;
      end Translate_Block_Declarations;

      --  Translate concurrent statements into subprograms.
      procedure Translate_Block_Subprograms (Block : Iir; Base_Block : Iir)
      is
         Stmts_List : Iir_List;
         Stmt : Iir;
         Mark : Id_Mark_Type;
         Block_Info : Block_Info_Acc;
         Decl_List : Iir_List;
         Base_Info : Block_Info_Acc;
      begin
         Decl_List := Get_Declaration_List (Block);
         Base_Info := Get_Info (Base_Block);
         Chap4.Translate_Declaration_List_Subprograms (Decl_List, Base_Block);

         Stmts_List := Get_Concurrent_Statement_List (Block);
         if Stmts_List = Null_Iir then
            return;
         end if;
         Block_Info := Get_Info (Block);
         for I in Natural loop
            Stmt := Get_Nth_Element (Stmts_List, I);
            exit when Stmt = Null_Iir;
            Push_Identifier_Prefix (Mark, Get_Identifier (Stmt));
            case Get_Kind (Stmt) is
               when Iir_Kind_Process_Statement
                 | Iir_Kind_Sensitized_Process_Statement =>
                  declare
                     Info : Proc_Info_Acc;
                  begin
                     Info := Get_Info (Stmt);
                     Push_Scope (Info.Process_Decls_Type,
                                 Info.Process_Parent_Field,
                                 Block_Info.Block_Decls_Type);
                     Chap4.Translate_Declaration_List_Subprograms
                       (Get_Declaration_List (Stmt), Base_Block);
                     Translate_Process_Statement (Stmt, Base_Info);
                     Pop_Scope (Info.Process_Decls_Type);
                  end;
               when Iir_Kind_Component_Instantiation_Statement =>
                  null;
               when Iir_Kind_Block_Statement =>
                  declare
                     Info : Block_Info_Acc;
                     Guard : Iir;
                  begin
                     Info := Get_Info (Stmt);
                     Push_Scope (Info.Block_Decls_Type,
                                 Info.Block_Parent_Field,
                                 Block_Info.Block_Decls_Type);
                     Guard := Get_Guard_Decl (Stmt);
                     if Guard /= Null_Iir then
                        Translate_Implicit_Guard_Signal (Guard, Base_Info);
                     end if;
                     Translate_Block_Subprograms (Stmt, Base_Block);
                     Pop_Scope (Info.Block_Decls_Type);
                  end;
               when Iir_Kind_Generate_Statement =>
                  declare
                     Info : Block_Info_Acc;
                  begin
                     Info := Get_Info (Stmt);
                     Push_Scope_Via_Field_Ptr (Block_Info.Block_Decls_Type,
                                               Info.Block_Origin_Field,
                                               Info.Block_Decls_Type);
                     Translate_Block_Subprograms (Stmt, Stmt);
                     Pop_Scope (Block_Info.Block_Decls_Type);
                  end;
               when others =>
                  Error_Kind ("translate_block_subprograms", Stmt);
            end case;
            Pop_Identifier_Prefix (Mark);
         end loop;
      end Translate_Block_Subprograms;

      --  Remove anonymous and implicit type definitions in a list of names.
      --  Such type definitions are created during slice translations, however
      --  variables created are defined in the translation scope.
      --  If the type is referenced again, the variables must be reachable.
      --  This is not the case for elaborator subprogram (which may references
      --  slices in the sensitivity or driver list) and the process subprg.
      procedure Destroy_Types_In_List (List : Iir_List)
      is
         El : Iir;
         Atype : Iir;
         Info : Type_Info_Acc;
      begin
         if List = Null_Iir_List then
            return;
         end if;
         for I in Natural loop
            El := Get_Nth_Element (List, I);
            exit when El = Null_Iir;
            loop
               Atype := Null_Iir;
               case Get_Kind (El) is
                  when Iir_Kind_Selected_Element
                    | Iir_Kind_Indexed_Name =>
                     El := Get_Prefix (El);
                  when Iir_Kind_Slice_Name =>
                     Atype := Get_Type (El);
                     El := Get_Prefix (El);
                  when Iir_Kind_Stable_Attribute
                    | Iir_Kind_Quiet_Attribute
                    | Iir_Kind_Delayed_Attribute
                    | Iir_Kind_Transaction_Attribute =>
                     El := Get_Prefix (El);
                  when Iir_Kind_Signal_Declaration
                    | Iir_Kind_Signal_Interface_Declaration
                    | Iir_Kind_Guard_Signal_Declaration =>
                     exit;
                  when others =>
                     Error_Kind ("destroy_types_in_list", El);
               end case;
               if Atype /= Null_Iir
                 and then Is_Anonymous_Type_Definition (Atype)
               then
                  Info := Get_Info (Atype);
                  if Info /= null then
                     Free_Type_Info (Info);
                     Set_Back_End_Info (Atype, System.Null_Address);
                  end if;
               end if;
            end loop;
         end loop;
      end Destroy_Types_In_List;

      --  PROC: the process to be elaborated
      --  BLOCK_INFO: info for the block containing the process
      --  BASE_INFO: info for the global block
      procedure Elab_Process (Proc : Iir;
                              Block_Info : Block_Info_Acc;
                              Base_Info : Block_Info_Acc)
      is
         Is_Sensitized : Boolean;
         Subprg : O_Dnode;
         Constr : O_Assoc_List;
         Info : Proc_Info_Acc;
         List : Iir_List;
      begin
         New_Debug_Line_Stmt (Get_Line_Number (Proc));

         Is_Sensitized :=
           Get_Kind (Proc) = Iir_Kind_Sensitized_Process_Statement;
         Info := Get_Info (Proc);

         --  Set instance name.
         Push_Scope (Info.Process_Decls_Type,
                     Info.Process_Parent_Field,
                     Block_Info.Block_Decls_Type);
         Chap6.Set_Instance_Name (Proc);

         --  Register process.
         if Is_Sensitized then
            Subprg := Ghdl_Sensitized_Process_Register;
         else
            Subprg := Ghdl_Process_Register;
         end if;

         Start_Association (Constr, Subprg);
         New_Association
           (Constr, New_Unchecked_Address
            (Get_Instance_Ref (Base_Info.Block_Decls_Type), Ghdl_Ptr_Type));
         New_Association
           (Constr,
            New_Subprogram_Address (Info.Process_Subprg, Ghdl_Ptr_Type));
         New_Association
           (Constr,
            New_Address (New_Selected_Element
                         (Get_Instance_Ref (Info.Process_Decls_Type),
                          Info.Process_Name),
                         Ghdl_Instance_Name_Acc));
         New_Procedure_Call (Constr);

         List := Get_Driver_List (Proc);
         Destroy_Types_In_List (List);
         Register_Signal_List (List, Ghdl_Process_Add_Driver);

         if Is_Sensitized then
            List := Get_Sensitivity_List (Proc);
            Destroy_Types_In_List (List);
            Register_Signal_List (List, Ghdl_Process_Add_Sensitivity);
         end if;

         Chap4.Elab_Declaration_List (Get_Declaration_List (Proc), Proc);

         Pop_Scope (Info.Process_Decls_Type);
      end Elab_Process;

      procedure Elab_Implicit_Guard_Signal
        (Block : Iir_Block_Statement; Block_Info : Block_Info_Acc)
      is
         Guard : Iir;
         Type_Info : Type_Info_Acc;
         Info : Object_Info_Acc;
         Constr : O_Assoc_List;
      begin
         --  Create the guard signal.
         Guard := Get_Guard_Decl (Block);
         Info := Get_Info (Guard);
         Type_Info := Get_Info (Get_Type (Guard));
         Start_Association (Constr, Ghdl_Signal_Create_Guard);
         New_Association
           (Constr, New_Unchecked_Address
            (Get_Instance_Ref (Block_Info.Block_Decls_Type), Ghdl_Ptr_Type));
         New_Association
           (Constr,
            New_Subprogram_Address (Info.Object_Function, Ghdl_Ptr_Type));
         New_Association (Constr, Chap6.Get_Instance_Name_Ref (Block));
         New_Assign_Stmt (Get_Var (Info.Object_Var),
                          New_Convert_Ov (New_Function_Call (Constr),
                                          Type_Info.Ortho_Type (Mode_Signal)));

         --  Register sensitivity list of the guard signal.
         Register_Signal_List (Get_Guard_Sensitivity_List (Guard),
                               Ghdl_Signal_Guard_Dependence);
      end Elab_Implicit_Guard_Signal;

      procedure Translate_Entity_Instantiation
        (Aspect : Iir; Mapping : Iir; Parent : Iir; Config_Override : Iir)
      is
         Entity_Unit : Iir_Design_Unit;
         Config : Iir;
         Arch : Iir;
         Entity : Iir_Entity_Declaration;
         Entity_Info : Block_Info_Acc;
         Arch_Info : Block_Info_Acc;

         Instance_Size : O_Lnode;
         Arch_Elab : O_Dnode;
         Arch_Config : O_Dnode;
         Arch_Config_Type : O_Tnode;

         Var_Sub : O_Lnode;
      begin
         case Get_Kind (Aspect) is
            when Iir_Kind_Entity_Aspect_Entity =>
               Entity_Unit := Get_Entity (Aspect);
               Arch := Get_Architecture (Aspect);
               if Flag_Elaborate and then Arch = Null_Iir then
                  --  This is valid only during elaboration.
                  Arch := Libraries.Get_Latest_Architecture
                    (Get_Library_Unit (Entity_Unit));
               end if;
            when Iir_Kind_Entity_Aspect_Configuration =>
               Config := Get_Library_Unit (Get_Configuration (Aspect));
               Entity_Unit := Get_Entity (Config);
               Arch := Get_Block_Specification
                 (Get_Block_Configuration (Config));
            when Iir_Kind_Entity_Aspect_Open =>
               return;
            when others =>
               Error_Kind
                 ("translate_entity_instantiation", Aspect);
         end case;
         Entity := Get_Library_Unit (Entity_Unit);
         Entity_Info := Get_Info (Entity);
         if Config_Override /= Null_Iir then
            Config := Config_Override;
            if Get_Kind (Arch) = Iir_Kind_Simple_Name then
               Arch := Get_Block_Specification
                 (Get_Block_Configuration (Config));
            end if;
         end if;

         --  1) Create instance for the arch
         if Arch /= Null_Iir then
            Arch_Info := Get_Info (Arch);
         else
            Arch_Info := null;
         end if;
         if Arch_Info = null or Config = Null_Iir then
            declare
               function Get_Arch_Name return String is
               begin
                  if Arch /= Null_Iir then
                     return Get_String (Arch);
                  else
                     return "LASTARCH";
                  end if;
               end Get_Arch_Name;

               Str : String :=
                 Get_String (Get_Parent (Get_Design_File (Entity_Unit)))
                 & "__" & Get_String (Entity) & "__"
                 & Get_Arch_Name & "__";
               Sub_Inter : O_Inter_List;
               Arg : O_Lnode;
            begin
               if Arch_Info = null then
                  New_Const_Decl
                    (Instance_Size, Get_Identifier (Str & "INSTSIZE"),
                     O_Storage_External, Ghdl_Index_Type, O_Enode_Null);

                  Start_Procedure_Decl
                    (Sub_Inter, Get_Identifier (Str & "ELAB"),
                     O_Storage_External);
                  New_Interface_Decl (Sub_Inter, Arg, Wki_Instance,
                                      Entity_Info.Block_Decls_Ptr_Type);
                  Finish_Subprogram_Decl (Sub_Inter, Arch_Elab);
               end if;

               if Config = Null_Iir then
                  Start_Procedure_Decl
                    (Sub_Inter, Get_Identifier (Str & "DEFAULT_CONFIG"),
                     O_Storage_External);
                  New_Interface_Decl (Sub_Inter, Arg, Wki_Instance,
                                      Entity_Info.Block_Decls_Ptr_Type);
                  Finish_Subprogram_Decl (Sub_Inter, Arch_Config);

                  Arch_Config_Type := Entity_Info.Block_Decls_Ptr_Type;
               end if;
            end;
         end if;

         if Arch_Info = null then
            if Config /= Null_Iir then
               -- Should have been handled before: arch must have been
               -- extracted from config.
               raise Internal_Error;
            end if;
         else
            Instance_Size := Arch_Info.Block_Instance_Size;
            Arch_Elab := Arch_Info.Block_Elab_Subprg;
            if Config /= Null_Iir then
               Arch_Config := Get_Info (Config).Config_Subprg;
               Arch_Config_Type := Arch_Info.Block_Decls_Ptr_Type;
            end if;
         end if;

         New_Var_Decl (Var_Sub, Get_Identifier ("SUB_INSTANCE"),
                       O_Storage_Local, Entity_Info.Block_Decls_Ptr_Type);

         New_Assign_Stmt
           (Var_Sub, Gen_Alloc (Alloc_Memory, New_Value (Instance_Size),
                                Entity_Info.Block_Decls_Ptr_Type));
         Push_Scope (Entity_Info.Block_Decls_Type, Var_Sub);

         --  2) Map generics.
         Chap5.Elab_Generic_Map_Aspect_List
           (Get_Generic_Map_Aspect_List (Mapping));

         --  Set bounds of unconstrained ports.
         Chap5.Elab_Unconstrained_Ports_Of_Port_List
           (Get_Port_Map_Aspect_List (Mapping));

         --  3) Elab instance.
         declare
            Assoc : O_Assoc_List;
         begin
            Start_Association (Assoc, Arch_Elab);
            New_Association (Assoc, New_Value (Var_Sub));
            New_Procedure_Call (Assoc);
         end;

         --  3.5) Link instance name.
         --  This must be done after elaborating instance, the elaboration
         --  initialize the fields.
         Chap6.Link_Instance_Name (Parent, Entity);
--          --  var_sub.name.parent := instance.name'address
--          New_Assign_Stmt
--            (New_Selected_Element
--             (New_Selected_Acc_Value (Var_Sub, Entity_Info.Block_Name),
--              Ghdl_Instance_Name_Parent),
--             New_Address
--             (New_Selected_Acc_Value (Parent_Instance, Parent_Name_Field),
--              Ghdl_Instance_Name_Acc));
--          --  instance.name.child := var_sub.name'address
--          New_Assign_Stmt
--            (New_Selected_Element (New_Selected_Acc_Value
--                                   (Parent_Instance, Parent_Name_Field),
--                                   Ghdl_Instance_Name_Child),
--             New_Address (New_Selected_Acc_Value
--                          (Var_Sub, Entity_Info.Block_Name),
--                          Ghdl_Instance_Name_Acc));

         --  4) Map ports.
         Chap5.Elab_Port_Map_Aspect_List (Get_Port_Map_Aspect_List (Mapping));

         --  5) Configure
         declare
            Assoc : O_Assoc_List;
         begin
            Start_Association (Assoc, Arch_Config);
            New_Association
              (Assoc, New_Convert_Ov (New_Value (Var_Sub), Arch_Config_Type));
            New_Procedure_Call (Assoc);
         end;

         Pop_Scope (Entity_Info.Block_Decls_Type);
      end Translate_Entity_Instantiation;

      procedure Elab_Component_Instantiation
        (Stmt : Iir; Parent_Info : Block_Info_Acc)
      is
         Comp : Iir;
         Comp_Info : Comp_Info_Acc;
         Block_Info : Block_Info_Acc;
         Map_List : Iir_List;
      begin
         New_Debug_Line_Stmt (Get_Line_Number (Stmt));

         Comp := Get_Instantiated_Unit (Stmt);
         if Get_Kind (Comp) /= Iir_Kind_Component_Declaration then
            --  This is either an entity or a configuration.
            Start_Declare_Stmt;
            Translate_Entity_Instantiation
              (Comp, Stmt, Get_Parent (Stmt), Null_Iir);
            Finish_Declare_Stmt;
            return;
         end if;

         Comp_Info := Get_Info (Comp);
         Block_Info := Get_Info (Stmt);
         Push_Scope (Comp_Info.Comp_Type, Block_Info.Block_Parent_Field,
                     Parent_Info.Block_Decls_Type);
         New_Debug_Line_Stmt (Get_Line_Number (Stmt));
         Chap6.Set_Instance_Name (Stmt);

         --  Elab generics and ports, and binding.
         Chap4.Elab_Generic_List (Get_Generic_List (Comp));

         --  Note: the generic map must be done before the elaboration of
         --  the ports, since a port subtype may depend on a generic.
         Chap5.Elab_Generic_Map_Aspect_List
           (Get_Generic_Map_Aspect_List (Stmt));

         Map_List := Get_Port_Map_Aspect_List (Stmt);
         Chap5.Elab_Unconstrained_Ports_Of_Port_List (Map_List);

         Chap4.Elab_Port_List (Get_Port_List (Comp), Stmt);

         Chap5.Elab_Port_Map_Aspect_List (Map_List);

         --  configuration specification ?

         Pop_Scope (Comp_Info.Comp_Type);
      end Elab_Component_Instantiation;

      procedure Elab_Conditionnal_Generate_Statement
        (Stmt : Iir_Generate_Statement; Parent : Iir; Base_Block : Iir)
      is
         Scheme : Iir;
         Info : Block_Info_Acc;
         Var : O_Lnode;
         Blk : O_If_Block;
         V : O_Lnode;
         Parent_Info : Block_Info_Acc;
         Base_Info : Block_Info_Acc;
      begin
         Parent_Info := Get_Info (Parent);
         Base_Info := Get_Info (Base_Block);
         Scheme := Get_Generation_Scheme (Stmt);
         Info := Get_Info (Stmt);
         Open_Temp;

         Var := Create_Temp (Info.Block_Decls_Ptr_Type);
         Start_If_Stmt (Blk, Chap7.Translate_Expression (Scheme));
         New_Assign_Stmt (Var, Gen_Alloc (Alloc_System,
                                          New_Sizeof (Info.Block_Decls_Type,
                                                      Ghdl_Index_Type),
                                          Info.Block_Decls_Ptr_Type));
         New_Else_Stmt (Blk);
         New_Assign_Stmt (Var, New_Null_Access (Info.Block_Decls_Ptr_Type));
         Finish_If_Stmt (Blk);

         --  Add a link to child in parent.
         V := Get_Instance_Ref (Parent_Info.Block_Decls_Type);
         V := New_Selected_Element (V, Info.Block_Parent_Field);
         New_Assign_Stmt (V, New_Value (Var));

         Start_If_Stmt
           (Blk,
            New_Compare_Op (ON_Neq,
                            New_Value (Var),
                            New_Null_Access (Info.Block_Decls_Ptr_Type),
                            Ghdl_Bool_Type_Node));
         --  Add a link to parent in child.
         New_Assign_Stmt
           (New_Selected_Acc_Value (Var, Info.Block_Origin_Field),
            Get_Instance_Access (Base_Block));
         --  Elaborate block
         Push_Scope (Info.Block_Decls_Type, Var);
         Push_Scope_Via_Field_Ptr (Base_Info.Block_Decls_Type,
                                   Info.Block_Origin_Field,
                                   Info.Block_Decls_Type);
         Elab_Block_Declarations (Stmt, Stmt);
         Pop_Scope (Base_Info.Block_Decls_Type);
         Pop_Scope (Info.Block_Decls_Type);
         Finish_If_Stmt (Blk);
         Close_Temp;
      end Elab_Conditionnal_Generate_Statement;

      procedure Elab_Iterative_Generate_Statement
        (Stmt : Iir_Generate_Statement; Parent : Iir; Base_Block : Iir)
      is
         Scheme : Iir;
         Iter_Type : Iir;
         Iter_Base_Type : Iir;
         Iter_Type_Info : Type_Info_Acc;
         Range_Ptr : O_Lnode;
         Info : Block_Info_Acc;
         Var_Inst : O_Lnode;
         Var_I : O_Lnode;
         Loop_Blk : O_Loop_Block;
         Label : O_Snode;
         V : O_Lnode;
         Var : O_Lnode;
         Parent_Info : Block_Info_Acc;
         Base_Info : Block_Info_Acc;
      begin
         Parent_Info := Get_Info (Parent);
         Base_Info := Get_Info (Base_Block);

         Scheme := Get_Generation_Scheme (Stmt);
         Iter_Type := Get_Type (Scheme);
         Iter_Base_Type := Get_Base_Type (Iter_Type);
         Iter_Type_Info := Get_Info (Iter_Base_Type);
         Info := Get_Info (Stmt);

         Open_Temp;

         --  Evaluate iterator range.
         Range_Ptr := Create_Temp_Ptr
           (Iter_Type_Info.T.Range_Ptr_Type,
            New_Selected_Element
            (Get_Instance_Ref (Parent_Info.Block_Decls_Type),
             Info.Block_Range_Field));
         Chap7.Translate_Range_Ptr
           (Range_Ptr, Get_Range_Constraint (Iter_Type), Iter_Base_Type);

         --  Allocate instances.
         Var_Inst := Create_Temp (Info.Block_Decls_Array_Ptr_Type);
         New_Assign_Stmt
           (Var_Inst,
            Gen_Alloc
            (Alloc_System,
             New_Dyadic_Op (ON_Mul_Ov,
                            New_Value_Selected_Acc_Value
                            (Range_Ptr, Iter_Type_Info.T.Range_Length),
                            New_Sizeof (Info.Block_Decls_Type,
                                        Ghdl_Index_Type)),
             Info.Block_Decls_Array_Ptr_Type));

         --  Add a link to child in parent.
         V := Get_Instance_Ref (Parent_Info.Block_Decls_Type);
         V := New_Selected_Element (V, Info.Block_Parent_Field);
         New_Assign_Stmt (V, New_Value (Var_Inst));

         --  Start loop.
         Var_I := Create_Temp (Ghdl_Index_Type);
         Init_Var (Var_I);
         Start_Loop_Stmt (Loop_Blk, Label);
         Gen_Exit_When
           (Label,
            New_Compare_Op (ON_Eq,
                            New_Value (Var_I),
                            New_Value_Selected_Acc_Value
                            (Range_Ptr, Iter_Type_Info.T.Range_Length),
                            Ghdl_Bool_Type_Node));

         Var := Create_Temp_Ptr (Info.Block_Decls_Ptr_Type,
                                 New_Indexed_Element (New_Acc_Value (Var_Inst),
                                                      New_Value (Var_I)));
         --  Add a link to parent in child.
         New_Assign_Stmt
           (New_Selected_Acc_Value (Var, Info.Block_Origin_Field),
            Get_Instance_Access (Base_Block));
         --  Mark the block as not (yet) configured.
         New_Assign_Stmt
           (New_Selected_Acc_Value (Var, Info.Block_Configured_Field),
            Ghdl_Bool_False_Node);

         --  Elaborate block
         Push_Scope (Info.Block_Decls_Type, Var);
         Push_Scope_Via_Field_Ptr (Base_Info.Block_Decls_Type,
                                   Info.Block_Origin_Field,
                                   Info.Block_Decls_Type);
         --  Set iterator value.
         --  FIXME: this could be slighly optimized...
         declare
            Val : O_Lnode;
            If_Blk : O_If_Block;
         begin
            Val := Create_Temp (Iter_Type_Info.Ortho_Type (Mode_Value));
            Start_If_Stmt
              (If_Blk,
               New_Compare_Op (ON_Eq,
                               New_Value_Selected_Acc_Value
                               (Range_Ptr, Iter_Type_Info.T.Range_Dir),
                               Ghdl_Dir_To_Node,
                               Ghdl_Bool_Type_Node));
            New_Assign_Stmt (Val, New_Value_Selected_Acc_Value
                             (Range_Ptr, Iter_Type_Info.T.Range_Left));
            New_Else_Stmt (If_Blk);
            New_Assign_Stmt (Val, New_Value_Selected_Acc_Value
                             (Range_Ptr, Iter_Type_Info.T.Range_Right));
            Finish_If_Stmt (If_Blk);

            New_Assign_Stmt
              (Get_Var (Get_Info (Scheme).Iterator_Var),
               New_Dyadic_Op
               (ON_Add_Ov,
                New_Value (Val),
                New_Convert_Ov (New_Value (Var_I),
                                Iter_Type_Info.Ortho_Type (Mode_Value))));
         end;

         --  Set iterator address.
         New_Assign_Stmt
           (New_Selected_Acc_Value (Var, Info.Block_Iterator_Addr_Field),
            New_Unchecked_Address (Get_Var (Get_Info (Scheme).Iterator_Var),
                                   Ghdl_Ptr_Type));

         --  set iterator type kind.
         New_Assign_Stmt
           (New_Selected_Acc_Value (Var, Info.Block_Iterator_TD_Field),
            Chap3.Get_Type_Desc_Addr (Get_Base_Type (Iter_Type)));

         --  Elaboration.
         Elab_Block_Declarations (Stmt, Stmt);

         Pop_Scope (Base_Info.Block_Decls_Type);
         Pop_Scope (Info.Block_Decls_Type);

         Inc_Var (Var_I);
         Finish_Loop_Stmt (Loop_Blk);
         Close_Temp;
      end Elab_Iterative_Generate_Statement;

      procedure Elab_Block_Declarations (Block : Iir; Base_Block : Iir)
      is
         Block_Info : Block_Info_Acc;
         Base_Info : Block_Info_Acc;
         Decl_List : Iir_List;
         Stmts_List : Iir_List;
         Stmt : Iir;
      begin
         Block_Info := Get_Info (Block);
         Base_Info := Get_Info (Base_Block);

         New_Debug_Line_Stmt (Get_Line_Number (Block));

         case Get_Kind (Block) is
            when Iir_Kind_Entity_Declaration =>
               Chap4.Elab_Port_List (Get_Port_List (Block), Block);
               Chap4.Elab_Generic_List (Get_Generic_List (Block));
            when Iir_Kind_Block_Statement =>
               declare
                  Header : Iir_Block_Header;
                  Guard : Iir;
               begin
                  Chap6.Set_Instance_Name (Block);
                  Guard := Get_Guard_Decl (Block);
                  if Guard /= Null_Iir then
                     New_Debug_Line_Stmt (Get_Line_Number (Guard));
                     Elab_Implicit_Guard_Signal (Block, Base_Info);
                  end if;
                  Header := Get_Block_Header (Block);
                  if Header /= Null_Iir then
                     New_Debug_Line_Stmt (Get_Line_Number (Header));
                     Chap4.Elab_Generic_List (Get_Generic_List (Header));
                     Chap4.Elab_Port_List (Get_Port_List (Header), Block);
                     Chap5.Elab_Generic_Map_Aspect_List
                       (Get_Generic_Map_Aspect_List (Header));
                     Chap5.Elab_Port_Map_Aspect_List
                       (Get_Port_Map_Aspect_List (Header));
                  end if;
               end;
            when Iir_Kind_Generate_Statement =>
               Chap6.Set_Instance_Name (Block);
            when others =>
               null;
         end case;

         Decl_List := Get_Declaration_List (Block);
         Chap4.Elab_Declaration_List (Decl_List, Block);

         Stmts_List := Get_Concurrent_Statement_List (Block);
         if Stmts_List /= Null_Iir then
            for I in Natural loop
               Stmt := Get_Nth_Element (Stmts_List, I);
               exit when Stmt = Null_Iir;
               case Get_Kind (Stmt) is
                  when Iir_Kind_Process_Statement
                    | Iir_Kind_Sensitized_Process_Statement =>
                     Elab_Process (Stmt, Block_Info, Base_Info);
                  when Iir_Kind_Component_Instantiation_Statement =>
                     Elab_Component_Instantiation (Stmt, Block_Info);
                  when Iir_Kind_Block_Statement =>
                     declare
                        Info : Block_Info_Acc;
                     begin
                        Info := Get_Info (Stmt);
                        Push_Scope (Info.Block_Decls_Type,
                                    Info.Block_Parent_Field,
                                    Block_Info.Block_Decls_Type);
                        Elab_Block_Declarations (Stmt, Base_Block);
                        Pop_Scope (Info.Block_Decls_Type);
                     end;
                  when Iir_Kind_Generate_Statement =>
                     if Get_Kind (Get_Generation_Scheme (Stmt))
                       = Iir_Kind_Iterator_Declaration
                     then
                        Elab_Iterative_Generate_Statement
                          (Stmt, Block, Base_Block);
                     else
                        Elab_Conditionnal_Generate_Statement
                          (Stmt, Block, Base_Block);
                     end if;
                  when others =>
                     Error_Kind ("elab_block_declarations", Stmt);
               end case;
            end loop;
         end if;
      end Elab_Block_Declarations;
   end Chap9;

   package body Chap10 is
      --  Identifiers.
      --  The following functions are helpers to create ortho identifiers.
      Identifier_Buffer : String (1 .. 512);
      Identifier_Len : Natural := 0;
      Identifier_Start : Natural := 1;
      Identifier_Local : Local_Identifier_Type := 0;


      Inst_Build : Inst_Build_Acc := null;
      procedure Unchecked_Deallocation is new Ada.Unchecked_Deallocation
        (Object => Inst_Build_Type, Name => Inst_Build_Acc);

      procedure Set_Global_Storage (Storage : O_Storage) is
      begin
         Global_Storage := Storage;
      end Set_Global_Storage;

      procedure Pop_Build_Instance
      is
         Old : Inst_Build_Acc;
      begin
         Old := Inst_Build;
         Identifier_Start := Old.Prev_Id_Start;
         Inst_Build := Old.Prev;
         Unchecked_Deallocation (Old);
      end Pop_Build_Instance;

--       procedure Push_Global_Factory (Storage : O_Storage)
--       is
--          Inst : Inst_Build_Acc;
--       begin
--          if Inst_Build /= null then
--             raise Internal_Error;
--          end if;
--          Inst := new Inst_Build_Type (Global);
--          Inst.Prev := Inst_Build;
--          Inst_Build := Inst;
--          Global_Storage := Storage;
--       end Push_Global_Factory;

--       procedure Pop_Global_Factory is
--       begin
--          if Inst_Build.Kind /= Global then
--             raise Internal_Error;
--          end if;
--          Pop_Build_Instance;
--          Global_Storage := O_Storage_Private;
--       end Pop_Global_Factory;

      procedure Push_Instance_Factory (Instance_Type : O_Tnode)
      is
         Inst : Inst_Build_Acc;
      begin
         if Inst_Build /= null and then Inst_Build.Kind /= Instance then
            raise Internal_Error;
         end if;
         Inst := new Inst_Build_Type (Instance);
         Inst.Prev := Inst_Build;

         Inst.Prev_Id_Start := Identifier_Start;
         Identifier_Start := Identifier_Len + 1;

         if Instance_Type /= O_Tnode_Null then
            Start_Uncomplete_Record_Type (Instance_Type, Inst.Elements);
         else
            Start_Record_Type (Inst.Elements);
         end if;
         Inst.Vars := null;
         Inst_Build := Inst;
      end Push_Instance_Factory;

      function Add_Instance_Factory_Field (Name : O_Ident; Ftype : O_Tnode)
        return O_Fnode
      is
         Res : O_Fnode;
      begin
         New_Record_Field (Inst_Build.Elements, Res, Name, Ftype);
         return Res;
      end Add_Instance_Factory_Field;

      procedure Pop_Instance_Factory (Instance_Type : out O_Tnode)
      is
         Res : O_Tnode;
         V : Var_Acc;
      begin
         if Inst_Build.Kind /= Instance then
            --  Not matching.
            raise Internal_Error;
         end if;
         Finish_Record_Type (Inst_Build.Elements, Res);
         --  Set type of all variable declared in this instance.
         V := Inst_Build.Vars;
         while V /= null loop
            V.I_Type := Res;
            V := V.I_Link;
         end loop;
         Pop_Build_Instance;
         Instance_Type := Res;
      end Pop_Instance_Factory;

      procedure Push_Local_Factory
      is
         Inst : Inst_Build_Acc;
      begin
         if Inst_Build /= null
           and then (Inst_Build.Kind /= Global and Inst_Build.Kind /= Local)
         then
            --  Cannot create a local factory on an instance.
            raise Internal_Error;
         end if;
         Inst := new Inst_Build_Type (Kind => Local);
         Inst.Prev := Inst_Build;
         Inst.Prev_Global_Storage := Global_Storage;

         Inst.Prev_Id_Start := Identifier_Start;
         Identifier_Start := Identifier_Len + 1;

         Inst_Build := Inst;
         case Global_Storage is
            when O_Storage_Public =>
               Global_Storage := O_Storage_Private;
            when O_Storage_Private
              | O_Storage_External =>
               null;
            when O_Storage_Local =>
               raise Internal_Error;
         end case;
      end Push_Local_Factory;

      --  Return TRUE is the current scope is local.
      function Is_Local_Scope return Boolean is
      begin
         if Inst_Build = null then
            return False;
         end if;
         case Inst_Build.Kind is
            when Local
              | Instance =>
               return True;
            when Global =>
               return False;
         end case;
      end Is_Local_Scope;

      procedure Pop_Local_Factory is
      begin
         if Inst_Build.Kind /= Local then
            --  Not matching.
            raise Internal_Error;
         end if;
         Global_Storage := Inst_Build.Prev_Global_Storage;
         Pop_Build_Instance;
      end Pop_Local_Factory;

      type Scope_Type;
      type Scope_Acc is access Scope_Type;
      type Scope_Type is record
         Is_Ptr : Boolean;
         Stype : O_Tnode;
         Field : O_Fnode;
         Parent : O_Tnode;
         Prev : Scope_Acc;
      end record;
      type Scope_Var_Type;
      type Scope_Var_Acc is access Scope_Var_Type;
      type Scope_Var_Type is record
         Svtype : O_Tnode;
         Var : O_Lnode;
         Prev : Scope_Var_Acc;
      end record;

      procedure Unchecked_Deallocation is new Ada.Unchecked_Deallocation
        (Object => Scope_Type, Name => Scope_Acc);

      Scopes : Scope_Acc := null;
      --  Chained list of unused scopes, in order to reduce number of
      --  dynamic allocation.
      Scopes_Old : Scope_Acc := null;

      Scopes_Var : Scope_Var_Acc := null;
      --  Chained list of unused var_scopes, to reduce number of allocations.
      Scopes_Var_Old : Scope_Var_Acc := null;

      --  Get a scope, either from the list of free scope or by allocation.
      function Get_A_Scope return Scope_Acc is
         Res : Scope_Acc;
      begin
         if Scopes_Old /= null then
            Res := Scopes_Old;
            Scopes_Old := Scopes_Old.Prev;
         else
            Res := new Chap10.Scope_Type;
         end if;
         return Res;
      end Get_A_Scope;

      procedure Push_Scope (Scope_Type : O_Tnode;
                            Scope_Field : O_Fnode; Scope_Parent : O_Tnode)
      is
         Res : Scope_Acc;
      begin
         Res := Get_A_Scope;
         Res.all := (Is_Ptr => False,
                     Stype => Scope_Type,
                     Field => Scope_Field,
                     Parent => Scope_Parent,
                     Prev => Scopes);
         Scopes := Res;
      end Push_Scope;

      procedure Push_Scope_Via_Field_Ptr
        (Scope_Type : O_Tnode;
         Scope_Field : O_Fnode; Scope_Parent : O_Tnode)
      is
         Res : Scope_Acc;
      begin
         Res := Get_A_Scope;
         Res.all := (Is_Ptr => True,
                     Stype => Scope_Type,
                     Field => Scope_Field,
                     Parent => Scope_Parent,
                     Prev => Scopes);
         Scopes := Res;
      end Push_Scope_Via_Field_Ptr;

      procedure Push_Scope (Scope_Type : O_Tnode; Scope_Param : O_Lnode)
      is
         Res : Scope_Var_Acc;
      begin
         if Scopes_Var_Old /= null then
            Res := Scopes_Var_Old;
            Scopes_Var_Old := Res.Prev;
         else
            Res := new Scope_Var_Type;
         end if;
         Res.all := (Svtype => Scope_Type,
                     Var => Scope_Param,
                     Prev => Scopes_Var);
         Scopes_Var := Res;
      end Push_Scope;

      procedure Pop_Scope (Scope_Type : O_Tnode)
      is
         Old : Scope_Acc;
         Var_Old : Scope_Var_Acc;
      begin
         --  Search in var scope.
         if Scopes_Var /= null and then Scopes_Var.Svtype = Scope_Type then
            Var_Old := Scopes_Var;
            Scopes_Var := Var_Old.Prev;
            Var_Old.Prev := Scopes_Var_Old;
            Scopes_Var_Old := Var_Old;
         elsif Scopes.Stype /= Scope_Type then
            --  Bad pop order.
            raise Internal_Error;
         else
            Old := Scopes;
            Scopes := Old.Prev;
            Old.Prev := Scopes_Old;
            Scopes_Old := Old;
         end if;
      end Pop_Scope;

      function Create_Global_Var
        (Name : O_Ident; Vtype : O_Tnode; Storage : O_Storage)
        return Var_Acc
      is
         Var : O_Lnode;
      begin
         New_Var_Decl (Var, Name, Storage, Vtype);
         return new Var_Type'(Kind => Var_Global, E => Var);
      end Create_Global_Var;

      function Create_Global_Const
        (Name : O_Ident;
         Vtype : O_Tnode;
         Storage : O_Storage;
         Initial_Value : O_Enode)
        return Var_Acc
      is
         Res : O_Lnode;
      begin
         New_Const_Decl (Res, Name, Storage, Vtype, Initial_Value);
         return new Var_Type'(Kind => Var_Global, E => Res);
      end Create_Global_Const;

      function Create_Var
        (Name : Var_Ident_Type;
         Vtype : O_Tnode;
         Storage : O_Storage := Global_Storage)
        return Var_Acc
      is
         Res : O_Lnode;
         Field : O_Fnode;
         V : Var_Acc;
         K : Inst_Build_Kind_Type;
      begin
         if Inst_Build = null then
            K := Global;
         else
            K := Inst_Build.Kind;
         end if;
         case K is
            when Global =>
               --  The global scope is in use...
               return Create_Global_Var (Name.Id, Vtype, Storage);
            when Local =>
               --  It is always possible to create a variable in a local scope.
               --  Create a var.
               New_Var_Decl (Res, Name.Id, O_Storage_Local, Vtype);
               return new Var_Type'(Kind => Var_Local, E => Res);
            when Instance =>
               --  Create a field.
               New_Record_Field (Inst_Build.Elements, Field, Name.Id, Vtype);
               V := new Var_Type'(Kind => Var_Scope, I_Field => Field,
                                  I_Type => O_Tnode_Null,
                                  I_Link => Inst_Build.Vars);
               Inst_Build.Vars := V;
               return V;
         end case;
      end Create_Var;

      function Find_Scope_Type (Stype : O_Tnode; First : Scope_Acc)
        return O_Lnode
      is
         S : Scope_Acc;
         Sv : Scope_Var_Acc;
      begin
         --  Find in var.
         Sv := Scopes_Var;
         while Sv /= null loop
            if Sv.Svtype = Stype then
               return New_Acc_Value (Sv.Var);
            end if;
            Sv := Sv.Prev;
         end loop;

         --  Find in fields.
         S := Scopes;
         while S /= null loop
            if S.Stype = Stype then
               if S.Is_Ptr then
                  return New_Access_Element
                    (New_Value
                     (New_Selected_Element (Find_Scope_Type (S.Parent, S.Prev),
                                            S.Field)));
               else
                  return New_Selected_Element
                    (Find_Scope_Type (S.Parent, S.Prev), S.Field);
               end if;
            end if;
            S := S.Prev;
         end loop;

         --  Not found.
         raise Internal_Error;
      end Find_Scope_Type;

      function Get_Instance_Access (Block : Iir) return O_Enode
      is
         Info : Block_Info_Acc;
      begin
         Info := Get_Info (Block);
         if Info.Block_Decls_Type = Scopes_Var.Svtype then
            return New_Value (Scopes_Var.Var);
         else
            return New_Address (Get_Instance_Ref (Info.Block_Decls_Type),
                                Info.Block_Decls_Ptr_Type);
         end if;
      end Get_Instance_Access;

      function Get_Instance_Ref (Itype : O_Tnode) return O_Lnode
      is
      begin
         --  Variables cannot be referenced if there is an instance being
         --  built.
         if Inst_Build /= null and then Inst_Build.Kind = Instance then
            raise Internal_Error;
         end if;
         return Find_Scope_Type (Itype, Scopes);
      end Get_Instance_Ref;

      function Get_Var (Var : Var_Acc) return O_Lnode
      is
      begin
         case Var.Kind is
            when Var_Local
              | Var_Global =>
               return Var.E;
            when Var_Scope =>
               null;
         end case;

         return New_Selected_Element (Get_Instance_Ref (Var.I_Type),
                                      Var.I_Field);
      end Get_Var;

      function Get_Alloc_Kind_For_Var (Var : Var_Acc) return Allocation_Kind is
      begin
         case Var.Kind is
            when Var_Local =>
               return Alloc_Stack;
            when Var_Global
              | Var_Scope =>
               return Alloc_Memory;
         end case;
      end Get_Alloc_Kind_For_Var;

      function Is_Var_Stable (Var : Var_Acc) return Boolean is
      begin
         case Var.Kind is
            when Var_Local
              | Var_Global =>
               return True;
            when Var_Scope =>
               return False;
         end case;
      end Is_Var_Stable;

      type Scope_Layer is record
         --  How to dereference a variable in the scope.
         --  O_Enode_Null if no there is no way to dereference an element of
         --   the scope, like during scope creation.
         This : O_Enode;

         --  Type of the scope; this must be a record type or NULL_TREE for
         --   the global scope.
         --  This is very important since a variable (in fact a FIELD_DECL)
         --   belong to a scope iff the type of the field context is
         --   Scope_Type.  As a consequence, Scope_Type of two different
         --   layers must be different.
         --  Note: scope_type is a type definition (such as RECORD_TYPE) and
         --   *not* a TYPE_DECL.
         --  NULL_TREE for a local scope.
         Scope_Type : O_Tnode;

         --  The scope just below this one.
         --Prev : Scope_Acc;
      end record;

      procedure Save_Local_Identifier (Id : out Local_Identifier_Type) is
      begin
         Id := Identifier_Local;
      end Save_Local_Identifier;

      procedure Restore_Local_Identifier (Id : Local_Identifier_Type) is
      begin
         if Identifier_Local > Id then
            --  If the value is restored with a smaller value, some identifiers
            --  will be reused.  This is certainly an internal error.
            raise Internal_Error;
         end if;
         Identifier_Local := Id;
      end Restore_Local_Identifier;

      --  Reset the identifier.
      procedure Reset_Identifier_Prefix is
      begin
         if Identifier_Len /= 0 or else Identifier_Local /= 0 then
            raise Internal_Error;
         end if;
      end Reset_Identifier_Prefix;

      procedure Pop_Identifier_Prefix (Mark : in Id_Mark_Type) is
      begin
         Identifier_Len := Mark.Len;
         Identifier_Local := Mark.Local_Id;
      end Pop_Identifier_Prefix;

      procedure Add_String (Len : in out Natural; Str : String)
      is
      begin
         Identifier_Buffer (Len + 1 .. Len + Str'Length) := Str;
         Len := Len + Str'Length;
      end Add_String;

      procedure Add_Nat (Len : in out Natural; Val : Natural)
      is
         Num : String (1 .. 10);
         V : Natural;
         P : Natural;
      begin
         P := Num'Last;
         V := Val;
         loop
            Num (P) := Character'Val (Character'Pos ('0') + V mod 10);
            V := V / 10;
            exit when V = 0;
            P := P - 1;
         end loop;
         Add_String (Len, Num (P .. Num'Last));
      end Add_Nat;

      procedure Add_Name (Len : in out Natural; Name : Name_Id) is
      begin
         Add_String (Len, Name_Table.Image (Name));
      end Add_Name;

      procedure Push_Identifier_Prefix (Mark : out Id_Mark_Type;
                                        Name : String;
                                        Val : Natural := 0)
      is
         P : Natural;
      begin
         Mark.Len := Identifier_Len;
         Mark.Local_Id := Identifier_Local;
         Identifier_Local := 0;
         P := Identifier_Len;
         Add_String (P, Name);
         if Val > 0 then
            Add_String (P, "O");
            Add_Nat (P, Val);
         end if;
         Add_String (P, "__");
         Identifier_Len := P;
      end Push_Identifier_Prefix;

      --  Add a suffix to the prefix (!!!).
      procedure Push_Identifier_Prefix
        (Mark : out Id_Mark_Type; Name : Name_Id; Val : Natural := 0)
      is
         use Name_Table;
      begin
         Image (Name);
         Push_Identifier_Prefix (Mark, Name_Buffer (1 .. Name_Length), Val);
      end Push_Identifier_Prefix;

      procedure Add_Local_Identifier (Len : in out Natural)
      is
         Str : String := Local_Identifier_Type'Image (Identifier_Local);
      begin
         Identifier_Local := Identifier_Local + 1;

         if Inst_Build = null then
            Str (1) := 'N';
         else
            case Inst_Build.Kind is
               when Local =>
                  Str (1) := 'L';
               when Global =>
                  Str (1) := 'G';
               when Instance =>
                  Str (1) := 'I';
            end case;
         end if;
         Add_String (Len, Str);
      end Add_Local_Identifier;

      procedure Push_Identifier_Prefix_Uniq (Mark : out Id_Mark_Type)
      is
         Str : String := Local_Identifier_Type'Image (Identifier_Local);
      begin
         Identifier_Local := Identifier_Local + 1;
         Str (1) := 'U';
         Push_Identifier_Prefix (Mark, Str, 0);
      end Push_Identifier_Prefix_Uniq;

      procedure Add_Identifier (Len : in out Natural; Id : Name_Id) is
      begin
         if Id = Null_Identifier then
            --Add_Local_Identifier (Len);
            null;
         else
            Add_Name (Len, Id);
         end if;
      end Add_Identifier;

      --  Create an identifier from IIR node ID without the prefix.
      function Create_Identifier_Without_Prefix (Id : Iir) return O_Ident
      is
      begin
         return Get_Identifier (Name_Table.Image (Get_Identifier (Id)));
      end Create_Identifier_Without_Prefix;

      function Create_Identifier_Without_Prefix (Id : Iir; Str : String)
        return O_Ident
      is
         Res : String := Name_Table.Image (Get_Identifier (Id)) & Str;
      begin
         return Get_Identifier (Res);
      end Create_Identifier_Without_Prefix;

      --  Create an identifier from IIR node ID with prefix.
      function Create_Id (Id : Name_Id; Str : String; Is_Local : Boolean)
        return O_Ident
      is
         L : Natural;
      begin
         L := Identifier_Len;
         Add_Identifier (L, Id);
         Add_String (L, Str);
         --Identifier_Buffer (L + Str'Length + 1) := Nul;
         if Is_Local then
            return Get_Identifier
              (Identifier_Buffer (Identifier_Start .. L));
         else
            return Get_Identifier (Identifier_Buffer (1 .. L));
         end if;
      end Create_Id;

      function Create_Identifier (Id : Name_Id; Str : String := "")
        return O_Ident
      is
      begin
         return Create_Id (Id, Str, False);
      end Create_Identifier;

      function Create_Identifier (Id : Iir; Str : String := "")
        return O_Ident
      is
      begin
         return Create_Id (Get_Identifier (Id), Str, False);
      end Create_Identifier;

      function Create_Identifier (Id : Iir; Val : Natural; Str : String := "")
                                  return O_Ident
      is
         Len : Natural;
      begin
         Len := Identifier_Len;
         Add_Identifier (Len, Get_Identifier (Id));

         if Val > 0 then
            Add_String (Len, "O");
            Add_Nat (Len, Val);
         end if;
         Add_String (Len, Str);
         return Get_Identifier (Identifier_Buffer (1 .. Len));
      end Create_Identifier;

      function Create_Identifier (Str : String)
        return O_Ident
      is
         Len : Natural;
      begin
         Len := Identifier_Len;
         Add_String (Len, Str);
         return Get_Identifier (Identifier_Buffer (1 .. Len));
      end Create_Identifier;

      function Create_Identifier return O_Ident
      is
      begin
         return Get_Identifier (Identifier_Buffer (1 .. Identifier_Len - 2));
      end Create_Identifier;

      function Create_Var_Identifier (Id : Iir; Str : String)
        return Var_Ident_Type
      is
         Res : Var_Ident_Type;
      begin
         Res.Id := Create_Id (Get_Identifier (Id), Str, Is_Local_Scope);
         return Res;
      end Create_Var_Identifier;

      function Create_Var_Identifier (Id : Name_Id; Str : String)
        return Var_Ident_Type
      is
         Res : Var_Ident_Type;
      begin
         Res.Id := Create_Id (Id, Str, Is_Local_Scope);
         return Res;
      end Create_Var_Identifier;

      function Create_Var_Identifier (Id : Iir)
        return Var_Ident_Type
      is
         Res : Var_Ident_Type;
      begin
         Res.Id := Create_Id (Get_Identifier (Id), "", Is_Local_Scope);
         return Res;
      end Create_Var_Identifier;

      function Create_Var_Identifier (Id : String)
        return Var_Ident_Type
      is
         Res : Var_Ident_Type;
      begin
         Res.Id := Create_Id (Null_Identifier, Id, Is_Local_Scope);
         return Res;
      end Create_Var_Identifier;

      function Create_Uniq_Identifier return Var_Ident_Type
      is
         Res : Var_Ident_Type;
      begin
         Res.Id := Create_Uniq_Identifier;
         return Res;
      end Create_Uniq_Identifier;
   end Chap10;

   package body Chap14 is
      function Translate_Array_Attribute_To_Range (Expr : Iir) return Mnode
      is
         Prefix : Iir;
         Arr : Mnode;
         Dim : Natural;
      begin
         Prefix := Get_Prefix (Expr);
         case Get_Kind (Prefix) is
            when Iir_Kind_Type_Declaration
              | Iir_Kind_Subtype_Declaration =>
               Arr := T2M (Get_Type (Prefix), Mode_Value);
            when others =>
               Arr := Chap6.Translate_Name (Prefix);
         end case;
         Dim := Natural (Get_Value (Get_Parameter (Expr)));
         return Chap3.Get_Array_Range (Arr, Get_Type (Prefix), Dim);
      end Translate_Array_Attribute_To_Range;

      function Translate_Range_Array_Attribute (Expr : Iir)
        return O_Lnode is
      begin
         return M2Lv (Translate_Array_Attribute_To_Range (Expr));
      end Translate_Range_Array_Attribute;

      function Translate_Length_Array_Attribute (Expr : Iir; Rtype : Iir)
        return O_Enode
      is
         Rng : Mnode;
         Val : O_Enode;
      begin
         Rng := Translate_Array_Attribute_To_Range (Expr);
         Val := M2E (Chap3.Range_To_Length (Rng));
         if Rtype /= Null_Iir then
            Val := New_Convert_Ov (Val, Get_Ortho_Type (Rtype, Mode_Value));
         end if;
         return Val;
      end Translate_Length_Array_Attribute;

      function Translate_High_Low_Array_Attribute (Expr : Iir;
                                                   Is_High : Boolean)
        return O_Enode
      is
         Prefix : Iir;
         Prefix_Type : Iir;
         Dim : Natural;
         Index_Type : Iir_List;
         Index_Info : Type_Info_Acc;
         If_Blk : O_If_Block;
         Res : O_Lnode;
         Op : ON_Op_Kind;
         Range_Var : Mnode;
      begin
         Prefix := Get_Prefix (Expr);
         Prefix_Type := Get_Type (Prefix);
         Dim := Natural (Get_Value (Get_Parameter (Expr)));
         Index_Type := Get_Nth_Element (Get_Index_List (Prefix_Type), Dim - 1);
         Index_Info := Get_Info (Get_Base_Type (Index_Type));

         Res := Create_Temp (Index_Info.Ortho_Type (Mode_Value));

         Open_Temp;
         Range_Var := Stabilize (Translate_Array_Attribute_To_Range (Expr));

         if Is_High then
            Op := ON_Neq;
         else
            Op := ON_Eq;
         end if;
         Start_If_Stmt (If_Blk,
                        New_Compare_Op (Op,
                                        M2E (Chap3.Range_To_Dir (Range_Var)),
                                        Ghdl_Dir_To_Node,
                                        Ghdl_Bool_Type_Node));
         New_Assign_Stmt (Res, M2E (Chap3.Range_To_Left (Range_Var)));
         New_Else_Stmt (If_Blk);
         New_Assign_Stmt (Res, M2E (Chap3.Range_To_Right (Range_Var)));
         Finish_If_Stmt (If_Blk);
         Close_Temp;
         return New_Value (Res);
      end Translate_High_Low_Array_Attribute;

      function Translate_Low_Array_Attribute (Expr : Iir)
        return O_Enode
      is
      begin
         return Translate_High_Low_Array_Attribute (Expr, False);
      end Translate_Low_Array_Attribute;

      function Translate_High_Array_Attribute (Expr : Iir)
        return O_Enode
      is
      begin
         return Translate_High_Low_Array_Attribute (Expr, True);
      end Translate_High_Array_Attribute;

      function Translate_Left_Array_Attribute (Expr : Iir)
        return O_Enode
      is
         Rng : Mnode;
      begin
         Rng := Translate_Array_Attribute_To_Range (Expr);
         return M2E (Chap3.Range_To_Left (Rng));
      end Translate_Left_Array_Attribute;

      function Translate_Right_Array_Attribute (Expr : Iir)
        return O_Enode
      is
         Rng : Mnode;
      begin
         Rng := Translate_Array_Attribute_To_Range (Expr);
         return M2E (Chap3.Range_To_Right (Rng));
      end Translate_Right_Array_Attribute;

      function Translate_Low_High_Type_Attribute
        (Atype : Iir; Is_Low : Boolean)
        return O_Enode
      is
         Range_Constr : Iir;
      begin
         if Get_Type_Staticness (Atype) = Locally then
            Range_Constr := Get_Range_Constraint (Atype);
            if Get_Direction (Range_Constr) = Iir_To xor Is_Low then
               --  TO and HIGH   or  DOWNTO and LOW  -> right
               return Chap7.Translate_Static_Range_Right
                 (Range_Constr, Atype);
            else
               --  TO and LOW   or   DOWNTO and HIGH  -> left
               return Chap7.Translate_Static_Range_Left
                 (Range_Constr, Atype);
            end if;
         else
            Error_Kind ("translate_high_low_type_attribute", Atype);
            --Info := Get_Info (Atype);
            --return New_Value
            --  (New_Selected_Element (Get_Var (Info.T.Range_Var),
            --                         Info.T.Range_Left));
         end if;
      end Translate_Low_High_Type_Attribute;

      function Translate_High_Type_Attribute (Atype : Iir) return O_Enode is
      begin
         return Translate_Low_High_Type_Attribute (Atype, False);
      end Translate_High_Type_Attribute;

      function Translate_Low_Type_Attribute (Atype : Iir) return O_Enode is
      begin
         return Translate_Low_High_Type_Attribute (Atype, True);
      end Translate_Low_Type_Attribute;

      function Translate_Left_Type_Attribute (Atype : Iir) return O_Enode
      is
         Info : Type_Info_Acc;
      begin
         if Get_Type_Staticness (Atype) = Locally then
            return Chap7.Translate_Static_Range_Left
              (Get_Range_Constraint (Atype), Atype);
         else
            Info := Get_Info (Atype);
            return New_Value
              (New_Selected_Element (Get_Var (Info.T.Range_Var),
                                     Info.T.Range_Left));
         end if;
      end Translate_Left_Type_Attribute;

      function Translate_Right_Type_Attribute (Atype : Iir) return O_Enode
      is
         Info : Type_Info_Acc;
      begin
         if Get_Type_Staticness (Atype) = Locally then
            return Chap7.Translate_Static_Range_Right
              (Get_Range_Constraint (Atype), Atype);
         else
            Info := Get_Info (Atype);
            return New_Value
              (New_Selected_Element (Get_Var (Info.T.Range_Var),
                                     Info.T.Range_Right));
         end if;
      end Translate_Right_Type_Attribute;

      function Translate_Dir_Type_Attribute (Atype : Iir) return O_Enode
      is
         Info : Type_Info_Acc;
      begin
         if Get_Type_Staticness (Atype) = Locally then
            return Chap7.Translate_Static_Range_Dir
              (Get_Range_Constraint (Atype));
         else
            Info := Get_Info (Atype);
            return New_Value
              (New_Selected_Element (Get_Var (Info.T.Range_Var),
                                     Info.T.Range_Right));
         end if;
      end Translate_Dir_Type_Attribute;

      function Translate_Val_Attribute (Attr : Iir) return O_Enode
      is
         T : O_Lnode;
         Prefix : Iir;
         Ttype : O_Tnode;
      begin
         Prefix := Get_Type (Get_Prefix (Attr));
         Ttype := Get_Ortho_Type (Prefix, Mode_Value);
         T := Create_Temp (Ttype);
         New_Assign_Stmt
           (T,
            New_Convert_Ov (Chap7.Translate_Expression (Get_Parameter (Attr)),
                            Ttype));
         Chap3.Check_Range (T, Attr, Prefix);
         return New_Value (T);
      end Translate_Val_Attribute;

      function Translate_Pos_Attribute (Attr : Iir; Res_Type : Iir)
        return O_Enode
      is
         T : O_Lnode;
         Ttype : O_Tnode;
      begin
         Ttype := Get_Ortho_Type (Res_Type, Mode_Value);
         T := Create_Temp (Ttype);
         New_Assign_Stmt
           (T,
            New_Convert_Ov (Chap7.Translate_Expression (Get_Parameter (Attr)),
                            Ttype));
         Chap3.Check_Range (T, Attr, Res_Type);
         return New_Value (T);
      end Translate_Pos_Attribute;

      function Translate_Succ_Pred_Attribute (Attr : Iir) return O_Enode
      is
         Expr_Type : Iir;
         Tinfo : Type_Info_Acc;
         Ttype : O_Tnode;
         Expr : O_Enode;
         List : Iir_List;
         Limit : Iir;
         Is_Succ : Boolean;
         Op : ON_Op_Kind;
      begin
         --  FIXME: should check bounds.
         Expr_Type := Get_Type (Attr);
         Tinfo := Get_Info (Expr_Type);
         Expr := Chap7.Translate_Expression (Get_Parameter (Attr), Expr_Type);
         Ttype := Tinfo.Ortho_Type (Mode_Value);
         Is_Succ := Get_Kind (Attr) = Iir_Kind_Succ_Attribute;
         if Is_Succ then
            Op := ON_Add_Ov;
         else
            Op := ON_Sub_Ov;
         end if;
         case Tinfo.Type_Mode is
            when Type_Mode_B2
              | Type_Mode_E8
              | Type_Mode_E32 =>
               --  Should check it is not the last.
               declare
                  L : O_Lnode;
               begin
                  List := Get_Enumeration_Literal_List (Get_Base_Type
                                                        (Expr_Type));
                  L := Create_Temp_Init (Ttype, Expr);
                  if Is_Succ then
                     Limit := Get_Last_Element (List);
                  else
                     Limit := Get_First_Element (List);
                  end if;
                  Chap6.Check_Bound_Error
                    (New_Compare_Op (ON_Eq,
                                     New_Value (L),
                                     Get_Ortho_Expr (Limit),
                                     Ghdl_Bool_Type_Node),
                     Attr, 0);
                  return New_Convert_Ov
                    (New_Dyadic_Op
                     (Op,
                      New_Convert_Ov (New_Value (L), Ghdl_Integer_Type),
                      New_Signed_Literal (Ghdl_Integer_Type, 1)),
                     Ttype);
               end;
            when Type_Mode_I32
              | Type_Mode_P64 =>
               return New_Dyadic_Op
                 (Op, Expr, New_Signed_Literal (Ttype, 1));
            when others =>
               raise Internal_Error;
         end case;
      end Translate_Succ_Pred_Attribute;

      --  Read the boolean attribute (active or event) FIELD of simple signal
      --  SIG.
      function Read_Bool_Signal_Attribute (Sig : O_Enode; Field : O_Fnode)
                                          return O_Enode
      is
         S : O_Enode;
      begin
         S := New_Convert_Ov (Sig, Ghdl_Signal_Ptr);
         return New_Value
           (New_Selected_Element (New_Access_Element (S), Field));
         --Ghdl_Signal_Event_Node));
      end Read_Bool_Signal_Attribute;

      type Bool_Sigattr_Data_Type is record
         Label : O_Snode;
         Field : O_Fnode;
      end record;

      procedure Bool_Sigattr_Non_Composite_Signal
        (Targ : Mnode; Targ_Type : Iir; Data : Bool_Sigattr_Data_Type)
      is
         If_Blk : O_If_Block;
      begin
         Gen_Exit_When (Data.Label,
                        Read_Bool_Signal_Attribute (New_Value (M2Lv (Targ)),
                                                    Data.Field));
      end Bool_Sigattr_Non_Composite_Signal;

      function Bool_Sigattr_Prepare_Data_Composite
        (Data : Bool_Sigattr_Data_Type; Targ_Type : Iir)
        return Bool_Sigattr_Data_Type is
      begin
         return Data;
      end Bool_Sigattr_Prepare_Data_Composite;

      function Bool_Sigattr_Update_Data_Array (Data : Bool_Sigattr_Data_Type;
                                               Targ_Type : Iir;
                                               Index : O_Lnode)
        return Bool_Sigattr_Data_Type
      is
      begin
         return Data;
      end Bool_Sigattr_Update_Data_Array;

      function Bool_Sigattr_Update_Data_Record (Data : Bool_Sigattr_Data_Type;
                                                Targ_Type : Iir;
                                                El : Iir_Element_Declaration)
        return Bool_Sigattr_Data_Type
      is
      begin
         return Data;
      end Bool_Sigattr_Update_Data_Record;

      procedure Bool_Sigattr_Foreach is new Foreach_Non_Composite
        (Data_Type => Bool_Sigattr_Data_Type,
         Composite_Data_Type => Bool_Sigattr_Data_Type,
         Do_Non_Composite => Bool_Sigattr_Non_Composite_Signal,
         Prepare_Data_Array => Bool_Sigattr_Prepare_Data_Composite,
         Update_Data_Array => Bool_Sigattr_Update_Data_Array,
         Prepare_Data_Record => Bool_Sigattr_Prepare_Data_Composite,
         Update_Data_Record => Bool_Sigattr_Update_Data_Record);

      function Translate_Bool_Signal_Attribute (Attr : Iir; Field : O_Fnode)
                                               return O_Enode
      is
         Data : Bool_Sigattr_Data_Type;
         Res : O_Lnode;
         Loop_Blk : O_Loop_Block;
         Name : Mnode;
         Prefix : Iir;
         Prefix_Type : Iir;
      begin
         Prefix := Get_Prefix (Attr);
         Prefix_Type := Get_Type (Prefix);

         if Get_Kind (Prefix_Type) in Iir_Kinds_Scalar_Type_Definition then
            --  Effecient handling for a scalar signal.
            Name := Chap6.Translate_Name (Prefix);
            return Read_Bool_Signal_Attribute (New_Value (M2Lv (Name)), Field);
         else
            --  Element per element handling for composite signals.
            Res := Create_Temp (Std_Boolean_Type_Node);
            Open_Temp;
            New_Assign_Stmt (Res, Std_Boolean_True_Node);
            Name := Chap6.Translate_Name (Prefix);
            Start_Loop_Stmt (Loop_Blk, Data.Label);
            Data.Field := Field;
            Bool_Sigattr_Foreach (Name, Prefix_Type, Data);
            New_Assign_Stmt (Res, Std_Boolean_False_Node);
            New_Exit_Stmt (Data.Label);
            Finish_Loop_Stmt (Loop_Blk);
            Close_Temp;
            return New_Value (Res);
         end if;
      end Translate_Bool_Signal_Attribute;

      function Translate_Event_Attribute (Attr : Iir) return O_Enode is
      begin
         return Translate_Bool_Signal_Attribute (Attr, Ghdl_Signal_Event_Node);
      end Translate_Event_Attribute;

      function Translate_Active_Attribute (Attr : Iir) return O_Enode is
      begin
         return Translate_Bool_Signal_Attribute
           (Attr, Ghdl_Signal_Active_Node);
      end Translate_Active_Attribute;

      function Read_Last_Value (Sig : O_Enode; Sig_Type : Iir) return O_Enode
      is
         S_Type : O_Tnode;
         T : O_Lnode;
      begin
         S_Type := Get_Ortho_Type (Sig_Type, Mode_Signal);
         T := New_Access_Element (New_Convert_Ov (Sig, Ghdl_Signal_Ptr));
         return New_Value
           (New_Access_Element
            (New_Unchecked_Address
             (New_Selected_Element (T, Ghdl_Signal_Last_Value_Node), S_Type)));
      end Read_Last_Value;

      function Translate_Last_Value is new Chap7.Translate_Signal_Value
        (Read_Value => Read_Last_Value);

      function Translate_Last_Value_Attribute (Attr : Iir) return O_Enode
      is
         Name : Mnode;
         Prefix : Iir;
         Prefix_Type : Iir;
      begin
         Prefix := Get_Prefix (Attr);
         Prefix_Type := Get_Type (Prefix);

         Name := Chap6.Translate_Name (Prefix);
         if Get_Object_Kind (Name) /= Mode_Signal then
            raise Internal_Error;
         end if;
         return Translate_Last_Value (M2E (Name), Prefix_Type);
      end Translate_Last_Value_Attribute;

      function Read_Last_Time (Sig : O_Enode; Field : O_Fnode) return O_Enode
      is
         T : O_Lnode;
      begin
         T := New_Access_Element (New_Convert_Ov (Sig, Ghdl_Signal_Ptr));
         return New_Value (New_Selected_Element (T, Field));
      end Read_Last_Time;

      type Last_Time_Data is record
         Var : O_Lnode;
         Field : O_Fnode;
      end record;

      procedure Translate_Last_Time_Non_Composite
        (Targ : Mnode; Targ_Type : Iir; Data : Last_Time_Data)
      is
         Val : O_Lnode;
         If_Blk : O_If_Block;
      begin
         Open_Temp;
         Val := Create_Temp_Init
           (Std_Time_Type,
            Read_Last_Time (New_Value (M2Lv (Targ)), Data.Field));
         Start_If_Stmt (If_Blk,
                        New_Compare_Op (ON_Gt,
                                        New_Value (Val),
                                        New_Value (Data.Var),
                                        Ghdl_Bool_Type_Node));
         New_Assign_Stmt (Data.Var, New_Value (Val));
         Finish_If_Stmt (If_Blk);
         Close_Temp;
      end Translate_Last_Time_Non_Composite;

      function Last_Time_Prepare_Data_Composite
        (Data : Last_Time_Data; Targ_Type : Iir)
        return Last_Time_Data is
      begin
         return Data;
      end Last_Time_Prepare_Data_Composite;

      function Last_Time_Update_Data_Array (Data : Last_Time_Data;
                                  Targ_Type : Iir;
                                  Index : O_Lnode)
        return Last_Time_Data is
      begin
         return Data;
      end Last_Time_Update_Data_Array;

      function Last_Time_Update_Data_Record (Data : Last_Time_Data;
                                             Targ_Type : Iir;
                                             El : Iir_Element_Declaration)
        return Last_Time_Data is
      begin
         return Data;
      end Last_Time_Update_Data_Record;

      procedure Translate_Last_Time is new Foreach_Non_Composite
        (Data_Type => Last_Time_Data,
         Composite_Data_Type => Last_Time_Data,
         Do_Non_Composite => Translate_Last_Time_Non_Composite,
         Prepare_Data_Array => Last_Time_Prepare_Data_Composite,
         Update_Data_Array => Last_Time_Update_Data_Array,
         Prepare_Data_Record => Last_Time_Prepare_Data_Composite,
         Update_Data_Record => Last_Time_Update_Data_Record);

      function Translate_Last_Time_Attribute (Prefix : Iir; Field : O_Fnode)
        return O_Enode
      is
         Prefix_Type : Iir;
         Name : Mnode;
         Info : Type_Info_Acc;
         Res : O_Enode;
         Var : O_Lnode;
      begin
         Prefix_Type := Get_Type (Prefix);
         Name := Chap6.Translate_Name (Prefix);
         Info := Get_Info (Prefix_Type);
         if Info.Type_Mode in Type_Mode_Scalar then
            Res := Read_Last_Time (M2E (Name), Field);
         else
            Var := Create_Temp (Std_Time_Type);
            Translate_Last_Time
              (Name, Prefix_Type, Last_Time_Data'(Var => Var, Field => Field));
            Res := New_Value (Var);
         end if;
         --  return NOW - Res
         return New_Dyadic_Op (ON_Sub_Ov, New_Value (Ghdl_Now), Res);
      end Translate_Last_Time_Attribute;
   end Chap14;

   --  Decorate the tree in order to be usable with the internal simulator.
   procedure Translate (Unit : Iir_Design_Unit; Main : Boolean)
   is
      Design_File : Iir_Design_File;
      Decl_List : Iir_List;
      El : Iir;
      Lib : Iir_Library_Declaration;
      Lib_Mark, Ent_Mark, Unit_Mark : Id_Mark_Type;
      El_Id : Name_Id;
   begin
      Design_File := Get_Design_File (Unit);
      --Current_Filename := Get_Address (Get_File_Name (Design_File));

      --  Create the variable containing the current file name.
      if Main and not Current_Filename_Set then
         New_Const_Decl
           (Current_Filename_Node,
            Get_Identifier ("__FILENAME"),
            O_Storage_Private,
            Ghdl_Str_Len_Type_Node,
            Create_String_Len (Get_Design_File_Filename (Design_File),
                               Get_Identifier ("__FILENAMESTR")));
         Current_Filename_Set := True;
      end if;

      Decl_List := Get_Context_Items (Unit);
      if Decl_List /= null then
         for I in Natural loop
            El := Get_Nth_Element (Decl_List, I);
            exit when El = Null_Iir;
            case Get_Kind (El) is
               when Iir_Kind_Use_Clause =>
                  null;
               when Iir_Kind_Library_Clause =>
                  null;
               when others =>
                  Error_Kind ("translate1", El);
            end case;
         end loop;
      end if;

      El := Get_Library_Unit (Unit);
      if Flags.Verbose then
         Ada.Text_IO.Put_Line ("translating " & Disp_Node (El));
      end if;

      --  Create the prefix for identifiers.
      Lib := Get_Parent (Get_Parent (Unit));
      Reset_Identifier_Prefix;
      Push_Identifier_Prefix (Lib_Mark, Get_Identifier (Lib));
      if Get_Kind (El) = Iir_Kind_Architecture_Declaration then
         Push_Identifier_Prefix (Ent_Mark, Get_Identifier (Get_Entity (El)));
      end if;
      El_Id := Get_Identifier (El);
      if El_Id /= Null_Identifier then
         Push_Identifier_Prefix (Unit_Mark, El_Id);
      end if;

      if Main then
         Set_Global_Storage (O_Storage_Public);
      else
         Set_Global_Storage (O_Storage_External);
      end if;

      case Get_Kind (El) is
         when Iir_Kind_Package_Declaration =>
            New_Debug_Comment_Decl ("package declaration " & Get_String (El));
            Chap2.Translate_Package_Declaration (El);
         when Iir_Kind_Package_Body =>
            New_Debug_Comment_Decl ("package body " & Get_String (El));
            --Push_Global_Factory (O_Storage_Private);
            Chap2.Translate_Package_Body (El);
            --Pop_Global_Factory;
         when Iir_Kind_Entity_Declaration =>
            New_Debug_Comment_Decl ("entity " & Get_String (El));
            --Set_Global_Storage (O_Storage_Private);
            Chap1.Translate_Entity_Declaration (El);
         when Iir_Kind_Architecture_Declaration =>
            New_Debug_Comment_Decl ("architecture " & Get_String (El));
            --Set_Global_Storage (O_Storage_Private);
            Chap1.Translate_Architecture_Declaration (El);
         when Iir_Kind_Configuration_Declaration =>
            New_Debug_Comment_Decl ("configuration " & Get_String (El));
            if El_Id = Null_Identifier then
               declare
                  Mark : Id_Mark_Type;
                  Mark_Entity : Id_Mark_Type;
                  Mark_Arch : Id_Mark_Type;
                  Arch : Iir;
                  Entity : Iir;
               begin
                  --  Note: this is done inside the architecture identifier.
                  Entity := Get_Library_Unit (Get_Entity (El));
                  Push_Identifier_Prefix
                    (Mark_Entity, Get_Identifier (Entity));
                  Arch := Get_Block_Specification
                    (Get_Block_Configuration (El));
                  Push_Identifier_Prefix (Mark_Arch, Get_Identifier (Arch));
                  Push_Identifier_Prefix
                    (Mark, Name_Table.Get_Identifier ("DEFAULT_CONFIG"));
                  Chap1.Translate_Configuration_Declaration (El);
                  Pop_Identifier_Prefix (Mark);
                  Pop_Identifier_Prefix (Mark_Entity);
                  Pop_Identifier_Prefix (Mark_Arch);
               end;
            else
               Chap1.Translate_Configuration_Declaration (El);
            end if;
         when others =>
            Error_Kind ("translate", El);
      end case;

      --Pop_Global_Factory;
      if El_Id /= Null_Identifier then
         Pop_Identifier_Prefix (Unit_Mark);
      end if;
      if Get_Kind (El) = Iir_Kind_Architecture_Declaration then
         Pop_Identifier_Prefix (Ent_Mark);
      end if;
      Pop_Identifier_Prefix (Lib_Mark);
   end Translate;

   procedure Initialize
   is
      Interfaces : O_Inter_List;
      Param : O_Lnode;
   begin
      New_Debug_Comment_Decl ("internal declarations, part 1");
      --  Give a name to sizetype.
      --Finish_Type_Decl (Sizetype, Get_Identifier ("__ghdl_size_t"));

      -- Create well known identifiers.
      Wki_This := Get_Identifier ("this");
      Wki_Size := Get_Identifier ("size");
      Wki_Res := Get_Identifier ("res");
      Wki_E := Get_Identifier ("_E");
      Wki_Dir_To := Get_Identifier ("dir_to");
      Wki_Dir_Downto := Get_Identifier ("dir_downto");
      Wki_Left := Get_Identifier ("left");
      Wki_Right := Get_Identifier ("right");
      Wki_Dir := Get_Identifier ("dir");
      Wki_Length := Get_Identifier ("length");
      Wki_Kind := Get_Identifier ("kind");
      Wki_Dim := Get_Identifier ("dim");
      Wki_I := Get_Identifier ("I");
      Wki_Instance := Get_Identifier ("INSTANCE");
      Wki_Arch_Instance := Get_Identifier ("ARCH_INSTANCE");
      Wki_Name := Get_Identifier ("NAME");

      Sizetype := New_Unsigned_Type (32);
      New_Type_Decl (Get_Identifier ("__ghdl_size_type"), Sizetype);

      --  Create __ghdl_index_type, which is the type for *all* array index.
      Ghdl_Index_Type := New_Unsigned_Type (32);
      New_Type_Decl (Get_Identifier ("__ghdl_index_type"), Ghdl_Index_Type);

      Ghdl_Integer_Type := New_Signed_Type (32);
      New_Type_Decl (Get_Identifier ("__ghdl_integer"), Ghdl_Integer_Type);

      Ghdl_Real_Type := New_Float_Type;
      New_Type_Decl (Get_Identifier ("__ghdl_real"), Ghdl_Real_Type);

      Ghdl_Phys_Type := New_Signed_Type (64);
      New_Type_Decl (Get_Identifier ("__ghdl_phys"), Ghdl_Phys_Type);

      --  File index for elaborated file object.
      File_Index_Type := New_Unsigned_Type (32);
      New_Type_Decl (Get_Identifier ("__ghdl_file_index"), File_Index_Type);
      File_Index_Ptr_Type := New_Access_Type (File_Index_Type);
      New_Type_Decl (Get_Identifier ("__ghdl_file_index_ptr"),
                     File_Index_Ptr_Type);

      --  Create const char and const char *.
      --Const_Char_Type_Node := Build_Qualified_Type
      --  (Char_Type_Node, TYPE_QUAL_CONST);
      Char_Type_Node := New_Unsigned_Type (8);
      New_Type_Decl (Get_Identifier ("__ghdl_char"), Char_Type_Node);

      Chararray_Type_Node := New_Array_Type (Char_Type_Node, Ghdl_Index_Type);
      New_Type_Decl (Get_Identifier ("__ghdl_chararray"), Chararray_Type_Node);

      Char_Ptr_Type_Node := New_Access_Type (Chararray_Type_Node);
      New_Type_Decl (Get_Identifier ("__ghdl_char_ptr"), Char_Ptr_Type_Node);

      Null_Const_Char_Ptr_Node := New_Null_Access (Char_Ptr_Type_Node);

      --  Generic pointer.
      Ghdl_Ptr_Type := New_Access_Type (Char_Type_Node);
      Const_Ptr_Type_Node := Ghdl_Ptr_Type;
      New_Type_Decl (Get_Identifier ("__ghdl_ptr"), Ghdl_Ptr_Type);

      --  Create record
      --     len : natural;
      --     str : C_String;
      --  end record;
      declare
         Constr : O_Element_List;
      begin
         Start_Record_Type (Constr);
         New_Record_Field (Constr, Ghdl_Str_Len_Type_Len_Field,
                           Get_Identifier ("len"), Ghdl_Index_Type);
         New_Record_Field
           (Constr, Ghdl_Str_Len_Type_Str_Field,
            Get_Identifier ("str"), Char_Ptr_Type_Node);
         Finish_Record_Type (Constr, Ghdl_Str_Len_Type_Node);
         New_Type_Decl (Get_Identifier ("__ghdl_str_len"),
                        Ghdl_Str_Len_Type_Node);
      end;

      Ghdl_Str_Len_Array_Type_Node := New_Array_Type
        (Ghdl_Str_Len_Type_Node, Ghdl_Index_Type);
      New_Type_Decl (Get_Identifier ("__ghdl_str_len_array"),
                     Ghdl_Str_Len_Array_Type_Node);

      -- Create type __ghdl_str_len_ptr is access all __ghdl_str_len
      Ghdl_Str_Len_Ptr_Node := New_Access_Type (Ghdl_Str_Len_Type_Node);
      New_Type_Decl (Get_Identifier ("__ghdl_str_len_ptr"),
                     Ghdl_Str_Len_Ptr_Node);

      -- Create type __ghdl_ada_bool is (false, true)
      New_Boolean_Type (Ghdl_Bool_Type_Node,
                        Get_Identifier ("false"),
                        Ghdl_Bool_False_Node,
                        Get_Identifier ("true"),
                        Ghdl_Bool_True_Node);
      New_Type_Decl (Get_Identifier ("__ghdl_bool_type"),
                     Ghdl_Bool_Type_Node);

      --  Create type ghdl_compare_type is (lt, eq, ge);
      declare
         Constr : O_Enum_List;
      begin
         Start_Enum_Type  (Constr, 8);
         New_Enum_Literal (Constr, Get_Identifier ("lt"), Ghdl_Compare_Lt);
         New_Enum_Literal (Constr, Get_Identifier ("eq"), Ghdl_Compare_Eq);
         New_Enum_Literal (Constr, Get_Identifier ("gt"), Ghdl_Compare_Gt);
         Finish_Enum_Type (Constr, Ghdl_Compare_Type);
         New_Type_Decl (Get_Identifier ("__ghdl_compare_type"),
                        Ghdl_Compare_Type);
      end;

      --  Create:
      --  type __ghdl_location is record
      --     file : __ghdl_str_len_ptr;
      --     line : integer;
      --     col : Integer;
      --  end record;
      declare
         Constr : O_Element_List;
      begin
         Start_Record_Type (Constr);
         New_Record_Field (Constr, Ghdl_Location_Filename_Node,
                           Get_Identifier ("filename"),
                           Ghdl_Str_Len_Ptr_Node);
         New_Record_Field (Constr, Ghdl_Location_Line_Node,
                           Get_Identifier ("line"),
                           Ghdl_Integer_Type);
         New_Record_Field (Constr, Ghdl_Location_Col_Node,
                           Get_Identifier ("col"),
                           Ghdl_Integer_Type);
         Finish_Record_Type (Constr, Ghdl_Location_Type_Node);
         New_Type_Decl (Get_Identifier ("__ghdl_location"),
                        Ghdl_Location_Type_Node);
      end;
      -- Create type __ghdl_location_ptr is access __ghdl_location;
      Ghdl_Location_Ptr_Node := New_Access_Type (Ghdl_Location_Type_Node);
      New_Type_Decl (Get_Identifier ("__ghdl_location_ptr"),
                     Ghdl_Location_Ptr_Node);


      --  Create type for ghdl_process_desc_type.
--       declare
--          Constr : O_Element_List;
--       begin
--          Start_Record_Type (Constr);
--          New_Record_Field
--            (Constr, Ghdl_Process_Desc_Name_Node,
--             Get_Identifier ("name"), Ghdl_Str_Len_Type_Node);
--          New_Record_Field
--            (Constr, Ghdl_Process_Desc_Sensitized_Node,
--             Get_Identifier ("is_sensitized"), Ghdl_Bool_Type_Node);
--          New_Record_Field
--            (Constr, Ghdl_Process_Desc_Passive_Node,
--             Get_Identifier ("is_passive"), Ghdl_Bool_Type_Node);
--          New_Record_Field
--            (Constr, Ghdl_Process_Desc_Proc_Node,
--             Get_Identifier ("proc"), Ghdl_Ptr_Type);
--          New_Record_Field
--            (Constr, Ghdl_Process_Desc_Signals_Node,
--             Get_Identifier ("signals"), Ghdl_Ptr_Type);
--          Finish_Record_Type (Constr, Ghdl_Process_Desc_Type_Node);
--          New_Type_Decl (Get_Identifier ("__ghdl_process_desc_type"),
--                         Ghdl_Process_Desc_Type_Node);
--       end;

--       -- Create:
--      -- type __ghdl_process_desc_ptr is access all __ghdl_process_desc_type.
--       Ghdl_Process_Desc_Ptr_Node :=
--         New_Access_Type (Ghdl_Process_Desc_Type_Node);
--       New_Type_Decl (Get_Identifier ("__ghdl_process_desc_type_ptr"),
--                      Ghdl_Process_Desc_Ptr_Node);

      --  Create type ghdl_dir_type is (dir_to, dir_downto);
      declare
         Constr : O_Enum_List;
      begin
         Start_Enum_Type (Constr, 8);
         New_Enum_Literal (Constr, Wki_Dir_To, Ghdl_Dir_To_Node);
         New_Enum_Literal (Constr, Wki_Dir_Downto, Ghdl_Dir_Downto_Node);
         Finish_Enum_Type (Constr, Ghdl_Dir_Type_Node);
         New_Type_Decl (Get_Identifier ("__ghdl_dir_type"),
                        Ghdl_Dir_Type_Node);
      end;

      --  Create type ghdl_type_kind is (ghdl_type_e8, ...)
      declare
         Constr : O_Enum_List;
      begin
         Start_Enum_Type (Constr, 8);
         New_Enum_Literal
           (Constr, Get_Identifier ("__ghdl_type_bool"), Ghdl_Kind_Bool_Node);
         New_Enum_Literal
           (Constr, Get_Identifier ("__ghdl_type_e8"), Ghdl_Kind_E8_Node);
         New_Enum_Literal
           (Constr, Get_Identifier ("__ghdl_type_e32"), Ghdl_Kind_E32_Node);
         New_Enum_Literal
           (Constr, Get_Identifier ("__ghdl_type_i32"), Ghdl_Kind_I32_Node);
         New_Enum_Literal
           (Constr, Get_Identifier ("__ghdl_type_i64"), Ghdl_Kind_I64_Node);
         New_Enum_Literal
           (Constr, Get_Identifier ("__ghdl_type_f64"), Ghdl_Kind_F64_Node);
         New_Enum_Literal
           (Constr, Get_Identifier ("__ghdl_type_p64"), Ghdl_Kind_P64_Node);
         New_Enum_Literal (Constr, Get_Identifier ("__ghdl_type_access"),
                           Ghdl_Kind_Access_Node);
         New_Enum_Literal (Constr, Get_Identifier ("__ghdl_type_array"),
                           Ghdl_Kind_Array_Node);
         New_Enum_Literal (Constr, Get_Identifier ("__ghdl_type_record"),
                           Ghdl_Kind_Record_Node);
         New_Enum_Literal (Constr, Get_Identifier ("__ghdl_subtype_array"),
                           Ghdl_Kind_Subarray_Node);
         New_Enum_Literal (Constr, Get_Identifier ("__ghdl_subtype_e8"),
                           Ghdl_Kind_Se8_Node);
         New_Enum_Literal (Constr, Get_Identifier ("__ghdl_subtype_i32"),
                           Ghdl_Kind_Si32_Node);
         New_Enum_Literal (Constr, Get_Identifier ("__ghdl_subtype_f64"),
                           Ghdl_Kind_Sf64_Node);
         New_Enum_Literal (Constr, Get_Identifier ("__ghdl_subtype_p64"),
                           Ghdl_Kind_Sp64_Node);
         Finish_Enum_Type (Constr, Ghdl_Type_Kind_Type_Node);
         New_Type_Decl (Get_Identifier ("__ghdl_type_kind"),
                        Ghdl_Type_Kind_Type_Node);
      end;

      --  type __ghdl_desc_ptr is access XXX;
      Ghdl_Desc_Ptr_Type := New_Access_Type (Char_Type_Node);
      New_Type_Decl (Get_Identifier ("__ghdl_desc_ptr"), Ghdl_Desc_Ptr_Type);


      --  type __ghdl_desc_ptr_array is array () of __ghdl_desc_ptr;
      Ghdl_Desc_Ptr_Array_Type := New_Array_Type
        (Ghdl_Desc_Ptr_Type, Ghdl_Index_Type);
      New_Type_Decl
        (Get_Identifier ("__ghdl_desc_ptr_array"), Ghdl_Desc_Ptr_Array_Type);

      --  type __ghdl_desc_ptr_array_ptr is access __ghdl_desc_ptr_array
      Ghdl_Desc_Ptr_Array_Ptr_Type :=
        New_Access_Type (Ghdl_Desc_Ptr_Array_Type);
      New_Type_Decl (Get_Identifier ("__ghdl_desc_ptr_array_ptr"),
                     Ghdl_Desc_Ptr_Array_Ptr_Type);

      --  Create:
      --  type __ghdl_desc_bool is record
      --      kind       : __ghdl_type_kind;
      --      false_name : __ghdl_str_len_type;
      --      true_name  : __ghdl_str_len_type;
      --  end record;
      declare
         Constr : O_Element_List;
      begin
         Start_Record_Type (Constr);
         New_Record_Field (Constr, Ghdl_Desc_Bool_Kind_Node,
                           Wki_Kind, Ghdl_Type_Kind_Type_Node);
         New_Record_Field
           (Constr, Ghdl_Desc_Bool_False_Node,
            Get_Identifier ("false_name"), Ghdl_Str_Len_Type_Node);
         New_Record_Field
           (Constr, Ghdl_Desc_Bool_True_Node,
            Get_Identifier ("true_name"), Ghdl_Str_Len_Type_Node);
         Finish_Record_Type (Constr, Ghdl_Desc_Bool_Type);
         New_Type_Decl (Get_Identifier ("__ghdl_desc_bool"),
                        Ghdl_Desc_Bool_Type);
      end;

      --  Create:
      --  type __ghdl_desc_e8 is record
      --      kind   : __ghdl_type_kind;
      --      erange : system.Address;
      --      values : __ghdl_str_len_ptr;
      --  end record;
      declare
         Constr : O_Element_List;
      begin
         Start_Record_Type (Constr);
         New_Record_Field (Constr, Ghdl_Desc_E8_Kind_Node,
                           Wki_Kind, Ghdl_Type_Kind_Type_Node);
         New_Record_Field (Constr, Ghdl_Desc_E8_Erange_Node,
                           Get_Identifier ("erange"), Ghdl_Ptr_Type);
         New_Record_Field
           (Constr, Ghdl_Desc_E8_Values_Node,
            Get_Identifier ("values"), Ghdl_Str_Len_Ptr_Node);
         Finish_Record_Type (Constr, Ghdl_Desc_E8_Type_Node);
         New_Type_Decl (Get_Identifier ("__ghdl_desc_e8"),
                        Ghdl_Desc_E8_Type_Node);
      end;

      --  Create:
      --  type __ghdl_desc_scalar_subtype is record
      --      kind   : __ghdl_type_kind;
      --      basetype : __ghdl_desc_ptr_type;
      --      range : system.Address;
      --  end record;
      declare
         Constr : O_Element_List;
      begin
         Start_Record_Type (Constr);
         New_Record_Field (Constr, Ghdl_Desc_Scalar_Kind_Node,
                           Wki_Kind, Ghdl_Type_Kind_Type_Node);
         New_Record_Field (Constr, Ghdl_Desc_Scalar_Basetype_Node,
                           Get_Identifier ("basetype"),
                           Ghdl_Desc_Ptr_Type);
         New_Record_Field
           (Constr, Ghdl_Desc_Scalar_Range_Node,
            Get_Identifier ("range"), Const_Ptr_Type_Node);
         Finish_Record_Type (Constr, Ghdl_Desc_Scalar_Subtype_Node);
         New_Type_Decl (Get_Identifier ("__ghdl_desc_scalar_subtype"),
                        Ghdl_Desc_Scalar_Subtype_Node);
      end;

      --  Create:
      --  type __ghdl_desc_i32 is record
      --      kind   : __ghdl_type_kind;
      --  end record;
      declare
         Constr : O_Element_List;
      begin
         Start_Record_Type (Constr);
         New_Record_Field (Constr, Ghdl_Desc_I32_Kind_Node,
                           Wki_Kind, Ghdl_Type_Kind_Type_Node);
         Finish_Record_Type (Constr, Ghdl_Desc_I32_Type_Node);
         New_Type_Decl (Get_Identifier ("__ghdl_desc_i32"),
                        Ghdl_Desc_I32_Type_Node);
      end;

      --  Create:
      --  type __ghdl_desc_f64 is record
      --      kind   : __ghdl_type_kind;
      --  end record;
      declare
         Constr : O_Element_List;
      begin
         Start_Record_Type (Constr);
         New_Record_Field (Constr, Ghdl_Desc_F64_Kind_Node,
                           Wki_Kind, Ghdl_Type_Kind_Type_Node);
         Finish_Record_Type (Constr, Ghdl_Desc_F64_Type_Node);
         New_Type_Decl (Get_Identifier ("__ghdl_desc_f64"),
                        Ghdl_Desc_F64_Type_Node);
      end;

      --  type __ghdl_p64 is signed (64)
      Ghdl_P64_Type := New_Signed_Type (64);
      New_Type_Decl (Get_Identifier ("__ghdl_p64"), Ghdl_P64_Type);

      --  type __ghdl_unit_desc is record
      --      name : ghdl_str_len_type;
      --      val  : __ghdl_p64;
      --  end record;
      declare
         Constr : O_Element_List;
      begin
         Start_Record_Type (Constr);
         New_Record_Field (Constr, Ghdl_Unit_Desc_Name_Node,
                           Get_Identifier ("name"), Ghdl_Str_Len_Type_Node);
         New_Record_Field (Constr, Ghdl_Unit_Desc_Val_Node,
                           Get_Identifier ("val"), Ghdl_P64_Type);
         Finish_Record_Type (Constr, Ghdl_Unit_Desc_Type);
         New_Type_Decl (Get_Identifier ("__ghdl_unit_desc"),
                        Ghdl_Unit_Desc_Type);
      end;

      --  type __ghdl_unit_desc_ptr is access __ghdl_unit_desc;
      Ghdl_Unit_Desc_Ptr := New_Access_Type (Ghdl_Unit_Desc_Type);
      New_Type_Decl (Get_Identifier ("__ghdl_unit_desc_ptr"),
                     Ghdl_Unit_Desc_Ptr);

      --  type __ghdl_unit_desc_array is array (__ghdl_index_type)
      --    of __ghdl_unit_desc.
      Ghdl_Unit_Desc_Array := New_Array_Type
        (Ghdl_Unit_Desc_Type, Ghdl_Index_Type);
      New_Type_Decl (Get_Identifier ("__ghdl_unit_desc_array"),
                     Ghdl_Unit_Desc_Array);

      --  Create:
      --  type __ghdl_desc_p64 is record
      --      kind      : __ghdl_type_kind;
      --      base_name : __ghdl_str_len_type;
      --      units     : __ghdl_unit_desc_ptr;
      --  end record;
      declare
         Constr : O_Element_List;
      begin
         Start_Record_Type (Constr);
         New_Record_Field (Constr, Ghdl_Desc_P64_Kind_Node,
                           Wki_Kind, Ghdl_Type_Kind_Type_Node);
         New_Record_Field (Constr, Ghdl_Desc_P64_Base_Name_Node,
                           Get_Identifier ("base_name"),
                           Ghdl_Str_Len_Type_Node);
         New_Record_Field (Constr, Ghdl_Desc_P64_Units_Node,
                           Get_Identifier ("units"), Ghdl_Unit_Desc_Ptr);
         Finish_Record_Type (Constr, Ghdl_Desc_P64_Type_Node);
         New_Type_Decl (Get_Identifier ("__ghdl_desc_p64"),
                        Ghdl_Desc_P64_Type_Node);
      end;

      --  type __ghdl_desc_access is record
      --      kind : __ghdl_type_kind;
      --      designated : system.address;
      --  end record;
      declare
         Constr : O_Element_List;
      begin
         Start_Record_Type (Constr);
         New_Record_Field
           (Constr, Ghdl_Desc_Access_Kind_Node,
            Wki_Kind, Ghdl_Type_Kind_Type_Node);
         New_Record_Field
           (Constr, Ghdl_Desc_Access_Designated_Node,
            Get_Identifier ("designated"), Ghdl_Desc_Ptr_Type);
         Finish_Record_Type (Constr, Ghdl_Desc_Access_Type_Node);
         New_Type_Decl (Get_Identifier ("__ghdl_desc_access"),
                        Ghdl_Desc_Access_Type_Node);
      end;

      --  type __ghdl_desc_array is record
      --      kind : __ghdl_type_kind;
      --      el_desc : system.address;
      --      el_size : sizetype;
      --      nbr_dim : natural;
      --      indexes : system.address;
      --  end record;
      declare
         Constr : O_Element_List;
      begin
         Start_Record_Type (Constr);
         New_Record_Field
           (Constr, Ghdl_Desc_Array_Kind_Node,
            Wki_Kind, Ghdl_Type_Kind_Type_Node);
         New_Record_Field
           (Constr, Ghdl_Desc_Array_El_Node,
            Get_Identifier ("el_desc"), Ghdl_Desc_Ptr_Type);
         New_Record_Field
           (Constr, Ghdl_Desc_Array_Elsize_Node,
            Get_Identifier ("el_size"), Sizetype);
         New_Record_Field
           (Constr, Ghdl_Desc_Array_Dim_Node,
            Get_Identifier ("nbr_dim"), Ghdl_Integer_Type);
         New_Record_Field
           (Constr, Ghdl_Desc_Array_Indexes_Node,
            Get_Identifier ("indexes"), Ghdl_Desc_Ptr_Array_Ptr_Type);
         Finish_Record_Type (Constr, Ghdl_Desc_Array_Type_Node);
         New_Type_Decl (Get_Identifier ("__ghdl_desc_array"),
                        Ghdl_Desc_Array_Type_Node);
      end;

      --  type __ghdl_desc_subarray is record
      --      kind : __ghdl_type_kind;
      --      basetype : __ghdl_desc_array_acc;
      --      trange : system.address;
      --  end record;
      declare
         Constr : O_Element_List;
      begin
         Start_Record_Type (Constr);
         New_Record_Field
           (Constr, Ghdl_Desc_Subarray_Kind_Node,
            Wki_Kind, Ghdl_Type_Kind_Type_Node);
         New_Record_Field
           (Constr, Ghdl_Desc_Subarray_Basetype_Node,
            Get_Identifier ("basetype"), Ghdl_Desc_Ptr_Type);
         New_Record_Field
           (Constr, Ghdl_Desc_Subarray_Range_Node,
            Get_Identifier ("trange"), Const_Ptr_Type_Node);
         Finish_Record_Type (Constr, Ghdl_Desc_Subarray_Type_Node);
         New_Type_Decl (Get_Identifier ("__ghdl_desc_subarray"),
                        Ghdl_Desc_Subarray_Type_Node);
      end;

      --  type __ghdl_desc_record is record
      --      kind : __ghdl_type_kind;
      --      nbr_el : natural;
      --      els : __ghdl_desc_ptr_array_ptr;
      --  end record;
      declare
         Constr : O_Element_List;
      begin
         Start_Record_Type (Constr);
         New_Record_Field (Constr, Ghdl_Desc_Record_Kind_Node,
                           Wki_Kind, Ghdl_Type_Kind_Type_Node);
         New_Record_Field (Constr, Ghdl_Desc_Record_Nbr_El_Node,
                           Get_Identifier ("nbr_el"), Ghdl_Integer_Type);
         New_Record_Field (Constr, Ghdl_Desc_Record_Els_Node,
                           Get_Identifier ("els"),
                           Ghdl_Desc_Ptr_Array_Ptr_Type);
         Finish_Record_Type (Constr, Ghdl_Desc_Record_Type_Node);
         New_Type_Decl (Get_Identifier ("__ghdl_desc_record"),
                        Ghdl_Desc_Record_Type_Node);
      end;

      --  type __ghdl_name_kind is
      --    (name_library, name_entity, name_architecture,
      --     name_block, name_component_instance, name_package,
      --     name_process, name_configuration);
      declare
         Constr : O_Enum_List;
      begin
         Start_Enum_Type (Constr, 32);
         New_Enum_Literal (Constr, Get_Identifier ("__ghdl_name_library"),
                           Ghdl_Name_Library);
         New_Enum_Literal (Constr, Get_Identifier ("__ghdl_name_entity"),
                           Ghdl_Name_Entity);
         New_Enum_Literal (Constr, Get_Identifier ("__ghdl_name_architecture"),
                           Ghdl_Name_Architecture);
         New_Enum_Literal (Constr, Get_Identifier ("__ghdl_name_block"),
                           Ghdl_Name_Block);
         New_Enum_Literal (Constr,
                           Get_Identifier ("__ghdl_name_generate_iterative"),
                           Ghdl_Name_Generate_Iterative);
         New_Enum_Literal (Constr,
                           Get_Identifier ("__ghdl_name_generate_conditional"),
                           Ghdl_Name_Generate_Conditional);
         New_Enum_Literal (Constr, Get_Identifier ("__ghdl_name_component"),
                           Ghdl_Name_Component);
         New_Enum_Literal (Constr, Get_Identifier ("__ghdl_name_package"),
                           Ghdl_Name_Package);
         New_Enum_Literal (Constr, Get_Identifier ("__ghdl_name_process"),
                           Ghdl_Name_Process);
         New_Enum_Literal
           (Constr, Get_Identifier ("__ghdl_name_configuration"),
            Ghdl_Name_Configuration);
         New_Enum_Literal (Constr, Get_Identifier ("__ghdl_name_signal"),
                           Ghdl_Name_Signal);
         New_Enum_Literal (Constr, Get_Identifier ("__ghdl_name_top"),
                           Ghdl_Name_Top);
         Finish_Enum_Type (Constr, Ghdl_Name_Kind_Type);
         New_Type_Decl (Get_Identifier ("__ghdl_name_kind"),
                        Ghdl_Name_Kind_Type);
      end;

      --  type ghdl_instance_name_acc is access;
      --  Incomplete type.
      New_Uncomplete_Record_Type (Ghdl_Instance_Name_Type);
      New_Type_Decl (Get_Identifier ("__ghdl_instance_name_type"),
                     Ghdl_Instance_Name_Type);
      Ghdl_Instance_Name_Acc := New_Access_Type (Ghdl_Instance_Name_Type);
      New_Type_Decl (Get_Identifier ("__ghdl_instance_name_ptr"),
                     Ghdl_Instance_Name_Acc);
      declare
         Constr : O_Element_List;
      begin
         Start_Uncomplete_Record_Type (Ghdl_Instance_Name_Type, Constr);
         New_Record_Field (Constr, Ghdl_Instance_Name_Kind,
                           Wki_Kind, Ghdl_Name_Kind_Type);
         New_Record_Field (Constr, Ghdl_Instance_Name_Name,
                           Get_Identifier ("name"), Ghdl_Str_Len_Ptr_Node);
         New_Record_Field (Constr, Ghdl_Instance_Name_Parent,
                           Get_Identifier ("parent"), Ghdl_Instance_Name_Acc);
         New_Record_Field (Constr, Ghdl_Instance_Name_Brother,
                           Get_Identifier ("brother"), Ghdl_Instance_Name_Acc);
         New_Record_Field (Constr, Ghdl_Instance_Name_Child,
                           Get_Identifier ("child"), Ghdl_Instance_Name_Acc);
         Finish_Record_Type (Constr, Ghdl_Instance_Name_Type);
      end;

      --  __ghdl_hierarchy_top : __ghdl_instance_name_type;
      New_Var_Decl
        (Ghdl_Hierarchy_Top, Get_Identifier ("__ghdl_hierarchy_top"),
         O_Storage_External, Ghdl_Instance_Name_Type);

      --  Create void* __ghdl_alloc (unsigned size);
      Start_Function_Decl (Interfaces, Get_Identifier ("__ghdl_alloc"),
                           O_Storage_External, Ghdl_Ptr_Type);
      New_Interface_Decl (Interfaces, Param, Get_Identifier ("size"),
                          Sizetype);
      Finish_Subprogram_Decl (Interfaces, Ghdl_Alloc_Ptr);

      --  procedure __ghdl_process_register (this : ghdl_ptr_type;
      --                                     proc : ghdl_ptr_type;
      --                                     name : __ghdl_instance_name_ptr);
      Start_Procedure_Decl
        (Interfaces, Get_Identifier ("__ghdl_process_register"),
         O_Storage_External);
      New_Interface_Decl (Interfaces, Param, Get_Identifier ("this"),
                          Ghdl_Ptr_Type);
      New_Interface_Decl (Interfaces, Param, Get_Identifier ("proc"),
                          Ghdl_Ptr_Type);
      New_Interface_Decl (Interfaces, Param, Wki_Name, Ghdl_Instance_Name_Acc);
      Finish_Subprogram_Decl (Interfaces, Ghdl_Process_Register);

      --  procedure __ghdl_sensitized_process_register
      --     (this : ghdl_ptr_type;
      --      proc : ghdl_ptr_type;
      --      name : __ghdl_instance_name_ptr);
      Start_Procedure_Decl
        (Interfaces, Get_Identifier ("__ghdl_sensitized_process_register"),
         O_Storage_External);
      New_Interface_Decl (Interfaces, Param, Get_Identifier ("this"),
                          Ghdl_Ptr_Type);
      New_Interface_Decl (Interfaces, Param, Get_Identifier ("proc"),
                          Ghdl_Ptr_Type);
      New_Interface_Decl (Interfaces, Param, Wki_Name, Ghdl_Instance_Name_Acc);
      Finish_Subprogram_Decl (Interfaces, Ghdl_Sensitized_Process_Register);

      --  procedure __ghdl_program_error -- (loc : __ghdl_location_acc);
      Start_Procedure_Decl
        (Interfaces, Get_Identifier ("__ghdl_program_error"),
         O_Storage_External);
      --New_Interface_Decl (Interfaces, Param,
      --                    Get_Identifier ("location"),
      --                    Ghdl_Location_Ptr_Node);
      Finish_Subprogram_Decl (Interfaces, Ghdl_Program_Error);

      --  procedure __ghdl_bound_check_failed_l0;
      Start_Procedure_Decl
        (Interfaces, Get_Identifier ("__ghdl_bound_check_failed_l0"),
         O_Storage_External);
      New_Interface_Decl (Interfaces, Param, Get_Identifier ("index"),
                          Ghdl_Index_Type);
      Finish_Subprogram_Decl (Interfaces, Ghdl_Bound_Check_Failed_L0);

      --  procedure __ghdl_bound_check_failed_l1 (loc : __ghdl_location_acc;
      --                                          dim : __ghdl_index_type);
      Start_Procedure_Decl
        (Interfaces, Get_Identifier ("__ghdl_bound_check_failed_l1"),
         O_Storage_External);
      New_Interface_Decl
        (Interfaces, Param, Get_Identifier ("loc"), Ghdl_Location_Ptr_Node);
      New_Interface_Decl
        (Interfaces, Param, Get_Identifier ("dim"), Ghdl_Index_Type);
      Finish_Subprogram_Decl (Interfaces, Ghdl_Bound_Check_Failed_L1);

      --  Secondary stack subprograms.
      --  function __ghdl_stack2_allocate (size : ghdl_index_type)
      --    return ghdl_ptr_type;
      Start_Function_Decl
        (Interfaces, Get_Identifier ("__ghdl_stack2_allocate"),
         O_Storage_External, Ghdl_Ptr_Type);
      New_Interface_Decl (Interfaces, Param, Get_Identifier ("size"),
                          Ghdl_Index_Type);
      Finish_Subprogram_Decl (Interfaces, Ghdl_Stack2_Allocate);

      --  function __ghdl_stack2_mark return ghdl_ptr_type;
      Start_Function_Decl (Interfaces, Get_Identifier ("__ghdl_stack2_mark"),
                           O_Storage_External, Ghdl_Ptr_Type);
      Finish_Subprogram_Decl (Interfaces, Ghdl_Stack2_Mark);

      --  procedure __ghdl_stack2_release (mark : ghdl_ptr_type);
      Start_Procedure_Decl
        (Interfaces, Get_Identifier ("__ghdl_stack2_release"),
         O_Storage_External);
      New_Interface_Decl (Interfaces, Param, Get_Identifier ("mark"),
                          Ghdl_Ptr_Type);
      Finish_Subprogram_Decl (Interfaces, Ghdl_Stack2_Release);

      --  procedure __ghdl_memcpy (dest : ghdl_ptr_type;
      --                           src  : ghdl_ptr_type;
      --                           length : ghdl_index_type);
      Start_Procedure_Decl
        (Interfaces, Get_Identifier ("__ghdl_memcpy"), O_Storage_External);
      New_Interface_Decl (Interfaces, Param, Get_Identifier ("dest"),
                          Ghdl_Ptr_Type);
      New_Interface_Decl (Interfaces, Param, Get_Identifier ("src"),
                          Ghdl_Ptr_Type);
      New_Interface_Decl (Interfaces, Param, Get_Identifier ("length"),
                          Ghdl_Index_Type);
      Finish_Subprogram_Decl (Interfaces, Ghdl_Memcpy);

      --  procedure __ghdl_deallocate (ptr : ghdl_ptr_type);
      Start_Procedure_Decl
        (Interfaces, Get_Identifier ("__ghdl_deallocate"), O_Storage_External);
      New_Interface_Decl (Interfaces, Param, Get_Identifier ("object"),
                          Ghdl_Ptr_Type);
      Finish_Subprogram_Decl (Interfaces, Ghdl_Deallocate);

      -- function __ghdl_malloc (length : ghdl_index_type)
      --    return ghdl_ptr_type;
      Start_Function_Decl
        (Interfaces, Get_Identifier ("__ghdl_malloc"), O_Storage_External,
         Ghdl_Ptr_Type);
      New_Interface_Decl (Interfaces, Param, Get_Identifier ("length"),
                          Ghdl_Index_Type);
      Finish_Subprogram_Decl (Interfaces, Ghdl_Malloc);

      --  function __ghdl_text_file_elaborate return file_index_type;
      Start_Function_Decl
        (Interfaces, Get_Identifier ("__ghdl_text_file_elaborate"),
         O_Storage_External, File_Index_Type);
      Finish_Subprogram_Decl (Interfaces, Ghdl_Text_File_Elaborate);

      --  function __ghdl_file_elaborate return file_index_type;
      Start_Function_Decl
        (Interfaces, Get_Identifier ("__ghdl_file_elaborate"),
         O_Storage_External, File_Index_Type);
      Finish_Subprogram_Decl (Interfaces, Ghdl_File_Elaborate);

      --  procedure __ghdl_signal_name
      --       (name : __ghdl_str_len_ptr;
      --        instance_name : __ghdl_instance_name_acc;
      --        mode : ghdl_integer_type;
      --        kind : ghdl_integer_type;
      --        type_desc : ghdl_desc_ptr_type);
      Start_Procedure_Decl
        (Interfaces, Get_Identifier ("__ghdl_signal_name"),
         O_Storage_External);
      New_Interface_Decl (Interfaces, Param, Get_Identifier ("name"),
                          Ghdl_Str_Len_Ptr_Node);
      New_Interface_Decl (Interfaces, Param, Get_Identifier ("instance_name"),
                          Ghdl_Instance_Name_Acc);
      New_Interface_Decl (Interfaces, Param, Get_Identifier ("mode"),
                          Ghdl_Integer_Type);
      New_Interface_Decl (Interfaces, Param, Get_Identifier ("kind"),
                          Ghdl_Integer_Type);
      New_Interface_Decl (Interfaces, Param, Get_Identifier ("type_desc"),
                          Ghdl_Desc_Ptr_Type);
      Finish_Subprogram_Decl (Interfaces, Ghdl_Signal_Name);
   end Initialize;

   procedure Create_Signal_Subprograms
     (Suffix : String;
      Val_Type : O_Tnode;
      Create_Signal : out O_Dnode;
      Simple_Assign : out O_Dnode;
      Start_Assign : out O_Dnode;
      Next_Assign : out O_Dnode)
   is
      Interfaces : O_Inter_List;
      Param : O_Lnode;
   begin
      --  function __ghdl_create_signal_XXX (init_val : VAL_TYPE)
      --                                     type_desc : TYPE_DESC_PTR;
      --                                     resolv_func : ghdl_ptr_type;
      --                                     resolv_inst : ghdl_ptr_type;
      --                                     return __ghdl_signal_ptr;
      Start_Function_Decl
        (Interfaces, Get_Identifier ("__ghdl_create_signal_" & Suffix),
         O_Storage_External, Ghdl_Signal_Ptr);
      New_Interface_Decl
        (Interfaces, Param, Get_Identifier ("init_val"), Val_Type);
      New_Interface_Decl (Interfaces, Param, Get_Identifier ("type_desc"),
                          Ghdl_Desc_Ptr_Type);
      New_Interface_Decl (Interfaces, Param, Get_Identifier ("resolv_func"),
                          Ghdl_Ptr_Type);
      New_Interface_Decl (Interfaces, Param, Get_Identifier ("resolv_inst"),
                          Ghdl_Ptr_Type);
      Finish_Subprogram_Decl (Interfaces, Create_Signal);

      --  procedure __ghdl_signal_simple_assign_XXX (sign : __ghdl_signal_ptr;
      --                                             val : VAL_TYPE);
      Start_Procedure_Decl
        (Interfaces, Get_Identifier ("__ghdl_signal_simple_assign_" & Suffix),
         O_Storage_External);
      New_Interface_Decl (Interfaces, Param, Get_Identifier ("signal"),
                          Ghdl_Signal_Ptr);
      New_Interface_Decl (Interfaces, Param, Get_Identifier ("val"), Val_Type);
      Finish_Subprogram_Decl (Interfaces, Simple_Assign);

      --  procedure __ghdl_signal_start_assign_XXX (sign : __ghdl_signal_ptr;
      --                                            reject : std_time;
      --                                            val : VAL_TYPE;
      --                                            after : std_time);
      Start_Procedure_Decl
        (Interfaces, Get_Identifier ("__ghdl_signal_start_assign_" & Suffix),
         O_Storage_External);
      New_Interface_Decl (Interfaces, Param, Get_Identifier ("signal"),
                          Ghdl_Signal_Ptr);
      New_Interface_Decl (Interfaces, Param, Get_Identifier ("reject"),
                          Std_Time_Type);
      New_Interface_Decl (Interfaces, Param, Get_Identifier ("val"),
                          Val_Type);
      New_Interface_Decl (Interfaces, Param, Get_Identifier ("after"),
                          Std_Time_Type);
      Finish_Subprogram_Decl (Interfaces, Start_Assign);

      --  procedure __ghdl_signal_next_assign_XXX (sign : __ghdl_signal_ptr;
      --                                            val : VAL_TYPE;
      --                                            after : std_time);
      Start_Procedure_Decl
        (Interfaces, Get_Identifier ("__ghdl_signal_next_assign_" & Suffix),
         O_Storage_External);
      New_Interface_Decl (Interfaces, Param, Get_Identifier ("signal"),
                          Ghdl_Signal_Ptr);
      New_Interface_Decl (Interfaces, Param, Get_Identifier ("val"),
                          Val_Type);
      New_Interface_Decl (Interfaces, Param, Get_Identifier ("after"),
                          Std_Time_Type);
      Finish_Subprogram_Decl (Interfaces, Next_Assign);
   end Create_Signal_Subprograms;

   --  Do internal declarations that need std.standard declarations.
   procedure Post_Initialize
   is
      Interfaces : O_Inter_List;
      Rec : O_Element_List;
      Param : O_Lnode;
      Integer_Otype : O_Tnode;
      Real_Otype : O_Tnode;
      Time_Otype : O_Tnode;
   begin
      New_Debug_Comment_Decl ("internal declarations, part 2");
      Std_String_Ptr_Node :=
        Get_Info (String_Type_Definition).Ortho_Ptr_Type (Mode_Value);
      Integer_Otype := Get_Ortho_Type (Integer_Type_Definition, Mode_Value);
      Real_Otype := Get_Ortho_Type (Real_Type_Definition, Mode_Value);
      Time_Otype := Get_Ortho_Type (Time_Type_Definition, Mode_Value);

      --  __ghdl_now : time;
      --  ??? maybe this should be a function ?
      New_Var_Decl (Ghdl_Now, Get_Identifier ("__ghdl_now"),
                    O_Storage_External, Time_Otype);

      --  procedure __ghdl_assert_failed (str : __ghdl_array_template;
      --                                  severity : ghdl_int);
      --                                  loc : __ghdl_location_acc);
      Start_Procedure_Decl
        (Interfaces, Get_Identifier ("__ghdl_assert_failed"),
         O_Storage_External);
      New_Interface_Decl
        (Interfaces, Param, Get_Identifier ("msg"), Std_String_Ptr_Node);
      New_Interface_Decl
        (Interfaces, Param, Get_Identifier ("severity"),
         Get_Ortho_Type (Severity_Level_Type_Definition, Mode_Value));
      New_Interface_Decl (Interfaces, Param, Get_Identifier ("location"),
                          Ghdl_Location_Ptr_Node);
      Finish_Subprogram_Decl (Interfaces, Ghdl_Assert_Failed);

      New_Var_Decl (Std_Assert_Default_Report_Node,
                    Get_Identifier ("__ghdl_assert_default_report"),
                    O_Storage_External,
                    Get_Info (String_Type_Definition).Ortho_Type (Mode_Value));

      --  procedure __ghdl_text_write (file : __ghdl_file_index;
      --                               str  : std_string_ptr);
      Start_Procedure_Decl
        (Interfaces, Get_Identifier ("__ghdl_text_write"), O_Storage_External);
      New_Interface_Decl (Interfaces, Param, Get_Identifier ("file"),
                          File_Index_Type);
      New_Interface_Decl (Interfaces, Param, Get_Identifier ("str"),
                          Std_String_Ptr_Node);
      Finish_Subprogram_Decl (Interfaces, Ghdl_Text_Write);

      --  function __ghdl_text_read_length (file : __ghdl_file_index;
      --                                    str : std_string_ptr)
      --     return std__standard_integer;
      Start_Function_Decl
        (Interfaces, Get_Identifier ("__ghdl_text_read_length"),
         O_Storage_External, Integer_Otype);
      New_Interface_Decl (Interfaces, Param, Get_Identifier ("file"),
                          File_Index_Type);
      New_Interface_Decl (Interfaces, Param, Get_Identifier ("str"),
                          Std_String_Ptr_Node);
      Finish_Subprogram_Decl (Interfaces, Ghdl_Text_Read_Length);

      --  procedure __ghdl_write_scalar (file : __ghdl_file_index;
      --                                 ptr : __ghdl_ptr_type;
      --                                 length : __ghdl_index_type);
      Start_Procedure_Decl
        (Interfaces, Get_Identifier ("__ghdl_write_scalar"),
         O_Storage_External);
      New_Interface_Decl (Interfaces, Param, Get_Identifier ("file"),
                          File_Index_Type);
      New_Interface_Decl (Interfaces, Param, Get_Identifier ("ptr"),
                          Ghdl_Ptr_Type);
      New_Interface_Decl (Interfaces, Param, Get_Identifier ("length"),
                          Ghdl_Index_Type);
      Finish_Subprogram_Decl (Interfaces, Ghdl_Write_Scalar);

      --  procedure __ghdl_read_scalar (file : __ghdl_file_index;
      --                                ptr : __ghdl_ptr_type;
      --                                length : __ghdl_index_type);
      Start_Procedure_Decl
        (Interfaces, Get_Identifier ("__ghdl_read_scalar"),
         O_Storage_External);
      New_Interface_Decl (Interfaces, Param, Get_Identifier ("file"),
                          File_Index_Type);
      New_Interface_Decl (Interfaces, Param, Get_Identifier ("ptr"),
                          Ghdl_Ptr_Type);
      New_Interface_Decl (Interfaces, Param, Get_Identifier ("length"),
                          Ghdl_Index_Type);
      Finish_Subprogram_Decl (Interfaces, Ghdl_Read_Scalar);

      --  function __ghdl_real_exp (left : std__standard__real;
      --                            right : std__standard__integer)
      --   return std__standard__real;
      Start_Function_Decl
        (Interfaces, Get_Identifier ("__ghdl_real_exp"), O_Storage_External,
         Real_Otype);
      New_Interface_Decl (Interfaces, Param, Get_Identifier ("left"),
                          Real_Otype);
      New_Interface_Decl (Interfaces, Param, Get_Identifier ("right"),
                          Integer_Otype);
      Finish_Subprogram_Decl (Interfaces, Ghdl_Real_Exp);

      --  function __ghdl_integer_exp (left : std__standard__integer;
      --                               right : std__standard__integer)
      --   return std__standard__integer;
      Start_Function_Decl
        (Interfaces, Get_Identifier ("__ghdl_integer_exp"), O_Storage_External,
         Integer_Otype);
      New_Interface_Decl (Interfaces, Param, Wki_Left, Integer_Otype);
      New_Interface_Decl (Interfaces, Param, Wki_Right, Integer_Otype);
      Finish_Subprogram_Decl (Interfaces, Ghdl_Integer_Exp);

      --  procedure __ghdl_text_file_open (file : file_index_type;
      --                                   mode : Ghdl_Integer_Type;
      --                                   str : std__standard__string_PTR);
      Start_Procedure_Decl
        (Interfaces, Get_Identifier ("__ghdl_text_file_open"),
         O_Storage_External);
      New_Interface_Decl (Interfaces, Param, Get_Identifier ("file"),
                          File_Index_Type);
      New_Interface_Decl (Interfaces, Param, Get_Identifier ("mode"),
                          Ghdl_Integer_Type);
      New_Interface_Decl (Interfaces, Param, Get_Identifier ("str"),
                          Std_String_Ptr_Node);
      Finish_Subprogram_Decl (Interfaces, Ghdl_Text_File_Open);

      --  procedure __ghdl_file_open (file : file_index_type;
      --                              mode : Ghdl_Integer_Type;
      --                              str : std__standard__string_PTR);
      Start_Procedure_Decl
        (Interfaces, Get_Identifier ("__ghdl_file_open"),
         O_Storage_External);
      New_Interface_Decl (Interfaces, Param, Get_Identifier ("file"),
                          File_Index_Type);
      New_Interface_Decl (Interfaces, Param, Get_Identifier ("mode"),
                          Ghdl_Integer_Type);
      New_Interface_Decl (Interfaces, Param, Get_Identifier ("str"),
                          Std_String_Ptr_Node);
      Finish_Subprogram_Decl (Interfaces, Ghdl_File_Open);

      --  function __ghdl_text_file_open_status
      --    (file : file_index_type;
      --     mode : Ghdl_Integer_Type;
      --     str : std__standard__string_PTR)
      --     return ghdl_integer_type;
      Start_Function_Decl
        (Interfaces, Get_Identifier ("__ghdl_text_file_open_status"),
         O_Storage_External, Ghdl_Integer_Type);
      New_Interface_Decl (Interfaces, Param, Get_Identifier ("file"),
                          File_Index_Type);
      New_Interface_Decl (Interfaces, Param, Get_Identifier ("mode"),
                          Ghdl_Integer_Type);
      New_Interface_Decl (Interfaces, Param, Get_Identifier ("str"),
                          Std_String_Ptr_Node);
      Finish_Subprogram_Decl (Interfaces, Ghdl_Text_File_Open_Status);

      --  function __ghdl_file_open_status (file : file_index_type;
      --                                    mode : Ghdl_Integer_Type;
      --                                    str : std__standard__string_PTR)
      --     return ghdl_integer_type;
      Start_Function_Decl
        (Interfaces, Get_Identifier ("__ghdl_file_open_status"),
         O_Storage_External, Ghdl_Integer_Type);
      New_Interface_Decl (Interfaces, Param, Get_Identifier ("file"),
                          File_Index_Type);
      New_Interface_Decl (Interfaces, Param, Get_Identifier ("mode"),
                          Ghdl_Integer_Type);
      New_Interface_Decl (Interfaces, Param, Get_Identifier ("str"),
                          Std_String_Ptr_Node);
      Finish_Subprogram_Decl (Interfaces, Ghdl_File_Open_Status);

      --  function __ghdl_file_endfile (file : file_index_type)
      --    return std_boolean_type_node;
      Start_Function_Decl (Interfaces, Get_Identifier ("__ghdl_file_endfile"),
                           O_Storage_External, Std_Boolean_Type_Node);
      New_Interface_Decl (Interfaces, Param, Get_Identifier ("file"),
                          File_Index_Type);
      Finish_Subprogram_Decl (Interfaces, Ghdl_File_Endfile);

      --  procedure __ghdl_text_file_close (file : file_index_type);
      Start_Procedure_Decl
        (Interfaces, Get_Identifier ("__ghdl_text_file_close"),
         O_Storage_External);
      New_Interface_Decl (Interfaces, Param, Get_Identifier ("file"),
                          File_Index_Type);
      Finish_Subprogram_Decl (Interfaces, Ghdl_Text_File_Close);

      --  procedure __ghdl_file_close (file : file_index_type);
      Start_Procedure_Decl (Interfaces, Get_Identifier ("__ghdl_file_close"),
                            O_Storage_External);
      New_Interface_Decl (Interfaces, Param, Get_Identifier ("file"),
                          File_Index_Type);
      Finish_Subprogram_Decl (Interfaces, Ghdl_File_Close);

      --  procedure __ghdl_signal_create_resolution
      --    (type_desc : TYPE_DESC_PTR;
      --     resolv_func : ghdl_ptr_type;
      --     resolv_inst : ghdl_ptr_type);
      Start_Procedure_Decl
        (Interfaces, Get_Identifier ("__ghdl_signal_create_resolution"),
         O_Storage_External);
      New_Interface_Decl  (Interfaces, Param, Get_Identifier ("type_desc"),
                           Ghdl_Desc_Ptr_Type);
      New_Interface_Decl (Interfaces, Param, Get_Identifier ("resolv_func"),
                          Ghdl_Ptr_Type);
      New_Interface_Decl (Interfaces, Param, Get_Identifier ("resolv_inst"),
                          Ghdl_Ptr_Type);
      Finish_Subprogram_Decl (Interfaces, Ghdl_Signal_Create_Resolution);

      --  Declarations for signals.
      --  Max length of a scalar type.
      --  type __ghdl_scalar_bytes is __ghdl_chararray (0 .. 8);
      Ghdl_Scalar_Bytes := New_Constrained_Array_Type
        (Chararray_Type_Node, New_Unsigned_Literal (Ghdl_Index_Type, 8));
      New_Type_Decl (Get_Identifier ("__ghdl_scalar_bytes"),
                     Ghdl_Scalar_Bytes);
      --  Type __signal_signal is record
      Start_Record_Type (Rec);
      New_Record_Field (Rec, Ghdl_Signal_Value_Node,
                        Get_Identifier ("value"),
                        Ghdl_Scalar_Bytes);
      New_Record_Field (Rec, Ghdl_Signal_Driving_Value_Node,
                        Get_Identifier ("driving_value"),
                        Ghdl_Scalar_Bytes);
      New_Record_Field (Rec, Ghdl_Signal_Last_Value_Node,
                        Get_Identifier ("last_value"),
                        Ghdl_Scalar_Bytes);
      New_Record_Field (Rec, Ghdl_Signal_Last_Event_Node,
                        Get_Identifier ("last_event"),
                        Time_Otype);
      New_Record_Field (Rec, Ghdl_Signal_Last_Active_Node,
                        Get_Identifier ("last_active"),
                        Time_Otype);
      New_Record_Field (Rec, Ghdl_Signal_Event_Node,
                        Get_Identifier ("event"),
                        Std_Boolean_Type_Node);
      New_Record_Field (Rec, Ghdl_Signal_Active_Node,
                        Get_Identifier ("active"),
                        Std_Boolean_Type_Node);
      Finish_Record_Type (Rec, Ghdl_Signal_Type);
      New_Type_Decl (Get_Identifier ("__ghdl_signal"), Ghdl_Signal_Type);
      Ghdl_Signal_Ptr := New_Access_Type (Ghdl_Signal_Type);
      New_Type_Decl (Get_Identifier ("__ghdl_signal_ptr"), Ghdl_Signal_Ptr);

      --  procedure __ghdl_signal_add_source (targ : __ghdl_signal_ptr;
      --                                      src : __ghdl_signal_ptr);
      Start_Procedure_Decl
        (Interfaces, Get_Identifier ("__ghdl_signal_add_source"),
         O_Storage_External);
      New_Interface_Decl (Interfaces, Param, Get_Identifier ("targ"),
                          Ghdl_Signal_Ptr);
      New_Interface_Decl (Interfaces, Param, Get_Identifier ("src"),
                          Ghdl_Signal_Ptr);
      Finish_Subprogram_Decl (Interfaces, Ghdl_Signal_Add_Source);

      --  procedure __ghdl_signal_effective_value (targ : __ghdl_signal_ptr;
      --                                           src : __ghdl_signal_ptr);
      Start_Procedure_Decl
        (Interfaces, Get_Identifier ("__ghdl_signal_effective_value"),
         O_Storage_External);
      New_Interface_Decl (Interfaces, Param, Get_Identifier ("targ"),
                          Ghdl_Signal_Ptr);
      New_Interface_Decl (Interfaces, Param, Get_Identifier ("src"),
                          Ghdl_Signal_Ptr);
      Finish_Subprogram_Decl (Interfaces, Ghdl_Signal_Effective_Value);

      --  procedure __ghdl_signal_simple_assign_error
      --              (sig : __ghdl_signal_ptr);
      Start_Procedure_Decl
        (Interfaces, Get_Identifier ("__ghdl_signal_simple_assign_error"),
         O_Storage_External);
      New_Interface_Decl
        (Interfaces, Param, Get_Identifier ("sig"), Ghdl_Signal_Ptr);
      Finish_Subprogram_Decl (Interfaces, Ghdl_Signal_Simple_Assign_Error);

      --  procedure __ghdl_signal_start_assign_error (sign : __ghdl_signal_ptr;
      --                                              reject : std_time;
      --                                              after : std_time);
      Start_Procedure_Decl
        (Interfaces, Get_Identifier ("__ghdl_signal_start_assign_error"),
         O_Storage_External);
      New_Interface_Decl (Interfaces, Param, Get_Identifier ("signal"),
                          Ghdl_Signal_Ptr);
      New_Interface_Decl (Interfaces, Param, Get_Identifier ("reject"),
                          Std_Time_Type);
      New_Interface_Decl (Interfaces, Param, Get_Identifier ("after"),
                          Std_Time_Type);
      Finish_Subprogram_Decl (Interfaces, Ghdl_Signal_Start_Assign_Error);

      --  procedure __ghdl_signal_next_assign_error (sig : __ghdl_signal_ptr;
      --                                             after : std_time);
      Start_Procedure_Decl
        (Interfaces, Get_Identifier ("__ghdl_signal_next_assign_error"),
         O_Storage_External);
      New_Interface_Decl (Interfaces, Param, Get_Identifier ("signal"),
                          Ghdl_Signal_Ptr);
      New_Interface_Decl (Interfaces, Param, Get_Identifier ("after"),
                          Std_Time_Type);
      Finish_Subprogram_Decl (Interfaces, Ghdl_Signal_Next_Assign_Error);

      --  procedure __ghdl_signal_start_assign_null (sign : __ghdl_signal_ptr;
      --                                             reject : std_time;
      --                                             after : std_time);
      Start_Procedure_Decl
        (Interfaces, Get_Identifier ("__ghdl_signal_start_assign_null"),
         O_Storage_External);
      New_Interface_Decl (Interfaces, Param, Get_Identifier ("signal"),
                          Ghdl_Signal_Ptr);
      New_Interface_Decl (Interfaces, Param, Get_Identifier ("reject"),
                          Std_Time_Type);
      New_Interface_Decl (Interfaces, Param, Get_Identifier ("after"),
                          Std_Time_Type);
      Finish_Subprogram_Decl (Interfaces, Ghdl_Signal_Start_Assign_Null);

      --  procedure __ghdl_signal_next_assign_null (sig : __ghdl_signal_ptr;
      --                                            after : std_time);
      Start_Procedure_Decl
        (Interfaces, Get_Identifier ("__ghdl_signal_next_assign_null"),
         O_Storage_External);
      New_Interface_Decl (Interfaces, Param, Get_Identifier ("signal"),
                          Ghdl_Signal_Ptr);
      New_Interface_Decl (Interfaces, Param, Get_Identifier ("after"),
                          Std_Time_Type);
      Finish_Subprogram_Decl (Interfaces, Ghdl_Signal_Next_Assign_Null);

      --  function __ghdl_create_signal_enum8 (init_val : ghdl_integer_type)
      --                                       return __ghdl_signal_ptr;
      --  procedure __ghdl_signal_simple_assign_e8 (sign : __ghdl_signal_ptr;
      --                                            val : __ghdl_integer);
      Create_Signal_Subprograms ("e8", Ghdl_Integer_Type,
                                 Ghdl_Create_Signal_E8,
                                 Ghdl_Signal_Simple_Assign_E8,
                                 Ghdl_Signal_Start_Assign_E8,
                                 Ghdl_Signal_Next_Assign_E8);

      --  function __ghdl_create_signal_b2 (init_val : ghdl_bool_type)
      --                                    return __ghdl_signal_ptr;
      --  procedure __ghdl_signal_simple_assign_b2 (sign : __ghdl_signal_ptr;
      --                                            val : ghdl_bool_type);
      Create_Signal_Subprograms ("b2", Ghdl_Bool_Type_Node,
                                 Ghdl_Create_Signal_B2,
                                 Ghdl_Signal_Simple_Assign_B2,
                                 Ghdl_Signal_Start_Assign_B2,
                                 Ghdl_Signal_Next_Assign_B2);

      Create_Signal_Subprograms ("i32", Ghdl_Integer_Type,
                                 Ghdl_Create_Signal_I32,
                                 Ghdl_Signal_Simple_Assign_I32,
                                 Ghdl_Signal_Start_Assign_I32,
                                 Ghdl_Signal_Next_Assign_I32);

      Create_Signal_Subprograms ("f64", Ghdl_Real_Type,
                                 Ghdl_Create_Signal_F64,
                                 Ghdl_Signal_Simple_Assign_F64,
                                 Ghdl_Signal_Start_Assign_F64,
                                 Ghdl_Signal_Next_Assign_F64);

      Create_Signal_Subprograms ("p64", Ghdl_Phys_Type,
                                 Ghdl_Create_Signal_P64,
                                 Ghdl_Signal_Simple_Assign_P64,
                                 Ghdl_Signal_Start_Assign_P64,
                                 Ghdl_Signal_Next_Assign_P64);

      --  procedure __ghdl_process_add_sensitivity (sig : __ghdl_signal_ptr);
      Start_Procedure_Decl
        (Interfaces, Get_Identifier ("__ghdl_process_add_sensitivity"),
         O_Storage_External);
      New_Interface_Decl (Interfaces, Param, Get_Identifier ("Sig"),
                          Ghdl_Signal_Ptr);
      Finish_Subprogram_Decl (Interfaces, Ghdl_Process_Add_Sensitivity);

      --  procedure __ghdl_process_add_driver (sig : __ghdl_signal_ptr);
      Start_Procedure_Decl
        (Interfaces, Get_Identifier ("__ghdl_process_add_driver"),
         O_Storage_External);
      New_Interface_Decl (Interfaces, Param, Get_Identifier ("Sig"),
                          Ghdl_Signal_Ptr);
      Finish_Subprogram_Decl (Interfaces, Ghdl_Process_Add_Driver);

      --  function __ghdl_create_stable_signal (val : std_time)
      --    return __ghdl_signal_ptr;
      Start_Function_Decl
        (Interfaces, Get_Identifier ("__ghdl_create_stable_signal"),
         O_Storage_External, Ghdl_Signal_Ptr);
      New_Interface_Decl (Interfaces, Param, Get_Identifier ("val"),
                          Std_Time_Type);
      Finish_Subprogram_Decl (Interfaces, Ghdl_Create_Stable_Signal);

      --  procedure __ghdl_signal_register_stable_prefix
      --    (sig : __ghdl_signal_ptr);
      Start_Procedure_Decl
        (Interfaces, Get_Identifier ("__ghdl_signal_register_stable_prefix"),
         O_Storage_External);
      New_Interface_Decl (Interfaces, Param, Get_Identifier ("sig"),
                          Ghdl_Signal_Ptr);
      Finish_Subprogram_Decl (Interfaces, Ghdl_Signal_Register_Stable_Prefix);

      --  function __ghdl_signal_create_guard
      --    (this : ghdl_ptr_type;
      --     proc : ghdl_ptr_type;
      --     instance_name : __ghdl_instance_name_acc)
      --    return __ghdl_signal_ptr;
      Start_Function_Decl
        (Interfaces, Get_Identifier ("__ghdl_signal_create_guard"),
         O_Storage_External, Ghdl_Signal_Ptr);
      New_Interface_Decl (Interfaces, Param, Get_Identifier ("this"),
                          Ghdl_Ptr_Type);
      New_Interface_Decl (Interfaces, Param, Get_Identifier ("proc"),
                          Ghdl_Ptr_Type);
      New_Interface_Decl (Interfaces, Param, Get_Identifier ("instance_name"),
                          Ghdl_Instance_Name_Acc);
      Finish_Subprogram_Decl (Interfaces, Ghdl_Signal_Create_Guard);

      --  procedure __ghdl_signal_guard_dependence (sig : __ghdl_signal_ptr);
      Start_Procedure_Decl
        (Interfaces, Get_Identifier ("__ghdl_signal_guard_dependence"),
         O_Storage_External);
      New_Interface_Decl (Interfaces, Param, Get_Identifier ("sig"),
                          Ghdl_Signal_Ptr);
      Finish_Subprogram_Decl (Interfaces, Ghdl_Signal_Guard_Dependence);

      --  procedure __ghdl_process_wait_exit (void);
      Start_Procedure_Decl
        (Interfaces, Get_Identifier ("__ghdl_process_wait_exit"),
         O_Storage_External);
      Finish_Subprogram_Decl (Interfaces, Ghdl_Process_Wait_Exit);

      --  void __ghdl_process_wait_timeout (time : std_time);
      Start_Procedure_Decl
        (Interfaces, Get_Identifier ("__ghdl_process_wait_timeout"),
         O_Storage_External);
      New_Interface_Decl (Interfaces, Param, Get_Identifier ("time"),
                          Std_Time_Type);
      Finish_Subprogram_Decl (Interfaces, Ghdl_Process_Wait_Timeout);

      --  void __ghdl_process_wait_set_timeout (time : std_time);
      Start_Procedure_Decl
        (Interfaces, Get_Identifier ("__ghdl_process_wait_set_timeout"),
         O_Storage_External);
      New_Interface_Decl (Interfaces, Param, Get_Identifier ("time"),
                          Std_Time_Type);
      Finish_Subprogram_Decl (Interfaces, Ghdl_Process_Wait_Set_Timeout);

      --  void __ghdl_process_wait_add_sensitivity (sig : __ghdl_signal_ptr);
      Start_Procedure_Decl
        (Interfaces, Get_Identifier ("__ghdl_process_wait_add_sensitivity"),
         O_Storage_External);
      New_Interface_Decl (Interfaces, Param, Get_Identifier ("Sig"),
                          Ghdl_Signal_Ptr);
      Finish_Subprogram_Decl (Interfaces, Ghdl_Process_Wait_Add_Sensitivity);

      --  function __ghdl_process_wait_suspend return __ghdl_bool_type;
      Start_Function_Decl
        (Interfaces, Get_Identifier ("__ghdl_process_wait_suspend"),
         O_Storage_External, Ghdl_Bool_Type_Node);
      Finish_Subprogram_Decl (Interfaces, Ghdl_Process_Wait_Suspend);

      --  void __ghdl_process_wait_close (void);
      Start_Procedure_Decl
        (Interfaces, Get_Identifier ("__ghdl_process_wait_close"),
         O_Storage_External);
      Finish_Subprogram_Decl (Interfaces, Ghdl_Process_Wait_Close);
   end Post_Initialize;

   procedure Translate_Std_Type_Declaration (Decl : Iir_Type_Declaration)
   is
   begin
      Chap4.Translate_Type_Declaration (Decl);
      if not Get_Info (Get_Type (Decl)).Type_Desc_Static then
         --  All std types description are static.
         raise Internal_Error;
      end if;
      Chap7.Translate_Implicit_Functions_List
        (Get_Implicit_Functions_List (Decl), Null_Iir_List);
   end Translate_Std_Type_Declaration;

   procedure Translate_Standard (Main : Boolean)
   is
      use Std_Package;
      Lib_Mark, Unit_Mark : Id_Mark_Type;
   begin
      New_Debug_Comment_Decl ("package std.standard");
      if Main then
         Set_Global_Storage (O_Storage_Public);
      else
         Set_Global_Storage (O_Storage_External);
      end if;

      Reset_Identifier_Prefix;
      Push_Identifier_Prefix
        (Lib_Mark, Get_Identifier (Libraries.Std_Library));
      Push_Identifier_Prefix
        (Unit_Mark, Get_Identifier (Standard_Package));

      Chap4.Translate_Bool_Type_Declaration (Boolean_Type);
      --  We need this type very early, for predefined functions.
      Std_Boolean_Type_Node :=
        Get_Ortho_Type (Boolean_Type_Definition, Mode_Value);
      Std_Boolean_True_Node := Get_Ortho_Expr (Boolean_True);
      Std_Boolean_False_Node := Get_Ortho_Expr (Boolean_False);

      Chap4.Translate_Bool_Type_Declaration (Bit_Type);

      Chap4.Translate_Type_Declaration (Character_Type);

      Chap4.Translate_Type_Declaration (Severity_Level_Type);

      Chap4.Translate_Type_Declaration (Universal_Integer_Type);

      Chap4.Translate_Type_Declaration (Universal_Real_Type);

      Chap4.Translate_Type_Declaration (Real_Type);

      Translate_Std_Type_Declaration (Integer_Type);
      Chap4.Translate_Subtype_Declaration (Natural_Subtype);
      Chap4.Translate_Subtype_Declaration (Positive_Subtype);

      Translate_Std_Type_Declaration (String_Type);

      Translate_Std_Type_Declaration (Bit_Vector_Type);

      declare
         Staticness : Iir_Staticness;
      begin
         --  With VHDL93 and later, time type is globally static.  As a result,
         --  it will be elaborated at run-time (and not statically).
         --  However, there is no elaboration of std.standard.  Furthermore,
         --  time type can be pre-elaborated without any difficulties.
         --  There is a kludge here:  set type staticess of time type locally
         --  and then revert it just after its translation.
         Staticness := Get_Type_Staticness (Time_Type_Definition);
         Set_Type_Staticness (Time_Type_Definition, Locally);
         Set_Type_Staticness (Time_Subtype_Definition, Locally);

         Translate_Std_Type_Declaration (Time_Type);

         Set_Type_Staticness (Time_Type_Definition, Staticness);
         Set_Type_Staticness (Time_Subtype_Definition, Staticness);
      end;
      Std_Time_Type := Get_Ortho_Type (Time_Type_Definition, Mode_Value);

      if Flags.Vhdl_Std > Vhdl_87 then
         Translate_Std_Type_Declaration (File_Open_Kind_Type);
         Translate_Std_Type_Declaration (File_Open_Status_Type);
         Std_File_Open_Status_Type :=
           Get_Ortho_Type (File_Open_Status_Type_Definition, Mode_Value);
      end if;

      Pop_Identifier_Prefix (Unit_Mark);
      Pop_Identifier_Prefix (Lib_Mark);

      Post_Initialize;
      --Pop_Global_Factory;
   end Translate_Standard;

   --  Create __ghdl_ELABORATE
   procedure Gen_Main (Entity : Iir_Entity_Declaration;
                       Arch : Iir_Architecture_Declaration;
                       Config_Subprg : O_Dnode)
   is
      Entity_Info : Block_Info_Acc;
      Arch_Info : Block_Info_Acc;
      Inter_List : O_Inter_List;
      Assoc : O_Assoc_List;
      Instance : O_Lnode;
      Arch_Instance : O_Lnode;
      Proc : O_Dnode;

   begin
      Arch_Info := Get_Info (Arch);
      Entity_Info := Get_Info (Entity);

      --  We need to create code.
      Set_Global_Storage (O_Storage_Private);

      --  Declare (but do not define):
      --  Variable for the hierarchy top instance.

      Start_Procedure_Decl (Inter_List, Get_Identifier ("__ghdl_ELABORATE"),
                            O_Storage_Public);
      Finish_Subprogram_Decl (Inter_List, Proc);

      Start_Subprogram_Body (Proc);
      New_Var_Decl (Arch_Instance, Wki_Arch_Instance,
                    O_Storage_External, Arch_Info.Block_Decls_Ptr_Type);

      New_Var_Decl (Instance, Wki_Instance, O_Storage_Local,
                    Entity_Info.Block_Decls_Ptr_Type);

      --  create instance for the architecture
      New_Assign_Stmt
        (Arch_Instance,
         Gen_Alloc (Alloc_Memory,
                    New_Sizeof (Arch_Info.Block_Decls_Type, Ghdl_Index_Type),
                    Arch_Info.Block_Decls_Ptr_Type));

      --  Set the top instance.
      New_Assign_Stmt
        (Instance,
         New_Address (New_Selected_Acc_Value (Arch_Instance,
                                              Arch_Info.Block_Parent_Field),
                      Entity_Info.Block_Decls_Ptr_Type));

      --  init instance
      Push_Scope (Entity_Info.Block_Decls_Type, Instance);
      Chap1.Translate_Entity_Init (Entity, Instance);

      --  elab instance
      Start_Association (Assoc, Arch_Info.Block_Elab_Subprg);
      New_Association (Assoc, New_Value (Instance));
      New_Procedure_Call (Assoc);

      Chap6.Link_Instance_Name (Null_Iir, Entity);

      --  configure instance.
      Start_Association (Assoc, Config_Subprg);
      New_Association (Assoc, New_Value (Arch_Instance));
      New_Procedure_Call (Assoc);

      Pop_Scope (Entity_Info.Block_Decls_Type);
      Finish_Subprogram_Body;
   end Gen_Main;

   --  Return TRUE iff ENTITY can be at the top of a hierarchy, ie:
   --  ENTITY has no generics or all generics have a default expression
   --  ENTITY has no ports or all ports type are constrained.
   procedure Check_Entity_Declaration_Top (Entity : Iir_Entity_Declaration)
   is
      Has_Error : Boolean := False;

      procedure Error (Msg : String; Loc : Iir) is
      begin
         if not Has_Error then
            Error_Msg_Elab
              (Disp_Node (Entity) & " cannot be at the top of a design");
            Has_Error := True;
         end if;
         Error_Msg_Elab (Msg, Loc);
      end Error;

      List : Iir_List;
      El : Iir;
   begin
      --  Check generics.
      List := Get_Generic_List (Entity);
      if List /= Null_Iir_List then
         for I in Natural loop
            El := Get_Nth_Element (List, I);
            exit when El = Null_Iir;
            if Get_Default_Value (El) = Null_Iir then
               Error ("(" & Disp_Node (El) & " has no default value)", El);
            end if;
         end loop;
      end if;

      --  Check port.
      List := Get_Port_List (Entity);
      if List /= Null_Iir_List then
         for I in Natural loop
            El := Get_Nth_Element (List, I);
            exit when El = Null_Iir;
            if Get_Kind (Get_Type (El)) in
              Iir_Kinds_Unconstrained_Array_Type_Definition
              and then Get_Default_Value (El) = Null_Iir
            then
               Error ("(" & Disp_Node (El)
                      & " is unconstrained and has no default value)", El);
            end if;
         end loop;
      end if;
   end Check_Entity_Declaration_Top;

   procedure Gen_Last_Arch (Entity : Iir_Entity_Declaration)
   is
      Entity_Info : Block_Info_Acc;

      Arch : Iir_Architecture_Declaration;
      Arch_Info : Block_Info_Acc;

      Lib : Iir_Library_Declaration;
      Lib_Mark, Entity_Mark, Arch_Mark : Id_Mark_Type;

      Config : Iir_Configuration_Declaration;
      Config_Info : Config_Info_Acc;

      Const : O_Lnode;
      Instance : O_Lnode;
      Inter_List : O_Inter_List;
      Constr : O_Assoc_List;
      Subprg : O_Dnode;
   begin
      Arch := Libraries.Get_Latest_Architecture (Entity);
      if Arch = Null_Iir then
         Error_Msg_Elab ("no architecture for " & Disp_Node (Entity));
      end if;
      Arch_Info := Get_Info (Arch);
      if Arch_Info = null then
         --  Nothing to do here, since the architecture is not used.
         return;
      end if;
      Entity_Info := Get_Info (Entity);

      --  Create trampoline for elab, default_architecture
      --  re-create instsize.
      Reset_Identifier_Prefix;
      Lib := Get_Parent (Get_Parent (Get_Design_Unit (Entity)));
      Push_Identifier_Prefix (Lib_Mark, Get_Identifier (Lib));
      Push_Identifier_Prefix (Entity_Mark, Get_Identifier (Entity));
      Push_Identifier_Prefix (Arch_Mark, "LASTARCH");

      --  Instance size.
      New_Const_Decl
        (Const, Create_Identifier ("INSTSIZE"), O_Storage_Public,
         Ghdl_Index_Type,
         New_Sizeof (Arch_Info.Block_Decls_Type, Ghdl_Index_Type));

      --  Elaborator.
      Start_Procedure_Decl
        (Inter_List, Create_Identifier ("ELAB"), O_Storage_Public);
      New_Interface_Decl
        (Inter_List, Instance, Wki_Instance,
         Entity_Info.Block_Decls_Ptr_Type);
      Finish_Subprogram_Decl (Inter_List, Subprg);

      Start_Subprogram_Body (Subprg);
      Start_Association (Constr, Arch_Info.Block_Elab_Subprg);
      New_Association (Constr, New_Value (Instance));
      New_Procedure_Call (Constr);
      Finish_Subprogram_Body;

      --  Default config.
      Config := Get_Library_Unit
        (Get_Default_Configuration_Declaration (Arch));
      Config_Info := Get_Info (Config);
      if Config_Info /= null then
         --  Do not create a trampoline for the default_config if it is not
         --  used.
         Start_Procedure_Decl
           (Inter_List, Create_Identifier ("DEFAULT_CONFIG"),
            O_Storage_Public);
         New_Interface_Decl (Inter_List, Instance, Wki_Instance,
                             Entity_Info.Block_Decls_Ptr_Type);
         Finish_Subprogram_Decl (Inter_List, Subprg);

         Start_Subprogram_Body (Subprg);
         Start_Association (Constr, Config_Info.Config_Subprg);
         New_Association (Constr, New_Value (Instance));
         New_Procedure_Call (Constr);
         Finish_Subprogram_Body;
      end if;

      Pop_Identifier_Prefix (Arch_Mark);
      Pop_Identifier_Prefix (Entity_Mark);
      Pop_Identifier_Prefix (Lib_Mark);
   end Gen_Last_Arch;

   procedure Elaborate
     (Primary : Ortho_Front.String_Acc;
      Secondary : Ortho_Front.String_Acc;
      Filelist : Ortho_Front.String_Acc)
   is
      use Name_Table;
      use Configuration;

      Primary_Id : Name_Id;
      Secondary_Id : Name_Id;
      Unit : Iir_Design_Unit;
      File : Iir_Design_File;
      Lib_Unit : Iir;
      Config : Iir_Design_Unit;
      Config_Lib : Iir_Configuration_Declaration;
      Entity : Iir_Entity_Declaration;
      Arch : Iir_Architecture_Declaration;
      Arch_Info : Block_Info_Acc;
      Conf_Info : Config_Info_Acc;
   begin
      Primary_Id := Get_Identifier (Primary.all);
      if Secondary /= null then
         Secondary_Id := Get_Identifier (Secondary.all);
      else
         Secondary_Id := Null_Identifier;
      end if;
      Config := Configure (Primary_Id, Secondary_Id);
      if Config = Null_Iir then
         return;
      end if;
      Config_Lib := Get_Library_Unit (Config);
      Entity := Get_Library_Unit (Get_Entity (Config_Lib));
      Arch := Get_Block_Specification (Get_Block_Configuration (Config_Lib));

      --  Be sure the entity can be at the top of a design.
      Check_Entity_Declaration_Top (Entity);

      --  Return now in case of errors.
      if Nbr_Errors /= 0 then
         return;
      end if;

      --  Translate all configurations needed.
      --  Also, set the ELAB_FLAG on package with body.
      for I in Design_Units.First .. Design_Units.Last loop
         Unit := Design_Units.Table (I);
         Lib_Unit := Get_Library_Unit (Unit);
         case Get_Kind (Lib_Unit) is
            when Iir_Kind_Configuration_Declaration =>
               --  Generate code only for default configuration.
               --  Non-default configuration declaration code is generated
               --  during compilation.
               if Get_Identifier (Lib_Unit) = Null_Identifier then
                  Translate (Unit, True);
               else
                  Translate (Unit, False);
               end if;
            when Iir_Kind_Entity_Declaration
              | Iir_Kind_Architecture_Declaration
              | Iir_Kind_Package_Declaration =>
               Set_Elab_Flag (Unit, False);
               Translate (Unit, False);
            when Iir_Kind_Package_Body =>
               Set_Elab_Flag (Get_Parent (Get_Package (Lib_Unit)), True);
            when others =>
               Error_Kind ("elaborate", Lib_Unit);
         end case;
      end loop;

      --  Generate code to elaboration body-less package.
      for I in Design_Units.First .. Design_Units.Last loop
         Unit := Design_Units.Table (I);
         Lib_Unit := Get_Library_Unit (Unit);
         case Get_Kind (Lib_Unit) is
            when Iir_Kind_Package_Declaration =>
               if not Get_Elab_Flag (Unit) then
                  Chap2.Elab_Package_Body (Lib_Unit, Null_Iir);
               end if;
            when Iir_Kind_Entity_Declaration =>
               Gen_Last_Arch (Lib_Unit);
            when Iir_Kind_Architecture_Declaration
              | Iir_Kind_Package_Body
              | Iir_Kind_Configuration_Declaration =>
               null;
            when others =>
               Error_Kind ("elaborate(2)", Lib_Unit);
         end case;
      end loop;

      --  Create main code.
      Arch_Info := Get_Info (Arch);
      Conf_Info := Get_Info (Config_Lib);
      Gen_Main (Entity, Arch, Conf_Info.Config_Subprg);

      --  Disp list of files needed.
      if Filelist /= null then
         declare
            use Interfaces.C_Streams;
            use System;
            Nul : constant Character := Character'Val (0);
            Fname : String := Filelist.all & Nul;
            Mode : constant String := "wt" & Nul;
            F : FILEs;
            R : int;
            S : size_t;
            Id : Name_Id;
            Lib : Iir_Library_Declaration;
         begin
            F := fopen (Fname'Address, Mode'Address);
            if F = NULL_Stream then
               Error_Msg_Elab ("cannot open " & Filelist.all);
            end if;
            for I in Design_Units.First .. Design_Units.Last loop
               Unit := Design_Units.Table (I);
               File := Get_Design_File (Unit);
               if not Get_Elab_Flag (File) then
                  Set_Elab_Flag (File, True);

                  Lib := Get_Parent (File);
                  R := fputc (Character'Pos ('>'), F);
                  Id := Get_Library_Directory (Lib);
                  S := fwrite (Get_Address (Id),
                               size_t (Get_Name_Length (Id)), 1, F);
                  R := fputc (10, F);

                  Id := Get_Design_File_Filename (File);
                  S := fwrite (Get_Address (Id),
                               size_t (Get_Name_Length (Id)), 1, F);
                  R := fputc (10, F);
               end if;
            end loop;
         end;
      end if;
   end Elaborate;
end Translation;
